/********************************************************************
        example1.c - example of use of XCDE APIs - July, 2002
	        Andrea Mastroianni - Paolo Ferragina 
       Dipartimento di Informatica - Universita' di Pisa (Italy)
                      ferragin@di.unipi.it
                      mastroia@di.unipi.it

   for permissions of distribution and use see the COPYRIGHT file
   contained in the XCDE package 
********************************************************************/

// header file that must be always included in a program
// that uses the XCDE library.
#include"xcde_lib.h"

#define MACRO_HEADER { \
fprintf(stdout, "\n\t\b\b-------------------------------------------") ; \
fprintf(stdout, "\n\t\tXCDE Library example of use") ;           \
fprintf(stdout, "\n\tby Paolo Ferragina - Andrea Mastroianni") ; \
fprintf(stdout, "\n\t\b\bferragin@di.unipi.it - mastroia@di.unipi.it") ; \
fprintf(stdout, "\n\t        Dipartimento di Informatica") ;     \
fprintf(stdout, "\n\t           Universita' di Pisa") ;          \
fprintf(stdout, "\n\t\b\b-------------------------------------------") ; \
fflush(stdout) ; \
} 

#define MACRO_HELP { \
fprintf(stdout, "\n\nSYNTAX: example1 [FILENAME] [TAG] [WORD]") ; \
fprintf(stdout, "\n\nDESCRIPTION: gets a filename, a tag and a word from the \ncommand line and returns the number of occurrences of\nthe word contained in the tag for that file; for each\noccurrence prints also its position in the body file.\n\n") ; \
fflush(stdout) ; \
}

#define MACRO_ERROR { \
XCDE_Finalize_Console(&console) ; \
exit(-1) ; \
}

#define MACRO_NO_OCCURRENCES { \
fprintf(stdout, "\nThere're no occurrences !!!\n") ; \
XCDE_Finalize_Console(&console) ; \
return(0) ; \
}

// -----------------------------------------------------------------  /
// | this program reads a filename, a tag and a word from the       | / 
// | command line (in this order) and finds all the occurrences in  | /
// | the source file where the word is in the tag                   | /
// ------------------------------------------------------------------ /

int main(int argc, char *argv[]) 
{ // XCDE console data structure, needed to deal XCDE data. The console must
  // be associated with an xml document that has before been compressed and 
  // indexed using XCDE_Build command. By using the XCDE_Init_Console
  // function the console is associated with an XML document that has been
  // compressed and indexed by the XCDE_Build command. Components of the console
  // are initialized with values regarding that document. All XCDE API functions 
  // get a console reference as a parameter.
  XCDE_Console_Type console ;

  // IL associated with the tag. It's another XCDE data type. An IL is
  // implemented using an array of element of type XCDE_IL_Couple_Type. 
  // This one is a struct with two integer component: XCDE_Position wich
  // contains the position of the occurrence in the body compressed file 
  // (or, better, the position of the codeword in Huffman coding in the
  // body file) and XCDE_Offset, which is different from 0 only when the
  // token at that position is derived from the expansion of an internal 
  // entity. With these two values is possible to know if a word
  // occurrence is contained in a tag occurrence. The IL of a tag has 
  // always a even number of  element because its contains the opening 
  // and closing positions. 
  XCDE_IL_Couple_Type *tag_IL ;

  // IL associated with the word. 
  XCDE_IL_Couple_Type *word_IL ;

  // value returned from a XCDE function: XCDE APIs always return a value 
  // that is 1 is all is correct or -1 otherwise. 
  int res=0 ;

  // rank of the tag in the tags dictionary. The APIs that search in 
  // different dictionaries (TAG, ATTR, WORD) returns the position (or
  // positions, because for attributes and words dictionaries its possibile
  // to return more values) of the element in the specific dictionary. This value
  // (or these values) is called the "rank in the dictionary" and can be a number
  // greater than 0 and less than the total number of elements in the dictionary. 
  // To obtain the IL of the element the rank in the dictionary must be traslated
  // to the general rank, that is the rank in the global dictionary, using a specific
  // API. 
  int tagdictrank=0 ;

  // rank of the tag in the general dictionary.
  int taggenrank=0 ;

  // rank of the word in the words dictionary. It's a pointer to an array because 
  // the occurrences may be more than one, even if in this case the search made 
  // on the words dictionary is exact and so the maximum number of occurrences is one.
  int *worddictrank=NULL ;

  // number of occurrence of the word in the words dictionary.
  int numworddictrank=0 ; 

  // rank of the words in the general dictionary. 
  int wordgenrank=0 ; 

  // pointer to the word positions list (that is an array of XCDE_IL_Couple_Type).
  int wordlistpos=0 ;

  // pointer to the tag list of positions.
  int taglistpos=0 ;

  // dimension of the tag list of positions.
  int tag_IL_length=0 ;

  // dimension of the word list of positions. 
  int word_IL_length=0 ; 

  // number of totaal occurrences of the word in the tag.
  int nocc=0 ; 

  // checks if the number of parameters is correct
  MACRO_HEADER

  if (argc == 1) {

    MACRO_HELP
    
    return(1) ;
  }

  if (argc != 4) {

    fprintf(stderr, "\n\nFatal Error: parameters aren't correct\n") ;

    MACRO_HELP

    exit(-1) ; 
  }

  // initializes the console: argv[1] must contain the
  // name of an XML file that has been compressed and indexed 
  // with the XCDE_Build command. If it isn't in the 
  // current directory it must also contain the path. Obviously
  // the source file hasn't to be contained in the directory. 
  // From this moment the console is associated with the 
  // document. 
  res=XCDE_Init_Console(argv[1],
			&console) ;
  
  if (res == -1) MACRO_ERROR 
  
  // argv[2] must contain the name of a tag. This API 
  // checks if the tag is in the tags dictionary and
  // in this case returns its rank in the dictionary.
  // A dictionary rank is always greater than 0.
  res=XCDE_Search_Tag_Dictionary(argv[2],
				 &tagdictrank,
				 &console) ;
fprintf(stdout, "\n\nrankdict: %d\n",tagdictrank) ;
  if (res == -1) MACRO_ERROR

  // checks if there's any occurrence of the tag.
  if (tagdictrank == 0) MACRO_NO_OCCURRENCES

  // if the program arrives here the tag is contained in the
  // tags dictionary.

  // argv[3] must contain the name of a word. This API
  // checks if the word is in the words dictionary: it is
  // more complex than the correspondent TAG API, because
  // different type of search can be made for the words dictionary
  // (for example exact, suffix, prefix, etc.). It's possible
  // that it returns more than one rank, so an array is passed
  // to contain ranks (worddictrank) and an integer variable
  // to contain the number of results (numworddictrank).
  res=XCDE_Search_Word_Dictionary(argv[3],
				  &worddictrank,
				  &numworddictrank,
				  XCDE_EXACT,
				  XCDE_CASE,
				  0,
				  &console) ;

  if (res == -1) MACRO_ERROR

  // traslates the rank in the tag dictionary to the general rank:
  // it's necessary to read the IL.
  res=XCDE_Dict2Rank(TAG,
		     tagdictrank,
		     &taggenrank,
		     &console) ;
fprintf(stdout, "\n\nrankassoluto: %d\n",taggenrank) ;
  if (res == -1) MACRO_ERROR

  // checks if the words dictionary contains the word.
  if (numworddictrank == 0) MACRO_NO_OCCURRENCES

  // if the program arrives here the word is contained in the
  // words dictionary.

//QUIIIIIIIIIIIIIIIIIIIII
  // traslates the rank in the word dictionary to the general rank:
  // it's necessary to read the IL.
  res=XCDE_Dict2Rank(WORD,
		     worddictrank[0],
		     &wordgenrank,
		     &console) ;
  
  if (res == -1) MACRO_ERROR
  
  // reads the IL associated to the tag and its length
  res=XCDE_Get_IL_ListPos(taggenrank,
			  &tag_IL,
			  &tag_IL_length,
			  &console) ;
  
  if (res == -1) MACRO_ERROR

  // checks if there's any occurrence of the tag. Also if the tag 
  // is contained in the dictionary it's possible that there're no
  // occurrences, because in the dictionary there're physical and
  // logical token (a physical token is necessary to decompress
  // the file but hasn't been indexed, for example tokens in 
  // comments).  		  
  if (tag_IL_length == 0) MACRO_NO_OCCURRENCES
    
  // if the program arrives here the IL associated with the tag 
  // isn't empty.

  // reads the IL associated to the word and its length.
  res=XCDE_Get_IL_ListPos(wordgenrank, 
			  &word_IL,
			  &word_IL_length,
			  &console) ;
  
  if (res == -1) MACRO_ERROR
  
  // checks if there's any occurrence of the word.
  if (word_IL_length == 0) MACRO_NO_OCCURRENCES 

  // if the program arrives here the IL associated with the tag			     
  // isn't empty.			     

  taglistpos=0 ; 
  
  wordlistpos=0 ;
  
  // loops over the inverted lists to find all word occurrences that are
  // contained in the tag: its possible to know that by analyzing positions
  // in the IL. If a word position is between a tag opening and closing 
  // positions so it's contained in the tag. The IL are sorted by increasing
  // position.
  while((taglistpos < tag_IL_length) && (wordlistpos < word_IL_length)) { 
      
      // the word occurrence is in the tag: increases the number of occurrences
      // and the word list pointer.
      if (((tag_IL[taglistpos].XCDE_Position < word_IL[wordlistpos].XCDE_Position) &&
	   (tag_IL[taglistpos+1].XCDE_Position > word_IL[wordlistpos].XCDE_Position)) || 
	  
	  ((tag_IL[taglistpos].XCDE_Position == word_IL[wordlistpos].XCDE_Position) &&
	   (tag_IL[taglistpos].XCDE_Offset < word_IL[wordlistpos].XCDE_Offset) &&
	   (tag_IL[taglistpos+1].XCDE_Position == word_IL[wordlistpos].XCDE_Position) &&
	   (tag_IL[taglistpos+1].XCDE_Offset > word_IL[wordlistpos].XCDE_Offset))) {
	
	++nocc ;

	fprintf(stdout, "\nOccurrence number %d position %d in the body file", nocc, word_IL[wordlistpos].XCDE_Position) ; 
	
	++wordlistpos ;
      }
      
      else
	
	// the word occurrence is before the tag: increases the word list pointer.
	if((tag_IL[taglistpos].XCDE_Position > word_IL[wordlistpos].XCDE_Position) ||
	   ((tag_IL[taglistpos].XCDE_Position == word_IL[wordlistpos].XCDE_Position) &&
	    (tag_IL[taglistpos].XCDE_Offset > word_IL[wordlistpos].XCDE_Offset))) 
	  
	  ++wordlistpos ;
      
	else
	  
	  // the word occurrence is after the tag: increases the tag list pointer.
	  if((tag_IL[taglistpos+1].XCDE_Position < word_IL[wordlistpos].XCDE_Position) ||
	     ((tag_IL[taglistpos+1].XCDE_Position == word_IL[wordlistpos].XCDE_Position) &&
	      (tag_IL[taglistpos+1].XCDE_Offset < word_IL[wordlistpos].XCDE_Offset))) 
	    
	    taglistpos+=2 ; 
      
	  else
	    
	    MACRO_ERROR
    }

  // prints the result and frees the memory allocated. 
  if (nocc > 0) fprintf(stdout, "\nNumber of occurrences: %d\n\n", nocc) ;
  
  else MACRO_NO_OCCURRENCES

  // frees the memory allocated: remember that memory for arrays is
  // allocated by the API, but it must be freed from the programmer.
  free(word_IL) ; 
  
  free(tag_IL) ; 

  free(worddictrank) ; 
  
  // finalizes the console. After this operation the memory
  // needed for the console is freed and temporary files 
  // removed. The console variable should be also reused
  // for another document. 
  XCDE_Finalize_Console(&console) ; 
  
  return(1) ; 
}
     
	  
    
    
