/********************************************************************
     	       example3.c - example of use of XCDE API
	        Andrea Mastroianni - Paolo Ferragina 
    Dipartimento di Informatica - Universita' di Pisa (Italy)
	               ferragin@di.unipi.it
                       mastroia@di.unipi.it

   for permissions of distribution and use see the COPYRIGHT file
   contained in the XCDE package
********************************************************************/

// header file that must be always included in a program
// that uses the XCDE library.
#include"xcde_lib.h"

#define MACRO_HEADER { \
fprintf(stdout, "\n\t\b\b-------------------------------------------") ; \
fprintf(stdout, "\n\t\tXCDE Library example of use") ;           \
fprintf(stdout, "\n\tby Paolo Ferragina - Andrea Mastroianni") ; \
fprintf(stdout, "\n\t\b\bferragin@di.unipi.it - mastroia@di.unipi.it") ; \
fprintf(stdout, "\n\t        Dipartimento di Informatica") ;     \
fprintf(stdout, "\n\t           Universita' di Pisa") ;          \
fprintf(stdout, "\n\t\b\b-------------------------------------------") ; \
fflush(stdout) ; \
} 

#define MACRO_HELP { \
fprintf(stdout, "\n\nSYNTAX: example3 [FILENAME] [WORD]") ; \
fprintf(stdout, "\n\nDESCRIPTION: gets a filename and a word from the \ncommand line and returns a well-formed snippet\nfor each of the first three occurrences of the\nword in that file.\n\n") ; \
fflush(stdout) ; \
}

#define MACRO_ERROR { \
XCDE_Finalize_Console(&console) ; \
exit(-1) ; \
}

#define MACRO_NO_OCCURRENCES { \
fprintf(stdout, "\nThere're no occurrences !!!\n") ; \
XCDE_Finalize_Console(&console) ; \
return(0) ; \
}

// -------------------------------------------------------------------  /
// | this programs reads a filename and a word from the command line  | /
// | and for the first three occurrences of the word included in the  | /
// | tag prints a well-formed snippet.                                | /
// -------------------------------------------------------------------- /

int main(int argc, char *argv[]) 
{ // XCDE console data structure, needed to deal XCDE data. The console must
  // be associated with an XML document that has before been compressed and 
  // indexed using XCDE_Build command. By using the XCDE_Init_Console
  // function the console is associated with an XML document that has been
  // compressed and indexed by the XCDE_Build command. Components of the console
  // are initialized with values regarding that document. All XCDE API functions 
  // get a console reference as a parameter.
  XCDE_Console_Type console ;

  // IL associated with the tag. It's another XCDE data type. An IL is
  // implemented using an array of element of type XCDE_IL_Couple_Type. 
  // This one is a struct with two integer component: XCDE_Position wich
  // contains the position of the occurrence in the body compressed file 
  // (or, better, the position of the codeword in Huffman coding in the
  // body file) and XCDE_Offset, which is different from 0 only when the
  // position associated with the token is derived from the expansion of 
  // an internal entity. With these two values is possible to determine 
  // if a word occurrence is contained in a tag occurrence. The IL of a 
  // tag has always a even number of element because its contains the 
  // opening and closing position. 
  XCDE_IL_Couple_Type *tag_IL=NULL ;

  // dimension of the IL associated with a tag in the tagstruct.
  int tag_IL_length=0 ;

  // IL associated with the word.
  XCDE_IL_Couple_Type *word_IL=NULL ;

  // dimension of the IL associated with the word.
  int word_IL_length=0 ;

  // value returned from a XCDE function: XCDE APIs always return a value 
  // that is 1 is all is correct or -1 otherwise. 
  int res=0 ;

  // pointer to an array that contains the ranks of the word in the 
  // words dictionary. It's an array because in general it can 
  // contain more than a value, because the type of search 
  // can be not only exact, but also for prefix, suffix, error and so
  // on. In this case the search is exact, so the array will contain
  // maximum one value. 
  int *worddictrank=NULL ;

  // variable that contains the number of occurrence of the word in the words dictionary 
  // (must be equal to 1 because the search is EXACT).
  int numwordindictrank=0 ; 

  // variable that contains the rank of the word in the general dictionary. It's obtained
  // traslating the rank in the dictionary by using an XCDE API and it's necessary to 
  // access to the IL. 
  int wordgenrank=0 ;

  // flag used to check is the tagstruct needed to create the snippets
  // has been created. To create the tagstruct is necessary to put the 
  // flag -t when the file is compressed and indexed by the XCDE_Build
  // comand. Otherwise the tagstruct isn't created and the snippets 
  // cannot be computed, so this program print an error message. 
  unsigned
    char tagstruct_created_flag=0u ;
  
  // flag used to check the termination condition for the loop
  unsigned 
    char done=0u ;

  // pointer to the array of integer returned from the API that computes
  // the tagstruct for a point in the body file. The array contains the
  // ranks in the general dictionary of the tags that form the tagstruct: 
  // the tag in the position i contains the tag in position i+1 and so on.
  int *tagsrank=NULL ;

  // number of tags in the previous array.
  int numtags=0 ;

  // pointer to the IL word. Remember that an IL is implemented as an 
  // array.
  int wordpointer=0 ;

  // pointer to a token string. A rank in the general dictionary can be
  // traslated to the string that represents the token. The string
  // can so be printed. 
  char *token=NULL ; 

  // variables used to read a codeword from the body file. The XCDE
  // APIs used to read the codeword returns also its length.
  int cw=0, cwlen=0, body_position=0 ;

  // variables used as counters in for loops
  register
    int k=0, j=0 ;

  // checks if the number of parameters is correct
  MACRO_HEADER

  if (argc == 1) {

    MACRO_HELP
    
    return(1) ;
  }

  if (argc != 3) {

    fprintf(stderr, "\n\nFatal Error: parameters aren't correct\n") ;

    MACRO_HELP

    exit(-1) ; 
  }
  // initializes the console: argv[1] must contain the
  // name of an xml file that has been compressed and indexed 
  // with the XCDE_Build command. If it isn't in the 
  // current directory it must also contain the path. Obviously
  // the source file hasn't to be contained in the directory. 
  // From this moment the console is associated with the 
  // document. 
  res=XCDE_Init_Console(argv[1],
			&console) ;
  
  if (res == -1) MACRO_ERROR 

  // this API returns a flag by which is possible to know
  // if tha tagstruct needed to this program has been 
  // created.
  res=XCDE_SettedTagstruct(&tagstruct_created_flag,
			   &console) ;
		   
  if (res == -1) MACRO_ERROR

  // if the tagstruct hasn't be created prints an error message and exits.
  if (!tagstruct_created_flag) { 

    fprintf(stderr, "\n\nError: the tagstruct is not present !!!\n\n") ;

    MACRO_ERROR }
  
  // if arrives here so the tagstruct has been created.

  // argv[3] must contain the name of a word. This API 
  // checks if the word passed is in the words dictionary: It's possible
  // that it returns more than one rank, so an array is passed
  // to contain ranks (word1dictrank) and an integer variable 
  // to contain the number of results (num1worddictrank). In this
  // case the search is made exact, so the maximum lenght of the 
  // array is one.  
  res=XCDE_Search_Word_Dictionary(argv[2],
				  &worddictrank,
				  &numwordindictrank,
				  XCDE_EXACT,
				  XCDE_CASE,
				  0,
				  &console) ; 

  if (res == -1) MACRO_ERROR 

  // checks if the word is contained in the dictionary.
  if (numwordindictrank == 0) MACRO_NO_OCCURRENCES 

  // if arrives here so the word is contained in the words dictionary.

  // traslates the rank of the word in the words dictionary to the
  // rank in the general dictionary. This one is needed to access
  // the IL.
  res=XCDE_Dict2Rank(WORD,
		     worddictrank[0],
		     &wordgenrank, 
		     &console) ;
  
  if (res == -1) MACRO_ERROR
		   
  // reads the IL associated with the word. 
  res=XCDE_Get_IL_ListPos(wordgenrank,
			  &word_IL,
			  &word_IL_length,
			  &console) ;
  
  if (res == -1) MACRO_ERROR

  // checks if the IL associated with the word is empty. In this 
  // case it means that the word is a physical token, needed to 
  // compress and decompress the file (for example a token in a comment) 
  // but there's not a IL associated with it.
  if (word_IL_length == 0) MACRO_NO_OCCURRENCES
    
  if (res == -1) MACRO_ERROR

  fprintf(stdout, "\nTotal number of occurrences: %d\n", word_IL_length) ;

  // loops over all occurrences of the word and print a snippet
  // for each of them until a maximum of three.
  for(wordpointer=0 ; ((wordpointer < word_IL_length) && (wordpointer < 3)); ++wordpointer) {

    fprintf(stdout, "---------------------------------------") ;

    fprintf(stdout, "\n snippet number %d\n", wordpointer+1) ;

    // reads the tagstruct for an occurrence of the word. The API
    // gets a position in the body file (a couple XCDE_Position -
    // XCDE_Offset) and returns an array of integer that will
    // contain the ranks in the general dictionary of tags
    // that form the tagstruct, and where the tag in the position i 
    // contains the tag in the position i+1 and so on. 
    res=XCDE_Get_Tagstruct_Tags_Enclosing(&word_IL[wordpointer],
					  &tagsrank,
					  &numtags,
					  &console) ;

    if (res == -1) MACRO_ERROR 
	
    // prints the first part of the snippet: loops over
    // all components of the tagstruct.
    for(k=0; k<numtags;++k) {

      // find the token associated with the rank in the 
      // general dictionary of the tag.
      res=XCDE_Rank2Token(tagsrank[k],
			  &token,
			  &console) ;
      
      if (res == -1) MACRO_ERROR
		       
      // prints spaces needed to output the snippet.
      for(j=0; j<=k;++j) fprintf(stdout, " ") ;

      // prints the TAG
      fprintf(stdout, "<%s>\n", token) ;
      
      // frees the memory needed to contain the TAG.
      free(token) ;
      
      token=NULL ;
    }
    
    // gets the IL of the most grafted tag in the 
    // tagstruct.
    res=XCDE_Get_IL_ListPos(tagsrank[numtags-1],
			    &tag_IL,
			    &tag_IL_length,
			    &console) ;
    
   if (res == -1) MACRO_ERROR
   
   // finds the occurrence of the most grafted tag that contains the word.
   // That happens when the word position is between tag opening and 
   // closing positions.
   for(k=0, done=0u; ((k<tag_IL_length) && (!done)); k+=2) {
	
     if (((tag_IL[k].XCDE_Position < word_IL[wordpointer].XCDE_Position) &&
	  (tag_IL[k+1].XCDE_Position > word_IL[wordpointer].XCDE_Position)) ||
	 ((tag_IL[k].XCDE_Position == word_IL[wordpointer].XCDE_Position) &&
	  (tag_IL[k].XCDE_Offset < word_IL[wordpointer].XCDE_Offset) &&
	  (tag_IL[k+1].XCDE_Position == word_IL[wordpointer].XCDE_Position) &&
	  (tag_IL[k+1].XCDE_Offset < word_IL[wordpointer].XCDE_Offset)))

       done=1u ;
   }

   // if the occurrence isn't found there's an error.
   if (!done) MACRO_ERROR

   // if arrives here the first part of the snippet was printed,
   // and the position of the most grafted tag that contain the 
   // word was found. The part of the document between its opening
   // and closing position is going to be printed. 
   body_position=tag_IL[k].XCDE_Position ;

   fprintf(stdout, "\n .... ") ;

   // prints words between in the most gradted tag.
   while(body_position <= tag_IL[k+1].XCDE_Position) {

     // reads the Huffman codeword at the position in the
     // body file passed as parameter. Also the codeword length 
     // will be read.
     res=XCDE_Get_Body_Forward_Codeword(body_position,
					&cw,
					&cwlen,
					&console) ;

     if (res == -1) MACRO_ERROR 
     
     // reads the TAG associated with the codeword.
     res=XCDE_Codeword2Token(cw,
			     cwlen,
			     &token,
			     &console) ;

     if (res == -1) MACRO_ERROR

     // prints the TAG
     fprintf(stdout, "%s", token) ;
	
     // frees the memory needed to contain the token. This memory
     // has been allocated by the API but must be explicitly freed
     // from the programmer. 
     free(token) ;

     // increaes the position in the body file using the codeword
     // length previous read. The next codeword will be read at
     // the new position.
     body_position+=cwlen ;
   }
			    
   fprintf(stdout, " .... \n") ;	  

   // free the memory needed to contain the most grafted tag IL. 
   // It's allocated from the API but must be explicitly freed
   // from the programmer.
   free(tag_IL) ; 

   tag_IL=NULL ;

   tag_IL_length=0 ;
	
   // prints the tagstruct in reverse order as the footer of the snippet.
   for(k=numtags-1; k>= 0;--k) {
     
     // gets the TAG corresponding to the tag rank in the
     // general dictionary. 
     res=XCDE_Rank2Token(tagsrank[k],
			 &token,
			 &console) ;

     if (res == -1) MACRO_ERROR

     // prints space to output the snippet. 
     for(j=0; j<=k;++j) fprintf(stdout, " ") ;

     // prints the TAG
     fprintf(stdout, "</%s>\n", token) ;

     // free the memory needed to contain the tag 
     // and allocated from the API.
     free(token) ;

     token=NULL ;
   }

   // frees the memory needed to contain the tagstruct
   // and allocated form the API. 
   free(tagsrank) ;

   numtags=0 ;
  }

  // frees the memory needed to contain the IL of the 
  // word and its occurrences in the dictionary. 
  free(word_IL) ; 
    
  free(worddictrank) ; 

  // finalizes the console. After this operation the memory
  // needed for the console is freed and temporary files 
  // removed. The console variable should be also reused
  // for another document. 
  XCDE_Finalize_Console(&console) ; 

  return(1) ; 
}
     
	  
    
    
