/********************************************************************
        example1.c - example of use of XCDE APIs - July, 2002
	        Andrea Mastroianni - Paolo Ferragina
       Dipartimento di Informatica - Universita' di Pisa (Italy)
                      ferragin@di.unipi.it
                      mastroia@di.unipi.it

   for permissions of distribution and use see the COPYRIGHT file
   contained in the XCDE package
********************************************************************/

// header file that must be always included in a program
// that uses the XCDE library.
#include"xcde_lib.h"

#define MACRO_HEADER { \
fprintf(stdout, "\n\t\b\b-------------------------------------------") ; \
fprintf(stdout, "\n\t\tXCDE Library example of use") ;           \
fprintf(stdout, "\n\tby Paolo Ferragina - Andrea Mastroianni") ; \
fprintf(stdout, "\n\t\b\bferragin@di.unipi.it - mastroia@di.unipi.it") ; \
fprintf(stdout, "\n\t        Dipartimento di Informatica") ;     \
fprintf(stdout, "\n\t           Universita' di Pisa") ;          \
fprintf(stdout, "\n\t\b\b-------------------------------------------") ; \
fflush(stdout) ; \
} 

#define MACRO_HELP { \
fprintf(stdout, "\n\nSYNTAX: example1 [FILENAME] [TAG] ") ; \
fprintf(stdout, "\n\nDESCRIPTION: gets a filename, a tag and a word from the \ncommand line and returns the number of occurrences of\nthe word contained in the tag for that file; for each\noccurrence prints also its position in the body file.\n\n") ; \
fflush(stdout) ; \
}

#define MACRO_ERROR { \
XCDE_Finalize_Console(&console) ; \
exit(-1) ; \
}

#define MACRO_NO_OCCURRENCES { \
fprintf(stdout, "\nThere're no occurrences !!!\n") ; \
XCDE_Finalize_Console(&console) ; \
return(0) ; \
}

// -----------------------------------------------------------------  /
// | this program reads a filename, a tag e calcola il rango nel diz del tag     | /
// | command line (in this order) and finds all the occurrences in  | /
// | the source file where the word is in the tag                   | /
// ------------------------------------------------------------------ /

int main(int argc, char *argv[]){
	 // XCDE console data structure, needed to deal XCDE data. The console must
	// be associated with an xml document that has before been compressed and
	// indexed using XCDE_Build command. By using the XCDE_Init_Console
	// function the console is associated with an XML document that has been
	// compressed and indexed by the XCDE_Build command. Components of the console
	// are initialized with values regarding that document. All XCDE API functions
	// get a console reference as a parameter.
	XCDE_Console_Type console ;

	// value returned from a XCDE function: XCDE APIs always return a value
	// that is 1 is all is correct or -1 otherwise.
	int res=0 ;

	// rank of the tag in the tags dictionary. The APIs that search in
	// different dictionaries (TAG, ATTR, WORD) returns the position (or
	// positions, because for attributes and words dictionaries its possibile
	// to return more values) of the element in the specific dictionary. This value
	// (or these values) is called the "rank in the dictionary" and can be a number
	// greater than 0 and less than the total number of elements in the dictionary.
	// To obtain the IL of the element the rank in the dictionary must be traslated
	// to the general rank, that is the rank in the global dictionary, using a specific
	// API.
	int tagdictrank=0 ;
	
	int genrank=0;

	int *depth=NULL;

	int number_depths;

	int i;

	// checks if the number of parameters is correct
	MACRO_HEADER

	if (argc == 1) {

		MACRO_HELP

		return(1) ;
	}

	if (argc != 3) {

		fprintf(stderr, "\n\nFatal Error: parameters aren't correct\n") ;

		MACRO_HELP

		exit(-1) ;
	}

	// initializes the console: argv[1] must contain the
	// name of an XML file that has been compressed and indexed
	// with the XCDE_Build command. If it isn't in the
	// current directory it must also contain the path. Obviously
	// the source file hasn't to be contained in the directory.
	// From this moment the console is associated with the
	// document.
	res = XCDE_Init_Console(argv[1], &console) ;

	if (res == -1) MACRO_ERROR

	// argv[2] must contain the name of a tag. This API
	// checks if the tag is in the tags dictionary and
	// in this case returns its rank in the dictionary.
	// A dictionary rank is always greater than 0.
	res=XCDE_Search_Tag_Dictionary(argv[2], &tagdictrank, &console) ;

	if (res == -1) MACRO_ERROR

	fprintf(stdout, "\n\n rango nel dizionario: %d\n",tagdictrank) ;

	
	res = XCDE_Dict2Rank(TAG, tagdictrank, &genrank, &console);

	if (res == -1) MACRO_ERROR

	fprintf(stdout, "\n\n rango nel dizionario globale: %d\n",genrank) ;


	res=XCDE_Get_Tag_Depth(tagdictrank, &depth, &number_depths, &console) ;


	if (res == -1) MACRO_ERROR

	fprintf(stdout, "\n\n il numero totale di profondit: %d la lista:\n", number_depths) ;

	for (i=0; i < number_depths; i++) {

		fprintf(stdout, "%d, \n",depth[i]) ;
	}

	// finalizes the console. After this operation the memory
	// needed for the console is freed and temporary files
	// removed. The console variable should be also reused
	// for another document.
	XCDE_Finalize_Console(&console) ;

	return(1) ;
}




