/********************************************************************
	  Hash Library for Xml Compressed Document Engine
	     Paolo Ferragina - Andrea Mastroianni 
       Dipartimento di Informatica - Universita' di Pisa 
	             ferragin@di.unipi.it
 	             mastroia@di.unipi.it
********************************************************************/

#define TAG     0
#define ATTR    1
#define WORD    2
#define NOTHING 3

#define HASH_CONT_BIT_MAX_LEN 10

#define XCDE_ENTITY_HASHTABLE ((0x00) | _INSERT_ASSOCIATED_ELEMENT)

#define _INCREASE_PHYSICAL_COUNT 0x01

#define _INSERT_PHYSICAL_OCCURRENCE 0x02

#define _INSERT_LOGICAL_OCCURRENCE 0x04

#define _WRITE_PHYSICAL_OCCURRENCE 0x08

#define _WRITE_LOGICAL_OCCURRENCE 0x20

#define _WRITE_PHYSIC_ESCAPE 0x80

#define _INSERT_ASSOCIATED_ELEMENT 0x10 

#define _INSERT_PHYSIC_ESCAPE 0x40 

#define CHECK_PARAM(par, PROPRIETY) ((par & PROPRIETY) == PROPRIETY)

/* ------ Hash table to store the tokens ----- */

typedef struct hnode *Hashptr;

typedef struct hnode {
  char	*str;     // token
  int len_str;    // length of the token (to manage also NULL)
  char *ass_el ; // element associated to the hash_func 
  int count_occ ; // number_of_occurrencies of the string
  int codeword;   // codeword for this token
  int cw_len;     // codeword for this token
  int tag_number_open_flag ; // if the token is a tag contains the number of times it has been opened
  unsigned short int type ; // type of the token (word, token, attribute)
  int num_rank_list; // number of elements in the rank list 
  int PosByteEscape; // position of the escape
  int num_byte_list; // number of elements in the byte list
  int cmp_rank_list_pointer; // position of the current element in the list of position buffer
  int cmp_pos_list_pointer; // position of the current element in the list of rank buffer 
  int cmp_rank_list_byte_length; // length of the compressed rank list 
  int cmp_pos_list_byte_length; // length of the compressed position list
  int cmp_rank_list_last_value; // last value inserted into the compressed rank list
  int cmp_pos_list_last_value; // last value inserted into the compressed position list 
  char *cmp_rank_list ; // pointer to the rank list buffer
  char *cmp_pos_list ; // pointer to the position list buffer  
  Hashptr next;   // pointer to the next item in the list
} Hashnode;

typedef struct 
{ Hashptr *Table;
 unsigned int Table_size; } HashTable  ;

/* ----------- Create hash table for tokens ------------ */
HashTable hcreate(unsigned int Table_size);

/* --------- Compute the hash value for the given token ---------- */
__inline__ int hashfunc(char *s, 
			int len, 
			HashTable ht);

/* ------ Checks wether the hash table contains the given token ------ */
Hashptr hsearch(char *str, 
		int len, 
		unsigned char type, 
		HashTable ht);

/* --- Insert token in the hash table and returns 1 if new, 0 otherwise --- */
int hinsert(char *str, 
	    int str_len, 
	    unsigned char type, 
	    char *ass_el, 
	    HashTable ht, 
	    int occur, 
	    int escape, 
	    unsigned char par) ; 

/* --- destroy the hash table created --- */
void hdestroy(HashTable ht) ;

/* --- sets the open flag on --- */
int hopentag(Hashptr tag) ;

/* --- sets the open flag off --- */
int hclosetag(Hashptr tag) ; 





