/********************************************************************
	             Xml Compressed Document Engine
		             API Library
	         Paolo Ferragina - Andrea Mastroianni - Rosanna Rossi
             Dipartimento di Informatica - Universita' di Pisa
	                ferragin@di.unipi.it
	                mastroia@di.unipi.it
		       rossiro@cli.di.unipi.it

		            xcde_lib.c
                       contains functions:

 	                 XCDE_Init_Console
  		       XCDE_Finalize_Console
		        XCDE_SettedProximity
		       XCDE_ConsoleInitialized
		       XCDE_Get_Source_Filesize
		       XCDE_Get_Source_Filename
		       XCDE_Get_Header_Filesize
		       XCDE_Get_Header_Filename
		        XCDE_Get_Body_Filesize
			XCDE_Get_Body_Filesize
			    XCDE_Body_Eof
 			 XCDE_Get_IL_Filesize
			 XCDE_Get_IL_Filename
  		      XCDE_Get_Dict_Cmp_Filename
		      XCDE_Get_Dict_Cmp_Filesize
	            XCDE_Get_Dict_Dcmptmp_Filename
		    XCDE_Get_Dict_Dcmptmp_Filesize
		           XCDE_Get_Maxcwlen
 		          XCDE_Get_Num_Tokens
			  XCDE_Get_Num_El_Dict
			  XCDE_Get_Num_El_Lev
 		        XCDE_Get_Num_El_Dictlev
		      XCDE_Decompress_Dictionary
		      XCDE_Search_Word_Dictionary
		          XCDE_DictEl2Lev
			  XCDE_Dict2Rank
 		        XCDE_Load_Dictionary
			  XCDE_Rank2Dict
 			XCDE_Get_IL_ListPos
	  	       XCDE_Get_IL_ListPosProx
		         XCDE_Codeword2Rank
		         XCDE_Codeword2Token
			   XCDE_Rank2Token
		     XCDE_Get_Body_Forward_Codeword
		     XCDE_Get_Body_Backward_Codeword
		       XCDE_Get_Body_TextWindow
		         XCDE_SettedTagstruct
 		     XCDE_Get_Tagstruct_Filesize
		     XCDE_Get_Tagstruct_Filename
  	          XCDE_Get_Tagstruct_Tags_Enclosing
		  	XCDE_SettedDepthstruct
		    XCDE_Get_Depthstruct_Filesize
		    XCDE_Get_Depthstruct_Filename
			  XCDE_Get_Tag_Depth
********************************************************************/
#include"xcde_consts.h"
#include"xcde_lib.h"
/*******************************************************************/


// function that initializes the console data structure.
// Example : ret=XCDE_Init_Console("example.xml", &Console) ;
// where "example.xml" is the name of the source file and
// console is the memory address of the console data structure.
// ret value is 1 if there aren't errors, or -1 otherwise.

int XCDE_Init_Console(const char *name_source,
		      XCDE_Console_Type *XCDE_Console)
{ struct
    stat sbuf ;

  // variables used as counters
  register
    unsigned int k=0 ;

  register
    unsigned int j=0 ;

  register
    unsigned int i=0 ;

  int function_result=0 ;

  int Header_Mem_Position=0 ;

  unsigned
    long len=0ul ;

  uchar *Code_Pointer=NULL ;

  unsigned
    int Deltacode_Bit_Offset=0 ;

  unsigned
    int Deltacode_Sum=0 ;

  unsigned
    int Deltacode_El=0 ;

  int IL_Number_Table_Element=0 ;

  /* ------------------------------------------------------------ */
  /*      checks the presence of the memory mapping library       */
  /* ------------------------------------------------------------ */

  #ifndef _POSIX_MAPPED_FILES
  return(-1) ;
  #endif

  /* ----------------------------------------- */
  /*            checks parameters              */
  /* ----------------------------------------- */

  if ( (XCDE_Console == NULL) || (name_source==NULL) )

    return(-1) ;

  /* ---------------------------------------- */
  /*             creates filenames            */
  /* ---------------------------------------- */

  XCDE_Console->XCDE_XML_Source_File_Name=(char *) malloc((strlen(name_source)+1)*sizeof(char)) ;

  strcpy(XCDE_Console->XCDE_XML_Source_File_Name,
	 name_source) ;

  XCDE_Console->XCDE_Body_File_Name=(char *) malloc((strlen(name_source)+strlen(XCDE_BODY_SUFFIX)+1)*sizeof(char)) ;

  strcpy(XCDE_Console->XCDE_Body_File_Name,
	 name_source) ;

  strcat(XCDE_Console->XCDE_Body_File_Name,
	 XCDE_BODY_SUFFIX) ;

  XCDE_Console->XCDE_Header_File_Name=(char *) malloc((strlen(name_source)+strlen(XCDE_HEADER_SUFFIX)+1)*sizeof(char)) ;

  strcpy(XCDE_Console->XCDE_Header_File_Name,
	 name_source) ;

  strcat(XCDE_Console->XCDE_Header_File_Name,
	 XCDE_HEADER_SUFFIX) ;


  XCDE_Console->XCDE_IL_File_Name=(char *) malloc((strlen(name_source)+strlen(XCDE_IL_SUFFIX)+1)*sizeof(char)) ;

  strcpy(XCDE_Console->XCDE_IL_File_Name,
	 name_source) ;

  strcat(XCDE_Console->XCDE_IL_File_Name,
	 XCDE_IL_SUFFIX) ;

  for(k=0 ; k<XCDE_NUMBER_DICT; ++k) {

    XCDE_Console->XCDE_Dict_Cmp_File_Names[k]=(char *) malloc((strlen(name_source)+15)*sizeof(char)) ;

    sprintf(XCDE_Console->XCDE_Dict_Cmp_File_Names[k], "%s.dict.%d.cmp", name_source, k) ;

    XCDE_Console->XCDE_Dict_File_Decompressed[k]=0 ;

    XCDE_Console->XCDE_Dict_File_Loaded[k]=0 ;
  }

  /* ------------- */
  /*  opens files  */
  /* ------------- */

  if ((XCDE_Console->XCDE_Header_File_Pointer=fopen(XCDE_Console->XCDE_Header_File_Name, "rb")) == NULL) {

    fprintf(stderr, "\n\tFatal Error: problems opening file %s\n", XCDE_Console->XCDE_Header_File_Name) ;

    if (XCDE_Console->XCDE_XML_Source_File_Name != NULL)

      free(XCDE_Console->XCDE_XML_Source_File_Name) ;

    if (XCDE_Console->XCDE_Body_File_Name != NULL)

      free(XCDE_Console->XCDE_Body_File_Name) ;

    if (XCDE_Console->XCDE_Header_File_Name)

      free(XCDE_Console->XCDE_Header_File_Name) ;

    if (XCDE_Console->XCDE_IL_File_Name)

      free(XCDE_Console->XCDE_IL_File_Name) ;

    for(k=0 ; k < XCDE_NUMBER_DICT; ++k)

      free(XCDE_Console->XCDE_Dict_Cmp_File_Names[k]) ;

    return(-1) ;
  }

  if ((XCDE_Console->XCDE_IL_File_Pointer=fopen(XCDE_Console->XCDE_IL_File_Name, "rb"))==NULL) {

    fprintf(stderr, "\n\tFatal Error: problems opening file %s\n", XCDE_Console->XCDE_IL_File_Name) ;

    if (XCDE_Console->XCDE_XML_Source_File_Name != NULL)

      free(XCDE_Console->XCDE_XML_Source_File_Name) ;

    if (XCDE_Console->XCDE_Body_File_Name != NULL)

      free(XCDE_Console->XCDE_Body_File_Name) ;

    if (XCDE_Console->XCDE_Header_File_Name)

      free(XCDE_Console->XCDE_Header_File_Name) ;

    if (XCDE_Console->XCDE_IL_File_Name)

      free(XCDE_Console->XCDE_IL_File_Name) ;

    for(k=0 ; k < XCDE_NUMBER_DICT; ++k)

      free(XCDE_Console->XCDE_Dict_Cmp_File_Names[k]) ;

    fclose(XCDE_Console->XCDE_Header_File_Pointer) ;

    return(-1) ;
  }

  if ((XCDE_Console->XCDE_Body_File_Pointer=fopen(XCDE_Console->XCDE_Body_File_Name, "rb")) == NULL) {

    fprintf(stderr, "\n\tFatal Error: problems opening file %s\n", XCDE_Console->XCDE_Body_File_Name) ;

    if (XCDE_Console->XCDE_XML_Source_File_Name != NULL)

      free(XCDE_Console->XCDE_XML_Source_File_Name) ;

    if (XCDE_Console->XCDE_Body_File_Name != NULL)

      free(XCDE_Console->XCDE_Body_File_Name) ;

    if (XCDE_Console->XCDE_Header_File_Name)

      free(XCDE_Console->XCDE_Header_File_Name) ;

    if (XCDE_Console->XCDE_IL_File_Name)

      free(XCDE_Console->XCDE_IL_File_Name) ;

    for(k=0 ; k < XCDE_NUMBER_DICT; ++k)

      free(XCDE_Console->XCDE_Dict_Cmp_File_Names[k]) ;

    fclose(XCDE_Console->XCDE_Header_File_Pointer) ;

    fclose(XCDE_Console->XCDE_IL_File_Pointer) ;

    return(-1) ;
  }

  /* -------------------------- */
  /* reads the body file length */
  /* -------------------------- */

  function_result=stat(XCDE_Console->XCDE_Body_File_Name,
		       &sbuf) ;

  if (function_result==-1) {

    fprintf(stderr, "\n\tFatal Error: problems in stat operation on file %s\n", XCDE_Console->XCDE_Body_File_Name) ;

    if (XCDE_Console->XCDE_XML_Source_File_Name != NULL)

      free(XCDE_Console->XCDE_XML_Source_File_Name) ;

    if (XCDE_Console->XCDE_Body_File_Name != NULL)

      free(XCDE_Console->XCDE_Body_File_Name) ;

    if (XCDE_Console->XCDE_Header_File_Name)

      free(XCDE_Console->XCDE_Header_File_Name) ;

    if (XCDE_Console->XCDE_IL_File_Name)

      free(XCDE_Console->XCDE_IL_File_Name) ;

    for(k=0 ; k < XCDE_NUMBER_DICT; ++k)

      free(XCDE_Console->XCDE_Dict_Cmp_File_Names[k]) ;

    fclose(XCDE_Console->XCDE_Header_File_Pointer) ;

    fclose(XCDE_Console->XCDE_IL_File_Pointer) ;

    return(-1) ;
    }

  XCDE_Console->XCDE_Body_Filesize=(int) sbuf.st_size ;

  /* --------------------------- */
  /* body file mapping operation */
  /* --------------------------- */

  XCDE_Console->XCDE_Body_Mapped_Pointer= (char *) mmap(NULL,
							XCDE_Console->XCDE_Body_Filesize,
							PROT_READ,
							MAP_SHARED,
							fileno(XCDE_Console->XCDE_Body_File_Pointer),
							0) ;

  if (XCDE_Console->XCDE_Body_Mapped_Pointer==MAP_FAILED) {

    if (XCDE_Console->XCDE_XML_Source_File_Name != NULL)

      free(XCDE_Console->XCDE_XML_Source_File_Name) ;

    if (XCDE_Console->XCDE_Body_File_Name != NULL)

      free(XCDE_Console->XCDE_Body_File_Name) ;

    if (XCDE_Console->XCDE_Header_File_Name)

      free(XCDE_Console->XCDE_Header_File_Name) ;

    if (XCDE_Console->XCDE_IL_File_Name)

      free(XCDE_Console->XCDE_IL_File_Name) ;

    for(k=0 ; k < XCDE_NUMBER_DICT; ++k)

      free(XCDE_Console->XCDE_Dict_Cmp_File_Names[k]) ;

    fprintf(stderr, "\n\tFatal Error: problems in map operation on file %s\n", XCDE_Console->XCDE_Body_File_Name) ;

    fclose(XCDE_Console->XCDE_Header_File_Pointer) ;

    fclose(XCDE_Console->XCDE_IL_File_Pointer) ;

    fclose(XCDE_Console->XCDE_Body_File_Pointer) ;

    return(-1) ;
    }

  /* ------------------------ */
  /* reads the IL file length */
  /* ------------------------ */

  function_result=stat(XCDE_Console->XCDE_IL_File_Name,
		       &sbuf) ;

  if (function_result==-1) {

    fprintf(stderr, "\n\tFatal Error: problems in stat operation on file %s\n", XCDE_Console->XCDE_IL_File_Name) ;

    if (XCDE_Console->XCDE_XML_Source_File_Name != NULL)

      free(XCDE_Console->XCDE_XML_Source_File_Name) ;

    if (XCDE_Console->XCDE_Body_File_Name != NULL)

      free(XCDE_Console->XCDE_Body_File_Name) ;

    if (XCDE_Console->XCDE_Header_File_Name)

      free(XCDE_Console->XCDE_Header_File_Name) ;

    if (XCDE_Console->XCDE_IL_File_Name)

      free(XCDE_Console->XCDE_IL_File_Name) ;

    for(k=0 ; k < XCDE_NUMBER_DICT; ++k)

      free(XCDE_Console->XCDE_Dict_Cmp_File_Names[k]) ;

    fclose(XCDE_Console->XCDE_Header_File_Pointer) ;

    fclose(XCDE_Console->XCDE_IL_File_Pointer) ;

    fclose(XCDE_Console->XCDE_Body_File_Pointer) ;

    munmap(XCDE_Console->XCDE_Body_Mapped_Pointer,
	   XCDE_Console->XCDE_Body_Filesize) ;

    return(-1) ;
  }

  XCDE_Console->XCDE_IL_Filesize=(int) sbuf.st_size ;

  /* ------------------------- */
  /* IL file mapping operation */
  /* ------------------------- */

  XCDE_Console->XCDE_IL_Mapped_Pointer= (char *) mmap(NULL,
						      XCDE_Console->XCDE_IL_Filesize,
						      PROT_READ,
						      MAP_SHARED,
						      fileno(XCDE_Console->XCDE_IL_File_Pointer),
						      0) ;

  if (XCDE_Console->XCDE_IL_Mapped_Pointer == MAP_FAILED) {

    fprintf(stderr, "\n\tFatal Error: problems in map operation on file %s\n", XCDE_Console->XCDE_IL_File_Name) ;

    if (XCDE_Console->XCDE_XML_Source_File_Name != NULL)

      free(XCDE_Console->XCDE_XML_Source_File_Name) ;

    if (XCDE_Console->XCDE_Body_File_Name != NULL)

      free(XCDE_Console->XCDE_Body_File_Name) ;

    if (XCDE_Console->XCDE_Header_File_Name)

      free(XCDE_Console->XCDE_Header_File_Name) ;

    if (XCDE_Console->XCDE_IL_File_Name)

      free(XCDE_Console->XCDE_IL_File_Name) ;

    for(k=0 ; k < XCDE_NUMBER_DICT; ++k)

      free(XCDE_Console->XCDE_Dict_Cmp_File_Names[k]) ;

    fclose(XCDE_Console->XCDE_Header_File_Pointer) ;

    fclose(XCDE_Console->XCDE_IL_File_Pointer) ;

    fclose(XCDE_Console->XCDE_Body_File_Pointer) ;

    munmap(XCDE_Console->XCDE_Body_Mapped_Pointer,
	   XCDE_Console->XCDE_Body_Filesize) ;

    return(-1) ;
  }

  /* ------------------------------- */
  /* reads the header file dimension */
  /* ------------------------------- */

  function_result=stat(XCDE_Console->XCDE_Header_File_Name,
		       &sbuf) ;

  if (function_result==-1) {

    fprintf(stderr, "\n\tFatal Error: problems in map operation on file %s\n", XCDE_Console->XCDE_IL_File_Name) ;

    if (XCDE_Console->XCDE_XML_Source_File_Name != NULL)

      free(XCDE_Console->XCDE_XML_Source_File_Name) ;

    if (XCDE_Console->XCDE_Body_File_Name != NULL)

      free(XCDE_Console->XCDE_Body_File_Name) ;

    if (XCDE_Console->XCDE_Header_File_Name)

      free(XCDE_Console->XCDE_Header_File_Name) ;

    if (XCDE_Console->XCDE_IL_File_Name)

      free(XCDE_Console->XCDE_IL_File_Name) ;

    for(k=0 ; k < XCDE_NUMBER_DICT; ++k)

      free(XCDE_Console->XCDE_Dict_Cmp_File_Names[k]) ;

    fclose(XCDE_Console->XCDE_Header_File_Pointer) ;

    fclose(XCDE_Console->XCDE_IL_File_Pointer) ;

    fclose(XCDE_Console->XCDE_Body_File_Pointer) ;

    munmap(XCDE_Console->XCDE_Body_Mapped_Pointer,
	   XCDE_Console->XCDE_Body_Filesize) ;

    munmap(XCDE_Console->XCDE_IL_Mapped_Pointer,
	   XCDE_Console->XCDE_IL_Filesize) ;

    return(-1) ;
  }

  XCDE_Console->XCDE_Header_Filesize=(int) sbuf.st_size ;

  /* ----------------------------- */
  /* header file mapping operation */
  /* ----------------------------- */

  XCDE_Console->XCDE_Header_Mapped_Pointer= (char *) mmap(NULL,
							  XCDE_Console->XCDE_Header_Filesize,
							  PROT_READ,
							  MAP_SHARED,
							  fileno(XCDE_Console->XCDE_Header_File_Pointer),
							  0) ;

  if (XCDE_Console->XCDE_Header_Mapped_Pointer==MAP_FAILED) {

    fprintf(stderr, "\nFatal Error: problems in map operation on file %s\n", XCDE_Console->XCDE_IL_File_Name) ;

    if (XCDE_Console->XCDE_XML_Source_File_Name != NULL)

      free(XCDE_Console->XCDE_XML_Source_File_Name) ;

    if (XCDE_Console->XCDE_Body_File_Name != NULL)

      free(XCDE_Console->XCDE_Body_File_Name) ;

    if (XCDE_Console->XCDE_Header_File_Name)
      
      free(XCDE_Console->XCDE_Header_File_Name) ; 
    
    if (XCDE_Console->XCDE_IL_File_Name) 
      
      free(XCDE_Console->XCDE_IL_File_Name) ;

    for(k=0 ; k < XCDE_NUMBER_DICT; ++k)
      
      free(XCDE_Console->XCDE_Dict_Cmp_File_Names[k]) ;

    fclose(XCDE_Console->XCDE_Header_File_Pointer) ;
  
    fclose(XCDE_Console->XCDE_IL_File_Pointer) ;
    
    fclose(XCDE_Console->XCDE_Body_File_Pointer) ; 

    munmap(XCDE_Console->XCDE_Body_Mapped_Pointer,
	   XCDE_Console->XCDE_Body_Filesize) ; 
    
    munmap(XCDE_Console->XCDE_IL_Mapped_Pointer,
	   XCDE_Console->XCDE_IL_Filesize) ; 
    
    return(-1) ;
  }

  /* --------------------- */
  /* reads the file header */
  /* --------------------- */

  for (k=1; k<=4; k++) { 

    for (XCDE_Console->Huffman_Firstcw[k]=0,j=0; j<4; j++) {

      XCDE_Console->Huffman_Firstcw[k] = ((XCDE_Console->Huffman_Firstcw[k] << 7) | fgetc(XCDE_Console->XCDE_Header_File_Pointer));
      
      ++Header_Mem_Position ;
    }
  }

  for (k=1; k<=4; k++)

    for (XCDE_Console->Huffman_Offsetcw[k]=0,j=0; j<4; j++) {

      XCDE_Console->Huffman_Offsetcw[k] = ((XCDE_Console->Huffman_Offsetcw[k] << 8) | fgetc(XCDE_Console->XCDE_Header_File_Pointer)) ;

      ++Header_Mem_Position ;
    }

  for (XCDE_Console->Huffman_Codeword_Max_Len=0, j=0; j<4; ++j) {

    XCDE_Console->Huffman_Codeword_Max_Len = ((XCDE_Console->Huffman_Codeword_Max_Len << 8) | fgetc(XCDE_Console->XCDE_Header_File_Pointer)) ;

    ++Header_Mem_Position ;
  }

  for(XCDE_Console->Huffman_Num_Tokens=0, k=0; (k<XCDE_Console->Huffman_Codeword_Max_Len); ++k) {

    for(i=0, XCDE_Console->Huffman_Dict_NumEl[k][i]=0; i<XCDE_NUMBER_DICT; ++i) {

      for(j=0; j<4; j++) {

	XCDE_Console->Huffman_Dict_NumEl[k][i]=((XCDE_Console->Huffman_Dict_NumEl[k][i] << 8) | fgetc(XCDE_Console->XCDE_Header_File_Pointer)) ;

	++Header_Mem_Position ;
      }

      XCDE_Console->Huffman_Num_Tokens+=XCDE_Console->Huffman_Dict_NumEl[k][i] ;
    }
  }

  continuation_bit_d((unsigned long *) &XCDE_Console->XCDE_XML_Source_Filesize,
		     (char *) (XCDE_Console->XCDE_Header_Mapped_Pointer+Header_Mem_Position),
		     (unsigned int) 0,
		     (unsigned long *) &len) ;

  Header_Mem_Position+=((int) (len/8)) ;

  continuation_bit_d((unsigned long *) &XCDE_Console->XCDE_Body_Size_Without_Zlib,
		     (char *) (XCDE_Console->XCDE_Header_Mapped_Pointer+Header_Mem_Position),
		     (unsigned int) 0,
		     (unsigned long *) &len) ;
    
  Header_Mem_Position+=((int) (len/8)) ;

  XCDE_Console->XCDE_Proximity=XCDE_Console->XCDE_Header_Mapped_Pointer[Header_Mem_Position] ; 
  
  Header_Mem_Position++ ; 
  
  XCDE_Console->XCDE_Tagstruct=XCDE_Console->XCDE_Header_Mapped_Pointer[Header_Mem_Position] ; 

  Header_Mem_Position++ ; 

  XCDE_Console->XCDE_Depthstruct=XCDE_Console->XCDE_Header_Mapped_Pointer[Header_Mem_Position] ; 

  Header_Mem_Position++ ; 

  continuation_bit_d((unsigned long *) &XCDE_Console->XCDE_Body_Num_Block, 
		     (char *) (XCDE_Console->XCDE_Header_Mapped_Pointer+Header_Mem_Position), 
		     (unsigned int) 0,
		     (unsigned long *) &len) ;
  
  Header_Mem_Position+=((int) (len/8)) ; 		     

  XCDE_Console->XCDE_Body_Table_Pointers=(int *) malloc(XCDE_Console->XCDE_Body_Num_Block*sizeof(int)) ; 

  XCDE_Console->XCDE_Body_Table_Pointers[0]=0 ; 

  Deltacode_Sum=0u ; 

  Deltacode_Bit_Offset=0u ; 

  Code_Pointer=(uchar *) (XCDE_Console->XCDE_Header_Mapped_Pointer+Header_Mem_Position) ; 

  for (k=1 ; k<XCDE_Console->XCDE_Body_Num_Block ; ++k) {

    continuation_bit_d((unsigned long *) &Deltacode_El, 
		       (char *) XCDE_Console->XCDE_Header_Mapped_Pointer+Header_Mem_Position,
		       (unsigned int) 0,
		       (unsigned long *) &len) ;

    Header_Mem_Position+=(len/8) ;

    Deltacode_Sum+=Deltacode_El ; 
    
    XCDE_Console->XCDE_Body_Table_Pointers[k]=Deltacode_Sum ; 
  }

  IL_Number_Table_Element=(XCDE_Console->Huffman_Num_Tokens/(XCDE_IL_POINTER_RATIO)) ; 

  if ((XCDE_Console->Huffman_Num_Tokens%(XCDE_IL_POINTER_RATIO))> 0)

    ++IL_Number_Table_Element ;
  
  XCDE_Console->XCDE_IL_Table_Pointers=(int *) malloc(IL_Number_Table_Element*sizeof(int)) ; 
  
  XCDE_Console->XCDE_IL_Table_Pointers_Size=IL_Number_Table_Element ; 

  Deltacode_Sum=0u ;

  XCDE_Console->XCDE_IL_Table_Pointers[0]=0 ;

  for (k=1 ; k<IL_Number_Table_Element ; ++k) {

    continuation_bit_d((unsigned long *) &Deltacode_El,
		       (char *) XCDE_Console->XCDE_Header_Mapped_Pointer+Header_Mem_Position,
		       (unsigned int) 0,
		       (unsigned long *) &len) ;

    Header_Mem_Position+=(len/8) ;

    Deltacode_Sum+=(Deltacode_El-1) ;

    XCDE_Console->XCDE_IL_Table_Pointers[k]=Deltacode_Sum ;
  }

  /* ------------------------------------- */
  /* if necessary opens the tagstruct file */
  /* ------------------------------------- */

	if (XCDE_Console->XCDE_Tagstruct) {

		XCDE_Console->XCDE_Tagstruct_File_Name= (char *) malloc((strlen(XCDE_Console->XCDE_XML_Source_File_Name)+strlen(XCDE_TAGSTRUCT_SUFFIX)+1)*sizeof(char)) ;

		if (XCDE_Console->XCDE_Tagstruct_File_Name == NULL) {

			fprintf(stderr, "\n\tFatal Error: problems allocating memory\n") ;

			if (XCDE_Console->XCDE_XML_Source_File_Name != NULL)

				free(XCDE_Console->XCDE_XML_Source_File_Name) ;

			if (XCDE_Console->XCDE_Body_File_Name != NULL)

				free(XCDE_Console->XCDE_Body_File_Name) ;

			if (XCDE_Console->XCDE_Header_File_Name)

				free(XCDE_Console->XCDE_Header_File_Name) ;

			if (XCDE_Console->XCDE_IL_File_Name)

				free(XCDE_Console->XCDE_IL_File_Name) ;

			for(k=0 ; k < XCDE_NUMBER_DICT; ++k)

				free(XCDE_Console->XCDE_Dict_Cmp_File_Names[k]) ;

			munmap(XCDE_Console->XCDE_Body_Mapped_Pointer,
				XCDE_Console->XCDE_Body_Filesize) ;

			munmap(XCDE_Console->XCDE_IL_Mapped_Pointer,
				XCDE_Console->XCDE_IL_Filesize) ;

			munmap(XCDE_Console->XCDE_Header_Mapped_Pointer,
				XCDE_Console->XCDE_Header_Filesize) ;

			free(XCDE_Console->XCDE_IL_Table_Pointers) ;

			free(XCDE_Console->XCDE_Body_Table_Pointers) ;

			fclose(XCDE_Console->XCDE_Header_File_Pointer) ;

			fclose(XCDE_Console->XCDE_IL_File_Pointer) ;

			fclose(XCDE_Console->XCDE_Body_File_Pointer) ;

			return(-1) ;
		}

		strcpy(XCDE_Console->XCDE_Tagstruct_File_Name,
			XCDE_Console->XCDE_XML_Source_File_Name) ;

		strcat(XCDE_Console->XCDE_Tagstruct_File_Name,
			XCDE_TAGSTRUCT_SUFFIX) ;

		if ((XCDE_Console->XCDE_Tagstruct_File_Pointer=fopen(XCDE_Console->XCDE_Tagstruct_File_Name, "rb")) == NULL) {

			fprintf(stderr, "\n\tFatal Error: problems opening file %s\n", XCDE_Console->XCDE_Tagstruct_File_Name) ;

			if (XCDE_Console->XCDE_XML_Source_File_Name != NULL)

				free(XCDE_Console->XCDE_XML_Source_File_Name) ;

			if (XCDE_Console->XCDE_Body_File_Name != NULL)

				free(XCDE_Console->XCDE_Body_File_Name) ;

			if (XCDE_Console->XCDE_Header_File_Name)

				free(XCDE_Console->XCDE_Header_File_Name) ;

			if (XCDE_Console->XCDE_IL_File_Name)

				free(XCDE_Console->XCDE_IL_File_Name) ;

			for(k=0 ; k < XCDE_NUMBER_DICT; ++k)

				free(XCDE_Console->XCDE_Dict_Cmp_File_Names[k]) ;

			munmap(XCDE_Console->XCDE_Body_Mapped_Pointer,
				XCDE_Console->XCDE_Body_Filesize) ;

			munmap(XCDE_Console->XCDE_IL_Mapped_Pointer,
				XCDE_Console->XCDE_IL_Filesize) ;

			munmap(XCDE_Console->XCDE_Header_Mapped_Pointer,
				XCDE_Console->XCDE_Header_Filesize) ;

			free(XCDE_Console->XCDE_IL_Table_Pointers) ;

			free(XCDE_Console->XCDE_Body_Table_Pointers) ;

			fclose(XCDE_Console->XCDE_Header_File_Pointer) ;

			fclose(XCDE_Console->XCDE_IL_File_Pointer) ;

			fclose(XCDE_Console->XCDE_Body_File_Pointer) ;

			return(-1) ;
		}

		function_result=stat(XCDE_Console->XCDE_Tagstruct_File_Name,
					&sbuf) ;

		if (function_result==-1) {

			fprintf(stderr, "\ntFatal Error: problems in map operation on file %s\n", XCDE_Console->XCDE_Tagstruct_File_Name) ;

			if (XCDE_Console->XCDE_XML_Source_File_Name != NULL)

				free(XCDE_Console->XCDE_XML_Source_File_Name) ;

			if (XCDE_Console->XCDE_Body_File_Name != NULL)

				free(XCDE_Console->XCDE_Body_File_Name) ;

			if (XCDE_Console->XCDE_Header_File_Name)

				free(XCDE_Console->XCDE_Header_File_Name) ;

			if (XCDE_Console->XCDE_IL_File_Name)

				free(XCDE_Console->XCDE_IL_File_Name) ;

			for(k=0 ; k < XCDE_NUMBER_DICT; ++k)

				free(XCDE_Console->XCDE_Dict_Cmp_File_Names[k]) ;

			munmap(XCDE_Console->XCDE_Body_Mapped_Pointer,
				XCDE_Console->XCDE_Body_Filesize) ;

			munmap(XCDE_Console->XCDE_IL_Mapped_Pointer,
				XCDE_Console->XCDE_IL_Filesize) ;

			munmap(XCDE_Console->XCDE_Header_Mapped_Pointer,
				XCDE_Console->XCDE_Header_Filesize) ;

			free(XCDE_Console->XCDE_IL_Table_Pointers) ;

			free(XCDE_Console->XCDE_Body_Table_Pointers) ;

			fclose(XCDE_Console->XCDE_Tagstruct_File_Pointer) ;

			fclose(XCDE_Console->XCDE_Header_File_Pointer) ;

			fclose(XCDE_Console->XCDE_IL_File_Pointer) ;

			fclose(XCDE_Console->XCDE_Body_File_Pointer) ;

			return(-1) ;
		}

		XCDE_Console->XCDE_Tagstruct_Filesize=(int) sbuf.st_size ;

		/* -------------------------------- */
		/* tagstruct file mapping operation */
		/* -------------------------------- */

		XCDE_Console->XCDE_Tagstruct_Mapped_Pointer= (char *) mmap(NULL,
									XCDE_Console->XCDE_Tagstruct_Filesize,
									PROT_READ,
									MAP_SHARED,
									fileno(XCDE_Console->XCDE_Tagstruct_File_Pointer),
									0) ;

		if (XCDE_Console->XCDE_Tagstruct_Mapped_Pointer==MAP_FAILED) {

			fprintf(stderr, "\n\tFatal Error: problems in map operation on file %s\n", XCDE_Console->XCDE_Tagstruct_File_Name) ;

			if (XCDE_Console->XCDE_XML_Source_File_Name != NULL)

				free(XCDE_Console->XCDE_XML_Source_File_Name) ;

			if (XCDE_Console->XCDE_Body_File_Name != NULL)

				free(XCDE_Console->XCDE_Body_File_Name) ;

			if (XCDE_Console->XCDE_Header_File_Name)

				free(XCDE_Console->XCDE_Header_File_Name) ;

			if (XCDE_Console->XCDE_IL_File_Name)

				free(XCDE_Console->XCDE_IL_File_Name) ;

			for(k=0 ; k < XCDE_NUMBER_DICT; ++k)

				free(XCDE_Console->XCDE_Dict_Cmp_File_Names[k]) ;

			munmap(XCDE_Console->XCDE_Body_Mapped_Pointer,
				XCDE_Console->XCDE_Body_Filesize) ;

			munmap(XCDE_Console->XCDE_IL_Mapped_Pointer,
				XCDE_Console->XCDE_IL_Filesize) ;

			munmap(XCDE_Console->XCDE_Header_Mapped_Pointer,
				XCDE_Console->XCDE_Header_Filesize) ;

			free(XCDE_Console->XCDE_IL_Table_Pointers) ;

			free(XCDE_Console->XCDE_Body_Table_Pointers) ;

			fclose(XCDE_Console->XCDE_Tagstruct_File_Pointer) ;

			fclose(XCDE_Console->XCDE_Header_File_Pointer) ;

			fclose(XCDE_Console->XCDE_IL_File_Pointer) ;

			fclose(XCDE_Console->XCDE_Body_File_Pointer) ;

			return(-1) ;
		}

		XCDE_Console->XCDE_Tagstruct_Windows_Information = NULL ;

		XCDE_Console->XCDE_Tagstruct_Number_Windows=0 ;

		XCDE_Console->XCDE_Tagstruct_Cache_Window=NULL ;

		XCDE_Console->XCDE_Tagstruct_Cache_Which_Block=-1 ;

		XCDE_Console->XCDE_Tagstruct_Cache_Cmp_Block_Size=0 ;

		XCDE_Console->XCDE_Tagstruct_Cache_Dcmp_Block_Size=0 ;
	}


	/* ------------------------------------- */
	/* if necessary open depthstruct file */
	/* ------------------------------------- */

	if (XCDE_Console->XCDE_Depthstruct) {

		XCDE_Console->XCDE_Depthstruct_File_Name= (char *) malloc((strlen(XCDE_Console->XCDE_XML_Source_File_Name)+strlen(XCDE_DEPTHSTRUCT_SUFFIX)+1)*sizeof(char)) ;

		if (XCDE_Console->XCDE_Depthstruct_File_Name == NULL) {

			fprintf(stderr, "\n\tFatal Error: problems allocating memory\n") ;

			if (XCDE_Console->XCDE_XML_Source_File_Name != NULL)

				free(XCDE_Console->XCDE_XML_Source_File_Name) ;

			if (XCDE_Console->XCDE_Body_File_Name != NULL)

				free(XCDE_Console->XCDE_Body_File_Name) ;

			if (XCDE_Console->XCDE_Header_File_Name)

				free(XCDE_Console->XCDE_Header_File_Name) ;

			if (XCDE_Console->XCDE_IL_File_Name)

				free(XCDE_Console->XCDE_IL_File_Name) ;

			for(k=0 ; k < XCDE_NUMBER_DICT; ++k)

				free(XCDE_Console->XCDE_Dict_Cmp_File_Names[k]) ;

			if (XCDE_Console->XCDE_Tagstruct_File_Name)

				free(XCDE_Console->XCDE_Tagstruct_File_Name) ;

			munmap(XCDE_Console->XCDE_Body_Mapped_Pointer,
				XCDE_Console->XCDE_Body_Filesize) ;

			munmap(XCDE_Console->XCDE_IL_Mapped_Pointer,
				XCDE_Console->XCDE_IL_Filesize) ;

			munmap(XCDE_Console->XCDE_Header_Mapped_Pointer,
				XCDE_Console->XCDE_Header_Filesize) ;

			free(XCDE_Console->XCDE_IL_Table_Pointers) ;

			free(XCDE_Console->XCDE_Body_Table_Pointers) ;

			fclose(XCDE_Console->XCDE_Header_File_Pointer) ;

			fclose(XCDE_Console->XCDE_IL_File_Pointer) ;

			fclose(XCDE_Console->XCDE_Body_File_Pointer) ;

			return(-1) ;
		}

		strcpy(XCDE_Console->XCDE_Depthstruct_File_Name,
			XCDE_Console->XCDE_XML_Source_File_Name) ;

		strcat(XCDE_Console->XCDE_Depthstruct_File_Name,
			XCDE_DEPTHSTRUCT_SUFFIX) ;

		if ((XCDE_Console->XCDE_Depthstruct_File_Pointer=fopen(XCDE_Console->XCDE_Depthstruct_File_Name, "rb")) == NULL) {

			fprintf(stderr, "\n\tFatal Error: problems opening file %s\n", XCDE_Console->XCDE_Depthstruct_File_Name) ;

			if (XCDE_Console->XCDE_XML_Source_File_Name != NULL)

				free(XCDE_Console->XCDE_XML_Source_File_Name) ;

			if (XCDE_Console->XCDE_Body_File_Name != NULL)

				free(XCDE_Console->XCDE_Body_File_Name) ;

			if (XCDE_Console->XCDE_Header_File_Name)

				free(XCDE_Console->XCDE_Header_File_Name) ;

			if (XCDE_Console->XCDE_IL_File_Name)

				free(XCDE_Console->XCDE_IL_File_Name) ;

			for(k=0 ; k < XCDE_NUMBER_DICT; ++k)

				free(XCDE_Console->XCDE_Dict_Cmp_File_Names[k]) ;

			if (XCDE_Console->XCDE_Tagstruct_File_Name)

				free(XCDE_Console->XCDE_Tagstruct_File_Name) ;

			munmap(XCDE_Console->XCDE_Body_Mapped_Pointer,
				XCDE_Console->XCDE_Body_Filesize) ;

			munmap(XCDE_Console->XCDE_IL_Mapped_Pointer,
				XCDE_Console->XCDE_IL_Filesize) ;

			munmap(XCDE_Console->XCDE_Header_Mapped_Pointer,
				XCDE_Console->XCDE_Header_Filesize) ;

			free(XCDE_Console->XCDE_IL_Table_Pointers) ;

			free(XCDE_Console->XCDE_Body_Table_Pointers) ;

			fclose(XCDE_Console->XCDE_Header_File_Pointer) ;

			fclose(XCDE_Console->XCDE_IL_File_Pointer) ;

			fclose(XCDE_Console->XCDE_Body_File_Pointer) ;

			return(-1) ;
		}

		if (fseek(XCDE_Console->XCDE_Depthstruct_File_Pointer, 0L, SEEK_END)) {

			fprintf(stderr,"seek error on file %s\n",XCDE_Console->XCDE_Depthstruct_File_Name);

			return(-1);
		}

		XCDE_Console->XCDE_Depthstruct_Filesize = ftell(XCDE_Console->XCDE_Depthstruct_File_Pointer);

		if (fseek(XCDE_Console->XCDE_Depthstruct_File_Pointer, 0L, SEEK_SET)) {

			fprintf(stderr,"ftell error on file %s\n",XCDE_Console->XCDE_Depthstruct_File_Name);

			return(-1);
		}

		/* -------------------------------- */
		/* depthstruct file mapping operation */
		/* -------------------------------- */

		XCDE_Console->XCDE_Depthstruct_Mapped_Pointer = (char *) mmap(NULL,
									XCDE_Console->XCDE_Depthstruct_Filesize,
									PROT_READ,
									MAP_SHARED,
									fileno(XCDE_Console->XCDE_Depthstruct_File_Pointer),
									0) ;

		if (XCDE_Console->XCDE_Depthstruct_Mapped_Pointer==MAP_FAILED) {

			fprintf(stderr, "\n\tFatal Error: problems in map operation on file %s\n", XCDE_Console->XCDE_Depthstruct_File_Name) ;

			if (XCDE_Console->XCDE_XML_Source_File_Name != NULL)

				free(XCDE_Console->XCDE_XML_Source_File_Name) ;

			if (XCDE_Console->XCDE_Body_File_Name != NULL)

				free(XCDE_Console->XCDE_Body_File_Name) ;

			if (XCDE_Console->XCDE_Header_File_Name)

				free(XCDE_Console->XCDE_Header_File_Name) ;

			if (XCDE_Console->XCDE_IL_File_Name)

				free(XCDE_Console->XCDE_IL_File_Name) ;

			for(k=0 ; k < XCDE_NUMBER_DICT; ++k)

				free(XCDE_Console->XCDE_Dict_Cmp_File_Names[k]) ;

			if (XCDE_Console->XCDE_Tagstruct_File_Name)

				free(XCDE_Console->XCDE_Tagstruct_File_Name) ;

			munmap(XCDE_Console->XCDE_Body_Mapped_Pointer,
				XCDE_Console->XCDE_Body_Filesize) ;

			munmap(XCDE_Console->XCDE_IL_Mapped_Pointer,
				XCDE_Console->XCDE_IL_Filesize) ;

			munmap(XCDE_Console->XCDE_Header_Mapped_Pointer,
				XCDE_Console->XCDE_Header_Filesize) ;

			free(XCDE_Console->XCDE_IL_Table_Pointers) ;

			free(XCDE_Console->XCDE_Body_Table_Pointers) ;

			fclose(XCDE_Console->XCDE_Tagstruct_File_Pointer) ;

			fclose(XCDE_Console->XCDE_Header_File_Pointer) ;

			fclose(XCDE_Console->XCDE_IL_File_Pointer) ;

			fclose(XCDE_Console->XCDE_Body_File_Pointer) ;

			return(-1) ;
		}
	
	}

	/* --------------------------- */
	/* initializes other variables */
	/* --------------------------- */

	XCDE_Console->XCDE_Body_Cache_Block=NULL ;

	XCDE_Console->XCDE_Body_Cache_Which_Block=-1 ;

	XCDE_Console->XCDE_Body_Cache_Block_File_Position=0 ;

	XCDE_Console->XCDE_Body_Cache_Dcmp_Block_Size=0 ;

	XCDE_Console->XCDE_Body_Cache_Cmp_Block_Size=0 ;

	XCDE_Console->XCDE_Console_Initialized=1 ;

	XCDE_Console->Huffman_Position_Token = NULL ;

	return(1) ;
}

/*******************************************************************/

// function that finalizes the console data structure and
// frees the memory allocated by the XCDE_Init_Console. 
// Example : ret=XCDE_Finalize_Console(&Console) ;
// where &Console is the memory address of the console data structure.
// ret value is 1 if there aren't errors, or -1 otherwise.

int XCDE_Finalize_Console(XCDE_Console_Type *XCDE_Console)
{ int function_result=0 ;

 register
   int k=0 ;

 /* ----------------------------------------- */
 /*            checks parameters              */
 /* ----------------------------------------- */

 if (XCDE_Console == NULL)

   return(-1) ;

 if (XCDE_Console->XCDE_Console_Initialized == 0u)

   return(-1) ;

 /* ----------------------------- */
 /* body file unmapping operation */
 /* ----------------------------- */

 function_result=munmap(XCDE_Console->XCDE_Body_Mapped_Pointer,
			XCDE_Console->XCDE_Body_Filesize) ;

 if (function_result == -1)

   return(-1) ;

 /* -------------------- */
 /* closes the body file */
 /* -------------------- */

 function_result=fclose(XCDE_Console->XCDE_Body_File_Pointer) ;

 if (function_result == -1)

   return(-1) ;

 /* -------------------------------------------- */
 /* frees the memory for the body table pointers */
 /* -------------------------------------------- */

 if (XCDE_Console->XCDE_Body_Table_Pointers != NULL)

   free(XCDE_Console->XCDE_Body_Table_Pointers) ;

 /* ------------------------------ */
 /* frees the memory for the cache */
 /* ------------------------------ */

 if (XCDE_Console->XCDE_Body_Cache_Block != NULL)

   free(XCDE_Console->XCDE_Body_Cache_Block) ;

 /* --------------------------------------- */
 /* frees the memory for the body file name */
 /* --------------------------------------- */

 if (XCDE_Console->XCDE_Body_File_Name != NULL)

   free(XCDE_Console->XCDE_Body_File_Name) ;

 /* --------------------------- */
 /* IL file unmapping operation */
 /* --------------------------- */

 function_result=munmap(XCDE_Console->XCDE_IL_Mapped_Pointer,
			XCDE_Console->XCDE_IL_Filesize) ;

 if (function_result == -1)

   return(-1) ;

 /* ------------------ */
 /* closes the IL file */
 /* ------------------ */

 function_result=fclose(XCDE_Console->XCDE_IL_File_Pointer) ;

 if (function_result == -1)

   return(-1) ;

 /* ------------------------------------------ */
 /* frees the memory for the IL table pointers */
 /* ------------------------------------------ */

 if (XCDE_Console->XCDE_IL_Table_Pointers != NULL)

   free(XCDE_Console->XCDE_IL_Table_Pointers) ;

 /* ------------------------------------- */
 /* frees the memory for the IL file name */
 /* ------------------------------------- */

 if (XCDE_Console->XCDE_IL_File_Name != NULL)

   free(XCDE_Console->XCDE_IL_File_Name) ;

 /* ------------------------------- */
 /* header file unmapping operation */
 /* ------------------------------- */

 function_result=munmap(XCDE_Console->XCDE_Header_Mapped_Pointer,
			XCDE_Console->XCDE_Header_Filesize) ;

 if (function_result == -1)

   return(-1) ;

 /* ----------------------- */
 /* closing the header file */
 /* ----------------------- */

 function_result=fclose(XCDE_Console->XCDE_Header_File_Pointer) ;

 if (function_result == -1)

   return(-1) ;

 /* ----------------------------------------- */
 /* frees the memory for the header file name */
 /* ----------------------------------------- */

 if (XCDE_Console->XCDE_Header_File_Name != NULL)

   free(XCDE_Console->XCDE_Header_File_Name) ;

 /* ---------------------------------------- */
 /* dictionaries files unmapping and closing */
 /* ---------------------------------------- */

 for (k=0 ; k < XCDE_NUMBER_DICT; ++k) {

   if (XCDE_Console->XCDE_Dict_File_Decompressed[k] == 1) {

     if (XCDE_Console->XCDE_Dict_Dcmptmp_Mapped_Pointers[k] != NULL) {

       function_result=munmap(XCDE_Console->XCDE_Dict_Dcmptmp_Mapped_Pointers[k],
			      XCDE_Console->XCDE_Dict_Dcmptmp_Filesize[k]) ;

       if (function_result == -1)

	 return(-1) ;
     }

     function_result=fclose(XCDE_Console->XCDE_Dict_Dcmptmp_File_Pointers[k]) ;

     if (function_result == -1)

       return(-1) ;

     function_result=unlink(XCDE_Console->XCDE_Dict_Dcmptmp_Names[k]) ;

     if (function_result == -1)

       return(-1) ;

     if (XCDE_Console->XCDE_Dict_Dcmptmp_Names[k] != NULL)

       free(XCDE_Console->XCDE_Dict_Dcmptmp_Names[k]) ;

     if (XCDE_Console->XCDE_Dict_Cmp_File_Names[k] != NULL)

       free(XCDE_Console->XCDE_Dict_Cmp_File_Names[k]) ;

     XCDE_Console->XCDE_Dict_File_Decompressed[k]=0u ;

     if (XCDE_Console->XCDE_Dict_File_Loaded[k] == 1u)

       XCDE_Console->XCDE_Dict_File_Loaded[k]=0u ;
   }
 }

 if (XCDE_Console->Huffman_Position_Token != NULL) {

   free(XCDE_Console->Huffman_Position_Token) ;

   XCDE_Console->Huffman_Position_Token = NULL ;
 }

 /* -------------------------------------- */
 /* if necessary closes the tagstruct file */
 /* -------------------------------------- */

 if (XCDE_Console->XCDE_Tagstruct) {

   function_result=munmap(XCDE_Console->XCDE_Tagstruct_Mapped_Pointer,
			  XCDE_Console->XCDE_Tagstruct_Filesize) ;

   if (function_result == -1)

     return(-1) ;

   function_result=fclose(XCDE_Console->XCDE_Tagstruct_File_Pointer) ;

   if (function_result == -1)

     return(-1) ;

   if (XCDE_Console->XCDE_Tagstruct_Cache_Window != NULL)

     free(XCDE_Console->XCDE_Tagstruct_Cache_Window) ;

   if (XCDE_Console->XCDE_Tagstruct_Windows_Information != NULL)

     free(XCDE_Console->XCDE_Tagstruct_Windows_Information) ;

   XCDE_Console->XCDE_Tagstruct_Cache_Dcmp_Block_Size=0 ;

   XCDE_Console->XCDE_Tagstruct_Cache_Cmp_Block_Size=0 ;

   if (XCDE_Console->XCDE_Tagstruct_File_Name != NULL)

     free(XCDE_Console->XCDE_Tagstruct_File_Name) ;
 }


 /* -------------------------------------- */
 /* if necessary closes the depthstruct file */
 /* -------------------------------------- */

  if (XCDE_Console->XCDE_Depthstruct) {

   function_result=munmap(XCDE_Console->XCDE_Depthstruct_Mapped_Pointer,
			  XCDE_Console->XCDE_Depthstruct_Filesize) ;

   if (function_result == -1)

     return(-1) ;

   function_result=fclose(XCDE_Console->XCDE_Depthstruct_File_Pointer) ;

   if (function_result == -1)

     return(-1) ;

   if (XCDE_Console->XCDE_Depthstruct_File_Name != NULL)

     free(XCDE_Console->XCDE_Depthstruct_File_Name) ;
 }

 XCDE_Console->XCDE_Proximity=0u ;

 XCDE_Console->XCDE_Console_Initialized=0u ;

 return(1) ;
}

/*******************************************************************/

// function that reads the proximity flag.
// Example : ret=XCDE_SettedProximity(&prox, &Console) ;
// where &prox is the memory address of the variable that will
// contain the proximity flag and &Console is the memory address of
// the console data structure.
// ret value is 1 if there aren't errors, or -1 otherwise.

int XCDE_SettedProximity(unsigned char *ProxFlag,
			 XCDE_Console_Type *XCDE_Console)  
{ 
  /* ----------------------------------------- */
  /*            checks parameters              */
  /* ----------------------------------------- */
  
  if (XCDE_Console == NULL)

    return(-1) ; 

  if (XCDE_Console->XCDE_Console_Initialized != 1)

    return(-1) ; 

  if (ProxFlag == NULL) 

    return(-1) ; 

  /* ------------------------ */
  /* reads the proximity flag */
  /* ------------------------ */ 

  *ProxFlag=XCDE_Console->XCDE_Proximity ; 

  return(1) ; 
}

/*******************************************************************/

// function that reads the initialized console flag.
// Example : ret=XCDE_ConsoleInitialized(&initflag, &Console) ; 
// where initflag is the memory address of the variable that will
// contain the console initialized flag and &Console is the memory 
// address of the console data structure.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_ConsoleInitialized(unsigned char *InitializeFlag, 
			    XCDE_Console_Type *XCDE_Console)
{ 
  /* ----------------------------------------- */
  /*            checks parameters              */
  /* ----------------------------------------- */
  
  if (XCDE_Console == NULL)

    return(-1) ; 

  if (InitializeFlag == NULL) 

    return(-1) ; 
  
  /* ---------------------------------- */
  /* reads the initialized console flag */
  /* ---------------------------------- */ 

  *InitializeFlag=XCDE_Console->XCDE_Console_Initialized ; 

  return(1) ; 
} 

/*******************************************************************/

// function that reads the size of the XML source file.
// Example : ret=XCDE_Get_Source_Filesize(&size, &Console)
// where &size is the memory address of the variable that will contain the 
// size, and &Console is the memory address of the console data structure 


int XCDE_Get_Source_Filesize(int *Source_Filesize,
			     XCDE_Console_Type *XCDE_Console) 
{ /* ----------------------------------------- */
  /*            checks parameters              */
  /* ----------------------------------------- */

  if ((XCDE_Console == NULL) || (Source_Filesize == NULL))   

    return(-1) ; 
  
  if (XCDE_Console->XCDE_Console_Initialized != 1)

    return(-1) ; 

  /* ------------------------- */ 
  /* gets the source file size */ 
  /* ------------------------- */ 

  *Source_Filesize=XCDE_Console->XCDE_XML_Source_Filesize ; 

  return(1) ; 
}

/*******************************************************************/

// function that reads the name of the XML source file.
// Example : ret=XCDE_Get_Source_Filename(&name, &Console)
// where name is the memory address of the variable that 
// will contain the name and Console is the memory address of the console 
// data structure.
// The function allocates the memory for the name.  
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_Source_Filename(char **Source_Filename, 
			     XCDE_Console_Type *XCDE_Console) 
{ 
  /* ----------------------------------------- */
  /*            checks parameters              */
  /* ----------------------------------------- */

  if (XCDE_Console == NULL)

    return(-1) ; 

  if (XCDE_Console->XCDE_Console_Initialized != 1)
    
    return(-1) ; 
  
  /* --------------------------------- */ 
  /* allocates the memory for the name */
  /* --------------------------------- */ 

  *Source_Filename=(char *) malloc(strlen(XCDE_Console->XCDE_XML_Source_File_Name)+1) ; 

  if (*Source_Filename == NULL) 

    return(-1) ; 

  /* -------------------- */ 
  /* copies the file name */
  /* -------------------- */ 

  strcpy(*Source_Filename, 
	 XCDE_Console->XCDE_XML_Source_File_Name) ; 

  return(1) ; 
}			     

/*******************************************************************/

// function that reads the size of the header file.
// Example : ret=XCDE_Get_Header_Filesize(&size, &Console)
// where &size is the memory address of the variable that will contain the 
// size, and &Console is the memory address of the console data structure 
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_Header_Filesize(int *Header_Filesize,
			     XCDE_Console_Type *XCDE_Console) 
{ 
  /* ----------------------------------------- */
  /*            checks parameters              */
  /* ----------------------------------------- */
  
  if ((XCDE_Console == NULL) || (Header_Filesize == NULL))   
    
    return(-1) ; 
  
  if (XCDE_Console->XCDE_Console_Initialized != 1)
    
    return(-1) ; 

  /* -------------------------- */ 
  /* reads the header file size */ 
  /* -------------------------- */ 

  *Header_Filesize=XCDE_Console->XCDE_Header_Filesize ; 

  return(1) ; 
}

/*******************************************************************/

// function that reads the name of the header file.
// Example : ret=XCDE_Get_Header_Filename(&name, &Console)
// where name is the memory address of the variable that 
// will contain the name and Console is the memory address of the console 
// data structure.
// The function allocates the memory for the name.  
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_Header_Filename(char **Header_Filename, 
			     XCDE_Console_Type *XCDE_Console) 
{ 
  /* ----------------------------------------- */
  /*            checks parameters              */
  /* ----------------------------------------- */
  
  if (XCDE_Console == NULL)

    return(-1) ; 

  if (XCDE_Console->XCDE_Console_Initialized != 1)

    return(-1) ; 

  if (Header_Filename == NULL) 

    return(-1) ; 

  /* --------------------------------- */ 
  /* allocates the memory for the name */
  /* --------------------------------- */ 

  *Header_Filename=malloc(strlen(XCDE_Console->XCDE_Header_File_Name)+1) ; 

  if (*Header_Filename==NULL) 

    return(-1) ; 

  /* ------------------------------ */ 
  /* copies the memory for the name */
  /* ------------------------------ */ 

  strcpy(*Header_Filename, 
	 XCDE_Console->XCDE_Header_File_Name) ; 

  return(1) ; 
}		

/*******************************************************************/

// function that reads the size of the the body file 
// Example : ret=XCDE_Get_Body_Filesize(&size, &Console)
// where &size is the memory address of the variable that will contain
// the size and &Console is the memory address of the console 
// data structure. 
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_Body_Filesize(int *Body_Filesize,
			   XCDE_Console_Type *XCDE_Console) 
{ 
  /* ----------------------------------------- */
  /*            checks parameters              */
  /* ----------------------------------------- */

  if ((XCDE_Console == NULL) || (Body_Filesize == NULL))   

    return(-1) ; 

  if (XCDE_Console->XCDE_Console_Initialized != 1)

    return(-1) ; 

  /* ------------------------ */ 
  /* reads the body file size */
  /* ------------------------ */

  *Body_Filesize=XCDE_Console->XCDE_Body_Filesize ; 

  return(1) ; 
}

/*******************************************************************/

// function that reads the name of the body file.
// Example : ret=XCDE_Get_Body_Filename(&name, &Console)
// where name is the memory address of the variable that 
// will contain the name and Console is the memory address of the console 
// data structure.
// The function allocates the memory for the name.  
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_Body_Filename(char **Body_Filename, 
			   XCDE_Console_Type *XCDE_Console) 
{ 
  /* ----------------------------------------- */
  /*            checks parameters              */
  /* ----------------------------------------- */

  if (XCDE_Console == NULL)

    return(-1) ; 

  if (XCDE_Console->XCDE_Console_Initialized != 1)

    return(-1) ; 

  if (Body_Filename == NULL) 

    return(-1) ; 

  /* --------------------------------- */ 
  /* allocates the memory for the name */
  /* --------------------------------- */ 

  *Body_Filename=malloc(strlen(XCDE_Console->XCDE_Body_File_Name)+1) ; 

  if (*Body_Filename == NULL) 

    return(-1) ; 

  /* -------------------- */ 
  /* copies the file name */
  /* -------------------- */ 

  strcpy(*Body_Filename, 
	 XCDE_Console->XCDE_Body_File_Name) ; 

  return(1) ; 
}		

// function that puts in Eof_Flag a flag which indicates
// if the position Body_Position is after the end of the Body,
// and so doesn't indicate a valid codeword. 
// Example : ret=XCDE_Body_Eof(body_position, &eofflag, &Console) 
// where body_position is the position we want to check, &eofflag is
// the memory address of the flag and &Console is the memory address
// of the console data structure.
// ret value is 1 if there aren't errors, or -1 otherwise.

int XCDE_Body_Eof(int Body_Position,
		  unsigned char *Eof_Flag, 
		  XCDE_Console_Type *XCDE_Console)

{ /* ----------------------------------------- */
  /*            checks parameters              */
  /* ----------------------------------------- */
  
  if (XCDE_Console == NULL)

    return(-1) ; 

  if (XCDE_Console->XCDE_Console_Initialized != 1)

    return(-1) ; 

  if (Eof_Flag == NULL)

    return(-1) ; 

  if (Body_Position < 0) 

    return(-1) ; 

  
  /* --------------------------------- */ 
  /* checks if the position is at end  */
  /* --------------------------------- */ 

  if (Body_Position >= XCDE_Console->XCDE_Body_Size_Without_Zlib) 

    *Eof_Flag=1 ;

  else

    *Eof_Flag=0 ; 

  return(1) ;
}


/*******************************************************************/

// function that reads the size of the the IL file 
// Example : ret=XCDE_Get_IL_Filesize(&size, &Console)
// where &size is the memory address of the variable that will contain
// the size and &Console is the memory address of the console 
// data structure. 
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_IL_Filesize(int *IL_Filesize,
			 XCDE_Console_Type *XCDE_Console) 
{ 
  /* ----------------------------------------- */
  /*            checks parameters              */
  /* ----------------------------------------- */

  if ((XCDE_Console == NULL) || (IL_Filesize == NULL))   
    
    return(-1) ; 
  
  if (XCDE_Console->XCDE_Console_Initialized != 1)
    
    return(-1) ; 

  /* --------------------- */ 
  /* gets the IL file size */ 
  /* --------------------- */ 

  *IL_Filesize=XCDE_Console->XCDE_IL_Filesize ; 
  
  return(1) ; 
}

/*******************************************************************/

// function that reads the name of the IL file.
// Example : ret=XCDE_Get_IL_Filename(&name, &Console)
// where name is the memory address of the variable that 
// will contain the name and Console is the memory address of the console 
// data structure.
// The function allocates the memory for the name.  
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_IL_Filename(char **IL_Filename, 
			 XCDE_Console_Type *XCDE_Console) 
{ 
  /* ----------------------------------------- */
  /*            checks parameters              */
  /* ----------------------------------------- */

  if (XCDE_Console == NULL)

    return(-1) ; 

  if (XCDE_Console->XCDE_Console_Initialized != 1)

    return(-1) ; 

  if (IL_Filename == NULL)

    return(-1) ; 

  /* --------------------------------- */ 
  /* allocates the memory for the name */
  /* --------------------------------- */ 

  *IL_Filename=malloc(strlen(XCDE_Console->XCDE_IL_File_Name)+1) ; 

  if (*IL_Filename == NULL) 

    return(-1) ; 

  /* -------------------- */ 
  /* copies the file name */
  /* -------------------- */ 

  strcpy(*IL_Filename, 
	 XCDE_Console->XCDE_IL_File_Name) ; 

  return(1) ; 
}	  

/*******************************************************************/

// function that reads the name of the specified compressed dictionary 
// file.
// Example : ret=XCDE_Get_Dict_Cmp_Filename(TAG, &name, &Console)
// where TAG (WORD, ATTR) is the constant for the dictionary of 
// TAG (WORD, ATTR), &name is the memory address of the variable 
// that will contain the name and &Console is the memory address of 
// the console data structure.
// The function allocates the memory for the name.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_Dict_Cmp_Filename(unsigned char Dict,
			       char **Dict_Cmp_Filename,
			       XCDE_Console_Type *XCDE_Console) 
{ 
  /* ----------------------------------------- */
  /*            checks parameters              */
  /* ----------------------------------------- */

  if ((XCDE_Console == NULL ) || (Dict_Cmp_Filename == NULL)) 

    return(-1) ; 

  if (XCDE_Console->XCDE_Console_Initialized != 1)

    return(-1) ; 

  if ((Dict != ATTR) && (Dict != WORD) && (Dict != TAG)) 
    
    return(-1) ; 

  /* --------------------------------- */ 
  /* allocates the memory for the name */
  /* --------------------------------- */ 

  *Dict_Cmp_Filename=malloc(strlen(XCDE_Console->XCDE_Dict_Cmp_File_Names[Dict])+1) ; 

  if (*Dict_Cmp_Filename == NULL) 
    
    return(-1) ; 

  /* -------------------- */ 
  /* copies the file name */
  /* -------------------- */ 

  strcpy(*Dict_Cmp_Filename, 
	 XCDE_Console->XCDE_Dict_Cmp_File_Names[Dict]) ;

  return(1) ; 
}	  

/*******************************************************************/

// function that gets the size of the specified compressed dictionary 
// file.
// Example : ret=XCDE_Get_Dict_Cmp_Filesize(TAG, &size, &Console)
// where TAG (WORD, ATTR) is the constant for the dictionary of 
// TAG (WORD, ATTR), &size is the memory address of the variable 
// that will contain the size and &Console is the memory address of 
// the console data structure.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_Dict_Cmp_Filesize(unsigned char Dict,
			       int *Dict_Cmp_Filesize,
			       XCDE_Console_Type *XCDE_Console) 
{ struct stat sbuf ; 

 int function_result=0 ; 

 /* ----------------------------------------- */
 /*            checks parameters              */
 /* ----------------------------------------- */
 
 if ((XCDE_Console == NULL ) || (Dict_Cmp_Filesize == NULL)) 
    
   return(-1) ; 

 if (XCDE_Console->XCDE_Console_Initialized != 1 )

   return(-1) ; 

  if ((Dict != ATTR) && (Dict != WORD) && (Dict != TAG)) 
    
    return(-1) ; 

  /* ---------------------------------------------------------- */
  /* reads the information about the compressed dictionary file */
  /* ---------------------------------------------------------- */

  function_result=stat(XCDE_Console->XCDE_Dict_Cmp_File_Names[Dict],
		       &sbuf) ;

  if (function_result == -1) 

    { fprintf(stderr, "\n\tFatal Error: problems in stat operation on file %s\n", XCDE_Console->XCDE_Dict_Cmp_File_Names[Dict]) ;

    return(-1) ; 
    }
  
  /* ---------------------------------------- */ 
  /* gets the compressed dictionary file size */
  /* ---------------------------------------- */
  
  *Dict_Cmp_Filesize=(int) sbuf.st_size ;   
  
  return(1) ; 
}

/*******************************************************************/

// function that reads the name of the temporary file where has been 
// decompressed the specified dictionary file.
// Example : ret=XCDE_Get_Dict_Dcmptmp_Filename(TAG, &name, &Console)
// where TAG (WORD, ATTR) is the constant for the dictionary of 
// TAG (WORD, ATTR), &name is the memory address of the variable 
// that will contain the name and &Console is the memory address of 
// the console data structure.
// The function allocates the memory for the name.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_Dict_Dcmptmp_Filename(unsigned char Dict,
				   char **Dict_Dcmptmp_Filename,
				   XCDE_Console_Type *XCDE_Console)
{ 
  /* ----------------------------------------- */
  /*            checks parameters              */
  /* ----------------------------------------- */
   
  if ((XCDE_Console == NULL ) || (Dict_Dcmptmp_Filename == NULL)) 
    
    return(-1) ; 

  if (XCDE_Console->XCDE_Console_Initialized != 1 )

    return(-1) ; 

  if ((Dict != ATTR) && (Dict != WORD) && (Dict != TAG)) 
    
    return(-1) ; 

  if (XCDE_Console->XCDE_Dict_File_Decompressed[Dict] != 1) 

    return(-1) ; 

  /* --------------------------------- */ 
  /* allocates the memory for the name */
  /* --------------------------------- */ 

  *Dict_Dcmptmp_Filename=malloc(strlen(XCDE_Console->XCDE_Dict_Dcmptmp_Names[Dict])+1) ; 

  if (*Dict_Dcmptmp_Filename == NULL) 

    return(-1) ;

  /* -------------------- */ 
  /* copies the file name */
  /* -------------------- */

  strcpy(*Dict_Dcmptmp_Filename, 
	 XCDE_Console->XCDE_Dict_Dcmptmp_Names[Dict]) ;
  
  return(1) ; 
}	

/*******************************************************************/

// function that gets the size of the temporary file where has been 
// decompressed the specified dictionary file.
// Example : ret=XCDE_Get_Dict_Dcmptmp_Filesize(TAG, &size, &Console)
// where TAG (WORD, ATTR) is the constant for the dictionary of 
// TAG (WORD, ATTR), &size is the memory address of the variable 
// that will contain the size and &Console is the memory address of 
// the console data structure.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_Dict_Dcmptmp_Filesize(unsigned char Dict,
				   int *Dict_Dcmptmp_Filesize, 
				   XCDE_Console_Type *XCDE_Console)
{ 
  /* ----------------------------------------- */
  /*            checks parameters              */
  /* ----------------------------------------- */

  if ((XCDE_Console == NULL) || (Dict_Dcmptmp_Filesize == NULL)) 

    return(-1) ; 

  if (XCDE_Console->XCDE_Console_Initialized != 1)

    return(-1) ; 

  if ((Dict != ATTR) && (Dict != WORD) && (Dict != TAG)) 
    
    return(-1) ; 

 if (XCDE_Console->XCDE_Dict_File_Decompressed[Dict] != 1) 

    return(-1) ; 

 /* ---------------------------- */ 
 /* gets the temporary file size */
 /* ---------------------------- */ 
 
 *Dict_Dcmptmp_Filesize=XCDE_Console->XCDE_Dict_Dcmptmp_Filesize[Dict] ; 
 
  return(1) ; 
}	    

/*******************************************************************/

// function that reads the maximum codeword length of the Huffman code.
// Example : ret=XCDE_Get_Maxcwlen(&Maxcwlength, &Console)
// where &Maxcwlenght is the memory address of the variable that
// will contain the maximum codeword length and &Console is the memory
// address of the console data structure.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_Maxcwlen(int *Maxcwlen,
		      XCDE_Console_Type *XCDE_Console) 
{ 
  /* ----------------------------------------- */
  /*            checks parameters              */
  /* ----------------------------------------- */
  
  if (Maxcwlen == NULL)

    return(-1) ; 

  if (XCDE_Console == NULL)
    
    return(-1) ; 

  if (XCDE_Console->XCDE_Console_Initialized != 1)

    return(-1) ; 
  
  /* ---------------------------------------------- */ 
  /* reads the value of the maximum codeword length */
  /* ---------------------------------------------- */

  *Maxcwlen=XCDE_Console->Huffman_Codeword_Max_Len ; 

  return(1) ; 
}

/*******************************************************************/

// function that reads the number of tokens that are in the general 
// dictionary.
// Example : ret=XCDE_Get_Num_Tokens(&num_tokens, &Console) 
// where &num_tokens is the memory address of the variable that will contain 
// the number of tokens and &Console is the memory address of the console
// data structure. 
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_Num_Tokens(int *Num_Tokens,
			XCDE_Console_Type *XCDE_Console) 
{ 
  /* ----------------------------------------- */
  /*            checks parameters              */
  /* ----------------------------------------- */

  if (Num_Tokens == NULL)

    return(-1) ;

  if (XCDE_Console == NULL)

    return(-1) ;

  if (XCDE_Console->XCDE_Console_Initialized != 1)

    return(-1) ;

  /* -------------------------- */
  /* reads the number of tokens */
  /* -------------------------- */

  *Num_Tokens=XCDE_Console->Huffman_Num_Tokens ;

  return(1) ;
}

/*******************************************************************/

// function that reads the number of tokens that are in the specified dictionary.
// Example : ret=XCDE_Num_El_Dict(TAG, &NumTokens, &Console)
// where TAG (WORD, ATTR) is the constant for the dictionary of
// TAG (WORD, ATTR), &NumTokens is the memory address of the variable
// that will contain the number of tokens and &Console is the memory
// address of the console data structure.
// ret value is 1 if there aren't errors, or -1 otherwise.

int XCDE_Get_Num_El_Dict(unsigned char Dict,
			 int *Num_El,
			 XCDE_Console_Type *XCDE_Console)
{ register
    int k=0 ;

 /* ----------------------------------------- */
 /*            checks parameters              */
 /* ----------------------------------------- */

 if (XCDE_Console == NULL)

   return(-1) ;

 if (XCDE_Console->XCDE_Console_Initialized != 1)

   return(-1) ;

 if ((Dict != ATTR) && (Dict != WORD) && (Dict != TAG))

   return(-1) ;

 if (Num_El == NULL)

   return(-1) ;

 /* ----------------------------------------------- */
 /* computes the number of tokens of the dictionary */
 /* ----------------------------------------------- */

  *Num_El=0 ;

  for(k=0; k<XCDE_Console->Huffman_Codeword_Max_Len; ++k)

    *Num_El+=XCDE_Console->Huffman_Dict_NumEl[k][Dict] ;

  return(1) ;
}

/*******************************************************************/

// function that gets the number of tokens which codeword
// lenght is lev.
// Example : ret=XCDE_Num_El_Lev(lev, &NumTokens, &Console)
// where lev is the level, &NumTokens is the memory address of
// the variable that will contain the number of tokens and &Console
// is the memory address of the console data structure.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_Num_El_Lev(int Lev,
			int *Num_El,
			XCDE_Console_Type *XCDE_Console) 
{ register 
    unsigned int k=0 ; 

 /* ----------------------------------------- */
 /*            checks parameters              */
 /* ----------------------------------------- */

  if (XCDE_Console == NULL)

    return(-1) ; 

  if (XCDE_Console->XCDE_Console_Initialized != 1)

    return(-1) ; 

  if (Num_El == NULL)

    return(-1) ; 

  if ((Lev<1) || (Lev > XCDE_Console->Huffman_Codeword_Max_Len))
    
    return(-1) ; 

  /* ---------------------------------------- */
  /* computes the number of tokens in a level */
  /* ---------------------------------------- */

  *Num_El=0 ;

  for(k=0; k< XCDE_NUMBER_DICT; ++k)

    *Num_El+=XCDE_Console->Huffman_Dict_NumEl[Lev-1][k] ; 

  return(1) ; 
}

/*******************************************************************/

// function that reads the number of tokens in a specified 
// dictionary wich codeword length is lev
// Example : ret=XCDE_Num_El_Dictlev(lev, TAG, &NumTokens, &Console) 
// where lev is the level,  TAG (WORD, ATTR) is the constant for 
// the dictionary of TAG (WORD, ATTR), &NumTokens is the memory 
// address of the variable that will contain the number of tokens 
// and &Console is the memory address of the console data structure.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_Num_El_Dictlev(int Lev,
			    unsigned char Dict,
			    int *Num_El,
			    XCDE_Console_Type *XCDE_Console) 
{ 
  /* ----------------------------------------- */
  /*            checks parameters              */
  /* ----------------------------------------- */

  if (XCDE_Console == NULL)

    return(-1) ; 

  if (XCDE_Console->XCDE_Console_Initialized != 1u)

    return(-1) ; 

  if (Num_El == NULL)

    return(-1) ; 

  if ((Lev<1) || (Lev > XCDE_Console->Huffman_Codeword_Max_Len))
    
    return(-1) ; 

  if ((Dict != ATTR) && (Dict != WORD) && (Dict != TAG)) 
    
    return(-1) ; 

  /* -------------------------------------------- */ 
  /* reads the number of tokens of the dictionary */
  /* -------------------------------------------- */
  
  *Num_El=XCDE_Console->Huffman_Dict_NumEl[Lev-1][Dict] ; 

  return(1) ; 
}

/*******************************************************************/

// function that decompress the specified dictionary if it hasn't been
// decompressed yet (in this case returns -1).
// Example : ret=XCDE_Decompress_Dictionary(TAG, &Console) 
// where TAG (WORD, ATTR) is the constant for the dictionary 
// of TAG (WORD, ATTR) and &Console is the memory address 
// of the console data structure. 
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Decompress_Dictionary(unsigned char Dict,
			       XCDE_Console_Type *XCDE_Console) 
{ register
    int i=0 ; 

 FILE *cmpdictionary=NULL ; 

 FILE *dcmpdictionary=NULL ;

 char zlib_decompress_buffer[2*XCDE_COMPRESS_BLOCK_SIZE] ; 
 
 char *dictcmpmappointer=NULL ;

 int dictcmppointer=0 ; 
 
 int dictcmpfilesize=0 ;

 unsigned
   long len=0 ;
  
 uLong comprLen=0 ;

 uLong uncomprLen=0 ; 

 int uncmp_err=0 ; 

 int result=0 ; 

 struct 
   stat sbuf ; 
  
 /* ----------------------------------------- */
 /*            checks parameters              */
 /* ----------------------------------------- */

  if (XCDE_Console == NULL)

    return(-1) ; 

  if (XCDE_Console->XCDE_Console_Initialized != 1)

    return(-1) ; 

  if ((Dict != ATTR) && (Dict != WORD) && (Dict != TAG)) 
    
    return(-1) ; 

  if (XCDE_Console->XCDE_Dict_File_Decompressed[Dict] != 0)

    return(-1) ; 

  /* ------------------------- */
  /* decompress the dictionary */ 
  /* ------------------------- */ 

  // allocates the memory needed by the name of decompressed dictionary
  XCDE_Console->XCDE_Dict_Dcmptmp_Names[Dict]=malloc(L_tmpnam) ; 
  
  if (XCDE_Console->XCDE_Dict_Dcmptmp_Names[Dict] == NULL) { 

    fprintf(stderr, "\n\tFatal Error: problems allocating memory\n") ; 

    return(-1) ; 
  }  
    
  // gets a temporary file name for the decompressed dictionary 
  XCDE_Console->XCDE_Dict_Dcmptmp_Names[Dict]=tmpnam(XCDE_Console->XCDE_Dict_Dcmptmp_Names[Dict]) ; 
  
  // open the temporary file name that will contain the decompressed dictionary 
  if ((dcmpdictionary=fopen(XCDE_Console->XCDE_Dict_Dcmptmp_Names[Dict], "wb")) == NULL) { 

    fprintf(stderr, "\n\tFatal Error: problems opening file %s\n", XCDE_Console->XCDE_Dict_Dcmptmp_Names[Dict]) ;

    if (XCDE_Console->XCDE_Dict_Dcmptmp_Names[Dict] != NULL) 

      free(XCDE_Console->XCDE_Dict_Dcmptmp_Names[Dict]) ; 

    XCDE_Console->XCDE_Dict_Dcmptmp_Names[Dict]=NULL ;

    return(-1) ; 
  }

  // open the compressed dictionary file 
  if ((cmpdictionary=fopen(XCDE_Console->XCDE_Dict_Cmp_File_Names[Dict], "rb")) == NULL) { 

    fprintf(stderr, "\n\tFatal Error: problems opening file %s\n", XCDE_Console->XCDE_Dict_Cmp_File_Names[Dict]) ;

    unlink(XCDE_Console->XCDE_Dict_Dcmptmp_Names[Dict]) ; 

    if (XCDE_Console->XCDE_Dict_Dcmptmp_Names[Dict] != NULL) 

      free(XCDE_Console->XCDE_Dict_Dcmptmp_Names[Dict]) ; 

    XCDE_Console->XCDE_Dict_Dcmptmp_Names[Dict]=NULL ;

    return(-1) ; 
  }
  
  // found the file size of the compressed dictionary 
  result=stat(XCDE_Console->XCDE_Dict_Cmp_File_Names[Dict],
	      &sbuf) ;
  
  if (result == -1) { 

    fprintf(stderr, "\n\tFatal Error: problems in stat operation on file %s\n", XCDE_Console->XCDE_Dict_Cmp_File_Names[Dict]) ;

    fclose(cmpdictionary) ;

    fclose(dcmpdictionary) ; 
    
    unlink(XCDE_Console->XCDE_Dict_Dcmptmp_Names[Dict]) ; 
    
    if (XCDE_Console->XCDE_Dict_Dcmptmp_Names[Dict] != NULL) 
      
      free(XCDE_Console->XCDE_Dict_Dcmptmp_Names[Dict]) ; 
    
    XCDE_Console->XCDE_Dict_Dcmptmp_Names[Dict]=NULL ;

    return(-1) ; 
  }

  dictcmpfilesize=(int) sbuf.st_size ;

  // maps the compressed dictionary 
  dictcmpmappointer = (char *) mmap(NULL,
				    dictcmpfilesize, 
				    PROT_READ,
				    MAP_SHARED,
				    fileno(cmpdictionary), 
				    0) ; 

  if (dictcmpmappointer == NULL) { 

    fprintf(stderr, "\n\tFatal Error: problems in mapping operation on file %s\n", XCDE_Console->XCDE_Dict_Cmp_File_Names[Dict]) ;

    fclose(cmpdictionary) ;

    fclose(dcmpdictionary) ; 

    unlink(XCDE_Console->XCDE_Dict_Dcmptmp_Names[Dict]) ; 

    if (XCDE_Console->XCDE_Dict_Dcmptmp_Names[Dict] != NULL) 
      
      free(XCDE_Console->XCDE_Dict_Dcmptmp_Names[Dict]) ; 
    
    XCDE_Console->XCDE_Dict_Dcmptmp_Names[Dict]=NULL ;

    return(-1) ; 
  }

  // decompress the dictionary 
  while(dictcmppointer < dictcmpfilesize) { 

    continuation_bit_d((unsigned long *) &comprLen, 
		       (char *) (dictcmppointer+dictcmpmappointer),
		       (unsigned int) 0,
		       (unsigned long *) &len) ;
    
    dictcmppointer+=(len/8) ; 
    
    uncomprLen=2*XCDE_COMPRESS_BLOCK_SIZE ;
    
    uncmp_err=uncompress((Bytef *) zlib_decompress_buffer, 
			 (uLong *) &uncomprLen, 
			 (Bytef *) (dictcmppointer+dictcmpmappointer), 
			 (uLong) comprLen) ; 
    
    if (uncmp_err != Z_OK) {
      
      fprintf(stderr, "\n\tFatal Error: problems decompressing dictionary\n") ;
      
      fclose(cmpdictionary) ; 
      
      fclose(dcmpdictionary) ;
      
      munmap(dictcmpmappointer, 
	     dictcmpfilesize) ; 
      
      unlink(XCDE_Console->XCDE_Dict_Dcmptmp_Names[Dict]) ; 
      
      if (XCDE_Console->XCDE_Dict_Dcmptmp_Names[Dict] != NULL) 
	
	free(XCDE_Console->XCDE_Dict_Dcmptmp_Names[Dict]) ; 
      
      XCDE_Console->XCDE_Dict_Dcmptmp_Names[Dict]=NULL ;

      return(-1) ; 
    }
  
    dictcmppointer+=comprLen ;

    for(i=0; i< uncomprLen; ++i) 
    
      fputc(zlib_decompress_buffer[i], 
	    dcmpdictionary) ; 
  }

  fclose(cmpdictionary) ;

  fclose(dcmpdictionary) ; 

  munmap(dictcmpmappointer, 
	 dictcmpfilesize) ; 

  // open the decompress dictionary to map it 
  if ((XCDE_Console->XCDE_Dict_Dcmptmp_File_Pointers[Dict]=fopen(XCDE_Console->XCDE_Dict_Dcmptmp_Names[Dict], "rb")) == NULL) { 

    fprintf(stderr, "\n\tFatal Error: problems opening file %s\n", XCDE_Console->XCDE_Dict_Dcmptmp_Names[Dict]) ; 

    unlink(XCDE_Console->XCDE_Dict_Dcmptmp_Names[Dict]) ;

    if (XCDE_Console->XCDE_Dict_Dcmptmp_Names[Dict] != NULL) 
      
      free(XCDE_Console->XCDE_Dict_Dcmptmp_Names[Dict]) ; 
    
    XCDE_Console->XCDE_Dict_Dcmptmp_Names[Dict]=NULL ;

    return(-1) ; 
  }
  
  // found the decompress dictionary filesize 
  result=stat(XCDE_Console->XCDE_Dict_Dcmptmp_Names[Dict],
	      &sbuf) ;

  if (result == -1) { 

    fprintf(stderr, "\n\tFatal Error: problems in stat operation on file %s\n", XCDE_Console->XCDE_Dict_Dcmptmp_Names[Dict]) ;

    fclose(XCDE_Console->XCDE_Dict_Dcmptmp_File_Pointers[Dict]) ; 

    unlink(XCDE_Console->XCDE_Dict_Dcmptmp_Names[Dict]) ; 

    if (XCDE_Console->XCDE_Dict_Dcmptmp_Names[Dict] != NULL) 
      
      free(XCDE_Console->XCDE_Dict_Dcmptmp_Names[Dict]) ; 
    
    XCDE_Console->XCDE_Dict_Dcmptmp_Names[Dict] = NULL ; 

    return(-1) ; 
  }
  
  XCDE_Console->XCDE_Dict_Dcmptmp_Filesize[Dict]=(int) sbuf.st_size ;   
  
  XCDE_Console->XCDE_Dict_Dcmptmp_Mapped_Pointers[Dict] = (char *) mmap(NULL,
									XCDE_Console->XCDE_Dict_Dcmptmp_Filesize[Dict],
									PROT_READ,
									MAP_SHARED,
									fileno(XCDE_Console->XCDE_Dict_Dcmptmp_File_Pointers[Dict]),
									0) ; 

  if (XCDE_Console->XCDE_Dict_Dcmptmp_Mapped_Pointers[Dict] == NULL) { 

    fprintf(stderr, "\n\tFatal Error: problems in mapping operation on file %s\n", XCDE_Console->XCDE_Dict_Dcmptmp_Names[Dict]) ;

    fclose(XCDE_Console->XCDE_Dict_Dcmptmp_File_Pointers[Dict]) ; 

    unlink(XCDE_Console->XCDE_Dict_Dcmptmp_Names[Dict]) ; 

    if (XCDE_Console->XCDE_Dict_Dcmptmp_Names[Dict] != NULL) 
      
      free(XCDE_Console->XCDE_Dict_Dcmptmp_Names[Dict]) ; 
    
    XCDE_Console->XCDE_Dict_Dcmptmp_Names[Dict] = NULL ; 

    return(-1) ; 
    }
  
  XCDE_Console->XCDE_Dict_File_Decompressed[Dict]=1u ; 

  return(1) ; 
}

/*******************************************************************/

// functions that traslates the position of a token in a specified 
// dictionary to its codeword length. 
// Example : ret=XCDE_Get_DictEl2Lev(TAG, position, &length, &Console)
// where TAG (WORD, ATTR) is the constant for the dictionary 
// of TAG (WORD, ATTR), position if the position of the element in
// the specified dictionary, &length is the memory address of the 
// variable that will contain the lenght, and &Console is the
// memory address of the console data structure.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_DictEl2Lev(unsigned char Dict,
		    int NDictElem,
		    int *Lev,
		    XCDE_Console_Type *XCDE_Console)
{ int function_result=0 ; 

  int n_elem_dict=0 ; 

  int sum_elem=0 ; 

  register 
    int k=0 ; 

  /* ----------------------------------------- */
  /*            checks parameters              */
  /* ----------------------------------------- */
 
  if ((XCDE_Console == NULL) || (Lev == NULL)) 

    return(-1) ; 

  if (XCDE_Console->XCDE_Console_Initialized != 1)

    return(-1) ; 
  
  if ((Dict != ATTR) && (Dict != WORD) && (Dict != TAG)) 
    
    return(-1) ; 
  
  /* ------------------------------------------------- */ 
  /* computes the level of the token in the dictionary */
  /* ------------------------------------------------- */

  function_result=XCDE_Get_Num_El_Dict(Dict,
				       &n_elem_dict,
				       XCDE_Console) ; 
  
  if (function_result == -1) 

    return(-1) ; 

  if ((NDictElem > n_elem_dict) || (NDictElem <= 0))

    return(-1) ; 

  k=XCDE_Console->Huffman_Codeword_Max_Len-1 ; 

  sum_elem=0 ;

  while ((k > 0) && (NDictElem > (sum_elem+XCDE_Console->Huffman_Dict_NumEl[k][Dict])))
    { sum_elem+=XCDE_Console->Huffman_Dict_NumEl[k][Dict] ; 

    --k ; 
    }

  *Lev=(k+1) ; 

  return(1) ; 
}

/*******************************************************************/

// functions that traslates the position of a token in a specified 
// dictionary to its rank in the general dictionary (the dictionary 
// of all the tokens).
// Example : ret=XCDE_Get_Dict2Rank(TAG, position, &rank, &Console)
// where TAG (WORD, ATTR) is the constant for the dictionary 
// of TAG (WORD, ATTR), position if the position of the element in
// the specified dictionary, &rank is the memory address of the 
// variable that will contain the rank computed and &Console is the
// memory address of the console data struct.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Dict2Rank(unsigned char Dict,
		   int NDictElem,
		   int *Rank,
		   XCDE_Console_Type *XCDE_Console) 
{ int function_result=0 ; 

  int Lev=0 ; 

  int Num_El=0 ; 

  int sum_el=0 ; 

  int mod_el=0 ; 

  register 
    int k=0 ; 

  /* ----------------------------------------- */
  /*            checks parameters              */
  /* ----------------------------------------- */

  if (XCDE_Console == NULL)

    return(-1) ; 

  if (XCDE_Console->XCDE_Console_Initialized != 1)

    return(-1) ; 

  if ((Dict != ATTR) && (Dict != WORD) && (Dict != TAG)) 
    
    return(-1) ; 

  /* -------------------------------- */ 
  /* computes the rank of the element */
  /* -------------------------------- */

  function_result=XCDE_DictEl2Lev(Dict,
				  NDictElem,
				  &Lev,
				  XCDE_Console) ;

  if (function_result == -1)

    return(-1) ; 
  
  mod_el=NDictElem ; 

  sum_el=0 ; 
  
  // computes the sum of the elements at the previous level 
  for (k=XCDE_Console->Huffman_Codeword_Max_Len-1; k>Lev-1 ; --k)

    { function_result=XCDE_Get_Num_El_Lev(k+1,
					  &Num_El,
					  XCDE_Console) ; 
    
    if (function_result == -1) 
      
      return(-1) ; 
    
    sum_el+=Num_El ; 
    }

  // computes the sum of the elements of the previous dictionary
  // at the same level of the specified one
  for(k=0 ; k<Dict; ++k) 

    sum_el+=XCDE_Console->Huffman_Dict_NumEl[Lev-1][k] ; 
  
  for(k=XCDE_Console->Huffman_Codeword_Max_Len-1; k>Lev-1 ; --k) 

    mod_el-=XCDE_Console->Huffman_Dict_NumEl[k][Dict] ; 

  sum_el+=mod_el ; 

  *Rank=(sum_el-1) ; 
  
  return(1) ; 
}

/*******************************************************************/

// function that loads the tokens of the specified dictionary 
// (that must be decompressed, otherwise returns -1) in the
// memory. 
// Example : ret=XCDE_Load_Dictionary(TAG, &Console) 
// where TAG (WORD, ATTR) is the constant for the dictionary 
// of TAG (WORD, ATTR) and Console is the memory address of the
// console data struct.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Load_Dictionary(unsigned char Dict,
			 XCDE_Console_Type *XCDE_Console)
{ int num_el_dict=0 ; 

  int function_result=0 ; 

  char *dict_position=NULL ; 
  
  int Rank=0 ; 

  register 
    int k=0 ; 

  /* ----------------------------------------- */
  /*            checks parameters              */
  /* ----------------------------------------- */
  
  if (XCDE_Console == NULL)

    return(-1) ; 

  if (XCDE_Console->XCDE_Console_Initialized != 1)

    return(-1) ; 

  if ((Dict != ATTR) && (Dict != WORD) && (Dict != TAG)) 
    
    return(-1) ; 
  
  if (XCDE_Console->XCDE_Dict_File_Decompressed[Dict] != 1)

    return(-1) ; 

  /* ----------------------------------------------- */
  /* loads the decompressed dictionary in the memory */
  /* ----------------------------------------------- */ 

  // computes the number of elements of the dictionary 
  function_result=XCDE_Get_Num_El_Dict(Dict,
				       &num_el_dict,
				       XCDE_Console) ; 

  if (function_result == -1) 
    
    return(-1) ; 

  // allocates the memory needed to contain the tokens 
  if (XCDE_Console->Huffman_Position_Token == NULL)

    { XCDE_Console->Huffman_Position_Token=calloc(XCDE_Console->Huffman_Num_Tokens, sizeof(char *)) ; 
    
    if (XCDE_Console->Huffman_Position_Token == NULL)

      return(-1) ; 
    }

  // initializes the pointer to the decompressed dictionary 
  dict_position=XCDE_Console->XCDE_Dict_Dcmptmp_Mapped_Pointers[Dict] ; 

  dict_position++ ; 

  for (k=1 ; k<=num_el_dict ; ++k)

    { function_result=XCDE_Dict2Rank(Dict,
				     k,
				     &Rank,
				     XCDE_Console) ; 
    
    if (function_result == -1) 

      return(-1) ; 

    // puts the token in the position corresponding to its rank 
    XCDE_Console->Huffman_Position_Token[Rank]=dict_position ; 

    for( ; ((int) (*dict_position)) != 0; ++dict_position) ; 
 
    dict_position++  ; 
    }

  // put on the flag for dictionary loaded 
  XCDE_Console->XCDE_Dict_File_Loaded[Dict]=1 ; 

  return(1) ; 
}

/*******************************************************************/

// function that translates the rank of a token in the general 
// dictionary to the position in the dictionary where it's in. 
// Example : XCDE_Rank2Dict(rank, &Dict, &DictPos, &Console)
// where rank is the rank of the token in the general dictionary,
// &Dict is the memory address of the variable that will contain
// the dictionary, &DictPos is the memory address of the variable
// that will contain the position and &Console is the memory
// address of the console data struct
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Rank2Dict(int Rank,
		   unsigned char *Dict,
		   int *NDictElem,
		   XCDE_Console_Type *XCDE_Console) 
{ register 
    int k=0 ; 

  register
    int j=0 ; 

  register
    int i=0 ;

  int sum_elem=0 ; 

  /* ----------------------------------------- */
  /*            checks parameters              */
  /* ----------------------------------------- */
  
  if (XCDE_Console == NULL)

    return(-1) ; 

  if (XCDE_Console->XCDE_Console_Initialized != 1)

    return(-1) ; 

  if ((Dict == NULL) || (NDictElem == NULL))

    return(-1) ; 

  if ((Rank < 0) || (Rank >= XCDE_Console->Huffman_Num_Tokens))

    return(-1) ; 

  /* ----------------------------------------------------- */
  /* computes the dictionary and the position of the token */
  /* ----------------------------------------------------- */
  
  sum_elem=0 ; 

  for (j=(XCDE_Console->Huffman_Codeword_Max_Len-1); j>=0; --j)

    for (k=0 ; k< XCDE_NUMBER_DICT; ++k) 

      if ((sum_elem+XCDE_Console->Huffman_Dict_NumEl[j][k]) > Rank) 

	{ *Dict=k ; 

	*NDictElem=(Rank-sum_elem+1) ; 

	for(i=(XCDE_Console->Huffman_Codeword_Max_Len-1) ; i>j ; --i)

	  (*NDictElem)+=XCDE_Console->Huffman_Dict_NumEl[i][k] ; 
	
	return(1) ; 
	}
    
      else sum_elem+=(XCDE_Console->Huffman_Dict_NumEl[j][k]) ; 

  return(-1) ; 
}

/*******************************************************************/

// function that gets the IL of positions of the token with the specified Rank.
// Example : ret=XCDE_Get_IL_ListPos(Rank, &IL, &IL_Num_El, &Console)
// where Rank is the rank of the element, &IL is the memory address of
// the variable that will contain the IL, &IL_Num_El is the memory address of
// the variable that will contain the number of elements of the IL and &Console
// is the memory address of the console data struct.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_IL_ListPos(int Rank,
			XCDE_IL_Couple_Type **IL,
			int *IL_Num_El,
			XCDE_Console_Type *XCDE_Console) 
{ register 
    int j=0 ; 

  int num_pointer_IL_Index=0 ; 

  int num_IL_List=0 ; 

  int counter_num_IL_Block=0 ; 

  int number_byte_IL=0 ; 

  int position=0 ; 

  unsigned 
    long len_continuation=0 ; 

  unsigned char escape=0 ; 

  int delta_diff=0 ;

  int function_result=0 ; 

  unsigned 
    long int read_el=0 ; 

  /* ----------------------------------------- */
  /*            checks parameters              */
  /* ----------------------------------------- */

  if (XCDE_Console == NULL)

    return(-1) ; 

  if (XCDE_Console->XCDE_Console_Initialized != 1)

    return(-1) ; 

  if ((Rank < 0) || (Rank >= XCDE_Console->Huffman_Num_Tokens))

    return(-1) ; 

  if ((IL == NULL) || (IL_Num_El == NULL)) 

    return(-1) ; 
  
  /* ---------------- */ 
  /* accessing the IL */ 
  /* ---------------- */

  num_pointer_IL_Index=((Rank) / (XCDE_IL_POINTER_RATIO)); 
  
  num_IL_List=((Rank) % (XCDE_IL_POINTER_RATIO)) ; 

  position=XCDE_Console->XCDE_IL_Table_Pointers[num_pointer_IL_Index] ; 


  while(counter_num_IL_Block<num_IL_List) 
  
    if (XCDE_Console->XCDE_IL_Mapped_Pointer[position] == 0) 
      
      { ++counter_num_IL_Block ; 

      ++position ; 
      }
  
    else
      
      { continuation_bit_d((unsigned long *) &number_byte_IL, 
			   (char *) XCDE_Console->XCDE_IL_Mapped_Pointer+position,
			   (unsigned int) 0,
			   (unsigned long *) &len_continuation) ;
	
      ++counter_num_IL_Block ; 
	
      position+=((len_continuation/8)+number_byte_IL) ; 
      }
  
  if (XCDE_Console->XCDE_IL_Mapped_Pointer[position] == 0) 
    
    { *IL=NULL ; 

    *IL_Num_El=0 ; 
    
    return(0) ; 
    }

  continuation_bit_d((unsigned long *) &number_byte_IL, 
		     (char *) XCDE_Console->XCDE_IL_Mapped_Pointer+position,
		     (unsigned int) 0,
		     (unsigned long *) &len_continuation) ;
       
  position+=(len_continuation/8) ; 

  continuation_bit_d((unsigned long *) IL_Num_El, 
		     (char *) XCDE_Console->XCDE_IL_Mapped_Pointer+position,
		     (unsigned int) 0,
		     (unsigned long *) &len_continuation) ;
       
  position+=(len_continuation/8) ; 

  (*IL)=(XCDE_IL_Couple_Type *) malloc((*IL_Num_El)*sizeof(XCDE_IL_Couple_Type)) ;

  delta_diff=0 ; 

  j=0 ; 

  escape=0 ; 

  while(j<(*IL_Num_El)) 
    
    { continuation_bit_d((unsigned long *) &read_el, 
			 (char *) XCDE_Console->XCDE_IL_Mapped_Pointer+position,
			 (unsigned int) 0,
			 (unsigned long *) &len_continuation) ;
       
    position+=(len_continuation/8) ; 

    read_el-=2 ; 

    if ((read_el==-1) && (!escape))
      
      { escape=1 ;
      
      continuation_bit_d((unsigned long *) &read_el, 
			 (char *) XCDE_Console->XCDE_IL_Mapped_Pointer+position,
			 (unsigned int) 0,
			 (unsigned long *) &len_continuation) ;
      
      position+=(len_continuation/8) ;

      delta_diff+=(read_el-2) ; 
      }

    else
      
      if ((read_el!=-1) && (!escape))

	{ delta_diff+=read_el ;
	
	(*IL)[j].XCDE_Position=delta_diff ; 

	(*IL)[j].XCDE_Offset=0 ; 

	j++ ; 
	}
    
      else
	
	if ((read_el==-1) && (escape))
	  
	  escape=0 ; 
	
	else
	  
	  if ((read_el!=-1) && (escape))
	  
	    { (*IL)[j].XCDE_Position=delta_diff ;
	  
	    (*IL)[j].XCDE_Offset=read_el ; 
	   
	    ++j ; 
	    }
    }
 
 return(0) ; 
}  

/*******************************************************************/

// function that reads the IL whit positions and proximities of the token with 
// the specified Rank.
// Example : ret=XCDE_Get_IL_ListPosProx(Rank, &IL, &IL_Num_El, &Console)
// where Rank is the rank of the element, &IL is the memory address of
// the variable that will contain the IL, &IL_Num_El is the memory address of
// the variable that will contain the number of elements of the IL and &Console
// is the memory address of the console data struct.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_IL_ListPosProx(int Rank,
			    XCDE_IL_Triple_Type **IL,
			    int *IL_Num_El,
			    XCDE_Console_Type *XCDE_Console) 
{ int num_pointer_IL_Index=0 ; 

  int num_IL_List=0 ; 

  int counter_num_IL_Block=0 ; 

  int number_byte_IL=0 ; 

  int position=0 ; 

  unsigned 
    long len_continuation=0 ; 

  unsigned char escape=0 ; 

  int delta_diff=0 ;

  int function_result=0 ; 

  unsigned char Dict=0 ; 

  int numElDict=0 ;

  register 
    int j=0 ; 

  unsigned 
    long int read_el=0 ; 

  /* ----------------------------------------- */
  /*            checks parameters              */
  /* ----------------------------------------- */

  if (XCDE_Console == NULL)

    return(-1) ; 

  if (XCDE_Console->XCDE_Console_Initialized != 1)

    return(-1) ; 

  if ((Rank < 0) || (Rank >= XCDE_Console->Huffman_Num_Tokens))

    return(-1) ; 

  if ((IL == NULL) || (IL_Num_El == NULL)) 

    return(-1) ; 

  if ((XCDE_Console->XCDE_Proximity) != 1) 

    return(-1) ; 

  // checks if the element is into the WORD dictionary. 

  // controlla che l'elemento faccia parte del dizionario 
  // delle parole, visto che e' stata richiesta la proximity 
  function_result=XCDE_Rank2Dict(Rank,
				 &Dict,
				 &numElDict,
				 XCDE_Console) ; 

  if (function_result == -1) 

    return(-1) ; 

  if (Dict != WORD)

    return(-1) ; 

  /* --------------- */ 
  /* accesing the IL */
  /* --------------- */

  num_pointer_IL_Index=((Rank)/(XCDE_IL_POINTER_RATIO)); 
  
  num_IL_List=((Rank)%(XCDE_IL_POINTER_RATIO)) ; 
  
  position=XCDE_Console->XCDE_IL_Table_Pointers[num_pointer_IL_Index] ; 

  while(counter_num_IL_Block<num_IL_List) 
  
    if (XCDE_Console->XCDE_IL_Mapped_Pointer[position] == 0) 
      
      { ++counter_num_IL_Block ; 
	
      ++position ; 
      }
  
    else
      
      {  continuation_bit_d((unsigned long *) &number_byte_IL, 
			    (char *) XCDE_Console->XCDE_IL_Mapped_Pointer+position,
			    (unsigned int) 0,
			    (unsigned long *) &len_continuation) ;
      
	++counter_num_IL_Block ; 
	
	position+=((len_continuation/8)+number_byte_IL) ; 
      }
  
  if (XCDE_Console->XCDE_IL_Mapped_Pointer[position] == 0) 
    
    { *IL=NULL ; 

    *IL_Num_El=0 ; 
    
    return(0) ; 
    }

  continuation_bit_d((unsigned long *) &number_byte_IL, 
		     (char *) XCDE_Console->XCDE_IL_Mapped_Pointer+position,
		     (unsigned int) 0,
		     (unsigned long *) &len_continuation) ;
  
  position+=(len_continuation/8) ; 
  
  continuation_bit_d((unsigned long *) IL_Num_El, 
		     (char *) XCDE_Console->XCDE_IL_Mapped_Pointer+position,
		     (unsigned int) 0,
		     (unsigned long *) &len_continuation) ;
  
  position+=(len_continuation/8) ; 
  
  (*IL)=(XCDE_IL_Triple_Type *) malloc((*IL_Num_El)*sizeof(XCDE_IL_Triple_Type)) ;

  delta_diff=0 ; 

  j=0 ; 

  escape=0 ; 

  while(j<(*IL_Num_El)) 
  
    {  continuation_bit_d((unsigned long *) &read_el, 
			  (char *) XCDE_Console->XCDE_IL_Mapped_Pointer+position,
			  (unsigned int) 0,
			  (unsigned long *) &len_continuation) ;
    
    position+=(len_continuation/8) ; 

    read_el-=2 ; 

    if ((read_el==-1) && (!escape))
      
      { escape=1 ;
      
      continuation_bit_d((unsigned long *) &read_el, 
			 (char *) XCDE_Console->XCDE_IL_Mapped_Pointer+position,
			 (unsigned int) 0,
			 (unsigned long *) &len_continuation) ;
    
      position+=(len_continuation/8) ; 
      
      delta_diff+=(read_el-2) ; 
      }

    else
      
      if ((read_el!=-1) && (!escape))

	{ delta_diff+=read_el ;
	
	(*IL)[j].XCDE_Position=delta_diff ; 

	(*IL)[j].XCDE_Offset=0 ; 

	j++ ; 
	}
    
      else
	
	if ((read_el==-1) && (escape))
	  
	  escape=0 ; 
	
	else
	  
	  if ((read_el!=-1) && (escape))
	  
	    { (*IL)[j].XCDE_Position=delta_diff ;
	  
	    (*IL)[j].XCDE_Offset=read_el ; 
	   
	    ++j ; 
	    }
    }

  delta_diff=0 ; 

  j=0 ; 

  while(j<(*IL_Num_El)) 
  
    { continuation_bit_d((unsigned long *) &read_el, 
			 (char *) XCDE_Console->XCDE_IL_Mapped_Pointer+position,
			 (unsigned int) 0,
			 (unsigned long *) &len_continuation) ;
    
    position+=(len_continuation/8) ; 

    read_el-=1 ; 

    delta_diff+=read_el ; 

    (*IL)[j].XCDE_Proximity=delta_diff ; 

    ++j ; 
    }
 return(1) ; 
}  

/*******************************************************************/

// internal function used to load a body zlib block 
int XCDE_LoadCacheBlock(const int position, 
			const char *source,
			char **destination, 
			int *destination_Dcmp_len,
			int *destination_Cmp_len)
{ int number_byte=0 ; 

  unsigned long 
    int len=0ul ;  

  int uncmp_err=0 ; 

  int pos=0 ; 

  number_byte=(XCDE_COMPRESS_BLOCK_SIZE) ; 
  
  pos+=(len/8) ; 
  
  if ((*destination_Dcmp_len) < number_byte)
  
    { if (destination != NULL) free(*destination) ; 

      *destination=malloc(number_byte*sizeof(char)*2) ; 

      if (*destination == NULL)

	return(-1) ; 
    }

  uncmp_err=uncompress((Bytef *) *destination, 
		       (uLong *) &number_byte, 
		       (Bytef *) source+pos+position, 
		       (uLong) *destination_Cmp_len) ; 
  
  if (uncmp_err != Z_OK) 
    
    { free(*destination) ; 
    
    fprintf(stderr, "\nErrore Decompressione") ; 

    return(-1) ; 
    }

  // number of decompressed characters 
  *destination_Dcmp_len=number_byte ; 

  return(1) ; 
}

/*******************************************************************/

// function that traslates the codeword to its rank.
// Example : ret=XCDE_Codeword2Rank(cw, cwlen, &Rank, &Console)
// where cw is the codeword, cwlen is its length, &Rank is the 
// memory address of the variable that will contain the Rank and
// &Console is the memory address of the console data struct.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Codeword2Rank(int Codeword,
		       int Cwlen, 
		       int *Rank,
		       XCDE_Console_Type *XCDE_Console)
{ /* ----------------------------------------- */
  /*            checks parameters              */
  /* ----------------------------------------- */
  
  if (XCDE_Console == NULL)

    return(-1) ; 

  if (XCDE_Console->XCDE_Console_Initialized != 1)

    return(-1) ; 

  if (Rank == NULL) 

    return(-1) ; 

  if ((Cwlen <= 0) || (Cwlen > XCDE_Console->Huffman_Codeword_Max_Len))  

    return(-1) ; 

  /* ------------------------------------------- */
  /* computes the rank in the general dictionary */
  /* ------------------------------------------- */

  *Rank=XCDE_Console->Huffman_Offsetcw[Cwlen]+Codeword-XCDE_Console->Huffman_Firstcw[Cwlen] ; 

  return(1) ; 
}

/*******************************************************************/

// function that traslates the codeword to the token corresponding it. 
// Example : ret=XCDE_Codeword2Token(cw, cwlen, &Token, &Console) 
// where cw is the codeword to traslate, cwlen is its length, &Token
// is the memory address of the variable that will contain the token
// and &Console is the memory address of the console data struct.
// The function allocates the memory for the token.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Codeword2Token(int Codeword,
			int Cwlen,
			char **Token,
			XCDE_Console_Type *XCDE_Console) 
{ int function_result=0 ; 
  
  int rank=0 ; 

  unsigned 
    char dict=0u ; 

  int ndictelem=0 ; 

  /* ----------------------------------------- */
  /*            checks parameters              */
  /* ----------------------------------------- */

  if (XCDE_Console == NULL)

    return(-1) ; 

  if (XCDE_Console->XCDE_Console_Initialized != 1)

    return(-1) ; 

  if (Token == NULL)

    return(-1) ; 

  if ((Cwlen <= 0) || (Cwlen > XCDE_Console->Huffman_Codeword_Max_Len)) 

    return(-1) ; 
  
  /* ------------------------------ */
  /* computes the rank of the token */
  /* ------------------------------ */

  rank=XCDE_Console->Huffman_Offsetcw[Cwlen]+Codeword-XCDE_Console->Huffman_Firstcw[Cwlen] ; 

  /* ---------------------------------------------------- */ 
  /* computes the dictionary where the token is contained */
  /* ---------------------------------------------------- */ 

  function_result=XCDE_Rank2Dict(rank,
				 &dict,
				 &ndictelem,
				 XCDE_Console) ; 
  
  if (function_result == -1) 

    return(-1) ; 

  /* --------------------------------------------------------------- */ 
  /* checks if the dictionary is compressed and loaded in the memory */
  /* --------------------------------------------------------------- */
  
  if (!XCDE_Console->XCDE_Dict_File_Decompressed[dict])

    { function_result=XCDE_Decompress_Dictionary(dict,
						 XCDE_Console) ; 

    if (function_result == -1) 

      return(-1) ; 
    }
  
  if (!XCDE_Console->XCDE_Dict_File_Loaded[dict])

    { function_result=XCDE_Load_Dictionary(dict,
					   XCDE_Console) ; 

    if (function_result == -1)

      return(-1) ; 
    }

  /* ------------------- */ 
  /* accessing the token */
  /* ------------------- */ 

  *Token=malloc(strlen(XCDE_Console->Huffman_Position_Token[rank])+1) ;

  if ((strlen(XCDE_Console->Huffman_Position_Token[rank]) > 0) && (*Token == NULL))

    return(-1) ; 

  strcpy(*Token, 
	 XCDE_Console->Huffman_Position_Token[rank]) ; 

  (*Token)[strlen(XCDE_Console->Huffman_Position_Token[rank])]='\0' ;  

  return(1) ; 
}

/*******************************************************************/

// function that traslates a rank to the token corresponding it. 
// Example : ret=XCDE_Rank2Token(rank, &Token, &Console) 
// where rank is the rank of the token, &Token is the memory address 
// of the variable that will contain it and &Console is the memory 
// address of the console data struct.  
// The function allocates the memory for the token.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Rank2Token(int Rank,
		    char **Token,
		    XCDE_Console_Type *XCDE_Console) 
{ int function_result=0 ; 
  
  unsigned 
    char dict=0u ; 

  int ndictelem=0 ; 

  /* ----------------------------------------- */
  /*            checks parameters              */
  /* ----------------------------------------- */

  if (XCDE_Console == NULL)

    return(-1) ; 

  if (XCDE_Console->XCDE_Console_Initialized != 1)

    return(-1) ; 

   if (Token == NULL)

     return(-1) ; 

   if ((Rank < 0) || (Rank >= XCDE_Console->Huffman_Num_Tokens))

     return(-1) ; 

   /* ---------------------------------------------------- */
   /* computes the dictionary where the token is contained */
   /* ---------------------------------------------------- */

   function_result=XCDE_Rank2Dict(Rank,
				  &dict,
				  &ndictelem,
				  XCDE_Console) ; 

   if (function_result == -1) 

     return(-1) ; 

   /* ----------------------------------------------------------------- */
   /* checks if the dictionary is decompressed and loaded in the memory */
   /* ----------------------------------------------------------------- */

   if (!XCDE_Console->XCDE_Dict_File_Decompressed[dict])

     { function_result=XCDE_Decompress_Dictionary(dict,
						  XCDE_Console) ; 
     
     if (function_result == -1) 

       return(-1) ; 
     }

   if (!XCDE_Console->XCDE_Dict_File_Loaded[dict])

     { function_result=XCDE_Load_Dictionary(dict,
					    XCDE_Console) ; 

     if (function_result == -1)

       return(-1) ; 
     }  

   /* -------------------------------- */
   /* acceding the corresponding token */
   /* -------------------------------- */

   *Token=malloc(strlen(XCDE_Console->Huffman_Position_Token[Rank])+1) ;

   if ((strlen(XCDE_Console->Huffman_Position_Token[Rank]) > 0) && (*Token == NULL))

     return(-1) ; 

   strcpy(*Token, 
	  XCDE_Console->Huffman_Position_Token[Rank]) ; 

   return(1) ; 
 }

/*******************************************************************/

// function that reads the codeword that begins at the specified position.
// Example : ret=XCDE_Get_Body_Forward_Codeword(position, &Cw, &Cwlen, &Console)
// where position is the position where the reading begins, &Cw is the memory
// address of the variable that will contain the codeword, &Cwlen is the
// memory address of the variable that will contain the codeword length and 
// &Console is the memory address of the console data struct.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_Body_Forward_Codeword(int Position,
				   int *Codeword,
				   int *Cwlen,
				   XCDE_Console_Type *XCDE_Console) 
 { register
     int k=0 ;

   int number_body_block=0 ; 

   int offset_body_block=0 ; 

   unsigned 
     char done=0u ; 

   int function_result=0 ; 

   /* ----------------------------------------- */
   /*            checks parameters              */
   /* ----------------------------------------- */

   if (XCDE_Console == NULL)

     return(-1) ; 

   if (XCDE_Console->XCDE_Console_Initialized != 1)

     return(-1) ; 

   if ((Codeword == NULL) || (Cwlen == NULL))

     return(-1) ; 

   if ((Position < 0) || (Position >= XCDE_Console->XCDE_Body_Size_Without_Zlib))

     return(-1) ;

   /* ---------------------------- */ 
   /* acceding the body zlib block */ 
   /* ---------------------------- */ 

   number_body_block=(Position/(XCDE_COMPRESS_BLOCK_SIZE)) ; 

   if (number_body_block >= XCDE_Console->XCDE_Body_Num_Block) 

     return(-1) ;

   offset_body_block=(Position % (XCDE_COMPRESS_BLOCK_SIZE)) ; 

   // checks if the cache block is loaded in the memory
   if (XCDE_Console->XCDE_Body_Cache_Which_Block != number_body_block)

     { if (number_body_block < XCDE_Console->XCDE_Body_Num_Block-1)

       XCDE_Console->XCDE_Body_Cache_Cmp_Block_Size=XCDE_Console->XCDE_Body_Table_Pointers[number_body_block+1] - XCDE_Console->XCDE_Body_Table_Pointers[number_body_block] ; 

     else

       XCDE_Console->XCDE_Body_Cache_Cmp_Block_Size=XCDE_Console->XCDE_Body_Filesize - XCDE_Console->XCDE_Body_Table_Pointers[number_body_block] ; 

     function_result=XCDE_LoadCacheBlock(XCDE_Console->XCDE_Body_Table_Pointers[number_body_block],
					 XCDE_Console->XCDE_Body_Mapped_Pointer,
					 &XCDE_Console->XCDE_Body_Cache_Block,
					 &XCDE_Console->XCDE_Body_Cache_Dcmp_Block_Size,
					 &XCDE_Console->XCDE_Body_Cache_Cmp_Block_Size) ; 

     if (function_result == -1) 

       return(-1) ; 

     XCDE_Console->XCDE_Body_Cache_Which_Block=number_body_block ; 

     XCDE_Console->XCDE_Body_Cache_Block_File_Position=XCDE_Console->XCDE_Body_Table_Pointers[number_body_block] ; 
     }
   
   /* ------------------------------- */ 
   /* acceding the requested codeword */ 
   /* ------------------------------- */ 

   if (offset_body_block >= XCDE_Console->XCDE_Body_Cache_Dcmp_Block_Size)

     return(-1) ; 

   if ((XCDE_Console->XCDE_Body_Cache_Block[offset_body_block] & (0x80)) != (0x80))   
     return(-1) ; 

   *Codeword = (XCDE_Console->XCDE_Body_Cache_Block[offset_body_block] & (0x7F)) ;   
   ++offset_body_block ; 

   *Cwlen=1 ;   

   while(!done) 

     { if (offset_body_block >= XCDE_Console->XCDE_Body_Cache_Dcmp_Block_Size)

	 { ++number_body_block ; 

	 if ((number_body_block == XCDE_Console->XCDE_Body_Num_Block) && (offset_body_block == XCDE_Console->XCDE_Body_Cache_Dcmp_Block_Size)) 

	   { done=1u ; 
	     
	   return(1) ; 
	   }

	 else

	   if (number_body_block > XCDE_Console->XCDE_Body_Num_Block)

	     return(-1) ; 

	   else

	     if (number_body_block < XCDE_Console->XCDE_Body_Num_Block-1)

	       XCDE_Console->XCDE_Body_Cache_Cmp_Block_Size=XCDE_Console->XCDE_Body_Table_Pointers[number_body_block+1] - XCDE_Console->XCDE_Body_Table_Pointers[number_body_block] ; 

	     else

	       XCDE_Console->XCDE_Body_Cache_Cmp_Block_Size=XCDE_Console->XCDE_Body_Filesize - XCDE_Console->XCDE_Body_Table_Pointers[number_body_block] ; 
	 
	 function_result=XCDE_LoadCacheBlock(XCDE_Console->XCDE_Body_Table_Pointers[number_body_block],
					     XCDE_Console->XCDE_Body_Mapped_Pointer,
					     &XCDE_Console->XCDE_Body_Cache_Block,
					     &XCDE_Console->XCDE_Body_Cache_Dcmp_Block_Size,
					     &XCDE_Console->XCDE_Body_Cache_Cmp_Block_Size) ; 

	 if (function_result ==-1) 

	   return(-1) ;
	   
	 XCDE_Console->XCDE_Body_Cache_Which_Block=number_body_block ; 
	 
	 XCDE_Console->XCDE_Body_Cache_Block_File_Position=XCDE_Console->XCDE_Body_Table_Pointers[number_body_block] ;       
	 
	 offset_body_block=0 ; 
	 }

     if ((XCDE_Console->XCDE_Body_Cache_Block[offset_body_block] & (0x80)) != 0) 

       done=1u ; 

     else

       { *Codeword = (((*Codeword) << 7) | (XCDE_Console->XCDE_Body_Cache_Block[offset_body_block] & (0x7F))) ;

       ++(*Cwlen) ; 

       ++offset_body_block ; 
       }
     }
   
   return(1) ; 
 }

/*******************************************************************/

// function that reads the previous codeword at the specified position.
// Example : ret=XCDE_Get_Body_Backward_Codeword(position, &Cw, &Cwlen, &Console)
// where position is the specified position, &Cw is the memory address of the
// variable that will contain the codeword, &Cwlen is the memory address of the
// variable that will contain the codeword length and &Console is the memory
// address of the console data structure.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_Body_Backward_Codeword(int Position,
				    int *Codeword,
				    int *Cwlen,
				    XCDE_Console_Type *XCDE_Console) 
{ int number_body_block=0 ; 

 int offset_body_block=0 ; 

 unsigned 
   char done=0u ; 

 int Mask=0 ; 

 int function_result=0 ; 

 /* ----------------------------------------- */
 /*            checks parameters              */
 /* ----------------------------------------- */

 if (XCDE_Console == NULL)

   return(-1) ; 

 if (XCDE_Console->XCDE_Console_Initialized != 1)

   return(-1) ; 

 if ((Codeword == NULL) || (Cwlen == NULL))

   return(-1) ; 

 // the Position may be equal to XCDE_Body_Size_Without_Zlib, becaure the 
 // function reads in backward direction 
 if ((Position < 0) || (Position > XCDE_Console->XCDE_Body_Size_Without_Zlib))

   return(-1) ; 
 
 /* --------------------------------------------- */ 
 /* acceding the cache block of the next codeword */ 
 /* --------------------------------------------- */ 

 number_body_block=(Position/(XCDE_COMPRESS_BLOCK_SIZE)) ; 

 if (number_body_block >= XCDE_Console->XCDE_Body_Num_Block) 

   return(-1) ;

 offset_body_block=(Position % (XCDE_COMPRESS_BLOCK_SIZE)) ; 

 if (XCDE_Console->XCDE_Body_Cache_Which_Block != number_body_block)

   { if (number_body_block < XCDE_Console->XCDE_Body_Num_Block-1)

     XCDE_Console->XCDE_Body_Cache_Cmp_Block_Size=XCDE_Console->XCDE_Body_Table_Pointers[number_body_block+1] - XCDE_Console->XCDE_Body_Table_Pointers[number_body_block] ; 

   else

     XCDE_Console->XCDE_Body_Cache_Cmp_Block_Size=XCDE_Console->XCDE_Body_Filesize - XCDE_Console->XCDE_Body_Table_Pointers[number_body_block] ; 

   function_result=XCDE_LoadCacheBlock(XCDE_Console->XCDE_Body_Table_Pointers[number_body_block],
				       XCDE_Console->XCDE_Body_Mapped_Pointer,
				       &XCDE_Console->XCDE_Body_Cache_Block,
				       &XCDE_Console->XCDE_Body_Cache_Dcmp_Block_Size,
				       &XCDE_Console->XCDE_Body_Cache_Cmp_Block_Size) ; 

   if (function_result == -1) 

     return(-1) ; 

   XCDE_Console->XCDE_Body_Cache_Which_Block=number_body_block ; 

   XCDE_Console->XCDE_Body_Cache_Block_File_Position=XCDE_Console->XCDE_Body_Table_Pointers[number_body_block] ; 
   }
 
 /* ------------------------------- */ 
 /* acceding the requested codeword */
 /* ------------------------------- */ 
 
 if (offset_body_block >= XCDE_Console->XCDE_Body_Cache_Dcmp_Block_Size)
    
   return(-1) ; 

 if (!(XCDE_Console->XCDE_Body_Cache_Block[offset_body_block] & 0x80))
    
   return(-1) ; 
  
 --offset_body_block ; 

 if (offset_body_block < 0)
    
   { --number_body_block ; 
	  
   if (number_body_block < 0) 
	    
     { done=1u ; 
	    
     return(1) ; 
     }
   
   else
	    
     if (number_body_block >= XCDE_Console->XCDE_Body_Num_Block)

       return(-1) ; 

     else

       if (number_body_block < XCDE_Console->XCDE_Body_Num_Block-1)

	 XCDE_Console->XCDE_Body_Cache_Cmp_Block_Size=XCDE_Console->XCDE_Body_Table_Pointers[number_body_block+1] - XCDE_Console->XCDE_Body_Table_Pointers[number_body_block] ; 
      
       else

	 XCDE_Console->XCDE_Body_Cache_Cmp_Block_Size=XCDE_Console->XCDE_Body_Filesize - XCDE_Console->XCDE_Body_Table_Pointers[number_body_block] ; 
      
   function_result=XCDE_LoadCacheBlock(XCDE_Console->XCDE_Body_Table_Pointers[number_body_block],
				       XCDE_Console->XCDE_Body_Mapped_Pointer,
				       &XCDE_Console->XCDE_Body_Cache_Block,
				       &XCDE_Console->XCDE_Body_Cache_Dcmp_Block_Size,
				       &XCDE_Console->XCDE_Body_Cache_Cmp_Block_Size) ; 
      
   XCDE_Console->XCDE_Body_Cache_Which_Block=number_body_block ; 

   XCDE_Console->XCDE_Body_Cache_Block_File_Position=XCDE_Console->XCDE_Body_Table_Pointers[number_body_block] ;       
	
   offset_body_block=XCDE_Console->XCDE_Body_Cache_Dcmp_Block_Size-1 ; 
   }

 (*Codeword)=(XCDE_Console->XCDE_Body_Cache_Block[offset_body_block] & 0x7F);   

 (*Cwlen)=1 ; 

 if (XCDE_Console->XCDE_Body_Cache_Block[offset_body_block] & 0x80)

   done=1u ; 

 --offset_body_block ; 

 while(!done) 

   { if (offset_body_block < 0)
   
     { --number_body_block ; 
     
     if (number_body_block < 0) 

       { done=1u ; 
	    
       return(1) ; 
       }
	  
     else
	    
       if (number_body_block >= XCDE_Console->XCDE_Body_Num_Block)

	 return(-1) ; 

       else

	 if (number_body_block < XCDE_Console->XCDE_Body_Num_Block-1)

	   XCDE_Console->XCDE_Body_Cache_Cmp_Block_Size=XCDE_Console->XCDE_Body_Table_Pointers[number_body_block+1] - XCDE_Console->XCDE_Body_Table_Pointers[number_body_block] ; 
      
	 else

	   XCDE_Console->XCDE_Body_Cache_Cmp_Block_Size=XCDE_Console->XCDE_Body_Filesize - XCDE_Console->XCDE_Body_Table_Pointers[number_body_block] ; 
      
     function_result=XCDE_LoadCacheBlock(XCDE_Console->XCDE_Body_Table_Pointers[number_body_block],
					 XCDE_Console->XCDE_Body_Mapped_Pointer,
					 &XCDE_Console->XCDE_Body_Cache_Block,
					 &XCDE_Console->XCDE_Body_Cache_Dcmp_Block_Size,
					 &XCDE_Console->XCDE_Body_Cache_Cmp_Block_Size) ;       

     XCDE_Console->XCDE_Body_Cache_Which_Block=number_body_block ; 

     XCDE_Console->XCDE_Body_Cache_Block_File_Position=XCDE_Console->XCDE_Body_Table_Pointers[number_body_block] ;       
	
     offset_body_block=XCDE_Console->XCDE_Body_Cache_Dcmp_Block_Size-1 ; 
     }

   Mask=(XCDE_Console->XCDE_Body_Cache_Block[offset_body_block] & 0x7F) ; 

   Mask=(Mask << (7*(*Cwlen))) ; 

   (*Codeword)=((*Codeword) | Mask) ; 

   ++(*Cwlen) ; 

   if (XCDE_Console->XCDE_Body_Cache_Block[offset_body_block] & 0x80)

     done=1u ; 

   else
	    
     --offset_body_block ; 
   }
 
  return(1) ; 
}

/*******************************************************************/

// function that reads a text window. The reading begins at position and
// goes forward and backward until the function passed as parameter is 
// true Count times backward and forward.
// Example : ret=XCDE_Get_Body_TextWindow(position, &Window, &func, &Console) ;
// where position is the position the extraction begin, &Window is the 
// memory address where the window is copied, &func is the memory address of
// the function used to check the codeword (gets as parameters the codeword 
// read and its length), Count is the number of times the function must be 
// true.
// The function allocates the memory for the window.
// ret value is 1 if there aren't errors, or -1 otherwise. 
int XCDE_Get_Body_TextWindow(int Position, 
			     char **Window,
			     int Count, 
			     unsigned char direction,
			     int *backward_new_position,
			     int *forward_new_position,
			     int (*Counter_function)(int cw, int cwlen, XCDE_Console_Type *XCDE_Console), 
			     XCDE_Console_Type *XCDE_Console)
{ 
#define XCDE_F_B_INITIAL_WINDOW_DIMENSION 1000 
 
 register 
   int k=0 ; 
     
 char **forward_window=NULL ; 

 int forward_window_dimension=0 ; 

 int forward_window_pointer=0 ; 
 
 int forward_body_position=0 ; 
 
 char **backward_window=NULL ; 
 
 int backward_window_dimension=0 ;
 
 int backward_window_pointer=0 ; 
 
 int backward_body_position=0 ; 
 
 int codeword=0 ; 
 
 int codeword_len=0 ; 
 
 int window_counter=0 ; 
 
 int window_text_len=0 ; 

 int window_pointer=0 ; 
 
 char *token=0 ; 
 
 int function_result=0 ; 

 /* ----------------------------------------- */
 /*            checks parameters              */
 /* ----------------------------------------- */

 if (XCDE_Console == NULL)

    return(-1) ; 
 
 if (XCDE_Console->XCDE_Console_Initialized != 1)
    
   return(-1) ; 
  
 if (Count <= 0) 

   return(-1) ; 

 if ((Position < 0) || (Position > XCDE_Console->XCDE_Body_Size_Without_Zlib))
    
   return(-1) ; 
  
 if (Window == NULL) 

   return(-1) ; 

 if ((direction != XCDE_BACKWARD) && (direction != XCDE_FORWARD) && (direction != XCDE_BACKWARD_FORWARD))

   return(-1) ;

 if (backward_new_position == NULL)

   return(-1) ;

 if (forward_new_position == NULL)

   return(-1) ;

 /* ------------------------ */ 
 /* initializes data structs */
 /* ------------------------ */ 
 
 forward_window_dimension = backward_window_dimension = XCDE_F_B_INITIAL_WINDOW_DIMENSION ;

 forward_window = (char **) malloc(sizeof(char *)*forward_window_dimension) ; 

 if (forward_window == NULL)

   return(-1) ; 

 backward_window = (char **) malloc(sizeof(char *)*backward_window_dimension) ; 

 if (backward_window == NULL)

   { free(forward_window) ;

   return(-1) ; 
   }

 forward_window_pointer = backward_window_pointer = 0 ; 
  
 backward_body_position = forward_body_position = Position ; 

 window_counter=0 ; 

 /* ------------------------------- */ 
 /* reads window backward codewords */ 
 /* ------------------------------- */ 

 if ((direction == XCDE_BACKWARD) || (direction == XCDE_BACKWARD_FORWARD))

   while ((window_counter <= Count) && (backward_body_position > 0))

     { function_result=XCDE_Get_Body_Backward_Codeword(backward_body_position, 
						       &codeword,
						       &codeword_len,
						       XCDE_Console) ; 

     if (function_result == -1)
     
       { for(k=0 ; k < backward_window_pointer; ++k)
	 
	 free(backward_window[k]) ;
       
       free(backward_window) ; 
       
       free(forward_window) ; 
       
	return(-1) ; 
       }
     
     function_result=XCDE_Codeword2Token(codeword,
					 codeword_len,
					 &token,
					 XCDE_Console) ; 
    
     if (function_result == -1)

       { for(k=0 ; k < backward_window_pointer; ++k)
	 
	 free(backward_window[k]) ;
       
       free(backward_window) ; 
       
       free(forward_window) ; 
	
       return(-1) ; 
       }
    
     backward_body_position-=codeword_len ; 
     
     window_text_len+=strlen(token) ;
     
     fflush(stdout) ;

     backward_window[backward_window_pointer]=token ; 
     
     ++backward_window_pointer ; 
     
     if (backward_window_pointer > backward_window_dimension)
       
       { backward_window_dimension*=2 ; 
       
       backward_window = (char **) realloc(backward_window, 
					   backward_window_dimension) ; 
       }
     
     if (Counter_function(codeword,
			  codeword_len,
			  XCDE_Console) > 0)
       
       ++window_counter ; 
     }

 *backward_new_position=backward_body_position ;
 
 window_counter=0 ; 

 /* ------------------------------ */ 
 /* reads window forward codewords */ 
 /* ------------------------------ */ 

 if ((direction == XCDE_FORWARD) || (direction == XCDE_BACKWARD_FORWARD))

   while ((window_counter <= Count) && (forward_body_position < (XCDE_Console->XCDE_Body_Size_Without_Zlib)))
     
     { function_result=XCDE_Get_Body_Forward_Codeword(forward_body_position, 
						      &codeword,
						      &codeword_len,
						      XCDE_Console) ; 
     
     if (function_result == -1)
       
       { for(k=0 ; k < backward_window_pointer; ++k)
	 
	 free(backward_window[k]) ;
       
       for(k=0 ; k < forward_window_pointer; ++k)
	 
	 free(forward_window[k]) ;
       
       free(backward_window) ; 
       
       free(forward_window) ; 
       
       return(-1) ; 
       }
     
     function_result=XCDE_Codeword2Token(codeword,
					 codeword_len,
					 &token,
					 XCDE_Console) ; 
     
     if (function_result == -1)
       
       { for(k=0 ; k < backward_window_pointer; ++k)
	 
	 free(backward_window[k]) ;
       
       for(k=0 ; k < forward_window_pointer; ++k)
	 
	 free(forward_window[k]) ;
       
       free(backward_window) ; 
       
       free(forward_window) ; 
       
       return(-1) ; 
       }

     forward_body_position+=codeword_len ; 
     
     window_text_len+=strlen(token) ;
     
     forward_window[forward_window_pointer]=token ; 
     
     ++forward_window_pointer ; 
     
     if (forward_window_pointer > forward_window_dimension)
       
       { forward_window_dimension*=2 ; 
       
       forward_window = (char **) realloc(forward_window, 
					  forward_window_dimension) ; 
       }
     
     if (Counter_function(codeword,
			  codeword_len,
			  XCDE_Console) > 0)
       
       ++window_counter ; 
     }

 *forward_new_position=forward_body_position ;
 
 /* -------------------------- */ 
 /* builds the complete window */
 /* -------------------------- */ 
 
 *Window = (char *) malloc(window_text_len*sizeof(char)+1) ; 
 
 window_pointer ; 
 
 for(k=backward_window_pointer-1 ; k>=0; --k)
   
   { strcpy((*Window)+window_pointer, 
	    backward_window[k]) ; 
   
   window_pointer+=strlen(backward_window[k]) ; 
   
   free(backward_window[k]); 
   }
 
 free(backward_window) ; 
 
 for(k=0 ; k<forward_window_pointer; ++k) 
   
   { strcpy((*Window)+window_pointer,
	    forward_window[k]) ; 
   
   window_pointer+=strlen(forward_window[k]) ; 

   free(forward_window[k]) ; 
   }
 
 free(forward_window) ; 

#undef XCDE_F_B_INITIAL_WINDOW_DIMENSION 
 
 return(1) ; 
}

/*******************************************************************/

// functions that reads the tagstruct flag, that is on if the
// tagstruct file is present, and of otherwise. 
// Example : ret = XCDE_SettedTagstruct(&TagstructFlag, &Console) 
// where &TagstructFlag is the memory address of the variable that
// will contain the tagstruct flag, and &Console is the memory 
// address of the console data struct.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_SettedTagstruct(unsigned char *TagstructFlag,
			 XCDE_Console_Type *XCDE_Console) 
{ 
  /* ----------------------------------------- */
  /*            checks parameters              */
  /* ----------------------------------------- */    

  if (XCDE_Console == NULL)

    return(-1) ; 

  if (XCDE_Console->XCDE_Console_Initialized != 1)
    
    return(-1) ; 

  if (TagstructFlag == NULL) 

    return(-1) ; 
  
  /* ------------------------ */
  /* reads the tagstruct flag */ 
  /* ------------------------ */ 

  *TagstructFlag=XCDE_Console->XCDE_Tagstruct ; 

  return(1) ; 
}


/*******************************************************************/

// functions that reads the size of the tagstruct file (returns
// -1 if itsn't present).
// Example : ret=XCDE_Get_Tagstruct_Filesize(&Size, &Console) 
// where &size is the memory address of the variable that will
// contain the tagstruct file size, and &Console is the memory 
// address of the console data struct.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_Tagstruct_Filesize(int *Tagstruct_Filesize,
				XCDE_Console_Type *XCDE_Console) 
{ 
  /* ----------------------------------------- */
  /*            checks parameters              */
  /* ----------------------------------------- */
  
  if (XCDE_Console == NULL)
    
    return(-1) ; 

  if (XCDE_Console->XCDE_Console_Initialized != 1)

    return(-1) ; 

  if (XCDE_Console->XCDE_Tagstruct != 1) 

    return(-1) ; 

  if (Tagstruct_Filesize == NULL)

    return(-1) ; 

  /* ----------------------------- */ 
  /* reads the tagstruct file size */      
  /* ----------------------------- */

  *Tagstruct_Filesize=XCDE_Console->XCDE_Tagstruct_Filesize ; 

  return(1) ; 
}

/*******************************************************************/

// functions that reads the tagstruct file name (returns -1 
// if theresn't the tagstruct file). 
// Example : ret=XCDE_Get_Tagstruct_Filename(&TagstructFilename, &Console)
// where &TagstructFilename is the memory address of the variable that 
// will contain the file name, and &Console is the memory address of 
// the console data structure.
// The function allocates the memory needed by the name.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_Tagstruct_Filename(char **Tagstruct_Filename, 
				XCDE_Console_Type *XCDE_Console) 
{ 
  /* ----------------------------------------- */
  /*            checks parameters              */
  /* ----------------------------------------- */
  
  if (XCDE_Console == NULL)
    
    return(-1) ; 

  if (XCDE_Console->XCDE_Console_Initialized != 1)

    return(-1) ; 

  if (XCDE_Console->XCDE_Tagstruct != 1) 

    return(-1) ; 

  if (Tagstruct_Filename == NULL)

    return(-1) ; 

  /* ----------------------------- */
  /* reads the tagstruct file name */ 
  /* ----------------------------- */ 

  *Tagstruct_Filename=malloc(strlen(XCDE_Console->XCDE_Tagstruct_File_Name)+1) ; 

  if (*Tagstruct_Filename==NULL) 

    return(-1) ; 

  strcpy(*Tagstruct_Filename, 
	 XCDE_Console->XCDE_Tagstruct_File_Name) ; 

  return(1) ; 
}		

/*******************************************************************/

// function that accesses the tagstruct file and computes the tags 
// that contain the position specified (return -1 is theresn't the 
// tagstruct file).
// Example : ret=XCDE_Get_Tagstruct_Tags_Enclosing(&point, &Tags, &Num_tags, &Console)
// where point is the position (couple physical position, offset), &Tags is the
// memory address of the variable that will contain the array of tags, 
// &Num_tags is the numbers of tags that are in the array and &Console is the
// memory address of the console data struct.
// The function allocates the memory needed by the array of tags 
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_Tagstruct_Tags_Enclosing(XCDE_IL_Couple_Type *Stab_point,
				      int **Enclosing_tags_ranks,
				      int *Num_enclosing_tags,
				      XCDE_Console_Type *XCDE_Console)
{ 
#define XCDE_INITIAL_STACK_DIMENSION 1000

  register 
    int i=0 ; 
  
  register 
    int j=0 ; 

  uchar *Virtual_pointer ;

  static
    int position ;

  int dcmp_Len=0 ;

  char *pointer = NULL ; 
  
  int *stack;

  int stack_size=0;

  int stack_dimension=XCDE_INITIAL_STACK_DIMENSION; 

  int phis_tag=0 ; 

  int displ_tag=0 ;

  int len_cont=0 ;

  int uncmp_err=0 ;

  int tag_id=0 ;

  int offset=0 ;

  int read_el=0 ;

  /* ----------------------------------------- */
  /*            checks parameters              */
  /* ----------------------------------------- */

  if (XCDE_Console == NULL)

    return(-1) ;

  if (XCDE_Console->XCDE_Console_Initialized != 1)

    return(-1) ;

  if (XCDE_Console->XCDE_Tagstruct != 1)

    return(-1) ;

  if (Stab_point == NULL)

    return(-1) ;
  /* ----------------------------------------- */
  /*            checks parameters              */
  /* ----------------------------------------- */

  if (XCDE_Console == NULL)

    return(-1) ;

  if (XCDE_Console->XCDE_Console_Initialized != 1)

    return(-1) ;

  if (XCDE_Console->XCDE_Tagstruct != 1)

    return(-1) ;

  if (Stab_point == NULL)

    return(-1) ;

  if (Num_enclosing_tags == NULL)

    return(-1) ;

  /* --------------------- */
  /* initializes the stack */
  /* --------------------- */

  stack=malloc(stack_dimension*sizeof(int)) ;

  /* -------------------------- */
  /* loads the tagstruct header */
  /* -------------------------- */

  if (XCDE_Console->XCDE_Tagstruct_Windows_Information == NULL)

    { position=0 ;

    continuation_bit_d((unsigned long *) &XCDE_Console->XCDE_Tagstruct_Number_Windows,
		       (char *) XCDE_Console->XCDE_Tagstruct_Mapped_Pointer+position,
		       (unsigned int) 0,
		       (unsigned long *) &len_cont) ;

    position+=(len_cont/8) ;

    XCDE_Console->XCDE_Tagstruct_Windows_Information = (XCDE_Tagstruct_Windows_Information_Type *) malloc(sizeof(XCDE_Tagstruct_Windows_Information_Type)*(XCDE_Console->XCDE_Tagstruct_Number_Windows)) ;

    if (XCDE_Console->XCDE_Tagstruct_Windows_Information == NULL)

      return(-1) ;

    for(i=0;i<XCDE_Console->XCDE_Tagstruct_Number_Windows;i++)

      { continuation_bit_d((unsigned long *) &XCDE_Console->XCDE_Tagstruct_Windows_Information[i].XCDE_Tagstruct_Window_Size,
			   (char *) XCDE_Console->XCDE_Tagstruct_Mapped_Pointer+position,
			   (unsigned int) 0,
			   (unsigned long *) &len_cont) ;

      position+=(len_cont/8) ;

      continuation_bit_d((unsigned long *) &XCDE_Console->XCDE_Tagstruct_Windows_Information[i].XCDE_Tagstruct_Window_Byte_Displacement,
			 (char *) XCDE_Console->XCDE_Tagstruct_Mapped_Pointer+position,
			 (unsigned int) 0,
			 (unsigned long *) &len_cont) ;

      position+=(len_cont/8) ;

      continuation_bit_d((unsigned long *) &XCDE_Console->XCDE_Tagstruct_Windows_Information[i].XCDE_Tagstruct_Window_Start,
			 (char *) XCDE_Console->XCDE_Tagstruct_Mapped_Pointer+position,
			 (unsigned int) 0,
			 (unsigned long *) &len_cont) ;

      position+=(len_cont/8) ;

      continuation_bit_d((unsigned long *) &XCDE_Console->XCDE_Tagstruct_Windows_Information[i].XCDE_Tagstruct_Window_Displ,
			 (char *) XCDE_Console->XCDE_Tagstruct_Mapped_Pointer+position,
			 (unsigned int) 0,
			 (unsigned long *) &len_cont) ;

      position+=(len_cont/8) ;
      }
    }

  /* ------------- */
  /* searches tags */
  /* ------------- */

  for(i=0 ; i<XCDE_Console->XCDE_Tagstruct_Number_Windows-1; ++i)

    { if ((Stab_point->XCDE_Position < XCDE_Console->XCDE_Tagstruct_Windows_Information[i+1].XCDE_Tagstruct_Window_Start) ||
	  ((Stab_point->XCDE_Position == XCDE_Console->XCDE_Tagstruct_Windows_Information[i+1].XCDE_Tagstruct_Window_Start) && (Stab_point->XCDE_Offset < XCDE_Console->XCDE_Tagstruct_Windows_Information[i+1].XCDE_Tagstruct_Window_Displ)))

      break ;
    }

  if (i != XCDE_Console->XCDE_Tagstruct_Cache_Which_Block)

    { pointer = XCDE_Console->XCDE_Tagstruct_Mapped_Pointer+XCDE_Console->XCDE_Tagstruct_Windows_Information[i].XCDE_Tagstruct_Window_Byte_Displacement+position ;

    if (i==(XCDE_Console->XCDE_Tagstruct_Number_Windows-1))

      XCDE_Console->XCDE_Tagstruct_Cache_Cmp_Block_Size = XCDE_Console->XCDE_Tagstruct_Filesize-(XCDE_Console->XCDE_Tagstruct_Windows_Information[i].XCDE_Tagstruct_Window_Byte_Displacement+position) ;

    else

      XCDE_Console->XCDE_Tagstruct_Cache_Cmp_Block_Size = XCDE_Console->XCDE_Tagstruct_Windows_Information[i+1].XCDE_Tagstruct_Window_Byte_Displacement - XCDE_Console->XCDE_Tagstruct_Windows_Information[i].XCDE_Tagstruct_Window_Byte_Displacement ;

    dcmp_Len=(XCDE_Console->XCDE_Tagstruct_Windows_Information[i].XCDE_Tagstruct_Window_Size*3*2*(sizeof(int)+1)) ;

    if (dcmp_Len > XCDE_Console->XCDE_Tagstruct_Cache_Dcmp_Block_Size)

      {
	if (XCDE_Console->XCDE_Tagstruct_Cache_Window != NULL)

	  free(XCDE_Console->XCDE_Tagstruct_Cache_Window) ;

	XCDE_Console->XCDE_Tagstruct_Cache_Dcmp_Block_Size = dcmp_Len ;

	XCDE_Console->XCDE_Tagstruct_Cache_Window = (char *) malloc(sizeof(char)*XCDE_Console->XCDE_Tagstruct_Cache_Dcmp_Block_Size) ;
	}

    uncmp_err=uncompress((Bytef *) XCDE_Console->XCDE_Tagstruct_Cache_Window,
			 (uLong *) &XCDE_Console->XCDE_Tagstruct_Cache_Dcmp_Block_Size,
			 (Bytef *) pointer,
			 (uLong) XCDE_Console->XCDE_Tagstruct_Cache_Cmp_Block_Size) ;

    if (uncmp_err != Z_OK)

      { free(XCDE_Console->XCDE_Tagstruct_Cache_Window) ;

      XCDE_Console->XCDE_Tagstruct_Cache_Which_Block=-1 ;

      XCDE_Console->XCDE_Tagstruct_Cache_Dcmp_Block_Size=0 ;

      XCDE_Console->XCDE_Tagstruct_Cache_Cmp_Block_Size=0 ;

      fprintf(stderr, "\n\tFatal Error : tagstruct decompression\n") ;

      return(-1) ;
	}

    XCDE_Console->XCDE_Tagstruct_Cache_Which_Block = i ;
    }

  Virtual_pointer=XCDE_Console->XCDE_Tagstruct_Cache_Window ;

  offset=0;

  for(j=0; j< XCDE_Console->XCDE_Tagstruct_Windows_Information[i].XCDE_Tagstruct_Window_Size; j++)

    {
      if (j==0)
	{ continuation_bit_d((unsigned long *) &read_el,
			     (char *) Virtual_pointer,
			     (unsigned int) 0,
			     (unsigned long *) &len_cont) ;

	Virtual_pointer+=(len_cont/8) ;

	phis_tag=read_el-1 ;
	}

      else
	{ continuation_bit_d((unsigned long *) &read_el,
			     (char *) Virtual_pointer,
			     (unsigned int) 0,
			     (unsigned long *) &len_cont) ;

	Virtual_pointer+=(len_cont/8) ;

	phis_tag += (read_el - 1);
	}

      continuation_bit_d((unsigned long *) &read_el,
			 (char *) Virtual_pointer,
			 (unsigned int) 0,
			 (unsigned long *) &len_cont) ;

      Virtual_pointer+=(len_cont/8) ;

      displ_tag = (read_el - 1);

      continuation_bit_d((unsigned long *) &read_el,
			 (char *) Virtual_pointer,
			 (unsigned int) 0,
			 (unsigned long *) &len_cont) ;

      Virtual_pointer+=(len_cont/8) ;

      tag_id = (read_el- 1);

      if(Stab_point->XCDE_Position < phis_tag)

	break;

      if((Stab_point->XCDE_Position == phis_tag) &&
	 (Stab_point->XCDE_Offset < displ_tag))

	break;    

      if (open_tag2(stack,
		    stack_size,
		    tag_id)) 
	{ char *token=NULL ; 


	  stack[stack_size++] = tag_id;  

	  if (stack_size > stack_dimension) 

	    { stack_dimension*=2 ; 

	    stack=realloc(stack,
			  stack_dimension) ; 
	    }
	}

      else
	{
	  if((Stab_point->XCDE_Position == phis_tag)
	     && (Stab_point->XCDE_Offset == displ_tag))

	    {break;}

	  stack_size--;
	}
    }

  /* ---------------- */ 
  /* produces results */
  /* ---------------- */ 

  *Enclosing_tags_ranks = (int *) malloc(stack_size*sizeof(int)) ;

  *Num_enclosing_tags=stack_size ; 

  for(i=0; i<stack_size; i++)

    (*Enclosing_tags_ranks)[i]=stack[i] ;

  free(stack) ; 

#undef XCDE_INITIAL_STACK_DIMENSION

  return(1) ;

}

/*******************************************************************/
// internal function used by function 
// XCDE_Get_Tagstruct_Tags_Enclosing 
int open_tag2(int *stack,
	      int size,
	      int tag)
{
  if (size==0) return(1);

  if (stack[size-1] != tag) return(1);

  return(0);
}


/*******************************************************************/

// functions that reads the depthstruct flag, that is on if the
// depthstruct file is present, and of otherwise.
// Example : ret = XCDE_SettedDepthstruct(&DepthstructFlag, &Console)
// where &DepthstructFlag is the memory address of the variable that
// will contain the depthstruct flag, and &Console is the memory
// address of the console data struct.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_SettedDepthstruct(unsigned char *DepthstructFlag,
			 XCDE_Console_Type *XCDE_Console) 
{
  /* ----------------------------------------- */
  /*            checks parameters              */
  /* ----------------------------------------- */    

  if (XCDE_Console == NULL)

    return(-1) ;

  if (XCDE_Console->XCDE_Console_Initialized != 1)
    
    return(-1) ; 

  if (DepthstructFlag == NULL) 

    return(-1) ; 
  
  /* ------------------------ */
  /* reads the depthstruct flag */ 
  /* ------------------------ */ 

  *DepthstructFlag=XCDE_Console->XCDE_Depthstruct ; 

  return(1) ;
}


/*******************************************************************/

// functions that reads the size of the depthstruct file (returns
// -1 if itsn't present).
// Example : ret=XCDE_Get_Depthstruct_Filesize(&Size, &Console)
// where &size is the memory address of the variable that will
// contain the depthstruct file size, and &Console is the memory
// address of the console data struct.
// ret value is 1 if there aren't errors, or -1 otherwise.

int XCDE_Get_Depthstruct_Filesize(int *Depthstruct_Filesize,
				XCDE_Console_Type *XCDE_Console)
{
  /* ----------------------------------------- */
  /*            checks parameters              */
  /* ----------------------------------------- */

  if (XCDE_Console == NULL)

    return(-1) ;

  if (XCDE_Console->XCDE_Console_Initialized != 1)

    return(-1) ;

  if (XCDE_Console->XCDE_Depthstruct != 1)

    return(-1) ;

  if (Depthstruct_Filesize == NULL)

    return(-1) ;

  /* ----------------------------- */
  /* reads the depthstruct file size */
  /* ----------------------------- */

  *Depthstruct_Filesize=XCDE_Console->XCDE_Depthstruct_Filesize ;

  return(1) ;
}

/*******************************************************************/

// functions that reads the depthstruct file name (returns -1
// if theresn't the depthstruct file).
// Example : ret=XCDE_Get_DEpthstruct_Filename(&DepthstructFilename, &Console)
// where &DepthstructFilename is the memory address of the variable that
// will contain the file name, and &Console is the memory address of 
// the console data structure.
// The function allocates the memory needed by the name.
// ret value is 1 if there aren't errors, or -1 otherwise. 

int XCDE_Get_Depthstruct_Filename(char **Depthstruct_Filename,
				XCDE_Console_Type *XCDE_Console)
{ 
  /* ----------------------------------------- */
  /*            checks parameters              */
  /* ----------------------------------------- */

  if (XCDE_Console == NULL)

    return(-1) ;

  if (XCDE_Console->XCDE_Console_Initialized != 1)

    return(-1) ;

  if (XCDE_Console->XCDE_Depthstruct != 1)

    return(-1) ;

  if (Depthstruct_Filename == NULL)

    return(-1) ;

  /* ----------------------------- */
  /* reads the depthstruct file name */
  /* ----------------------------- */

  *Depthstruct_Filename=malloc(strlen(XCDE_Console->XCDE_Depthstruct_File_Name)+1) ;

  if (*Depthstruct_Filename==NULL)

    return(-1) ;

  strcpy(*Depthstruct_Filename,
	 XCDE_Console->XCDE_Depthstruct_File_Name) ;

  return(1) ; 
}




int read_integer(int *destinazione, unsigned char *buffer) {

	int i;

	*destinazione = buffer[0];
	
	for (i=1; i < sizeof(int); i++) {

		*destinazione <<= 8;
		*destinazione += buffer[i];

	}


	return 1;

}

/*******************************************************************/

// functions that reads the depths of the input tag.
// Example : res=XCDE_Get_Tag_Depth(rank, &depth, &num_depth, &Console)
// where rank is the rank the tag in the dictionary, &depth is the memory address of the variable
// that will contain the tag's depths, &num_depth is the memory address of the variable
// that will contain the total number of depths and &Console is the memory address of
// the console data struct.
// res value is 1 if there aren't errors, or -1 otherwise.

int XCDE_Get_Tag_Depth(int Rank_TagDict, int **Depth, int *Num_depth, XCDE_Console_Type *XCDE_Console)
{

	int function_result = 0;

	int total_tag = 0;

	unsigned char *Dict;

	int dictrank = 0;

	int byte_DEPTH_list = 0;

	int num_depths = 0;

	int depth_read = 0;

	static
	  int position = 0 ;

	int len_cont=0 ;

	int num_pointer_DEPTH_Index=0 ;

	int num_DEPTH_List=0 ;


	int i;

	/* ----------------------------------------- */
	/*            checks parameters              */
	/* ----------------------------------------- */

	if (XCDE_Console == NULL)

		return(-1) ;

	if (XCDE_Console->XCDE_Console_Initialized != 1)

		return(-1) ;

	if (XCDE_Console->XCDE_Depthstruct != 1)

		return(-1) ;

	if (Rank_TagDict == 0)

		return(-1) ;

  /* -------------------------- */
  /* loads the depthstruct header */
  /* -------------------------- */


	// search the total number of tag
	function_result = XCDE_Get_Num_El_Dict(TAG, &total_tag, XCDE_Console);

	if (function_result == -1)

	return(-1) ;


	/* ----------------------- */
	/* accessing the TAB_DEPTH */
	/* ----------------------- */

	num_pointer_DEPTH_Index = ((Rank_TagDict - 1) / SKIP);

	num_DEPTH_List=((Rank_TagDict - 1) % (SKIP)) ;

	position = num_pointer_DEPTH_Index * sizeof(int);

	read_integer(&byte_DEPTH_list, XCDE_Console->XCDE_Depthstruct_Mapped_Pointer + position);

	// find the rigth list of depth
	for (i=0 ; i < num_DEPTH_List; i++) {

		read_integer(&position, XCDE_Console->XCDE_Depthstruct_Mapped_Pointer + byte_DEPTH_list);
		
		byte_DEPTH_list += position + sizeof(int);

	}

	// add sizeof(int) to start the reading of the vector
	byte_DEPTH_list += sizeof(int);

	// decompress the total number of depths
	continuation_bit_d((unsigned long *) &num_depths,
			(char *) XCDE_Console->XCDE_Depthstruct_Mapped_Pointer + byte_DEPTH_list,
			(unsigned int) 0,
			(unsigned long *) &len_cont) ;

	byte_DEPTH_list += (len_cont/8) ;

	*Num_depth = num_depths - 1;

	// allocates the memory needed to contain the result
	*Depth = (int *) malloc(num_depths * sizeof(int)) ;

	for (i=0; i < num_depths; i++) {

		// decompress the depth vector
		continuation_bit_d((unsigned long *) &depth_read,
					(char *) XCDE_Console->XCDE_Depthstruct_Mapped_Pointer + byte_DEPTH_list,
					(unsigned int) 0,
					(unsigned long *) &len_cont) ;

		byte_DEPTH_list += (len_cont/8) ;

		(*Depth)[i] = depth_read - 1;

	}

	return (1);
}
/*******************************************************************/

// function that looks for tokens in the word dictionary and
// gets the numbers of the dictionary element where the each element
// is.
// Example : ret=XCDE_Search_Word_Dictionary(token, &token, &num_token, XCDE_PREFIX, XCDE_CAPUND, 0, &Console)
// where token is the pointer to the string to look for, &token is the
// memory address of the array  will contain elements in the dictionary,
// &num_token is the memory address of the variable that will contain the number of elements,
// XCDE_PREFIX is the type of search (may be XCDE_SUFFIX, XCDE_EXACT, XCDE_CONTAINED, XCDE_ERROR, XCDE_REGEXP),
// XCDE_CASE means that there's difference between capital and undercase letters (may be otherwise XCDE_NOCASE) in the search,
// 0 is the number of errors (may be 1 or 2, and can be used only with XCDE_ERROR) and Console is the memory address of the console data structure.
// ret value is 1 if there aren't errors, or -1 otherwise.

int XCDE_Search_Word_Dictionary(char *String_To_Search,
				int **TokenDict,
				int *NumToken,
				unsigned char type,
				unsigned char lettercase,
				unsigned int nerror,
				XCDE_Console_Type *XCDE_Console)
{
#define XCDE_TOKEN_INITIAL_LENGTH 10

#define XCDE_SEARCH_RESULT_LINE   90

#define XCDE_SEARCH_DESCR_RES   1000

  register 
    int k=0 ;

  register
    int j=0 ;

  register
    int i=0 ;

  int function_result=0 ;

  char *sys_comand=NULL ; 
 
  char *search=NULL ;

  // pointer to the name of the temporary file 
  char *outputfilename=NULL ; 

  // pointer to memory space needed to read the result
  char *descr_res=NULL ;

  // pointer to the temporary file used to read the result
  FILE *outputfile ;

  int TokenDimension=0 ;

  /* ----------------------------------------- */
  /*            checks parameters              */
  /* ----------------------------------------- */

  if (XCDE_Console == NULL)

    return(-1) ; 

  if (XCDE_Console->XCDE_Console_Initialized != 1)

    return(-1) ; 
  
  if (TokenDict == NULL)

    return(-1) ;

  if (NumToken == NULL)
     
    return(-1) ; 

  if ((type != XCDE_SUFFIX) && (type != XCDE_EXACT) && (type != XCDE_CONTAINED) && (type != XCDE_ERROR) && 
      (type != XCDE_PREFIX) && (type != XCDE_REGEXP))

    return(-1) ;

  if ((lettercase != XCDE_CASE) && (lettercase != XCDE_NOCASE))

    return(-1) ;

  if ((type != XCDE_ERROR) && (nerror != 0)) 
    
    return(-1) ;

  if ((type == XCDE_ERROR) && (nerror == 0))

    return(-1) ;

  /* ----------------------------------------------- */ 
  /* if is necessary, decompress the WORD dictionary */ 
  /* ----------------------------------------------- */ 

  if (XCDE_Console->XCDE_Dict_File_Decompressed[WORD] != 1u)

    { 
      function_result=XCDE_Decompress_Dictionary(WORD,
						 XCDE_Console) ; 
      
      if (function_result == -1) 
       
	return(-1) ; 
    }

  /* ------------------------------ */
  /* creates the string to look for */ 
  /* -------------------------------*/

  if  ((type != XCDE_REGEXP) && (type != XCDE_ERROR))

    { search= (char *) malloc(((strlen(String_To_Search)*2)+1)*sizeof(char)) ;
      
      for (k=0, j=0 ; j<strlen(String_To_Search) ; ++j)

	if ((String_To_Search[j] == '^') || (String_To_Search[j] == '?') || (String_To_Search[j] == '*') ||
	    (String_To_Search[j] == '{') || (String_To_Search[j] == '}') || (String_To_Search[j] == '[') ||
	    (String_To_Search[j] == ']') || (String_To_Search[j] == '$') || (String_To_Search[j] == '^') || 
	    (String_To_Search[j] == '+') || (String_To_Search[j] == '.'))
	  
	  { search[k]='\\' ;
	  
	  search[k+1]=String_To_Search[j] ;
	  
	  k+=2 ;
	  }
      
	else 
	  
	  { search[k]=String_To_Search[j] ;
	  
	  ++k ;
	  }

      search[k] = '\0' ;
    }

  else 

    if (type == XCDE_ERROR)

      { search = (char *) malloc(strlen(String_To_Search)*sizeof(char)) ;

      k=0 ;

      i=0 ;

      while(k<strlen(String_To_Search))

	{ if (String_To_Search[k] == '&')

	  { j=k ;

	  while((j<strlen(String_To_Search)) && (String_To_Search[j] != ';'))

	    ++j ;

	  if (String_To_Search[j] == ';')

	    { search[i]=String_To_Search[k] ;

	    ++i ;

	    k=(j+1) ;
	    }

	  else

	    { search[i]=String_To_Search[k] ;

	    ++i ;

	    ++k ;
	    }
	  }

	else 

	  { search[i]=String_To_Search[k] ;
	  
	  ++i ;
	  
	  ++k ;
	  }
	}

      search[i]='\0' ; 
      }
  
    else
      
      search=String_To_Search ;
  
  /* ---------------------------------------- */ 
  /* looking for the string in the dictionary */
  /* ---------------------------------------- */
  
  // allocates the memory space needed by the temporary file name
  outputfilename=malloc(L_tmpnam*sizeof(char)) ; 
  
  // creates the temporary file name
  outputfilename=tmpnam(outputfilename) ; 

  // allocates the memory space needed by the string that will contain
  // the system comand 
  sys_comand=malloc((strlen(outputfilename)+strlen(XCDE_Console->XCDE_Dict_Dcmptmp_Names[WORD])+strlen(search)+256)*sizeof(char)) ;

  if ((type == XCDE_EXACT) && (lettercase == XCDE_CASE))

    // comand that will passed to the system function
    sprintf(sys_comand, "grep -n -x -z -s \"%s\" %s | tr \'\\000\' \'\\n\' > %s", search, XCDE_Console->XCDE_Dict_Dcmptmp_Names[WORD], outputfilename) ;

  else

    if ((type == XCDE_EXACT) && (lettercase == XCDE_NOCASE))

      // comand that will passed to the system function
      sprintf(sys_comand, "grep -n -i -x -z -s \"%s\" %s | tr \'\\000\' \'\\n\' > %s", search, XCDE_Console->XCDE_Dict_Dcmptmp_Names[WORD], outputfilename) ;

    else

      if ((type == XCDE_EXACT) && (lettercase == XCDE_CASE))

	// comand that will passed to the system function
	sprintf(sys_comand, "grep -n -z -s -i \"%s\" %s | tr \'\\000\' \'\\n\' > %s", search, XCDE_Console->XCDE_Dict_Dcmptmp_Names[WORD], outputfilename) ;

      else

	if ((type == XCDE_PREFIX) && (lettercase == XCDE_CASE))

	  // comand that will passed to the system function
	  sprintf(sys_comand, "grep -n -z -s \"^%s\" %s | tr \'\\000\' \'\\n\' > %s", search, XCDE_Console->XCDE_Dict_Dcmptmp_Names[WORD], outputfilename) ;

	else

	  if ((type == XCDE_PREFIX) && (lettercase == XCDE_NOCASE))

	    // comand that will passed to the system function
	    sprintf(sys_comand, "grep -n -z -s -i \"^%s\" %s | tr \'\\000\' \'\\n\' > %s", search, XCDE_Console->XCDE_Dict_Dcmptmp_Names[WORD], outputfilename) ;

	  else

	    if ((type == XCDE_SUFFIX) && (lettercase == XCDE_CASE))

	    // comand that will passed to the system function
	    sprintf(sys_comand, "grep -n -z -s \"%s$\" %s | tr \'\\000\' \'\\n\' > %s", search, XCDE_Console->XCDE_Dict_Dcmptmp_Names[WORD], outputfilename) ;
	    else

	      if ((type == XCDE_SUFFIX) && (lettercase == XCDE_NOCASE))

		// comand that will passed to the system function
		sprintf(sys_comand, "grep -n -z -s -i \"%s$\" %s | tr \'\\000\' \'\\n\' > %s", search, XCDE_Console->XCDE_Dict_Dcmptmp_Names[WORD], outputfilename) ;

	      else

		if ((type == XCDE_CONTAINED) && (lettercase == XCDE_CASE))

		  // comand that will passed to the system function
		  sprintf(sys_comand, "grep -n -z -s \"%s\" %s | tr \'\\000\' \'\\n\' > %s", search, XCDE_Console->XCDE_Dict_Dcmptmp_Names[WORD], outputfilename) ;

		else

		  if ((type == XCDE_CONTAINED) && (lettercase == XCDE_NOCASE))

		    // comand that will passed to the system function
		    sprintf(sys_comand, "grep -n -z -s -i \"%s\" %s | tr \'\\000\' \'\\n\' > %s", search, XCDE_Console->XCDE_Dict_Dcmptmp_Names[WORD], outputfilename) ;

		  else

		    if ((type == XCDE_REGEXP) && (lettercase == XCDE_CASE))

		      // comand that will passed to the system function
		      sprintf(sys_comand, "grep -n -z -s -e \"%s\" %s | tr \'\\000\' \'\\n\' > %s", search, XCDE_Console->XCDE_Dict_Dcmptmp_Names[WORD], outputfilename) ;

		    else

		      if ((type == XCDE_REGEXP) && (lettercase == XCDE_NOCASE))

			// comand that will passed to the system function
			sprintf(sys_comand, "grep -n -z -s -i \"%s\" %s | tr \'\\000\' \'\\n\' > %s", search, XCDE_Console->XCDE_Dict_Dcmptmp_Names[WORD], outputfilename) ;

		      else

			if (type == XCDE_ERROR)

			  // comand that will passed to the system function
			  sprintf(sys_comand, "_xcde_transf < %s | agrep -n -i -%d '^%s$' > %s", XCDE_Console->XCDE_Dict_Dcmptmp_Names[WORD], nerror, search, outputfilename) ;


  *TokenDict=(int *) malloc(sizeof(int)*XCDE_TOKEN_INITIAL_LENGTH) ;

  TokenDimension=XCDE_TOKEN_INITIAL_LENGTH ;

  *NumToken=0 ;

  if ((function_result=system(sys_comand)) == 0)

    if ((outputfile=fopen(outputfilename,
			  "r")) != NULL)

      { // allocates the memory needed to contain the result
	descr_res=malloc(XCDE_SEARCH_DESCR_RES*sizeof(char)) ;

	// reads the output of the comand
	while ((fgets(descr_res,
		      XCDE_SEARCH_RESULT_LINE,
		      outputfile) != NULL) && (strlen(descr_res) > 0))

	  { 

	    // trasform the output to an integer
	    (*TokenDict)[*NumToken]=atoi(descr_res)-1 ;

	    if ((*TokenDict)[*NumToken] > 0) ++(*NumToken) ;

	    if ((*NumToken) >= TokenDimension)

	      { *TokenDict=(int *) realloc(*TokenDict,
					   TokenDimension*2*sizeof(int)) ;

	      TokenDimension*=2 ;
	      }
	  }

	// releases the memory space allocated
	if (descr_res != NULL)

	  { free(descr_res) ;

	  descr_res=NULL ;
	  }

	if (sys_comand != NULL)

	  { free(sys_comand) ;

	  sys_comand=NULL ;
	  }

	// close the output file
	fclose(outputfile) ;
      }

    else

      { // removes the temporary file
	if (outputfilename != NULL)

	  { unlink(outputfilename) ;

	  // releases the memory for the temporary file name
	  free(outputfilename) ;

	  outputfilename=NULL ;
	  }

	// releases the memory for the system comand
	if (sys_comand != NULL)

	  { free(sys_comand) ;

	  sys_comand=NULL ;
	  }

	if ((type != XCDE_REGEXP) && (search != NULL))

	  { free(search) ;

	  search=NULL ;
	  }

	return(-1) ; 
      }
  
  // removes the temporary file 
  if (outputfilename != NULL)

    { unlink(outputfilename) ; 
    
    // releases the memory allocates 
    free(outputfilename) ; 

    outputfilename=NULL ;
    }

  if ((type != XCDE_REGEXP) && (search != NULL))

    { free(search) ;

    search=NULL ;
    }

  return(1) ; 
}

int XCDE_Search_Tag_Dictionary(char *String_To_Search,
			       int *dictrank,
			       XCDE_Console_Type *XCDE_Console) 
{ 
#define XCDE_SEARCH_RESULT_LINE   90 

#define XCDE_SEARCH_DESCR_RES   1000

  register 
    int k=0 ;

  register
    int j=0 ;

  register
    int i=0 ;
  
  int function_result=0 ; 

  char *search=NULL ;

  char *sys_comand=NULL ; 
 
  // pointer to the name of the temporary file 
  char *outputfilename=NULL ; 

  // pointer to memory space needed to read the result 
  char *descr_res=NULL ; 

  // pointer to the temporary file used to read the result 
  FILE *outputfile ; 

  /* ----------------------------------------- */
  /*            checks parameters              */
  /* ----------------------------------------- */

  if (XCDE_Console == NULL)

    return(-1) ; 

  if (XCDE_Console->XCDE_Console_Initialized != 1)

    return(-1) ; 
  
  if (dictrank == NULL)
    
    return(-1) ; 

  /* ---------------------------------------------- */ 
  /* if is necessary, decompress the tag dictionary */ 
  /* ---------------------------------------------- */ 

  if (XCDE_Console->XCDE_Dict_File_Decompressed[TAG] != 1u)
   
    { 
      function_result=XCDE_Decompress_Dictionary(TAG,
						 XCDE_Console) ; 
      
      if (function_result == -1) 
       
	return(-1) ; 
    }
 
  /* ------------------------------ */
  /* creates the string to look for */ 
  /* -------------------------------*/
  
  // inserts escape characters 
  search=(char *) malloc(((strlen(String_To_Search)*2)+1)*sizeof(char)) ;
  
  for (k=0, j=0 ; j<strlen(String_To_Search) ; ++j)

    if ((String_To_Search[j] == '^') || (String_To_Search[j] == '?') || (String_To_Search[j] == '*') || 
	(String_To_Search[j] == '{') || (String_To_Search[j] == '}') || (String_To_Search[j] == '[') ||
	(String_To_Search[j] == ']') || (String_To_Search[j] == '$') || (String_To_Search[j] == '^') || 
	(String_To_Search[j] == '+') || (String_To_Search[j] == '.'))
      
      { search[k]='\\' ; 

      search[k+1]=String_To_Search[j] ; 

      k+=2 ;
      }

    else

      { search[k]=String_To_Search[j] ; 

      k++ ;
      }

  search[k]='\0' ; 

  /* ---------------------------------------- */ 
  /* looking for the string in the dictionary */
  /* ---------------------------------------- */ 

  // allocates the memory space needed by the temporary file name 
  outputfilename=malloc((L_tmpnam+1)*sizeof(char)) ; 

  // creates the temporary file name 
  outputfilename=tmpnam(outputfilename) ; 

  // allocates the memory space needed by the string that will contain
  // the system comand 
  sys_comand=malloc((strlen(outputfilename)+strlen(XCDE_Console->XCDE_Dict_Dcmptmp_Names[TAG])+strlen(search)+2000)*sizeof(char)) ; 

  // comand that will passed to the system function 
  sprintf(sys_comand, "grep -n -z -x -s \"%s\" %s | tr \'\\000\' \'\\n\' > %s", search, XCDE_Console->XCDE_Dict_Dcmptmp_Names[TAG], outputfilename) ; 

  *dictrank=0 ; 

  if ((function_result=system(sys_comand)) == 0)

    if ((outputfile=fopen(outputfilename, 
			  "r")) != NULL)
      
      { // allocates the memory needed to contain the result 
	descr_res=malloc(XCDE_SEARCH_DESCR_RES*sizeof(char)) ; 

	// reads the output of the comand 
	if (fgets(descr_res, 
		  XCDE_SEARCH_RESULT_LINE,
		  outputfile) != NULL)
	  
	  // trasform the output to an integer 
	  *dictrank=atoi(descr_res)-1 ; 
	
	// releases the memory space allocated 
	if (descr_res != NULL) 

	  { free(descr_res) ; 
	
	  descr_res=NULL ;
	  }

	if (sys_comand != NULL) 
	  
	  { free(sys_comand) ; 

	  sys_comand=NULL ;
	  }

	// close the output file 
	fclose(outputfile) ; 
      }
  
    else 
      
      { // releases the memory for the temporary file name 
	if (outputfilename != NULL) 

	  { // removes the temporary file 
	    unlink(outputfilename) ; 

	    free(outputfilename) ; 
	  }
	
	if (search != NULL)

	  free(search) ;

	// releases the memory for the system comand 
	if (sys_comand != NULL) 

	  free(sys_comand) ; 

	return(-1) ; 
      }

  if (outputfilename != NULL) 
    
    {  // removes the temporary file 
      unlink(outputfilename) ; 
      
      free(outputfilename) ;
    }
  
  if (search != NULL)
    
    free(search) ; 

  return(1) ; 
}

int XCDE_Search_Attr_Dictionary(char *attributename,
				char *attributevalue, 
				int **TokenDict,
				int *NumToken,
				unsigned char typeattributesearch,
				unsigned char typevaluesearch, 
				unsigned char lettercaseattribute, 
				unsigned char lettercasevalue, 
				unsigned int nerrorattribute,
				unsigned int nerrorvalue, 
				XCDE_Console_Type *XCDE_Console) 
{ 
#define XCDE_TOKEN_INITIAL_LENGTH 10

#define XCDE_SEARCH_RESULT_LINE   90 

#define XCDE_SEARCH_DESCR_RES   1000

  unsigned 
    char istoken=0 ; 

  register 
    int k=0 ;
  
  register
    int j=0 ;
  
  char *attribute=NULL ;

  char *value=NULL ;

  char *searchattributename=NULL ;
  
  int function_result=0 ; 
  
  char *sys_comand=NULL ; 
 
  // pointer to the name of the temporary file 
  char *outputfilename=NULL ; 

  // pointer to memory space needed to read the result 
  char *descr_res=NULL ; 

  // pointer to the temporary file used to read the result 
  FILE *outputfile ; 

  int TokenDimension=0 ;

  /* ----------------------------------------- */
  /*            checks parameters              */
  /* ----------------------------------------- */

  if (XCDE_Console == NULL)

    return(-1) ; 

  if (XCDE_Console->XCDE_Console_Initialized != 1)

    return(-1) ; 
  
  if (TokenDict == NULL)
    
    return(-1) ; 

  if (NumToken == NULL)
     
    return(-1) ; 

  if ((typeattributesearch != XCDE_SUFFIX) && (typeattributesearch != XCDE_EXACT) && (typeattributesearch != XCDE_CONTAINED) &&  
      (typeattributesearch != XCDE_PREFIX) && (typeattributesearch != XCDE_REGEXP) && (typeattributesearch != XCDE_ANYVALUE)) 

    return(-1) ;

  if ((typevaluesearch != XCDE_SUFFIX) && (typevaluesearch != XCDE_EXACT) && (typevaluesearch != XCDE_CONTAINED) && (typevaluesearch != XCDE_ERROR) && 
      (typevaluesearch != XCDE_PREFIX) && (typevaluesearch != XCDE_REGEXP) && (typevaluesearch != XCDE_ANYVALUE) && (typevaluesearch != XCDE_TOKEN_EXACT) &&
      (typevaluesearch != XCDE_TOKEN_SUFFIX) && (typevaluesearch != XCDE_TOKEN_PREFIX))

    return(-1) ;

  if ((typeattributesearch == XCDE_ANYVALUE) && (typevaluesearch == XCDE_ANYVALUE)) 

    return(-1) ;

  if ((typevaluesearch == XCDE_ERROR) && (typeattributesearch != XCDE_EXACT))

    return(-1) ;

  if ((lettercaseattribute != XCDE_CASE) && (lettercaseattribute != XCDE_NOCASE))

    return(-1) ;

  if ((lettercasevalue != XCDE_CASE) && (lettercasevalue != XCDE_NOCASE))

    return(-1) ;

  if (((typeattributesearch == XCDE_ERROR) || (typeattributesearch == XCDE_REGEXP)) && (lettercaseattribute == XCDE_NOCASE))

     return(-1) ; 

  if (((typevaluesearch == XCDE_ERROR) || (typevaluesearch == XCDE_REGEXP)) && (lettercasevalue == XCDE_NOCASE))

     return(-1) ; 

  if ((typeattributesearch != XCDE_ERROR) && (nerrorattribute != 0)) 
    
    return(-1) ;

  if ((typevaluesearch != XCDE_ERROR) && (nerrorvalue != 0)) 

    return(-1) ;

  if ((typevaluesearch == XCDE_ERROR) && (nerrorvalue == 0)) 

    return(-1) ;

  /* ---------------------------------------------- */ 
  /* if is necessary, decompress the tag dictionary */ 
  /* ---------------------------------------------- */ 

  if (XCDE_Console->XCDE_Dict_File_Decompressed[ATTR] != 1u)
   
    { 
      function_result=XCDE_Decompress_Dictionary(ATTR,
						 XCDE_Console) ; 
      
      if (function_result == -1) 
       
	return(-1) ; 
    }

  /* ------------------------------------------------------ */ 
  /* creating the regular expression for the attribute name */
  /* ------------------------------------------------------ */ 

  if (typeattributesearch != XCDE_ANYVALUE) 

    attribute=(char *) malloc(strlen(attributename)*4*sizeof(char)+256) ;

  else
    
    attribute=(char *) malloc(sizeof(char)*256) ;

  k=0 ;

  attribute[0]='\0' ; 


  if ((typeattributesearch == XCDE_SUFFIX) || (typeattributesearch== XCDE_CONTAINED) || (typeattributesearch == XCDE_ANYVALUE))

    { strcat(attribute, "\\([^=]\\)*") ;

    k+=strlen("\\([^=]\\)*") ;
    }

  if (typeattributesearch != XCDE_ANYVALUE) 
    {  if (lettercaseattribute == XCDE_NOCASE)
  
      for(j=0 ; j<strlen(attributename); ++j) 

	{ if ((attributename[j] >= 65) && (attributename[j] <= 90))
	  
	{ attribute[k]='[' ;
	
	attribute[k+1]=attributename[j] ;
	
	attribute[k+2]=attributename[j]+32 ;

	attribute[k+3]=']' ;
	
	k+=4 ;
	}

	else

	if ((attributename[j] >= 97) && (attributename[j] <=122))

	  { attribute[k]='[' ;

	  attribute[k+1]=attributename[j] ;

	  attribute[k+2]=attributename[j]-32 ;
	  
	  attribute[k+3]=']' ;
	
	  k+=4 ;
	  }
	
	else
    
	  { attribute[k]=attributename[j] ;

	  ++k ;
	  }
	
	attribute[k]='\0' ;
	}

    else

      { for(j=0 ; j<strlen(attributename); ++j) 
	
	{ attribute[k]=attributename[j] ;

	++k ;
	}

      attribute[k]='\0' ;
      }
    }

  if ((typeattributesearch == XCDE_PREFIX) || (typeattributesearch== XCDE_CONTAINED))

    { strcat(attribute, "\\([^=]\\)*") ;
    
    k+=strlen("\\([^=]\\)*") ;
    }  
  
  attribute[k]='\0' ;
 
  /* -------------------------------------------------- */ 
  /* creating the regular expression for the value name */
  /* -------------------------------------------------- */ 

  if (typevaluesearch != XCDE_ANYVALUE) 

    value=(char *) malloc(strlen(attributevalue)*4*sizeof(char)+256) ;

  else

    value=(char *) malloc(sizeof(char)*256) ;

  value[0]='\0' ;

  k=0 ;

  if ((typevaluesearch == XCDE_TOKEN_PREFIX) || (typevaluesearch == XCDE_TOKEN_SUFFIX) || (typevaluesearch == XCDE_TOKEN_EXACT))

    istoken=1 ; 

  if ((!istoken) && (typevaluesearch !=XCDE_ERROR))
  
    { strcat(value, "\\([\\\'\\\"]\\)") ;
    
    k+=strlen("\\([\\\'\\\"]\\)") ; 
    }
   
  if ((typevaluesearch == XCDE_CONTAINED) || (typevaluesearch == XCDE_ANYVALUE) || (typevaluesearch == XCDE_SUFFIX) || (typevaluesearch == XCDE_TOKEN_SUFFIX)) 
    
    { strcat(value, "\\(.\\)*") ;
      
    k+=strlen("\\(.\\)*") ;
    }
  
  if (typevaluesearch != XCDE_ANYVALUE)

    { if(lettercasevalue == XCDE_NOCASE)
  
      { for(j=0 ; j<strlen(attributevalue); ++j) 
      
	if ((attributevalue[j] >= 65) && (attributevalue[j] <= 90))
	
	  { value[k]='[' ;

	value[k+1]=attributevalue[j] ;

	value[k+2]=attributevalue[j]+32 ;

	value[k+3]=']' ;
	
	k+=4 ;
	}
      
	else
	
	if ((attributevalue[j] >= 97) && (attributevalue[j] <= 122))
	  
	  { value[k]='[' ;

	  value[k+1]=attributevalue[j] ;

	  value[k+2]=attributevalue[j]-32 ;
	  
	  value[k+3]=']' ;
	
	  k+=4 ;
	  }
    
	else
	  
	  { value[k]=attributevalue[j] ;

	  ++k ;
	  }

      value[k]='\0' ;
      }

    else

      { for(j=0 ; j<strlen(attributevalue); ++j) 
	
	{ value[k]=attributevalue[j] ;
	
	++k ;
	}
      
      value[k]='\0' ;
      } 
    }
    
  if ((typevaluesearch == XCDE_CONTAINED) || (typevaluesearch == XCDE_PREFIX) || (typevaluesearch == XCDE_TOKEN_PREFIX))

    { strcat(value, "\\(.\\)*") ;
      
    k+=strlen("\\(.\\)*") ;
    }
  
  if ((!istoken) && (typevaluesearch !=XCDE_ERROR))

    { strcat(value, "\\([\\\'\\\"]\\)") ;
    
    k+=strlen("\\([\\\'\\\"]\\)") ; 
    }

  value[k]='\0' ;

  // allocates the memory space needed by the temporary file name 
  outputfilename=malloc(L_tmpnam*sizeof(char)) ; 
 
  // creates the temporary file name 
  outputfilename=tmpnam(outputfilename) ; 

  // allocates the memory space needed by the string that will contain
  // the system comand 
  sys_comand=malloc((strlen(outputfilename)+strlen(XCDE_Console->XCDE_Dict_Dcmptmp_Names[ATTR])+strlen(attribute)+strlen(value)+256)*sizeof(char)) ; 

  // comand that will passed to the system function 
  if (typevaluesearch != XCDE_ERROR)

    sprintf(sys_comand, "grep -n -z -s \"^%s[=]%s$\" %s | tr \'\\000\' \'\\n\' > %s", attribute, value, XCDE_Console->XCDE_Dict_Dcmptmp_Names[ATTR], outputfilename) ; 
  
  else

    if (typevaluesearch == XCDE_ERROR)

      sprintf(sys_comand, "_xcde_transf < %s | agrep -%d -n \"^%s[=]%s$\" | grep \"^\\([0-9]\\)*[:][ ]%s[=]\" > %s", XCDE_Console->XCDE_Dict_Dcmptmp_Names[ATTR], nerrorvalue,attribute, value, attribute, outputfilename)  ;

  *TokenDict=(int *) malloc(sizeof(int)*XCDE_TOKEN_INITIAL_LENGTH) ;
  
  TokenDimension=XCDE_TOKEN_INITIAL_LENGTH ;
  
  *NumToken=0 ;

  if ((function_result=system(sys_comand)) == 0)
    
    if ((outputfile=fopen(outputfilename, 
			  "r")) != NULL)
      
      { // allocates the memory needed to contain the result 
	descr_res=malloc(XCDE_SEARCH_DESCR_RES*sizeof(char)) ; 

	// reads the output of the comand 
	while ((fgets(descr_res, 
		      XCDE_SEARCH_RESULT_LINE,
		      outputfile) != NULL) && (strlen(descr_res) > 0))
	  
	  {  // trasform the output to an integer 
	    (*TokenDict)[*NumToken]=atoi(descr_res)-1 ; 
	
	    if ((*TokenDict)[*NumToken] > 0) ++(*NumToken) ;

	    if ((*NumToken) >= TokenDimension)

	      {  *TokenDict=(int *) realloc(*TokenDict,
					    TokenDimension*2*sizeof(int)) ;

	      TokenDimension*=2 ;
	      }
	    
	  }

	// releases the memory space allocated 
	if (descr_res != NULL)

	  free(descr_res) ; 
	
	if (sys_comand != NULL) 

	  free(sys_comand) ; 

	// close the output file 
	fclose(outputfile) ;
      }
  
    else 

      { // releases the memory for the temporary file name 
	if (outputfilename != NULL) 

	{ unlink(outputfilename) ;
	
	free(outputfilename) ; 
	}

	if (attribute != NULL)

	  free(attribute) ;

	if (value != NULL) 

	  free(value) ;

	// releases the memory for the system comand 
	if (sys_comand != NULL) 

	  free(sys_comand) ; 

	return(-1) ; 
      }
  
  if (outputfilename != NULL) 
  
  { // removes the temporary file 
    unlink(outputfilename) ; 

    // releases the memory allocates 
    free(outputfilename) ; 
  }

  return(1) ; 
}

