 /********************************************************************
			       xcde_libsearch.h
                 header file needed by the XCDE 2.0 API library
	         Xml Compressed Document Engine Query Language
	             Alessandro Perrucci - Rosanna Rossi
             Dipartimento di Informatica - Universita' di Pisa
	                   alexperrucci@libero.it
	                  rosannarossi79@libero.it
 ********************************************************************/

#include "xmlparse.h"
#include "xcde_lib.h"
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include<sys/mman.h>
#include<sys/stat.h>
 

/********************************************************************
			DATA STRUCTURE
*********************************************************************/

// struttura del node dell'albero della query
typedef struct query_tree_node {

  char *node_name;
  unsigned int node_type;
  struct query_tree_node **childs_vector;
  unsigned int num_childs;
  union {
	  XCDE_IL_Couple_Type *tag_positions;		// usato se il node ha tipo == "TAG"
	  XCDE_IL_Triple_Type *word_positions;		// usato se il node ha tipo == "WORD"
  };
  unsigned int num_positions;
  int *depth_vector;
  unsigned int num_depth;

  int *range_results;							// vettore contenente per ogni coppia di posizioni il numero di
												//	risultati associati a tale posizione
  unsigned char is_setted_number_result;		//

  char **attributes_list;  						//  un vettore di stringhe dove nome e valore di un attributo
												// 	occupano posizioni contigue del vettore
  unsigned int num_attributes;
  unsigned char is_potable;
} query_tree_node;


// struttura dell'elemento del vettore pila_parsing
typedef struct {
  query_tree_node *node;
  unsigned int childs_vector_dimension;
} node_pila;


/* -------------------------------------------------------- */
/* ---------  strutture dati per la tabella hash ---------- */
/* -------------------------------------------------------- */

/*! Element of the catenating lists in the hash table.  */
typedef struct Hash_node {
  char	*variable_name;          	//!< name of the variable
  int index;        			//!< index of the variable
  char *node_name;			//!< name of node of the variable
  struct Hash_node *next;         	//!< pointer to the next item in the list
} Hash_node;

typedef Hash_node **Hash_nodeptr_array; //!< array of pointers to Hash_nodes

/*! Data structure for an hash table managed by catenating lists and
  MTF. The field "size" indicates the table size, whereas the field
  "card" indicates the number of objects stored into the catenating
  lists. */
typedef struct {
  int size;             		//!< hash table size
  int card;             		//!< number of stored items
  Hash_nodeptr_array table;  		//!< pointer to the hash table
} HHash_table;


/*******************************************************************
	                   XCDE QUERY CONSOLE
 *******************************************************************/

typedef struct XCDE_Console_Query_Type_Struct {

  /* ------------- information about query ----------- */
  char *select_string;
  char **file_vector;
  int number_of_files;
  char *return_string;
  int first_range;
  int last_range;

  /* ------------- information about results ----------- */
  query_tree_node *select_tree;
  query_tree_node **trees_vector;
  int number_of_trees;
  query_tree_node *return_tree;
  HHash_table *hash_table;

  /* ------------- information about query state ----------- */
  unsigned char XCDE_query_parsed;
  unsigned char XCDE_query_executed;

} XCDE_Console_Query_Type;





/*******************************************************************/

// function that gets the IL of positions of the token with the specified Rank.
// Example : ret=XCDE_Get_IL_ListPos(Rank, &IL, &IL_Num_El, &Console)
// where Rank is the rank of the element, &IL is the memory address of
// the variable that will contain the IL, &IL_Num_El is the memory address of
// the variable that will contain the number of elements of the IL and &Console
// is the memory address of the console data struct.
// ret value is 1 if there aren't errors, or -1 otherwise.



/* -------------------------------------------------------- */
/* ---------  		API FUNCTIONS		 ---------- */
/* -------------------------------------------------------- */


/*******************************************************************/

// function that parse query_string and fill some Console_Query fields.
//
// Example : ret=XCDE_Parse_Query_String(query_string, &Console_Query)
// where query_string is the string that contain the query and &Console_Query
// is the memory address of the console_query data struct.
// ret value is 1 if there aren't errors, or -2 if there are problem in query sintax,
// -1 otherwise.

int XCDE_Parse_Query_String(char *query_string, XCDE_Console_Query_Type *Console_Query);



/*******************************************************************/

// function that gets the number of results joined to tree without build result table.
// Example : ret=XCDE_Get_Number_Results(ptrTree, &num_res, &Console_Query)
// where ptrTree is the pointer to tree and &num_res is the memory address of the
// integer variable that will contain the number of result joined to tree and &Console_Query
// is the memory address of the console_query data struct.
// ret value is 1 if there aren't errors, or -1 otherwise.

int XCDE_Get_Number_Results(query_tree_node *tree, int *number_results, XCDE_Console_Query_Type *Console_Query);



/*******************************************************************/

// function that executes the query in Console_Query and fill the position
// of results building one tree for each file in query.
// Example : ret=XCDE_Execute_Query(&Console_Query)
// where Console_Query is the memory address of the console_query data struct.
// ret value is 1 if there aren't errors, or -2 if there are problem in query sintax,
// -1 otherwise.

int XCDE_Execute_Query(XCDE_Console_Query_Type *Console_Query);



/*******************************************************************/

// function that gets the table result for one of the trees build from
// XCDE_Execute_Query.
// Example : ret=XCDE_Compute_TableResult(index, &result_table, &size_result_table, &Console_Query)
// where index is the number of tree (in the console_query data struct) for which compute the table
// results, result_table is the memory address of the variable that will contain results, size_result_table
// is the memory address of the variable that will contain the number of results and Console_Query is
// the memory address of the console_query data struct.
// ret value is 1 if there aren't errors, or -2 if there are problem in query sintax,
// -1 otherwise.

int XCDE_Compute_TableResult(int index_albero, XCDE_IL_Couple_Type ***risultato, int *size_vettore_risultato, XCDE_Console_Query_Type *Console_Query);




int XCDE_View_Result(int index_file, XCDE_IL_Couple_Type **risultato, int size_vettore_risultato, unsigned char flag_position, XCDE_Console_Query_Type *Console_Query);


// Funzione che scarica il contenuto del vettore result e della tabella hash in un file
//
// Write in a file ".result" the result vector and hash table with an HEADER
//
// Header contains:
// 	1. number of result in risultato
// 	2. number of stored objects in hash table (number of variables)
//	3. pointer vector (which contain (number of result)/SKIP_RESULT integer)
//	4. pointer to hash table byte
//	5. pointer to name of file byte
//
// Body contains:
//
// 	for each result in risultato:
//		1. length in byte of comressed result
//		2. for each stored variable
//			1. open position (compressed)
//			2. open offset (compressed)
//			3. close position (compressed)
//			4. close offset (compressed)
//
// 	for each object stored in hash table:
// 		1. variable name length
//		2. variable name
// 		3. variable index (compressed)
// 		4. name of associated node length
// 		5. name of associated node

int XCDE_Write_ResultFile(XCDE_IL_Couple_Type **risultato, int size_vettore_risultato, char *filename, XCDE_Console_Query_Type *Console_Query);


int XCDE_Read_Range_ResultFile(char *filename, int first_range, int last_range, XCDE_IL_Couple_Type ***risultato, int *size_vettore_risultato, XCDE_Console_Query_Type *Console_Query);


int XCDE_Read_ResultFile(char *filename, XCDE_IL_Couple_Type ***risultato, int *size_vettore_risultato, XCDE_Console_Query_Type *Console_Query);


int XCDE_Search(char *query_string, unsigned char flag_position);
