/********************************************************************
		          xcde_build.c 
                      contains functions: 
                        main, output_IL
	          Xml Compressed Document Engine
	                 Building Program
	        Andrea Mastroianni - Paolo Ferragina
          Dipartimento di Informatica - Universita' di Pisa
	               ferragin@di.unipi.it
                       mastroia@di.unipi.it
********************************************************************/
#include"xcde_consts.h"
#include"xcde_build.h"
//#include "xcde_lib.h"
/*******************************************************************
                   DECLARING GLOBAL VARIABLES
 *******************************************************************/

// FILES POINTERS
FILE *inputFile=NULL ; 

FILE *configFile=NULL ;

FILE *dictsFile[XCDE_NUMBER_DICT] ;

FILE *cmpdictFile=NULL ;

FILE *bodyFile=NULL ;

FILE *headerFile=NULL ;

FILE *IL_File=NULL ;

FILE *tagstructFile=NULL ;

FILE *depthstructFile=NULL ;

FILE *debugPhysicView=NULL ;

FILE *debugIndexFile=NULL ;

FILE *debugStatsFile=NULL ;

// FILES' NAMES POINTERS
char *inputFilename=NULL ;

char *dictionaryFilename[XCDE_NUMBER_DICT] ;

char *cmpdictionaryFilename=NULL ;

char *headerFilename=NULL ;

char *bodyFilename=NULL ;

char *IL_Filename=NULL ;

char *tagstructFilename=NULL ;

char *depthstructFilename=NULL ;

char *debugPhysicFilename=NULL ;

char *debugStatsFilename=NULL ;

char *debugIndexFilename=NULL ;

char *configFilename=NULL ;

// input file dimension
int inputFilesize=0 ;

// buffer used to parsing the source file
char expat_parsing_buffer[XCDE_EXPAT_PARSING_BUFFER_SIZE] ;

// buffer used to decompress zlib body blocks
char zlib_decompress_buffer[2*XCDE_COMPRESS_BLOCK_SIZE] ;

// buffer used to compress zlib body blocks
char zlib_compress_buffer[2*XCDE_COMPRESS_BLOCK_SIZE] ;

// pointer to the previous buffer
int pointer_decompress_buffer=0 ;

// flag which is on if a CDATA section is being parsed
unsigned
char expat_CDATA_section_flag=0 ;

// flag which is on if a DTD declaration is being parsed
unsigned
char expat_DTD_flag=0u ;

// flag which is on if the CDATA section has been derived from an
// internal entity
unsigned
char expat_CDATA_expansion_flag=0 ;

// flag which is on if a tag is empty (es. <TAG />)
unsigned
char empty_tag_flag=0u ;

// flag which is on if the verbosing option is present
unsigned
char verbosing_flag=0u ;

// flag which is on if the proximity option is present
unsigned
char proximity_flag=1u ;

// flag which is on if the tagstruct option is present
unsigned
char tagstruct_flag=0u ;

// flag which is on if the depthstruct option is present
unsigned
char depthstruct_flag=0u ;

// flag which is on if the user used a different configuration
// file
unsigned
char different_configfile_flag=0u ;

// flag wich is on if the parser must ignore a recursive tag
unsigned
char recursivetag_flag=0u ;

// variable used to compute the word proximity
int proximity_counter=0 ;

// variable used to compute the element displacement into an internal
// entity
int element_displacement=0 ;

// number of tokens into the dictionary
int number_tokens=0 ;

// pointer to the body file (HUFFMAN)
int last_body_position_written=0 ;

int bodyFileSize=0 ;

int body_size_without_zlib=0 ;

int dictionaryFileSize=0 ;

int expat_number_tag_opened=0 ;

// matrix which contains the dictionary division into
// tags with the codeword length
// elem[0][0] number of TAG with codeword length 1
// elem[1][0] number of TAG with codeword length 2
// elem[2][0] number of TAG with codeword length 3
// ...
// elem[0][1] number of ATTR with codeword length 1
// elem[1][1] number of ATTR with codeword length 2
// elem[2][1] number of ATTR with codeword length 3
// ...
// elem[0][2] number of WORD with codeword length 1
// elem[1][2] number of WORD with codeword length 2
// elem[2][2] number of WORD with codeword length 3
// ...
int dict_elem[XCDE_HUFF_MAX_LEVEL][XCDE_NUMBER_DICT] ;

// hash table which contains the tokens found during the parsing
HashTable ht ;

// hash table which contains general internal entities
HashTable expat_general_declared_entity ;

// hash table which contains the separators declared
HashTable expat_separators_table ;

// hash table which contains the separators for attribute
HashTable expat_attribute_separators_table ;

// Huffman tree
Hashptr *Huffman_tree=NULL ;

// data structure used by Expat
XML_Parser parser ;

char *expat_source_pointer=NULL ;

// variables used to traduce the physical text into the logical view
char *xml_code=NULL ;

long xml_dimension=0 ;

long xml_position=0 ;

// last physical position where an entity has been expanded ;
long last_expanded_position=-1 ;

// flag which is on if the position parsed has changed
unsigned
char changed_position=0 ;

// flag which indicates if there is a correspondence between
// the physical view and the logical one
unsigned
char physical_correspondence=0u ;

// last value of the physical_correspondence flag
unsigned
char last_physical_correspondence=0u ;

// rank needed to compute proximity into entity expansion
int elementRank=1 ;

// used by the stat function
struct stat sbuf ;

// buffer used to create physical and logical tokens
Textbuffer textbuffer ;

// word separators list
char separators[XCDE_MAX_NUMBER_SEPARATORS]=" <>?!,.;:()[]{}\t\n\"" ;

// attribute separators list
char attribute_separators[XCDE_MAX_NUMBER_SEPARATORS]=" <>?!,.;:()[]{}\t\n\"" ;

// array used to contain compressed body blocks pointers
char *body_pointers_cmp_buffer=NULL ;

// pointer to the previous buffer
int body_pcmp_point=0 ;

// initial dimension of the previous buffer
int body_pointers_cmp_buffer_dimension=2*K ;

// number of pointers into the previous buffer
int number_body_points=1 ;

// array used to contain compressed inverted lists pointers
char *IL_pointers_buff=NULL ;

int IL_pointers_buff_length=0 ;

int number_IL_table_pointer=0 ;

int diff_position=0 ;

int new_diff_position=0 ;

int new_body_pointers_point=0 ;

int number_tagstruct_intervals=0 ;

 //XCDE_Console_Type XCDE_Console ;

//int res=0;
/*******************************************************************
                             FUNCTIONS
 *******************************************************************/

int
main(int argc,
     char *argv[])
{ // variables used as counters
  int j=0 ;
  int k=0 ;
  int i=0 ;

  // variable used to count the number of tag
  int tag_elem=0;

  // variable used to check functions results
  int result=0 ;

  // variables used to construct the Huffman code
  int offsetcw[5] ;
  int firstcw[5] ;
  int count_depth[5] ;

  // variable used to compute the number of levels
  // of the Huffman tree
  int current_level=0 ;

  // used in the Huffman code
  int count=0 ;

  // used to write integer in the header file
  int temp=0 ;

  // maximum length of the Huffman tree
  int max_length=0 ;

  // counter of the number of level in the Huffman Tree
  int number_levels=0 ;

  int len=0 ;

  // buffer used to compress number with the continuation bit algorithm
  char continuation_bit_buffer[XCDE_CONT_BIT_BUFFER_MAX_LENGTH]="\0" ;

  // number of byte of a number compressed by the continuation bit
  unsigned
    long int len_continuation_bit=0 ;

  unsigned int
    integer_coding_bit_offset=0 ;

  // used to compress integer
  uchar *pp ;

  // ----------------------
  // VARIABLES USED BY ZLIB
  // ----------------------

  Byte *uncompr ;

  uLong comprLen ;

  uLong uncomprLen ;

  int cmp_err=0 ;

  /*******************************************************************
	                 creates data structures
   *******************************************************************/

  setbuf(stdout, NULL) ;

  setbuf(stderr, NULL) ;

  for(k=0; k<XCDE_NUMBER_DICT; ++k)

    dictsFile[k]=NULL ;

  body_pointers_cmp_buffer=malloc(body_pointers_cmp_buffer_dimension*sizeof(char)) ;

  if (body_pointers_cmp_buffer == NULL) {

    fprintf(stderr, "\n\tFatal Error: problems during memory allocation\n\tError code 1\n") ;

    exit(-1) ;
  }

  body_pcmp_point=0 ;

  expat_general_declared_entity=hcreate(XCDE_ENTITY_HT_DIMENSION) ;

  expat_separators_table=hcreate(XCDE_SEPARATORS_HT_DIMENSION) ;

  expat_attribute_separators_table=hcreate(XCDE_SEPARATORS_HT_DIMENSION) ;

  init_buffer(&textbuffer,
	      XCDE_TEXTBUFFER_SIZE) ;

  /*******************************************************************
       inserts XML predefined general entities into the entity table
   *******************************************************************/

  hinsert((char *) "&amp;",
	  strlen("&amp;"),
	  NOTHING,
	  (char *) "&",
	  expat_general_declared_entity,
	  0,
	  0,
	  XCDE_ENTITY_HASHTABLE) ;

  hinsert((char *) "&lt;",
	  strlen("&lt;"),
	  NOTHING,
	  (char *) "<",
	  expat_general_declared_entity,
	  0,
	  0,
	  XCDE_ENTITY_HASHTABLE) ;

  hinsert((char *) "&gt;",
	  strlen("&gt;"),
	  NOTHING,
	  (char *) ">",
	  expat_general_declared_entity,
	  0,
	  0,
	  XCDE_ENTITY_HASHTABLE) ;

  hinsert((char *) "&apos;",
	  strlen("&apos;"),
	  NOTHING,
	  (char *) "'",
	  expat_general_declared_entity,
	  0,
	  0,
	  XCDE_ENTITY_HASHTABLE) ;

  hinsert((char *) "&quot;",
	  strlen("&quot;"),
	  NOTHING,
	  (char *) "\"",
	  expat_general_declared_entity,
	  0,
	  0,
	  XCDE_ENTITY_HASHTABLE) ;

  /*******************************************************************
	 prints XCDE_BUILD informations and instructions
   *******************************************************************/

  fprintf(stdout, "\t------------------------------------------------------------------") ;

  fprintf(stdout, "\n\t\t       XCDE_BUILD : Version 1.0Beta 11-07-2002 ") ;

  fprintf(stdout, "\n\t\t         X.ml C.ompressed D.ocument E.ngine") ;

  fprintf(stdout, "\n\t\t        Paolo Ferragina - Andrea Mastroianni ") ;

  fprintf(stdout, "\n\t\t    Dipartimento di Informatica - Universita' di Pisa ") ;

  fprintf(stdout, "\n\t\t               XCDE File System Builder ") ;

  fprintf(stdout, "\n\t------------------------------------------------------------------\n\n") ;

  if (argc == 1) {

    fprintf(stdout, "\t xcde_build [-v] [-c configfile] [-t] [-p] [-f] sourcefile") ;

    fprintf(stdout, "\n\n\n\t-v : verbose") ;

    fprintf(stdout, "\n\n\t-c : configuration file") ;

    fprintf(stdout, "\n\n\t-t : insert tagstruct") ;

    fprintf(stdout, "\n\n\t-p : no proximity") ;

    fprintf(stdout, "\n\n\t-f : force recursive tags") ;

	fprintf(stdout, "\n\n\t-d : insert depthstruct \n\n") ;

    exit(-1) ;
  }

  /*******************************************************************
	      reads parameters passed by the user
   *******************************************************************/

  for (k=1 ; k<argc; ++k)

    if ((argv[k][0]=='-') && (argv[k][1]=='v'))

      verbosing_flag=1u ;

    else

      if ((argv[k][0]=='-') && (argv[k][1]=='c')) {

	if (k<(argc-1)) {

	  configFilename=(char *) malloc(strlen(argv[k+1])+1) ;

	  if (configFilename == NULL) {

	    fprintf(stderr, "\n\tFatal Error: problems during memory allocation\n\tError code 2\n") ;

	    DELETE_FILES ;

	    exit(-1) ;
	  }

	  strcpy(configFilename,
		 argv[k+1]) ;

	  different_configfile_flag=1u ;

	  ++k ;
	}

	else

	  // there's the -c option but not the name of the new configuration file
	  { fprintf(stderr, "\n\tFatal Error: no valid configfile name\n\tError code 3\n") ;

	  DELETE_FILES ;

	  exit(-1) ;
	  }
      }

      else

	if ((argv[k][0]=='-') && (argv[k][1]=='p'))

	  proximity_flag=0u ;

	else

	  if ((argv[k][0]=='-') && (argv[k][1]=='f'))

	    recursivetag_flag=1u ;

	  else

	    if ((argv[k][0]=='-') && (argv[k][1]=='t'))

	      tagstruct_flag=1u ;

		else

		  if ((argv[k][0]=='-') && (argv[k][1]=='d'))

			depthstruct_flag=1u;

		  else

	        if (inputFilename==NULL) {

		  inputFilename=(char *) malloc(strlen(argv[k])+1) ;

		  if (inputFilename == NULL) {

		  fprintf(stderr, "\n\tFatal Error: problems during memory allocation\n\tError code 4\n") ;

		  DELETE_FILES ;

		  exit(-1) ;
		}

		strcpy(inputFilename,
		       argv[k]) ;

	      }

	      else {

		fprintf(stderr, "\n\tFatal Error: no valid option\n\tError code 5\n") ;

		DELETE_FILES ;

		exit(-1) ;
	      }

  if (inputFilename == NULL) {

    DELETE_FILES ;

    exit(-1) ;
    }

  /*******************************************************************
                         opens or creates files
   *******************************************************************/

  // creates dictionaries names
  for(k=0;k<3; ++k) {

    dictionaryFilename[k]=(char *) malloc(strlen(inputFilename)+strlen(XCDE_DICTIONARY_SUFFIX)+2) ;

    if (dictionaryFilename[k] == NULL)  { 

      fprintf(stderr, "\n\tFatal Error: problems during memory allocation\n\tError code 6\n") ; 
      
      DELETE_FILES ;

      exit(-1) ;
    }

    sprintf(dictionaryFilename[k], "%s%s%d", inputFilename, XCDE_DICTIONARY_SUFFIX,k) ;

    if (verbosing_flag) {

      fprintf(stdout, "\nCreating file %s", dictionaryFilename[k]) ;

      if (k==0) fprintf(stdout, " TAG Dictionary ...  ") ;

      if (k==1) fprintf(stdout, " ATTR Dictionary ... ") ; 

      if (k==2) fprintf(stdout, " WORD Dictionary ... ") ;
    }

    // opening dictionary files 
    if ((dictsFile[k]=fopen(dictionaryFilename[k],"wb"))==NULL) {

      fprintf(stderr,"\n\tFatal Error: Dictionary file\n\tError code 7\n");

      DELETE_FILES ;
      
      exit(-1); 
      } 

    if (verbosing_flag)
      
      fprintf(stdout, "OK") ; 
  }
  
  headerFilename=(char *) malloc(strlen(inputFilename)+strlen(XCDE_HEADER_SUFFIX)+1) ; 

  if (headerFilename == NULL) {

    fprintf(stderr, "\n\tFatal Error: problems during memory allocation\n\tError code 8\n") ; 

    DELETE_FILES ;

    exit(-1) ;
    }

  strcpy(headerFilename,
	 inputFilename) ; 

  strcat(headerFilename,
	 XCDE_HEADER_SUFFIX) ;

  if (verbosing_flag)

    fprintf(stdout, "\nCreating Header file %s ... ", headerFilename) ;
  
  if ((headerFile=fopen(headerFilename,"wb"))==NULL) { 

    fprintf(stderr,"\n\tFatal Error: Header file\n\tError code 9\n") ;

    DELETE_FILES ;

    exit(-1); 
    }
  
  if (verbosing_flag) 

    fprintf(stdout, "OK") ; 

  bodyFilename=(char *) malloc(strlen(inputFilename)+strlen(XCDE_BODY_SUFFIX)+1) ;

  if (bodyFilename == NULL) { 

    fprintf(stderr, "\n\tFatal Error: problems during memory allocation\n\tError code 10\n") ; 

    DELETE_FILES ;

    exit(-1) ;
    }

  strcpy(bodyFilename, 
	 inputFilename) ;

  strcat(bodyFilename,
	 XCDE_BODY_SUFFIX) ;

  if (verbosing_flag)

    fprintf(stdout, "\nCreating Body file %s ... ", bodyFilename) ; 

  if ((bodyFile=fopen(bodyFilename,"wb"))==NULL) { 

    fprintf(stderr,"\n\tFatal Error: Body file\n\tError code 11\n");
    
    DELETE_FILES ; 
    
    exit(-1);
  }
  
  if (verbosing_flag)

    fprintf(stdout, "OK") ; 

  IL_Filename=(char *) malloc(strlen(inputFilename)+strlen(XCDE_IL_SUFFIX)+1) ; 
  
  if (IL_Filename == NULL) { 
    
    fprintf(stderr, "\n\tFatal Error: problems during memory allocation\n\tError code 12\n") ; 

    DELETE_FILES ; 

    exit(-1) ; 
  }

  strcpy(IL_Filename,
	 inputFilename) ;
  
  strcat(IL_Filename,
	 XCDE_IL_SUFFIX) ;

  if (verbosing_flag)
    
    fprintf(stdout, "\nCreating Inverted Lists file %s ... ", IL_Filename) ;

  if ((IL_File=fopen(IL_Filename,"wrb"))==NULL) { 

    fprintf(stderr,"\n\tFatal Error: Inverted Lists file\nError code 13\n");
    
    DELETE_FILES ;

    exit(-1);
  }
  
  if (verbosing_flag)

    fprintf(stdout, "OK") ;

  // executes these operations only if the tagstruct option is setted 
  if (tagstruct_flag) {

    tagstructFilename=(char *) malloc(strlen(inputFilename)+strlen(XCDE_TAGSTRUCT_SUFFIX)+1) ;

    if (tagstructFilename == NULL) { 
      
      fprintf(stderr,"\n\tFatal Error: problems during memory allocation\n\tError code 14\n") ; 
      
      DELETE_FILES ;
      
      exit(-1);
    }
    
    strcpy(tagstructFilename,
	   inputFilename) ; 

    strcat(tagstructFilename,
	   XCDE_TAGSTRUCT_SUFFIX) ; 
    
    if (verbosing_flag) 

      fprintf(stdout, "\nCreating Tagstruct file %s ... ", tagstructFilename) ;

    if ((tagstructFile=fopen(tagstructFilename,"wb"))==NULL) {

      fprintf(stderr,"\n\tFatal Error: Tagstruct file\n\tError code 15\n");
      
      DELETE_FILES ; 
      
      exit(-1);
    }

    if (verbosing_flag)

      fprintf(stdout, "OK") ; 
  }

    // executes these operations only if the depthstruct option is setted
  if (depthstruct_flag) { 

    depthstructFilename=(char *) malloc(strlen(inputFilename)+strlen(XCDE_DEPTHSTRUCT_SUFFIX)+1) ;
    
    if (depthstructFilename == NULL) {
      
      fprintf(stderr,"\n\tFatal Error: problems during memory allocation\n\tError code 14\n") ;
      
      DELETE_FILES ;

      exit(-1); 
    }
    
    strcpy(depthstructFilename,
	   inputFilename) ;
    
    strcat(depthstructFilename,
	   XCDE_DEPTHSTRUCT_SUFFIX) ; 
    
    if (verbosing_flag) 

      fprintf(stdout, "\nCreating Depthstruct file %s ... ", depthstructFilename) ;

    if ((depthstructFile=fopen(depthstructFilename,"wb"))==NULL) {

      fprintf(stderr,"\n\tFatal Error: Depthstruct file\n\tError code 15\n");
      
      DELETE_FILES ;

      exit(-1);
    }
    
    if (verbosing_flag)

      fprintf(stdout, "OK") ;
  }
  
  if (verbosing_flag)

    fprintf(stdout, "\nOpening Configuration file ... ") ; 

  if (different_configfile_flag == 0u)

    if (setenv(XCDE_CONFIGDIR_ENV,
	       NULL,
	       0)) {

      configFilename=(char *) malloc(strlen(XCDE_CONFIGFILE_STANDARD_NAME)+strlen(getenv(XCDE_CONFIGDIR_ENV))+2) ;

      if (configFilename == NULL) {

	fprintf(stderr,"\n\tFatal Error: problems during memory allocation\n\tError code 16\n") ;
	
	DELETE_FILES ;
	
	exit(-1);
      } 

      sprintf(configFilename, "%s/%s", getenv(XCDE_CONFIGDIR_ENV), XCDE_CONFIGFILE_STANDARD_NAME) ;
    }

    else

      { configFilename=(char *) malloc(strlen(XCDE_CONFIGFILE_STANDARD_NAME)+1) ;

      if (configFilename == NULL) { 

	fprintf(stderr,"\n\tFatal Error: problems during memory allocation\nError code 17\n") ;

	DELETE_FILES ;

	exit(-1);
      }

      sprintf(configFilename, "%s", XCDE_CONFIGFILE_STANDARD_NAME) ;
      }

  if ((configFile=fopen(configFilename, "r"))==NULL) {

    fprintf(stderr, "\n\tFatal Error: configuration file %s\nError code 18", configFilename) ;

    DELETE_FILES ;

    exit(-1) ;
  }

  // reads data contained in configFile
  read_conf_file(configFile) ; 

  fclose(configFile) ;

  configFile=NULL ;

  free(configFilename) ;

  configFilename=NULL ;

  if (verbosing_flag)

    fprintf(stdout, "OK") ;

  // debug files are created only if the verbosing option is setted
  if (verbosing_flag) {

    debugIndexFilename=(char *) malloc(strlen(inputFilename)+strlen(XCDE_DEBUG_INDEX_SUFFIX)+1) ;

    if (debugIndexFilename == NULL) {

      fprintf(stderr,"\n\tFatal Error: problems during memory allocation\n\tError code 19\n") ;

      DELETE_FILES ;

      exit(-1) ;
    }

    strcpy(debugIndexFilename,
	   inputFilename) ;

    strcat(debugIndexFilename,
	   XCDE_DEBUG_INDEX_SUFFIX) ;
    
    fprintf(stdout, "\nCreating Index Debugging file %s ... ", debugIndexFilename) ;

    if ((debugIndexFile=fopen(debugIndexFilename, "w"))==NULL) { 
      
      fprintf(stderr,"\n\tFatal Error: Index Debugging file\n\tError code 20\n") ; 
      
      DELETE_FILES ; 

      exit(-1);
    } 
    
    fprintf(stdout, "OK") ; 

    debugPhysicFilename=(char *) malloc(strlen(inputFilename)+strlen(XCDE_DEBUG_PHYSIC_SUFFIX)+1) ; 
    
    if (debugPhysicFilename == NULL) { 

      fprintf(stderr,"\n\tFatal Error: problems during memory allocation\n\tError code 21\n") ; 

      DELETE_FILES ; 

      exit(-1);
    } 

    strcpy(debugPhysicFilename,
	   inputFilename) ;

    strcat(debugPhysicFilename,
	   XCDE_DEBUG_PHYSIC_SUFFIX) ;

    fprintf(stdout, "\nCreating Physical View Debugging file %s ... ", debugPhysicFilename) ; 

    // apertura del file con la visione a livello fisico
    if ((debugPhysicView=fopen(debugPhysicFilename, "w")) == NULL) {

      fprintf(stderr, "\n\tFatal Error: Physic View Debugging file %s ... \n\tError code 22\n", debugPhysicFilename) ; 

      DELETE_FILES ;

      exit(-1) ;
    }
    
    fprintf(stdout, "OK") ;
    
    debugStatsFilename=(char *) malloc(strlen(inputFilename)+strlen(XCDE_DEBUG_STATS_SUFFIX)+1) ;
    
    if (debugStatsFilename == NULL) { 

      fprintf(stderr,"\n\tFatal Error: problems during memory allocation\n\tError code 23\n") ; 

      DELETE_FILES ; 

      exit(-1) ; 
    }
    
    strcpy(debugStatsFilename,
	   inputFilename) ; 

    strcat(debugStatsFilename, 
	   XCDE_DEBUG_STATS_SUFFIX) ;

    fprintf(stdout, "\nCreating Statistic Debugging file %s ... ", debugStatsFilename) ; 

    // apertura del file con le statistiche 
    if ((debugStatsFile=fopen(debugStatsFilename, "w")) == NULL) { 

      fprintf(stderr, "\nFatal Error: Statistic Debugging file %s ...\n\tError code 24\n", debugStatsFilename) ;

      DELETE_FILES ;
      
      exit(-1) ;
    }

    fprintf(stdout, "OK") ;
  }

  if (verbosing_flag)

    fprintf(stdout, "\nOpening XML source file %s ... ", inputFilename) ;
    
  if ((inputFile=fopen(inputFilename, "r"))==NULL) {

    fprintf(stderr, "\n\tFatal Error: XML Source file %s\n\tError code 25\n", inputFilename) ;

    DELETE_FILES ; 
    
    exit(-1) ; 
  }
  
  if (verbosing_flag) 

    fprintf(stdout, "OK") ;

  /*******************************************************************
                       maps source file into the memory
   *******************************************************************/    

  result=stat(inputFilename ,
	      &sbuf);

  if (result == -1) { 

    fprintf(stderr, "\n\tFatal Error: problems in stat function on file %s\nError code 26\n", inputFilename) ; 

    DELETE_FILES ;
    
    exit(-1) ;
  }

  inputFilesize = (int) sbuf.st_size;

  // create the hash table
  ht = hcreate((inputFilesize/XCDE_HASHTABLE_RATIO == 0) ? XCDE_HASHTABLE_RATIO : inputFilesize / XCDE_HASHTABLE_RATIO) ;

  // maps the XML source file into the memory
  expat_source_pointer = (char *) mmap(NULL,
				       inputFilesize,
				       PROT_READ,
				       MAP_SHARED,
				       fileno(inputFile),
				       0);

  if (expat_source_pointer == NULL) { 
    
    fprintf(stderr, "\n\tFatal Error: problems in mapping file %s\n\tError code 27\n", inputFilename) ;
    
    DELETE_FILES ; 
    
    exit(-1) ; 
  }
  
  /*******************************************************************
                          configures EXPAT system
   *******************************************************************/

  // create the EXPAT parser
  parser = XML_ParserCreate(NULL) ; 

  if (!parser) {

    fprintf(stderr, "\n\tFatal Error: problems with EXPAT Parser\n\tError code 28\n") ;
    
    DELETE_FILES ;

    exit(-1) ;
  }

  // handlers for the DTD declaration start and end
  XML_SetDoctypeDeclHandler(parser,
			    doctypestart,
			    doctypeend) ;

  XML_SetParamEntityParsing(parser,
			    XML_PARAM_ENTITY_PARSING_ALWAYS) ;

  // handlers for the tag start and end
  XML_SetElementHandler(parser,
			start, 
			end) ;

  // handler for the comments
  XML_SetCommentHandler(parser,
			comment) ; 

  // handler for the text 
  XML_SetCharacterDataHandler(parser, 
			      text)  ;

  // handlers for the CDATA section start and end
  XML_SetCdataSectionHandler(parser, 
			     cdstart, 
			     cdend) ;

  // handler for the declaration of and internal entity into the DTD
  XML_SetInternalParsedEntityDeclHandler(parser,
					 InternalEntity) ;

  // handler for the default elements
  XML_SetDefaultHandlerExpand(parser,
			      defa) ;

  /*******************************************************************
       		       parsing the XML source
    First Pass: finds physical and logical tokens, computes 
                their probabilities and words proximities
   *******************************************************************/

  if (verbosing_flag) 

    fprintf(stdout, "\n\nParsing the XML source ... ") ;

  for ( ; ; ) {

    int done=0 ;

    int bytes_read=0 ;
    
    bytes_read=fread(expat_parsing_buffer, 
		     1, 
		     XCDE_EXPAT_PARSING_BUFFER_SIZE, 
		     inputFile) ;
    
    if (ferror(inputFile)) { 
      
      fprintf(stderr, "\n\tFatal Error: problems in reading source file\n\tError code 29\n") ;
      
      DELETE_FILES ; 

      exit(-1) ; 
    }
	  
    done=feof(inputFile) ;

    if (!XML_Parse(parser,
		   expat_parsing_buffer,
		   bytes_read,
		   done)) {

      fprintf(stderr, "\n\tFatal Error: problems in parsing raw %d %s\n\tError code 30\n",
	      
	      XML_GetCurrentLineNumber(parser),

	      XML_ErrorString(XML_GetErrorCode(parser))) ;

      DELETE_FILES ;

      exit(-1) ;
    }

    if (done) break ;
  }

  if (verbosing_flag)

    fprintf(stdout, "OK") ;

  /*******************************************************************
     builds the Huffman tree with the tokens found during parsing
                        and writes dictionaries
  *******************************************************************/

  fprintf(stdout, "\nNumber of Tokens %d\n", number_tokens) ; fflush(stdout) ;

  if (verbosing_flag)

    fprintf(stdout, "\n\nBuilds Huffman Tree ... ") ;

  Huffman_tree=Huffman_byte_code(firstcw,
				 offsetcw,
				 count_depth);

  for(k=number_tokens; k<((number_tokens)+(128 - (number_tokens % 127))); ++k)

    { if (Huffman_tree[k] != NULL)

      free(Huffman_tree[k]) ;

    Huffman_tree[k]=NULL ;
    }

  if (verbosing_flag)

    fprintf(stdout, "OK") ;

  if (verbosing_flag)

    fprintf(stdout, "\nWriting Header file %s ... ", headerFilename) ;

  // writes Huffman tree informations into the Header file
  for (k=1; k<=4; k++)
      for (temp=firstcw[k],j=3; j>=0; j--)
	fputc(((temp >> 7*j) & 0x7F), headerFile) ;

  for (k=1; k<=4; k++)
    for (temp=offsetcw[k],j=3; j>=0; j--)
      fputc((temp >> 8*j) & 0xFF, headerFile) ;

  if (verbosing_flag)

    fprintf(stdout, "OK") ;


  for(k=0; k<3; ++k) {

    // dictionary files begin with a NULL value
    fputc(0, dictsFile[k]) ;

    if (verbosing_flag)

      fprintf(stdout, "\nWriting Dictionary file %s ... OK", dictionaryFilename[k]) ;
  }

  // initializes the variable used to compute the Huffman tree
  // maximum length
  max_length=0 ;

  // initializes the counter of the Huffman tree levels
  current_level=5 ;

  count=0 ;

  // loops over tokens
  for(i=0 ; i<number_tokens; i++) {

    if (Huffman_tree[i]->cw_len != current_level) {

      // copies the level of the token
      current_level = Huffman_tree[i]->cw_len ;

      // computes the maximum level of the Huffman tree
      if (current_level > max_length)

	max_length=current_level ;

      ++number_levels ;

      count=0 ;
    }

    if (Huffman_tree[i]->cw_len <= 0) {

      fprintf(stderr,"\n\tFatal Error: problems with token %s\n\tError code 31\n", Huffman_tree[i]->str) ;

      DELETE_FILES ;

      exit(-1) ;
    }

    // computes the token codeword: needed to create the body file
    Huffman_tree[i]->codeword = firstcw[current_level]+count ;

    // increases the number of elements with current_level codeword length
    ++dict_elem[current_level-1][Huffman_tree[i]->type] ;

    // increases the number of element with current_level codeword length
    count++;

    // checks if it's a void token
    if (Huffman_tree[i]->len_str==0) ;

    else

      // writes the token into the dictionary
      for(j=0; j<Huffman_tree[i]->len_str; j++)

	fputc(Huffman_tree[i]->str[j],
	      dictsFile[Huffman_tree[i]->type]) ;

    fputc(0,dictsFile[Huffman_tree[i]->type]) ;
  }

  // writes the maximum length of a codeword in the header file
  for (temp=max_length,j=3; j>=0; j--)

    fputc((temp >> 8*j) & 0xFF,
	  headerFile) ;

  // writes informations about dictionaries in the header file
  for(k=0; k<max_length; ++k)
    for (i=0; i<3; i++)
      for (temp=dict_elem[k][i],j=3; j>=0; j--)
	fputc((temp >> 8*j) & 0xFF,
	      headerFile) ;

  for(k=0; k<3; ++k) {

    // close the dictionary
    fclose(dictsFile[k]);

    if (verbosing_flag)

      fprintf(stdout, "\nCompressing file %s ... ", dictionaryFilename[k]) ;

    if ((dictsFile[k]=fopen(dictionaryFilename[k],"rb"))==NULL) {

      fprintf(stderr,"\n\tFatal Error: Dictionary file\n\tError code 32\n");

      DELETE_FILES ;

      exit(-1);
    }

    cmpdictionaryFilename=(char *) malloc(strlen(dictionaryFilename[k])+strlen(".cmp")+1) ;

    strcpy(cmpdictionaryFilename, dictionaryFilename[k]) ;

    strcat(cmpdictionaryFilename, ".cmp") ;

    if ((cmpdictFile=fopen(cmpdictionaryFilename,"wb"))==NULL) {

      fprintf(stderr,"\n\tFatal Error: Dictionary file\n\tError code 32\n");

      DELETE_FILES ;

      exit(-1);
    }

    uncompr=(Byte *) zlib_decompress_buffer ;

    while((uncomprLen=fread(zlib_decompress_buffer,
			    1,
			    XCDE_COMPRESS_BLOCK_SIZE,
			    dictsFile[k]))>0) {

      comprLen=2*XCDE_COMPRESS_BLOCK_SIZE ;

      cmp_err=compress2((Bytef *) zlib_compress_buffer,
			(uLongf *) &comprLen,
			(const Bytef *) uncompr,
			(uLong) uncomprLen,
			Z_BEST_COMPRESSION) ;

      if (cmp_err != Z_OK) { fprintf(stderr, "\n\tFatal Error: problems with zlib compression\n\tError code 36\n") ;

      DELETE_FILES ;

      exit(-1) ;
      }

      continuation_bit_c((unsigned long) comprLen,
			 (char *) continuation_bit_buffer,
			 (unsigned int) 0,
			 (unsigned long *) &len_continuation_bit) ;

      // writes the dimension of the compressed block to the compressed dictionary file
      for (j=0 ; j<(len_continuation_bit/8); ++j)

	fputc((char) continuation_bit_buffer[j],
	      cmpdictFile) ;

      for (i=0 ; i<comprLen; ++i)

	fputc((char) zlib_compress_buffer[i],
	      cmpdictFile) ;
    }

    if (cmpdictionaryFilename != NULL) {

      free(cmpdictionaryFilename) ;

      cmpdictionaryFilename=NULL ;
    }

    if (cmpdictFile != NULL) {

      fclose(cmpdictFile) ;

      cmpdictFile=NULL ;
    }

    if (dictionaryFilename[k] != NULL) {

      unlink(dictionaryFilename[k]) ;

      free(dictionaryFilename[k]) ;

      dictionaryFilename[k]=NULL ;
    }

    if (verbosing_flag)

      fprintf(stdout, "OK") ;
  }

  /*******************************************************************
   Second pass of EXPAT: writes the Body file and computes IL positions
   *******************************************************************/

  XML_ParserFree(parser) ;

  xml_code=NULL ;

  xml_dimension=0 ;

  xml_position=0 ;

  last_expanded_position=-1 ;

  expat_number_tag_opened=0 ;

  parser=XML_ParserCreate(NULL) ;

  if (!parser) {

    fprintf(stderr, "\n\tFatal Error: problems with EXPAT Parser\n\tError code 33\n") ;

    DELETE_FILES ;

    exit(-1) ;
  }

  XML_SetParamEntityParsing(parser,
			    XML_PARAM_ENTITY_PARSING_ALWAYS) ;

  // handlers for DTD declaration start and end
  XML_SetDoctypeDeclHandler(parser,
			    doctypestart,
			    doctypeend) ;

  // handlers for TAG open and close
  XML_SetElementHandler(parser,
			startcod,
			endcod) ;

  // handler for comments
  XML_SetCommentHandler(parser,
			commentcod) ;

  // handler for the text
  XML_SetCharacterDataHandler(parser,
			      textcod)  ;

  // handlers for CDATA sections start and end
  XML_SetCdataSectionHandler(parser,
			     cdstartcod,
			     cdendcod) ;

  // handler for default elements
  XML_SetDefaultHandlerExpand(parser,
			      defacod) ;

  // rewind the source file pointer
  fseek(inputFile,
	0L,
	SEEK_SET) ;

  if (verbosing_flag)

    fprintf(stdout, "\nWriting Body file %s ... ", bodyFilename) ;

  for ( ; ; ) {

    int done=0 ;

    int bytes_read=0 ;

    bytes_read=fread(expat_parsing_buffer,
		     1,
		     XCDE_EXPAT_PARSING_BUFFER_SIZE,
		     inputFile) ;

    if (ferror(inputFile)) {

      fprintf(stderr, "\n\tFatal Error: problems in reading source file\n\tError code 34\n") ;

      DELETE_FILES ;

      exit(-1) ;
    }

    done = feof(inputFile) ;

    if (!XML_Parse(parser,
		   expat_parsing_buffer,
		   bytes_read,
		   done)) {

      fprintf(stderr, "\n\tFatal Error: problems in parsing raw %d %s\n\tError code 35\n",

	      XML_GetCurrentLineNumber(parser),

	      XML_ErrorString(XML_GetErrorCode(parser))) ;

      DELETE_FILES ;

      exit(-1) ;
    }

    if (done) break ;
  }

  XML_ParserFree(parser) ;

  // compress and write the last block of the body file
  if (pointer_decompress_buffer > 0) {

    if (diff_position > 0)

      { // checks if there is enough space in the buffer to compress block point
	if ((body_pcmp_point) > (body_pointers_cmp_buffer_dimension-10)) {

	  body_pointers_cmp_buffer=realloc(body_pointers_cmp_buffer,
					   body_pointers_cmp_buffer_dimension*2) ;

	  body_pointers_cmp_buffer_dimension*=2 ;
	}

	continuation_bit_c((unsigned long) new_diff_position,
			   (char *) body_pointers_cmp_buffer+body_pcmp_point,
			   (unsigned int) 0,
			   (unsigned long *) &len_continuation_bit) ;

	body_pcmp_point+=(len_continuation_bit/8) ;

	++number_body_points ;
      }

    uncompr=(Byte *) zlib_decompress_buffer ;

    uncomprLen=pointer_decompress_buffer ;

    comprLen=2*XCDE_COMPRESS_BLOCK_SIZE ;

    cmp_err=compress2((Bytef *) zlib_compress_buffer,
		      (uLongf *) &comprLen,
		      (const Bytef *) uncompr,
		      (uLong) uncomprLen,
		      Z_BEST_COMPRESSION) ;

    if (cmp_err != Z_OK) { fprintf(stderr, "\n\tFatal Error: problems with zlib compression\n\tError code 36\n") ;

    DELETE_FILES ;

    exit(-1) ;
    }

    for (i=0 ; i<comprLen; ++i)

      fputc((char) zlib_compress_buffer[i],
	    bodyFile) ;
  }

  pointer_decompress_buffer=0 ;

  if (verbosing_flag)

    fprintf(debugStatsFile,
	    "\nNumber of pointers to Body File %d ", number_body_points) ;

  continuation_bit_c((unsigned long) inputFilesize,
		     (char *) continuation_bit_buffer,
		     (unsigned int) 0,
		     (unsigned long *) &len_continuation_bit) ;

  // writes the source XML file dimension in the header file
  for (j=0 ; j<(len_continuation_bit/8); ++j)

    fputc((char) continuation_bit_buffer[j],
	  headerFile) ;

  continuation_bit_c((unsigned long) body_size_without_zlib,
		     (char *) continuation_bit_buffer,
		     (unsigned int) 0,
		     (unsigned long *) &len_continuation_bit) ;

  // writes the body size without zlib compression (only with
  // Huffman code compression) in the hader file
  for (j=0 ; j<(len_continuation_bit/8) ; ++j)

    fputc((char) continuation_bit_buffer[j],
	  headerFile) ;

  // writes the proximity flag in the header file
  fputc((unsigned char) proximity_flag,
	headerFile) ;

  // writes the tagstruct flag in the header file
  fputc((unsigned char) tagstruct_flag,
	headerFile) ;

    // writes the depthstruct flag in the header file
  fputc((unsigned char) depthstruct_flag,
	headerFile) ;
  
  continuation_bit_c((unsigned long) number_body_points, 
		     (char *) continuation_bit_buffer, 
		     (unsigned int) 0,
		     (unsigned long *) &len_continuation_bit) ; 	

  // writes the number of body pointers in the header file 
  for(j=0 ; j<(len_continuation_bit/8); ++j) 
    
    fputc((char) continuation_bit_buffer[j],
	  headerFile) ; 

  // if the number of body is greater than 1 writes the
  // body pointers buffer in the header file 
  if (number_body_points > 1)
    
    for(j=0 ; j<body_pcmp_point; ++j)

      fputc((char) body_pointers_cmp_buffer[j],
	    headerFile) ;

  if (verbosing_flag)

    fprintf(debugStatsFile, 
	    "\nLength of the body pointers list %d", body_pcmp_point) ;

  if (body_pointers_cmp_buffer != NULL) {

    free(body_pointers_cmp_buffer) ;

    body_pointers_cmp_buffer=NULL ;
  }

  // computes the number of pointers to the Inverted Lists
  number_IL_table_pointer = number_tokens / XCDE_IL_POINTER_RATIO ;
  
  if ((number_tokens % XCDE_IL_POINTER_RATIO) > 0)

    ++number_IL_table_pointer ; 

  if (verbosing_flag)

    fprintf(stdout, "OK") ;

  // allocates the memory needed for the compressed lists 
  for(k=0 ; k<number_tokens; ++k) {

    if (Huffman_tree[k]->num_byte_list > 0) {

      Huffman_tree[k]->cmp_pos_list=(char *) malloc(Huffman_tree[k]->cmp_pos_list_byte_length*sizeof(char)) ; 

      Huffman_tree[k]->PosByteEscape=-1 ;

      Huffman_tree[k]->cmp_pos_list_pointer=0 ;
 
      Huffman_tree[k]->cmp_pos_list_last_value=0 ;
    }

    else Huffman_tree[k]->cmp_pos_list=NULL ;

    if ((Huffman_tree[k]->type == WORD) && (Huffman_tree[k]->num_rank_list > 0) && (proximity_flag)) {

      Huffman_tree[k]->cmp_rank_list=(char *) malloc(Huffman_tree[k]->cmp_rank_list_byte_length*sizeof(char)) ; 
      
      Huffman_tree[k]->cmp_rank_list_pointer=0 ;

      Huffman_tree[k]->cmp_rank_list_last_value=0 ; 
    }

    else Huffman_tree[k]->cmp_rank_list=NULL ; 
  }
  
  if (verbosing_flag)

    fprintf(stdout, "\nWriting IL file %s ... ", IL_Filename) ; 

  /*******************************************************************
                       Third pass of EXPAT: writes the IL
  *******************************************************************/

  xml_code=NULL ;

  xml_dimension=0 ;

  xml_position=0 ;

  proximity_counter=0 ;

  last_expanded_position=-1 ;

  expat_number_tag_opened=0 ;

  last_body_position_written=0 ;

  body_size_without_zlib=0 ;

  parser=XML_ParserCreate(NULL) ;

  if (!parser) {

    fprintf(stderr, "\n\tFatal Error: problems with EXPAT Parser\n\tError code 39\n") ;

    DELETE_FILES ;

    exit(-1) ;
  }

  XML_SetParamEntityParsing(parser,
			    XML_PARAM_ENTITY_PARSING_ALWAYS) ; 
  
  // handlers for DTD declaration start and end 
  XML_SetDoctypeDeclHandler(parser,
			    doctypestart,
			    doctypeend) ;

  // handlers for TAG open and close
  XML_SetElementHandler(parser, 
			startwrite, 
			endwrite) ;
  
  // handler for comments
  XML_SetCommentHandler(parser,
			commentwrite) ;
  
  // handler for the text
  XML_SetCharacterDataHandler(parser,
			      textwrite)  ;

  // handlers for CDATA sections start and end 
  XML_SetCdataSectionHandler(parser,
			     cdstartwrite,
			     cdendwrite) ;

  // handler for default elements
  XML_SetDefaultHandlerExpand(parser,
			      defawrite) ;
  
  // rewind the input file
  fseek(inputFile,
	0,
	SEEK_SET) ; 

  for ( ; ; ) {

    int done=0 ;

    int bytes_read=0 ;

    bytes_read=fread(expat_parsing_buffer,
		     1,
		     XCDE_EXPAT_PARSING_BUFFER_SIZE,
		     inputFile) ;
    
    if (ferror(inputFile)) {

      fprintf(stderr, "\n\tFatal Error: problems in reading source file\n\tError code 34\n") ; 
      
      DELETE_FILES ; 

      exit(-1) ; 
    }

    done = feof(inputFile) ;

    if (!XML_Parse(parser, 
		   expat_parsing_buffer,
		   bytes_read,
		   done)) {

      fprintf(stderr, "\n\tFatal Error: problems in parsing raw %d %s\n\tError code 35\n",

	      XML_GetCurrentLineNumber(parser),

	      XML_ErrorString(XML_GetErrorCode(parser))) ;

      DELETE_FILES ;

      exit(-1) ;
    }

    if (done) break ;
  }

  if (verbosing_flag)

    fprintf(stdout, "OK\n") ;

  XML_ParserFree(parser) ;

  // writes the Inverted Lists file
  output_IL(Huffman_tree,
	    IL_File,
	    number_tokens,
	    &IL_pointers_buff,
	    &IL_pointers_buff_length) ;

  // writes the IL table in the header file
  for(k=0; k<IL_pointers_buff_length; ++k)

    fputc(IL_pointers_buff[k],
	  headerFile) ;

  if (IL_pointers_buff != NULL)

    free(IL_pointers_buff) ;

  // creates the tagstruct file
  if (tagstruct_flag)

    Build_Context_File(tagstructFile,
   			Huffman_tree,
    		        number_tokens,
    		        number_tagstruct_intervals,
			depthstruct_flag) ;



  // creates the depthstruct file
  if (depthstruct_flag){


	 // compute the number of elements of the tag's dictionary
	for(k=0; k<max_length; ++k) {

		tag_elem += dict_elem[k][0];

	}


  	// create the depth data structure
	Build_Depth_File(depthstructFile,
			Huffman_tree,
			tag_elem,
		        number_tokens,
    		        number_tagstruct_intervals,
			inputFilename) ;
  }


  hdestroy(ht) ;

  hdestroy(expat_general_declared_entity) ;

  hdestroy(expat_separators_table) ;

  hdestroy(expat_attribute_separators_table) ;

  // frees allocated memory and closes files
  if (inputFile != NULL)

    fclose(inputFile) ;

  if (headerFile != NULL)

   fclose(headerFile) ;

  if (headerFilename != NULL)

    free(headerFilename) ;

  if (bodyFile != NULL)

    fclose(bodyFile) ;

  if (bodyFilename != NULL)

    free(bodyFilename) ;

  if (IL_File != NULL)

    fclose(IL_File) ;

  if (IL_Filename != NULL)

    free(IL_Filename) ;

  if (inputFilename != NULL)

    free(inputFilename) ;

  if (tagstruct_flag) {

    if (tagstructFile != NULL) 

      fclose(tagstructFile) ; 

    if (tagstructFilename != NULL)

      free(tagstructFilename) ; 
  }

  if (depthstruct_flag) {

	 if (depthstructFile != NULL)

		fclose(depthstructFile) ; 

	 if (depthstructFilename != NULL)

		free(depthstructFilename) ;
  }

  if (verbosing_flag) {

    if (debugPhysicView != NULL)  

      fclose(debugPhysicView) ; 

    if (debugPhysicFilename != NULL)

      free(debugPhysicFilename) ;

    if (debugStatsFile != NULL) 

      fclose(debugStatsFile) ;

    if (debugStatsFilename != NULL)

      free(debugStatsFilename) ;

    if (debugIndexFile != NULL)

      fclose(debugIndexFile) ;

    if (debugIndexFilename != NULL)

      free(debugIndexFilename) ;
  }

  return(1) ;
}

// function that writes the IL file and computes the IL index table
void output_IL(Hashptr *table,
	       FILE *invdex,
	       int dimension,
	       char **IL_Index_Buffer,
	       int *IL_Index_Byte_Length)
{ // variables used as counters
  register
    int k=0 ;

  register
    int i=0 ;

  // variable that contains the position of a list in the IL file
  int position=0 ;

  // variables used to compute the delta difference within positions
  int positions_diff=0 ;

  int previous_position=0 ;

  // variables used to manage the IL pointers buffer
  int IL_Index_Buffer_Dimension=5*K ;

  int IL_Index_Buffer_Pointer=0 ;

  // total length in bytes of an IL
  int total_list_length=0 ;

  // variable used to compress IL buffer pointers with continuation bit algorithm
  unsigned
    long length_continuation_bit=0 ;

  // variable used to read element from compressed lists
  int read_el=0 ;

  // variable used to decompress the compressed lists
  int elem=0 ;

  // variable used to decompress the compressed lists
  unsigned
    long len=0 ;

  // variable used to compress IL lengths with continuation bit algorithm
  unsigned
    long length_continuation_bit_IL_length=0 ;

  // buffer used to compress IL lengths with the continuation bit algorithm
  char continuation_bit_IL_length_buffer[XCDE_CONT_BIT_BUFFER_MAX_LENGTH] ;

  // variable used to compress position lists lengths with continuation bit algorithm
  unsigned
    long length_continuation_bit_byte_length=0 ;

  // buffer used to compress position lists lengths with continuation bit algorithm
  char continuation_bit_byte_length_buffer[XCDE_CONT_BIT_BUFFER_MAX_LENGTH] ;

  // flag used to check the escape condition in the lists
  unsigned
    char escape=0u ;

  // pointer to compressed lists
  int pointer=0 ;

  /*******************************************************************
	                 initializes variables
  *******************************************************************/

  (*IL_Index_Buffer) = (char *) malloc(IL_Index_Buffer_Dimension*sizeof(char)) ;

  IL_Index_Buffer_Pointer=0 ;

  position=0 ;

  previous_position=0 ;

  /*******************************************************************
	                 loops over all tokens
  *******************************************************************/

  for(k=0; k<dimension; ++k) { // sets the IL table index

    // puts the pointer in the IL pointers buffer
    if (((k % (XCDE_IL_POINTER_RATIO))==0) && (k>0)) {

      positions_diff=position-previous_position ;

      previous_position=position ;

      // checks if there is enough space, else allocates much memory
      if (IL_Index_Buffer_Pointer > (IL_Index_Buffer_Dimension-5)) {

	*IL_Index_Buffer = (char *) realloc(*IL_Index_Buffer,
					    IL_Index_Buffer_Dimension*2) ;

	IL_Index_Buffer_Dimension*=2 ;
      }

      // adds 1 to positions difference to avoid the zero problem
      continuation_bit_c((unsigned long) positions_diff+1,
			 (char *) (*IL_Index_Buffer)+IL_Index_Buffer_Pointer,
			 (unsigned int) 0,
			 (unsigned long *) &length_continuation_bit) ;

      IL_Index_Buffer_Pointer+=(length_continuation_bit/8) ;
    }

    if (verbosing_flag) {

      fprintf(debugIndexFile, "\n---------------------------------------------------------------------------") ;

      fprintf(debugIndexFile, "\nRank %d ", k) ;

      fprintf(debugIndexFile, "\nToken %s ", table[k]->str) ;

      fprintf(debugIndexFile, "\nString Length %d ", table[k]->len_str) ;

      fprintf(debugIndexFile, "\nCodeword Length %d ", table[k]->cw_len) ;

      if ((table[k]->len_str) == 1)

	fprintf(debugIndexFile, "\nAscii Code Character %d ", table[k]->str[0]) ;

      fprintf(debugIndexFile, "\nPhysical Occurrences %d\nType %d\nPositions list elements %d\n", table[k]->count_occ-1, table[k]->type, table[k]->num_byte_list) ;
    }

    // the number of elements in tag lists must be even
    if ((table[k]->type == TAG) &&
	((table[k]->num_byte_list % 2) != 0)) {

      fprintf(stderr, "\n\tFatal Error: unexpected value in TAG rank %d number of token %d %s\n\tError code 37\n",
	      k,
	      table[k]->num_byte_list,
	      table[k]->str) ;

      DELETE_FILES ;

      exit(-1) ;
    }

    // computes the number of intervals for tagstruct construction
    if (((table[k]->type == TAG) && (tagstruct_flag) && (!depthstruct_flag)) || ((table[k]->type == TAG) && (!tagstruct_flag) && (depthstruct_flag)) ||((table[k]->type == TAG) && (tagstruct_flag) && (depthstruct_flag)))

      number_tagstruct_intervals+=(table[k]->num_byte_list/2) ;

   
    // computes the number of lists
    if (table[k]->num_byte_list > 0) {

      total_list_length=0 ;

      continuation_bit_c((unsigned long) table[k]->num_byte_list,
			 (char *) continuation_bit_byte_length_buffer,
			 (unsigned int) 0,
			 (unsigned long *) &length_continuation_bit_byte_length) ;
      
      total_list_length+=(length_continuation_bit_byte_length/8) ; 

      total_list_length+=(table[k]->cmp_pos_list_byte_length) ; 

      if ((table[k]->type == WORD) && (table[k]->num_rank_list > 0) && (proximity_flag))

	total_list_length+=(table[k]->cmp_rank_list_byte_length) ; 

      // writes in the file the total list length 
      continuation_bit_c((unsigned long) total_list_length,
			 (char *) continuation_bit_IL_length_buffer,
			 (unsigned int) 0,
			 (unsigned long *) &length_continuation_bit_IL_length) ;
      
      for(i=0; i<(length_continuation_bit_IL_length/8); ++i)

	fputc(continuation_bit_IL_length_buffer[i],
	      invdex) ; 

      position+=(length_continuation_bit_IL_length/8) ; 

      for(i=0; i<(length_continuation_bit_byte_length/8); ++i)

	fputc(continuation_bit_byte_length_buffer[i], 
	      invdex) ;

      position+=(length_continuation_bit_byte_length/8) ;
      
      table[k]->cmp_pos_list_last_value=0 ;
      
      table[k]->PosByteEscape=-1 ;

      position+=table[k]->cmp_pos_list_byte_length ;
      
      fwrite(table[k]->cmp_pos_list,
	     table[k]->cmp_pos_list_byte_length,
	     1,
	     IL_File) ;

      if (verbosing_flag) { 
	
	pointer=0 ; 
	
	elem=0 ;
	
	escape=0u ;
	
	while(pointer < table[k]->cmp_pos_list_byte_length) {

	  continuation_bit_d((unsigned long *) &read_el, 
			     (char *) (table[k]->cmp_pos_list+pointer),
			     (unsigned int) 0,
			     (unsigned long *) &len) ;
	  
	  pointer+=(len/8) ;

	  if ((read_el == 1) && (!escape)) { 

	    escape=1 ;
	    
	    continuation_bit_d((unsigned long *) &read_el, 
			       (char *) (table[k]->cmp_pos_list+pointer), 
			       (unsigned int) 0,
			       (unsigned long *) &len) ;

	    pointer+=(len/8) ;

	    elem+=(read_el+2) ;
	    
	    fprintf(debugIndexFile, " ESCAPE_START %d ", elem) ; 
	  }

	  else

	    if ((read_el == 1) && (escape)) { 

	      escape=0 ;
	      
	      fprintf(debugIndexFile, " ESCAPE_END ") ; 
	    }

	  else

	    if (escape) 

	      fprintf(debugIndexFile, " %d ", (read_el-2)) ; 

	    else { 
	      
	      elem+=(read_el-2) ;
	      
	      fprintf(debugIndexFile, " %d ", elem) ;
	    }
	}
      }

      if (((!tagstruct_flag) && (!depthstruct_flag)) ||
	  ((tagstruct_flag) && (table[k]->type != TAG))||
	  ((depthstruct_flag) && (table[k]->type != TAG))) {

	if (table[k]->cmp_pos_list != NULL) 

	  free(table[k]->cmp_pos_list) ; 

	table[k]->cmp_pos_list = NULL ;

	table[k]->cmp_pos_list=0 ; 

	table[k]->cmp_pos_list_pointer=0 ; 
	
	table[k]->cmp_pos_list_last_value=0 ;
      }
      
      if ((table[k]->type == WORD) && (table[k]->num_rank_list > 0) && (proximity_flag)) {  
		
	fwrite(table[k]->cmp_rank_list, 
	       table[k]->cmp_rank_list_byte_length,
	       1,
	       IL_File) ; 

	position+=table[k]->cmp_rank_list_byte_length ; 
	 
	if (verbosing_flag) {

	  fprintf(debugIndexFile, "\n\n            PROXIMITIES ") ; 

	  pointer=0 ; 

	  elem=0 ;
	  
	  while(pointer < table[k]->cmp_rank_list_byte_length) {

	    continuation_bit_d((unsigned long *) &read_el, 
			       (char *) (table[k]->cmp_rank_list+pointer),
			       (unsigned int) 0,
			       (unsigned long *) &len) ;

	    pointer+=(len/8) ;

	    elem+=(read_el-1) ;

	    fprintf(debugIndexFile, " %d ", elem) ;
	  }
	}

	if (table[k]->cmp_rank_list != NULL) { 

	  free(table[k]->cmp_rank_list) ; 
	  
	  table[k]->cmp_rank_list=NULL ;

	  table[k]->cmp_rank_list_byte_length=0 ; 

	  table[k]->cmp_rank_list_pointer=0 ; 
	}
      }
    }
      
    else // the list is empty 
	
      { fputc(0, 
	      invdex) ;
      
      ++position ;
      }
  }
  
  *IL_Index_Byte_Length=IL_Index_Buffer_Pointer ;
  
  return ;
}




