 /********************************************************************
 		        xcde_build_utility.c
                        contains functions : 
	             phy_corr, get_next_token, 
	             read_conf_file, undercase 
		   Xml Compressed Documents Engine
		         Building Program
	      Andrea Mastroianni - Paolo Ferragina 
         Dipartimento di Informatica - Universita' di Pisa 
	               ferragin@di.unipi.it
                      mastroi@cli.di.unipi.it
 ********************************************************************/
#include"xcde_consts.h"
#include"xcde_build.h"

extern
FILE *inputFile ; 

extern
FILE *configFile ; 

extern
FILE *dictsFile[XCDE_NUMBER_DICT] ; 

extern
FILE *bodyFile ;

extern
FILE *headerFile ; 

extern
FILE *IL_File ; 

extern
FILE *tagstructFile ;

extern
FILE *depthstructFile ;

extern
FILE *debugPhysicView ;

extern
FILE *debugIndexFile ;

extern
FILE *debugStatsFile ;

extern
char *inputFilename ;

extern
char *dictionaryFilename[XCDE_NUMBER_DICT] ;

extern
char *headerFilename ;

extern
char *bodyFilename ;

extern
char *IL_Filename ;

extern
char *tagstructFilename ;

extern
char *depthstructFilename ;

extern
char *debugPhysicFilename ;

extern
char *debugStatsFilename ;

extern
char *debugIndexFilename ;

extern
char *configFilename ;

extern
unsigned
char expat_CDATA_section_flag ;

extern
unsigned
char expat_DTD_flag ;

extern
unsigned
char expat_CDATA_expansion_flag ;

extern
unsigned
char verbosing_flag ;

extern
int number_tokens ;

// hash table which contains the tokens found during the parsing
extern
HashTable ht ;

// hash table which contains general internal entities
extern
HashTable expat_general_declared_entity ;

// hash table which contains the separators
extern
HashTable expat_separators_table ;

// hash table which contains the separators for attribute
extern
HashTable expat_attribute_separators_table ;

extern
Hashptr *Huffman_tree ;

// data structure used by Expat
extern
XML_Parser parser ;

extern
char *expat_source_pointer ;

// variables used to traduce the physical text into the logical view
extern
char *xml_code ;

extern
long xml_dimension ;

extern
long xml_position ;

// last physical position where an entity has been expanded ;
extern
long last_expanded_position ;

// flag which indicates if the position parsed has changed
extern
unsigned
char changed_position ;

// flag which indicates if there is a corrispondence between
// the physical view and the logical one
extern
unsigned
char physical_corrispondence ;

extern
char separators[XCDE_MAX_NUMBER_SEPARATORS] ;

extern
char attribute_separators[XCDE_MAX_NUMBER_SEPARATORS] ;

extern
int number_tagstruct_intervals ;

extern
int inputFilesize ;

 /*******************************************************************
   gets a token starting at point text, until a separator is met
 *******************************************************************/

int get_next_token(char *text,
		   const char *separator,
		   char *sepa,
		   int maxlen)
{ register
    int j=0 ;

  int count=0 ;

  while(1==1) {

    for (j=0;j<strlen(separator);++j)
    
      if (text[count]==separator[j]) { 

	*sepa=*(separator+j) ; 

	return(count) ; 
      }
      
    ++count ; 
    
    if (count==maxlen) { 

      *sepa='\0' ; 
      
      return(maxlen) ; 
    }
  } 
  
  return(-1) ;
}  

 /*******************************************************************
     checks the correspondence between physic and logic levels
 *******************************************************************/

unsigned char phy_corr(char **entity, 
		       int *entity_len, 
		       unsigned char mem) 
{ int k=0 ;

  long bytecount ; 

  int token_len=0 ;

  char sep[1]="\0"; 

  static 
    char *tok=NULL ; 

  Hashptr search=NULL ; 

  bytecount = XML_GetCurrentByteIndex(parser);

  // checks if it's into a CDATA section 
  if (((expat_source_pointer[bytecount] == '&') && (!expat_CDATA_section_flag)) ||
     
      // checks if it's into a parametric entity expansion
      ((expat_source_pointer[bytecount] == '%') && (expat_source_pointer[bytecount+1] != 32) && (expat_DTD_flag)) || 
      
      // checks if it's into a CDATA section that has been expanded from 
      // a general internal entity 
      ((expat_source_pointer[bytecount] == '&') && (expat_CDATA_section_flag) && (expat_CDATA_expansion_flag)))
   
      // theresn't corrispondence beetween physic and logic level { 

    // checks if it's into the same expansion of the last entity checked 
    if (bytecount==last_expanded_position) { 

      *entity=(char *) malloc((strlen(tok)+1)*sizeof(char)) ; 
      
      *entity_len=strlen(tok) ; 
      
      strcpy(*entity, 
	     tok) ; 
      
      changed_position=0 ; 
      
      return(0) ; 
    }
  
    else
      
      // it's into a new expansion 
      { changed_position=1 ; 
      
      token_len=get_next_token(expat_source_pointer+bytecount,
			       ";",
			       sep,
			       inputFilesize-bytecount) ; 
      
      ++token_len ; 
      
      if (tok != NULL)
	
	free(tok) ; 
      
      last_expanded_position=bytecount ; 
      
      *entity=(char *) malloc(sizeof(char)*(token_len+1)) ; 
      
      tok=(char *) malloc(sizeof(char)*(token_len+1)) ; 
      
      for (k=0; k<token_len; ++k)
	
	tok[k]=(*entity)[k]=expat_source_pointer[bytecount+k] ; 
      
      tok[k]=(*entity)[k]='\0' ; 
      
      *entity_len=token_len ; 
      
      if (mem == 1) { 

	number_tokens+=hinsert((char *) *entity,
			       strlen(*entity),
			       WORD, 
			       (char *) NULL,
			       ht,
			       0,
			       0,
			       ((0x00) | _INCREASE_PHYSICAL_COUNT)) ; 
	
	if (verbosing_flag)
	  
	  fprintf(debugPhysicView, "%s", *entity) ; 
      }
      
      else 
	
	if (mem == 0) output_token(*entity, 
				   strlen(*entity),
				   WORD) ; 
      
	else
	  
	  if (mem == 2) 
		
	    nooutput_token(*entity, 
			   strlen(*entity),
			   WORD) ;
      
      // checks if it's an external general entity
      if (!(search=hsearch((char *) *entity,
			   strlen(*entity),
			   NOTHING, 
			   expat_general_declared_entity)))
	
	return(0) ; 
      
      else 
	
	// points to the entity expantion 
	{ xml_code = search->ass_el ; 
	
	xml_position = 0 ; 
	
	xml_dimension = strlen(search->ass_el) ;
	}
      
      return(0) ; 
      }

  else

    // points to the original file 
    { xml_code = expat_source_pointer+bytecount ; 
    
    xml_position = 0 ; 
    
    xml_dimension = inputFilesize-bytecount ; 
    
    return(1) ; 
    }
}

 /*******************************************************************
           functions that reads the configuration file contain       
  *******************************************************************/

void read_conf_file(FILE *configFile) 
{ register
    int k=0; 

 int number_element_separators_inserted=0 ; 
 
 int number_attribute_separators_inserted=0 ; 

 int ascii_code_separator=0 ; 
 
 int number_line=0 ; 

 char line[XCDE_MAX_CONF_FILE_LINE_LENGTH]="\0" ; 

 char entity[XCDE_MAX_CONF_FILE_LINE_LENGTH]="\0" ; 

 while(fgets(line, 
	     XCDE_MAX_CONF_FILE_LINE_LENGTH, 
	     configFile) != NULL) { 

   ++number_line ; 
   
   // the line contains an external general entity or and internal general entity
   // declared as a separator 
   if (((line[0] == ASCII_SPACE) || (line[0] == ASCII_E) || (line[0] == ASCII_A)) 
       && (line[1] == _StartConfEntDecl)) { 

     for(k=1; ( (k<strlen(line)) && (line[k] != _EndConfEntDecl)); ++k) 

       entity[k-1]=line[k] ; 
     
     if (k>strlen(line)) { 

       fprintf(stderr, "\n\tFatal Error: syntax error in configuration file al line %d : %s\n", number_line, line) ; 
       
       DELETE_FILES ; 
       
       exit(-1) ; 
     }
     
     else 
       
       { entity[k-1]=line[k] ; 

       entity[k]='\0' ; 
       }
	
     if ((line[0] == ASCII_SPACE) || (line[0] == ASCII_E)) { 

       // checks if the entity has been inserted into the separator table 
       if (!hsearch(line,
		    strlen(line),
		    NOTHING,
		    expat_separators_table)) 
	 
	 hinsert((char *) line,
		 strlen(line), 
		 NOTHING, 
		 (char *) NULL, 
		 expat_separators_table,
		 0,
		 0,
		 0x00) ; 
       
       else
	 
	 { fprintf(stderr, "\n\tFatal Error: duplicated declaration in configuration file at line %d\n", number_line) ; 
	 
	 DELETE_FILES ; 
	 
	 exit(-1) ; 
	 }
     }
     
     if ((line[0] == ASCII_SPACE) || (line[0] == ASCII_A))
       
       { // checks if the entity has been inserted into the separator table 
	 if (!hsearch(line,
		      strlen(line),
		      NOTHING,
		      expat_attribute_separators_table)) 
	   
	   hinsert((char *) line,
		   strlen(line), 
		   NOTHING, 
		   (char *) NULL, 
		   expat_attribute_separators_table, 
		   0,
		   0,
		   0x00) ; 
	 
	 else
	   
	   { fprintf(stderr, "\n\tFatal Error: duplicated declaration in configuration file at line %d\n", number_line) ; 
	   
	   DELETE_FILES ; 
	   
	   exit(-1) ; 
	   }
       }
   }     
   
   else

     if ((line[0] != _StartConfComment) && 
	 (line[1] > 47) && 
	 (line[1] < 58))
       
       { // checks if the separator has been inserted yet
	 if ((line[0] == ASCII_SPACE) || (line[0] == ASCII_E)) { 

	   line[0]=' ' ;
	   
	   sscanf(line, "%d", &ascii_code_separator) ;
	   
	   for(k=0; k<number_element_separators_inserted; ++k) 
	     
	     if (separators[k] == ascii_code_separator) { 

	       fprintf(stderr, "\n\tFatal Error: duplicated declaration in configuration file at line %d\n", number_line) ; 
	       
	       DELETE_FILES ; 
	       
	       exit(-1) ; 
	     }	      
	   
	   separators[number_element_separators_inserted] = ascii_code_separator ; 
	   
	   ++number_element_separators_inserted ; 
	 }

	 if ((line[0] == ASCII_SPACE) || (line[0] = ASCII_A)) {  

	   line[0]=' ' ;
	   
	   sscanf(line, "%d", &ascii_code_separator) ;
	   
	   for(k=0; k<number_attribute_separators_inserted; ++k)
	     
	     if (attribute_separators[k] == ascii_code_separator) { 

	       fprintf(stderr, "\n\tFatal Error: duplicated declaration in configuration file at line %d\n", number_line) ; 
	       
	       DELETE_FILES ; 
	       
	       exit(-1) ; 
	     }	      
	   
	   attribute_separators[number_attribute_separators_inserted] = ascii_code_separator ; 
	   
	   ++number_attribute_separators_inserted ; 
	 }
	 
	 if ((line[0] != ASCII_SPACE) && (line[0] != ASCII_A) && (line[0] != ASCII_E)) { 

	   fprintf(stderr, "\n\tFatal Error: syntax error in configuration file al line %d : %s\n", number_line, line) ; 
	   
	   DELETE_FILES ; 
	   
	   exit(-1) ; 
	 }
       }       
   
     else 
       
       if (line[0] != _StartConfComment) { 

	 fprintf(stderr, "\n\tFatal Error: syntax error in configuration file al line %d : %s\n", number_line, line) ; 
	 
	 DELETE_FILES ; 
	 
	 exit(-1) ; 
       }
 }
 
 separators[number_element_separators_inserted]='\0'; 
 
 attribute_separators[number_attribute_separators_inserted]='\0' ;  
 
 return ; 
}

/*******************************************************************
      function that trasform a string to the undercase form
*******************************************************************/

void undercase(char *string, 
	       int string_len) 
{ register
    int k ; 

 for(k=0 ; k<string_len; ++k) 

   if ((string[k] >= 65) && (string[k] <= 90))

     string[k]+=32 ; 

 return ;
}
