/********************************************************************
						xcde_depth.c
	          Xml Compressed Document Engine
	                 Building Program
			Paolo Ferragina - Rosanna Rossi
           Dipartimento di Informatica - Universita' di Pisa
	               ferragin@di.unipi.it
                       rossiro@di.unipi.it
********************************************************************/

// *****************************************************************
// Author: Rosanna Rossi
// Date: 2003
// *****************************************************************

#include "stdlib.h"
#include "stdio.h"
#include "math.h"
#include <unistd.h>
#include <sys/mman.h>
#include "xcde_consts.h"
#include "xcde_lib.h"

#define MACRO_ERROR { \
XCDE_Finalize_Console(&console) ; \
exit(-1) ; \
}
// ======================= GLOBALS TYPES ==========================================

XCDE_Console_Type XCDE_Console;

typedef struct {
  uint phis_pos;
  uint displ;
} pos;

typedef struct {
  pos position;
  uint tag_pos_dict;
} int_endpoint;

typedef struct {
 	 uint rank;
	 int depth_sum;
	 int *depth;
} int_depth;

int res = 0;

// ============ AUXILIARY FUNCTIONS  ====================================

// Comparison function for ranks sorting
int cmp_ranks(const void *a, const void *b) {

	int_depth *aa, *bb;
	uint a_val, b_val;
	int cmp;

	aa = (int_depth *) a;
	bb = (int_depth *) b;

	return( aa->rank - bb->rank);
}

// write in the file the integer in bytes
int fwrite_integer (FILE *ptrFile, int number) {

	int i;
	char temp_number = 0;

	for (i = sizeof(int) - 1; i >= 0 ; i--) {

		temp_number = number >> (8 * i);

		fwrite(&temp_number, sizeof(char), 1, ptrFile);

	}
}

// Comparison function for interval sorting
int compare_interv(const void *a, const void *b) {

	int_endpoint *aa, *bb;
	uint a_val, b_val;
	int cmp;

	aa = (int_endpoint *) a;
	bb = (int_endpoint *) b;
	cmp = (aa->position.phis_pos == bb->position.phis_pos);
	a_val = (cmp ? aa->position.displ : aa->position.phis_pos);
	b_val = (cmp ? bb->position.displ : bb->position.phis_pos);

	return ( a_val - b_val );
}

// Check if an endpoint is an opening (Pila on triples)
int tag_open(int_endpoint *Pila, int size, int_endpoint *A, int i) {

	if (size==0) return(1);

	if (Pila[size-1].tag_pos_dict != A[i].tag_pos_dict) return(1);

	return(0);

}


// ================== BUILDING FUNCTION  ====================================

// This procedure builds the Pila of tags from int_array, and foreach tag compute the Depth Table.
//
// A contains endpoints (phis_pos, displ, tag_id) of input
// Num_Tag_Dict is the number of tags in the tag dictionary

int Build_Tab_Depth (FILE *depthstructFile, int_endpoint *A, int n, int Num_Tag_Dict) {


      	int dimension_Pila = 5000 ;
	int_endpoint *Pila ;
	int Pila_size = 0;

	int dimension_Tab_depth;
	int_depth *Tab_depth;
	int Tab_size = 0;

	int depth = 0;

	int i = 0;
	int j,k;


	dimension_Tab_depth = Num_Tag_Dict;

	// allocates the memory needed by the Pila
	Pila = (int_endpoint *) malloc(dimension_Pila * sizeof(int_endpoint)) ;

	// check of the right allocation
	if (Pila == NULL) {

		fprintf(stderr, "\n\tFatal Error: problems during memory allocation for Pila in xcde_depth.c - Build_Tab_Depth()\n") ;

		fflush(stderr) ;

		exit(-1) ;
	}

	// allocates the memory needed by the Tab_depth
	Tab_depth = (int_depth *) malloc(dimension_Tab_depth * sizeof(int_depth));

	// check of the right allocation
	if (Tab_depth == NULL) {

		fprintf(stderr, "\n\tFatal Error: problems during memory allocation for Tab_depth in xcde_depth.c - Build_Tab_Depth()\n") ;

		fflush(stderr) ;

		exit(-1) ;
	}

	for(i=0; i<dimension_Tab_depth; i++) {

		Tab_depth[i].depth_sum = 0;
	}
	
	// We cycle over all the endpoints
	for(i=0; i<n; i++) {

		if ( tag_open(Pila, Pila_size, A, i) ) {

		// Endpoint denoting an open-tag

		// insert its position on the Pila
		// checks if there is enough space, else allocates much memory

			if (Pila_size >= dimension_Pila) {

				dimension_Pila *= 2;

				Pila = realloc(Pila,dimension_Pila);

				// check of the right allocation
				if (Pila == NULL) {

					fprintf(stderr, "\n\tFatal Error: problems during memory allocation for Pila in xcde_depth.c - Build_Tab_Depth()\n") ;

					fflush(stderr) ;

					exit(-1) ;
				}
			}

			Pila[Pila_size].position.phis_pos = A[i].position.phis_pos;
			Pila[Pila_size].position.displ = A[i].position.displ;
			Pila[Pila_size].tag_pos_dict = A[i].tag_pos_dict;

			// insert the position of tag with its depth
			res = Insert_Depth(Tab_depth, A[i].tag_pos_dict, depth, &Tab_size);


			if (res == -1) {

				fprintf(stderr, "\nProblems with the Insert_Depth function\n") ;

				exit(-1);
			}

			Pila_size++;

			depth++;

		} else	{

			// Case of a closing endpoint

			Pila_size--;

			depth--;
		}
	}
	
	
	//Sort the absolutes ranks
	qsort(Tab_depth, dimension_Tab_depth, sizeof(int_depth), cmp_ranks);

  	// Write the file for the depth data structure
  	Write_Depth_File(depthstructFile, Tab_depth, dimension_Tab_depth);

	free(Pila) ;

  	return(1);
}





// insert the depth of the tag in the Tab_Depth (rank_dict_tag, sum_depth, depth)

int Insert_Depth(int_depth *Tab_depth, uint tag_dict_gen, int depth, int *size_Tab_depth) {

	int i = 0;

	int dimension_depth_vector = 2000;
	int current_sum_depth = 0;
	int current_size;


	current_size = *size_Tab_depth;

	for (i=0; i<current_size; i++) {

		// the rank is in the Table, must only insert the new depth and increase the total number of depth
		if (Tab_depth[i].rank == tag_dict_gen) {

			current_sum_depth = Tab_depth[i].depth_sum;

			if (current_sum_depth >= dimension_depth_vector) {

				dimension_depth_vector *= 2;

				// checks if there is enough space, else allocates much memory
				Tab_depth[i].depth = realloc(Tab_depth[i].depth, dimension_depth_vector);

				// check of the right allocation
				if (Tab_depth[i].depth == NULL) {

					fprintf(stderr, "\n\tFatal Error: problems during memory allocation for Tab_depth[i].depth in xcde_depth.c - Insert_Depth()\n") ;

 					fflush(stderr) ;

 					exit(-1) ;

				}
			}

			(Tab_depth[i].depth)[current_sum_depth] = depth;

			current_sum_depth++;

			Tab_depth[i].depth_sum = current_sum_depth;

			return(1);

		}
	}

	// the rank is not in the Table, must insert the new rank with its depth and increase the total number of depth
	Tab_depth[current_size].rank = tag_dict_gen;

	Tab_depth[current_size].depth = (int *) malloc(dimension_depth_vector * sizeof(int));

	(Tab_depth[current_size].depth)[0] = depth;

	Tab_depth[current_size].depth_sum++;

	current_size++;

	(*size_Tab_depth) = current_size;

	return(1);
}


// ====================== I/Os FUNCTION  ====================================

// Header (its length is (dimension_Tab_depth / SKIP) * 4byte ) contains:
//
//	1. vector of pointers with length (dimension_Tab_depth / SKIP * 4byte)
//
// Body contains:
//
//	foreach tag in the Tab_depth
//		1. the length (bytes) of the depth list
//		2. the number of depth (compress)
//		3. the list of the depth (compress)

int Write_Depth_File (FILE *depthstructfile, int_depth *Tab_depth, int dimension_Tab_depth) {

	int *pointer_vector = NULL;
	int dimension_pointer = 0;		// the block size is 10

	int Depth_Buffer_Dimension = XCDE_CONT_BIT_BUFFER_MAX_LENGTH;
	char *continuation_bit_DEPTH_buffer = NULL;
	unsigned long cont_bit_len = 0ul;

	int pointer = 0;
	unsigned long counter_byte = 0ul;
	int i, j;
	int k = 0;

	dimension_pointer = ((dimension_Tab_depth -1) / SKIP) + 1;

	// allocates the memory needed by the pointer_vector
	pointer_vector = (int *) calloc (dimension_pointer, sizeof(int));

	// check of the right allocation
	if (pointer_vector == NULL) {

		fprintf(stderr, "\n\tFatal Error: problems during memory allocation for pointer_vector in xcde_depth.c - Write_Depth_File()\n") ;

 		fflush(stderr) ;

 		exit(-1) ;

	}

	if (fseek(depthstructfile, 0L, SEEK_SET)) {

		fprintf(stderr,"ftell error on file .depth\n");

		return(-1);
	}

	// Write the Header of the .depthstruct file
	// write the empty vector_pointer (one pointer foreach 10 tag)
	fwrite (pointer_vector, sizeof(int), dimension_pointer, depthstructfile);

	counter_byte = (sizeof(int) * dimension_pointer);

	// allocates the memory needed by the continuation_bit_DEPTH_buffer
	continuation_bit_DEPTH_buffer = (char *) malloc (XCDE_CONT_BIT_BUFFER_MAX_LENGTH * sizeof(int));

	// check of the right allocation
	if (continuation_bit_DEPTH_buffer == NULL) {

		fprintf(stderr, "\n\tFatal Error: problems during memory allocation for continuation_bit_DEPTH_buffer in xcde_depth.c - Write_Depth_File()\n") ;

 		fflush(stderr) ;

 		exit(-1) ;

	}

	for (i=0 ; i<dimension_Tab_depth ; i++) {

		// puts the pointer in the DEPTH pointers vector
		if ( (i % (SKIP)) == 0 ) {

			pointer_vector[k] = counter_byte;

			k++;
		}

		// checks if there is enough space, else allocates much memory
		if (pointer > (Depth_Buffer_Dimension - 5)) {

			Depth_Buffer_Dimension *= 2 ;

			continuation_bit_DEPTH_buffer = (char *) realloc(continuation_bit_DEPTH_buffer, Depth_Buffer_Dimension);

			// check of the right allocation
			if (continuation_bit_DEPTH_buffer == NULL) {

				fprintf(stderr, "\n\tFatal Error: problems during memory allocation for continuation_bit_DEPTH_buffer in xcde_depth.c - Write_Depth_File()\n") ;

				fflush(stderr) ;

				exit(-1) ;

			}
		}

		// adds 1 to depth to avoid the zero problem
		// the number of depth
		continuation_bit_c((unsigned long) Tab_depth[i].depth_sum + 1,
					(char *) continuation_bit_DEPTH_buffer + pointer,
					(unsigned int) 0,
					(unsigned long *) &cont_bit_len);

		pointer += cont_bit_len/8;

		for (j = 0 ; j<Tab_depth[i].depth_sum ; j++) {

			// checks if there is enough space, else allocates much memory
			if (pointer > (Depth_Buffer_Dimension - 5)) {

				Depth_Buffer_Dimension *= 2 ;

				continuation_bit_DEPTH_buffer = (char *) realloc(continuation_bit_DEPTH_buffer, Depth_Buffer_Dimension) ;

				// check of the right allocation
				if (continuation_bit_DEPTH_buffer == NULL) {

					fprintf(stderr, "\n\tFatal Error: problems during memory allocation for continuation_bit_DEPTH_buffer in xcde_depth.c - Write_Depth_File()\n") ;

					fflush(stderr) ;

					exit(-1) ;

				}
			}

			// adds 1 to depth to avoid the zero problem
			// the depth
			continuation_bit_c((unsigned long) (Tab_depth[i].depth[j]) + 1,
						(char *) continuation_bit_DEPTH_buffer + pointer,
						(unsigned int) 0,
						(unsigned long *) &cont_bit_len);

			pointer += cont_bit_len/8;

		}

		// write the length (bytes) of (sum_depth, depth,.....,depth)
		fwrite_integer (depthstructfile, pointer);

		// write the sum_depth and the depths
		fwrite(continuation_bit_DEPTH_buffer, sizeof(char), pointer, depthstructfile);

		counter_byte += pointer + sizeof(int);

		pointer = 0;
	}

	fseek(depthstructfile, 0L, SEEK_SET);

	// write the pointer vector
	for (i=0; i < dimension_pointer; i++) {

		// write the update pointer vector
		fwrite_integer(depthstructfile, pointer_vector[i]);

	}

	free(Tab_depth) ;

	return(1);
}


// ================== BUILDING PROCEDURES ===============================
void Build_Depth_File(FILE *depthstructFile,
		      Hashptr *Huffman_tree,
		      int num_tag,
		      int ntokens,
		      int num_intervals,
		      char *inputFilename){

	int_endpoint *int_array ;   	// Array of 2n endpoints

	int res=0;

	int i=0 ;

	uchar escape = 0;

	int pointer = 0;

	int elem = 0;

	int read_el = 0;

	unsigned long len = 0l;

	int int_array_position=0 ;

	// allocates the memory needed by the int_array
	int_array = (int_endpoint *) calloc((2*num_intervals), sizeof(int_endpoint));

	// check of the right allocation
	if (int_array == NULL) {

		fprintf(stderr, "\n\tFatal Error: problems during memory allocation for int_array in xcde_depth.c - Build_Depth_File()\n") ;

		fflush(stderr) ;

		exit(-1) ;
	}

	for(i=0 ; i<ntokens; ++i) {

		if (Huffman_tree[i]->type == TAG) {

			escape=0 ;

			pointer = 0 ;

			elem=0 ;


			while(pointer < Huffman_tree[i]->cmp_pos_list_byte_length) {

				continuation_bit_d((unsigned long *) &read_el,
						(char *) (Huffman_tree[i]->cmp_pos_list+pointer),
						(unsigned int) 0,
						(unsigned long *) &len) ;

				pointer+=(len/8) ;

				if ((read_el == 1) && (!escape)) {

					escape=1 ;

					continuation_bit_d((unsigned long *) &read_el,
							(char *) (Huffman_tree[i]->cmp_pos_list+pointer),
							(unsigned int) 0,
							(unsigned long *) &len) ;

					pointer+=(len/8) ;

				} else {

					if ((read_el == 1) && (escape)) {

						escape=0;

					} else {

						if ((read_el != 1) && (escape)) {

							int_array[int_array_position].position.phis_pos=elem ;

							int_array[int_array_position].position.displ=(read_el-2) ;

							int_array[int_array_position].tag_pos_dict=i ;

							++int_array_position ;

						} else {

							elem+=(read_el-2) ;

							int_array[int_array_position].position.phis_pos=elem ;

							int_array[int_array_position].position.displ=0 ;

							int_array[int_array_position].tag_pos_dict=i;

							++int_array_position ;
						}
					}
				}
			}

			if (Huffman_tree[i]->cmp_pos_list != NULL) {

				free(Huffman_tree[i]->cmp_pos_list) ;

				Huffman_tree[i]->cmp_pos_list=NULL ;

				Huffman_tree[i]->cmp_pos_list_pointer=0 ;

				Huffman_tree[i]->cmp_pos_list_last_value=0 ;
			}
		}
	}

	// Sort the endpoints
	qsort(int_array, 2 * num_intervals, sizeof(int_endpoint), compare_interv) ;

	//compute the depth data structure
	res = Build_Tab_Depth(depthstructFile, int_array, 2 * num_intervals, num_tag);

	if (res == -1) {

		 fprintf(stderr, "\nProblems with the Build_Tab_Depth function\n") ;

		 exit(-1);
 	}

	free(int_array) ;

}

