 /********************************************************************
			       xcde_search_new.c
                 header file needed by the XCDE 2.0 API library
	         Xml Compressed Document Engine Query Language
	             Alessandro Perrucci - Rosanna Rossi
             Dipartimento di Informatica - Universita' di Pisa
	                   alexperrucci@libero.it
	                  rosannarossi79@libero.it
 ********************************************************************/

#include "xcde_libsearch.h"

#define MACRO_FATAL_ERROR { \
	fflush(stderr) ; \
	exit(-1) ; \
}

int read_query_from_file(FILE *f, char **string)
{
#define INIT_QUERY_SIZE 100;

	int query_size;
	int carattere;
	int sting_len = 0;


	query_size = INIT_QUERY_SIZE;

	/* --- alloc memory for text --- */
	*string = (char *) malloc(query_size);

	if(*string == NULL) {

		fprintf(stderr, "\n\t<xml_error> problem during memory allocation for *string </xml_error>\n\n") ;

		fflush(stderr) ;

		return(-1);

	}

	// si posiziona il puntatore all'inizio del file
	rewind(f);

	while ( (carattere = fgetc(f)) != EOF) {

		(*string)[sting_len] = (char) carattere;
		sting_len++;

		if ( sting_len == (query_size) ) {

			query_size *= 2;

			*string = (char *) realloc((*string), query_size);

			if(*string == NULL) {

				fprintf(stderr, "\n\t<xml_error> problem during memory allocation for *string </xml_error>\n\n") ;

				fflush(stderr) ;

				return(-1);

			}
		}
	}

	(*string)[sting_len] = '\0';
	sting_len++;

	return 1;

}





/*******************************************************************
                          MAIN function
*******************************************************************/

int main(int argc, char *argv[])
{

#define INIT_SIZE_QUERY_STRING 100

	int getopt(int argc, char * const *argv, const char *options);
	int i;
	int res;

	// variabili utilizzate per memorizzare la query
	char *query_string = NULL;
	int dimension_query_string = 0;

	// variabili utilizzate in caso di lettura della query dal file
	char * nome_file_query = NULL;
	FILE *ptrQueryFile = NULL;

	// variabili utilizzate per la lettura delle opzioni dalla riga di comando
	extern char *optarg;
	extern int optind, opterr, optopt;
	char c;

	/* ------------- informazioni sulle opzioni inserite ----------- */
	unsigned char flag_position = 0;
	unsigned char flag_query_from_file = 0;


	/*******************************************************************
			analisi della riga di comando
	*******************************************************************/


	/* ------------- controllo sulla correttezza del numero di argomenti ----------- */
	if (argc < 2) {

		/*******************************************************************
			prints XCDE_SEARCH informations and instructions
		*******************************************************************/

		fprintf(stdout, "\n\n\t<!-----------------------------------------------------------------") ;

		fprintf(stdout, "\n\t       	    XCDE_SEARCH2 : Version 1.0Beta 17-10-2003		") ;

		fprintf(stdout, "\n\t       	       XCDE2: kernel XCDE + Query Language		") ;

		fprintf(stdout, "\n\t       		   Supervisor:  Paolo Ferragina 		") ;

		fprintf(stdout, "\n\t       		     Kernel: Andrea Mastroianni 		") ;

		fprintf(stdout, "\n\t       		  Developper: Alessandro Perrucci 		") ;

		fprintf(stdout, "\n\t       		     Developper: Rosanna Rossi			") ;

		fprintf(stdout, "\n\t	Dipartimento di Informatica - Universita' di Pisa 		") ;

		fprintf(stdout, "\n\t                         XCDE Search Engine			") ;

		fprintf(stdout, "\n\t            for copyright information see COPYRIGHT file   	") ;

		fprintf(stdout, "\n\t------------------------------------------------------------------>\n\n\n") ;

		fprintf(stderr, "\n\t FATAL ERROR: missing query string!") ;

		fprintf(stdout, "\n\n\t Usage: %s [-p] [-f nomefile] \"SELECT select_string FROM from_string RETURN return_string [RANGE number number]\"", argv[0]) ;

		fprintf(stdout, "\n\n\t\t -p : view positions instand of snippets") ;

		fprintf(stdout, "\n\n\t\t -f : take query from file 'nomefile'\n\n\n") ;

		MACRO_FATAL_ERROR

	}


	/* ------------- lettura delle opzioni dalla riga di comando ----------- */
	opterr = 0;

	while ((c=getopt(argc, argv, "f:p")) != -1) {

		switch (c) {
			case 'p':
				if (argc == 2) {

					/*******************************************************************
						prints XCDE_SEARCH informations and instructions
					*******************************************************************/

					fprintf(stdout, "\n\n\t<!-----------------------------------------------------------------") ;

					fprintf(stdout, "\n\t       	    XCDE_SEARCH2 : Version 1.0Beta 17-10-2003		") ;

					fprintf(stdout, "\n\t       	       XCDE2: kernel XCDE + Query Language		") ;

					fprintf(stdout, "\n\t       		   Supervisor:  Paolo Ferragina 		") ;

					fprintf(stdout, "\n\t       		     Kernel: Andrea Mastroianni 		") ;

					fprintf(stdout, "\n\t       		  Developper: Alessandro Perrucci 		") ;

					fprintf(stdout, "\n\t       		     Developper: Rosanna Rossi			") ;

					fprintf(stdout, "\n\t	Dipartimento di Informatica - Universita' di Pisa 		") ;

					fprintf(stdout, "\n\t                         XCDE Search Engine			") ;

					fprintf(stdout, "\n\t            for copyright information see COPYRIGHT file   	") ;

					fprintf(stdout, "\n\t------------------------------------------------------------------>\n\n\n") ;

					fprintf(stderr, "\n\t FATAL ERROR: missing query string!") ;

					fprintf(stdout, "\n\n\t Usage: %s [-p] [-f nomefile] \"SELECT select_string FROM from_string RETURN return_string [RANGE number number]\"", argv[0]) ;

					fprintf(stdout, "\n\n\t\t -p : view positions instand of snippets") ;

					fprintf(stdout, "\n\n\t\t -f : take query from file 'nomefile'\n\n\n") ;

					MACRO_FATAL_ERROR

				}

				flag_position = 1;
				break;

			case 'f':

				nome_file_query = optarg;
				flag_query_from_file = 1;
				break;

			case '?':

				/*******************************************************************
					prints XCDE_SEARCH informations and instructions
				*******************************************************************/

				fprintf(stdout, "\n\n\t<!-----------------------------------------------------------------") ;

				fprintf(stdout, "\n\t       	    XCDE_SEARCH2 : Version 1.0Beta 17-10-2003		") ;

				fprintf(stdout, "\n\t       	       XCDE2: kernel XCDE + Query Language		") ;

				fprintf(stdout, "\n\t       		   Supervisor:  Paolo Ferragina 		") ;

				fprintf(stdout, "\n\t       		     Kernel: Andrea Mastroianni 		") ;

				fprintf(stdout, "\n\t       		  Developper: Alessandro Perrucci 		") ;

				fprintf(stdout, "\n\t       		     Developper: Rosanna Rossi			") ;

				fprintf(stdout, "\n\t	Dipartimento di Informatica - Universita' di Pisa 		") ;

				fprintf(stdout, "\n\t                         XCDE Search Engine			") ;

				fprintf(stdout, "\n\t            for copyright information see COPYRIGHT file   	") ;

				fprintf(stdout, "\n\t------------------------------------------------------------------>\n\n\n") ;

				if (optopt == 'f')
					fprintf(stderr, "\n\t FATAL ERROR: missing filename");
				else
					fprintf(stderr, "\n\t FATAL ERROR: Unknown option: -%c -main-", optopt);

				fprintf(stdout, "\n\n\t Usage: %s [-p] [-f nomefile] \"SELECT select_string FROM from_string RETURN return_string [RANGE number number]\"", argv[0]) ;

				fprintf(stdout, "\n\n\t\t -p : view positions instand of snippets") ;

				fprintf(stdout, "\n\n\t\t -f : take query from file 'nomefile'\n\n\n") ;

				MACRO_FATAL_ERROR
		}
	}


	if (flag_query_from_file) {


		/* ------------- leggo la query da file ----------- */


		ptrQueryFile = fopen( nome_file_query, "r");

		if(ptrQueryFile == NULL) {

			fprintf(stderr,"\n\t<xml_error> Unable to open file %s </xml_error>\n\n", nome_file_query);

			MACRO_FATAL_ERROR

		}

		read_query_from_file(ptrQueryFile, &query_string);

	} else {

		/* ------------- leggo la query dalla riga di comando ----------- */

		query_string = (char *) malloc (INIT_SIZE_QUERY_STRING * sizeof(char));

		// controllo sulla corretta allocazione della memoria
		if (query_string == NULL) {

			fprintf(stderr, "\n\t<xml_error> problems during memory allocation for query_string </xml_error>\n\n") ;

			MACRO_FATAL_ERROR
		}

		dimension_query_string = INIT_SIZE_QUERY_STRING;

		strcpy(query_string, "");


		if(optind<argc){

			for (i=optind; i<argc; i++) {

				if ( (strlen(query_string) + strlen(argv[i]) + 2) > dimension_query_string) {

					dimension_query_string *= 2;

					query_string = (char *) realloc (query_string, dimension_query_string * sizeof(char));

					// controllo sulla corretta allocazione della memoria
					if (query_string == NULL) {

						fprintf(stderr, "\n\t<xml_error> problems during memory allocation for query_string </xml_error>\n\n") ;

						MACRO_FATAL_ERROR
					}

				}

				strcat(query_string, argv[i]);
				strcat(query_string, " ");

			}

		}

	}

	res = XCDE_Search(query_string, flag_position);

	if (res == -1) {

		exit(-1);
	}

	if (res == -2) {

		exit(-2);
	}

	
	return 1;

}

