/********************************************************************
		           xcde_view.c 
	         Xml Compressed Documents Engine
                  snippets visualization program
		 Paolo Ferragina - Andrea Mastroianni
         Dipartimento di Informatica - Universita' di Pisa 
	               ferragin@di.unipi.it
                      mastroi@cli.di.unipi.it
********************************************************************/
#include"xcde_lib.h"

// prints the header of a well-formed snippet
int print_header(const char *filename) ; 

// prints the foot of a well-formed snippet
int print_foot(void) ;

// prints the context at the beginning point of the textwindow extracted
int print_backward_context(int *backward_rank_context, 
			   int backward_rank_context_num,
			   int backward_position,
			   char *backward_window,
			   XCDE_Console_Type *XCDE_Console) ;

// prints the context at the ending point of the textwindow extracted 
int print_forward_context(int *forward_rank_context,
			  int forward_rank_context_num,
			  int forward_position,
			  char *forward_window,
			  XCDE_Console_Type *XCDE_Console) ;


// function used by XCDE_Get_Body_Textwindow to count words 
int Counter_function(int cw,
		     int cwlen,
		     XCDE_Console_Type *XCDE_Console) ; 

int main(int argc,
	 char **argv[]) 

{ register 
    int k=0 ;

 char *token=NULL ; 
 
 int res=0 ;
  
 int number_par=0 ;
 
 int *parameters=NULL ;
 
 
 // checks if there're parameters
 if (argc == 1) { // if enters this block there're no parameters. Prints 
                  // an error message and exits

   fprintf(stderr, "\nFatal error : there are no parameters\n") ;
   
   exit(-1) ;
 }
 
 // checks if the number of parameters is correct (it must be a multiple of three, excluded the
 // filename)
 if (((argc-2) > 0) && (((argc-2)%3) != 0)) { // if enters this block the number of parameters
                                              // isn't correct. It isn't a multiple of three. 
                                              // Prints an error message and exits
   
   fprintf(stderr, "\nFatal error : parameters aren't correct\n") ;
   
   fflush(stderr) ;
   
   exit(-1) ;
 }
  
 // gets the number of parameters  
 number_par=(argc-2) ;
  
 // allocates the memory needed to contain the parameters
 parameters = (int *) malloc(sizeof(int)*number_par) ;
  
 // copies the parameters in the memory allocated 
 for(k=0 ; k<number_par; ++k) { 

   // reads the number corresponding to a parameter 
   parameters[k]=atoi((char *) argv[k+2]) ;
   
   // checks if the parameters are correct
   if (parameters[k]<0) { // if enters this block a parameter isn't correct. Prints an error message and exits
     
     fprintf(stderr, "\nFatal error : parameters aren't correct\n") ;
     
     fflush(stderr) ;
     
     if (parameters != NULL)
       
       free(parameters) ;
     
     exit(-1) ;
   }
 }
 
 // calls the print_snippets function that prints well formed snippets from the list of triples 
 // passed as parameter. The list is in the format produced by the xcde_search command 
 // (start_win, end_win, num_word) 
 res=print_snippets(argv[1],
		    parameters,
		    number_par) ;
  
  if (res==-1) { 

    fprintf(stderr, "\nFatal error : problems with printing snippets function\n") ;
    
    fflush(stderr) ;
    
    if (parameters != NULL)
      
      free(parameters) ;
    
    exit(-1) ;
  }
 
  if (parameters != NULL)

    free(parameters) ;
 
  exit(0) ;
}

// prints on the standard output the well-formed snippets from the list of
// triples passed as a parameter
int print_snippets(const char *filename,
		   int *parms,
		   int num_par)
{ int res=0 ;

 int position=0 ;

 int cw=0 ;

 int cwlen=0 ;

 char *token=NULL ;

 char *backwardtextwindow=NULL ;

 char *forwardtextwindow=NULL ;

 int ncars_back=0 ;

 int ncars_forw=0 ;

 int backwardposition=0 ;

 int forwardposition=0 ;

 int *backward_rank_context=NULL ;

 int backward_rank_context_num=0 ;

 int *forward_rank_context=NULL ;
 
 int forward_rank_context_num=0 ;
 
 XCDE_IL_Couple_Type tagpos ;

 XCDE_Console_Type Console ;		   
 
 register 
   int k=0 ;
 
 // initializes the console 
 res=XCDE_Init_Console(filename,
		       &Console) ;
 
 if (res==-1) 

   return(-1) ;
 
 // loops over the list of triples 
 for(k=0 ; k<num_par; k+=3) { 

   // prints the header of the snippets
   res=print_header(filename) ; 

   if (res==-1) { 

     XCDE_Finalize_Console(&Console) ;
     
     return(-1) ;
   }
     
   // calls an XCDE API that extract from the body a window
   // going in backward direction and ending to extract at
   // the starting position of the window 
   res=XCDE_Get_Body_TextWindow(parms[k],
				&backwardtextwindow,
				parms[k+2],
				XCDE_BACKWARD,
				&backwardposition,
				&forwardposition,
				Counter_function, 
				&Console) ;
   
     if (res == -1) { 

       XCDE_Finalize_Console(&Console) ;

       if (backwardtextwindow != NULL)
	 
	 free(backwardtextwindow) ;
       
       return(-1) ;
     }

     // the variable backwardposition contain the position in which the previous
     // XCDE_Get_Body_TextWindow function has stopped the text window extraction 
     tagpos.XCDE_Position=backwardposition ;

     tagpos.XCDE_Offset=0 ;
     
     // gets the tagstruct computed at the point indicated by backwardposition 
     res=XCDE_Get_Tagstruct_Tags_Enclosing(&tagpos, 
					   &backward_rank_context,
					   &backward_rank_context_num, 
					   &Console) ;

     if (res == -1) { 

       XCDE_Finalize_Console(&Console) ;
       
       if (backwardtextwindow != NULL)
	 
	 free(backwardtextwindow) ;
       
       if (backward_rank_context != NULL)
	 
	 free(backward_rank_context) ;
       
       return(-1) ;
     }
     
     // prints the context at the beginning point of the textwindow extracted
     res=print_backward_context(backward_rank_context,
				backward_rank_context_num,
				backwardposition, 
				backwardtextwindow, 
				&Console) ;
     

     if (res == -1)
       
       { XCDE_Finalize_Console(&Console) ;
       
       if (backwardtextwindow != NULL)
	 
	 free(backwardtextwindow) ;     
       
       if (backward_rank_context != NULL)
	 
	 free(backward_rank_context) ;
       
       return(-1) ;
       }		

     // calls an XCDE API that extract from the body a window
     // going in forward direction and ending to extract at
     // the starting position of the window 
     res=XCDE_Get_Body_TextWindow(parms[k+1],
				  &forwardtextwindow,
				  parms[k+2],
				  XCDE_FORWARD,
				  &backwardposition,
				  &forwardposition,
				  Counter_function, 
				  &Console) ;
     
     if (res == -1) { 

       XCDE_Finalize_Console(&Console) ;
       
       if (backwardtextwindow != NULL)
	 
	 free(backwardtextwindow) ;
       
       if (backward_rank_context != NULL)
	 
	 free(backward_rank_context) ;

       if (forwardtextwindow != NULL)
	 
	 free(forwardtextwindow) ;
       
       return(-1) ;
     }		
     
     // copies the position it has ended to extract 
     tagpos.XCDE_Position=forwardposition ;
     
     tagpos.XCDE_Offset=0 ; 

     // gets the tagstruct at ending position 
     res=XCDE_Get_Tagstruct_Tags_Enclosing(&tagpos,
					   &forward_rank_context,
					   &forward_rank_context_num, 
					   &Console) ;   
     
     if (res == -1) { 

       XCDE_Finalize_Console(&Console) ;
       
       if (backwardtextwindow != NULL)
	 
	 free(backwardtextwindow) ;
       
       if (backward_rank_context != NULL)
	 
	 free(backward_rank_context) ;
       
       if (forwardtextwindow != NULL)
	 
	 free(forwardtextwindow) ;
       
       if (forward_rank_context != NULL)
	 
	 free(forwardtextwindow) ;
       
       return(-1) ;
     }		

     // copies the starting position of the window 
     position=parms[k] ;
     
     // extract the text window as received 
     while(position < parms[k+1]) { // extract the token of the textwindow until it reachs
                                    // the ending position of the window

       // extract the next codeword of the window in the body file 
       res=XCDE_Get_Body_Forward_Codeword(position,
					  &cw,
					  &cwlen,
					  &Console) ;
       
       if (res == -1) { 

	 XCDE_Finalize_Console(&Console) ;
	 
	 if (backwardtextwindow != NULL)
	   
	   free(backwardtextwindow) ;
	 
	 if (backward_rank_context != NULL)
	   
	   free(backward_rank_context) ;
	 
	 if (forwardtextwindow != NULL)
	   
	   free(forwardtextwindow) ;
	 
	 if (forward_rank_context != NULL)
	   
	   free(forwardtextwindow) ;
	 
	 return(-1) ;
       }	

       // traslates the codeword to the token 
       res=XCDE_Codeword2Token(cw,
			       cwlen,
			       &token,
			       &Console) ;
       
     
     if (res == -1) { 

       XCDE_Finalize_Console(&Console) ;
       
       if (backwardtextwindow != NULL)
	 
	 free(backwardtextwindow) ;
       
       if (backward_rank_context != NULL)
	 
	 free(backward_rank_context) ;
       
       if (forwardtextwindow != NULL)

	 free(forwardtextwindow) ;
       
       if (forward_rank_context != NULL)
	 
	 free(forwardtextwindow) ;
       
       return(-1) ;
       }	
     
     // increases the position in the body file 
     position+=cwlen ;
     
     // prints the token extracted to the standard output 
     fprintf(stdout, "%s", token) ;

     // frees the memory used to contain the token extracted from the body file 
     if (token != NULL) { 
       
       free(token) ;

       token=NULL ;
       }
     }

     // prints the context at the ending point of the textwindow extracted 
     res=print_forward_context(forward_rank_context,
			       forward_rank_context_num,
			       forwardposition, 
			       forwardtextwindow, 
			       &Console) ;
     
     if (res == -1) { 

       XCDE_Finalize_Console(&Console) ;
       
       if (backwardtextwindow != NULL)
	 
	 free(backwardtextwindow) ;
       
       if (backward_rank_context != NULL)
	 
	 free(backward_rank_context) ;
       
       if (forwardtextwindow != NULL)
	 
	 free(forwardtextwindow) ;
       
       if (forward_rank_context != NULL)
	 
	 free(forwardtextwindow) ;
       
       return(-1) ;
     }		
     
     // print the foot of the well-formed snippet 
     res=print_foot() ;

     if (res == -1) { 

       XCDE_Finalize_Console(&Console) ;
       
       if (backwardtextwindow != NULL)
	 
	 free(backwardtextwindow) ;
       
       if (backward_rank_context != NULL)
	 
	 free(backward_rank_context) ;
       
       if (forwardtextwindow != NULL)
	 
	 free(forwardtextwindow) ;
       
       if (forward_rank_context != NULL)
	 
	 free(forwardtextwindow) ;
       
       return(-1) ;
     }	
  
     // frees the memory allocated to print the well-formed snippet 
     if (backwardtextwindow != NULL) { 

       free(backwardtextwindow) ;
       
       backwardtextwindow=NULL ;
     }
     
     if (backward_rank_context != NULL) { 

       free(backward_rank_context) ;
       
       backward_rank_context=NULL ;
     }
     
   if (forwardtextwindow != NULL) { 

     free(forwardtextwindow) ;
     
     forwardtextwindow=NULL ;
   }
   
   if (forward_rank_context != NULL) { 

     free(forward_rank_context) ;
     
     forward_rank_context=NULL ;
   }
 }
 
 // finalizes the console 
 res=XCDE_Finalize_Console(&Console) ;

 if (res == -1)
   
   return(-1) ;
 
 return(1) ;
}
  
// prints the header of a well-formed snippet 
int print_header(const char *filename) 
{ 
  fprintf(stdout, "\n<xml_snippet>") ;

  fprintf(stdout, "\n<xml_filename>%s</xml_filename>", filename) ;

  fprintf(stdout, "\n<xml_text>\n") ;
  
  return(1) ;
}

// prints the foot of a well-formed snippet  
int print_foot(void) 
{ fprintf(stdout, "</xml_text>") ;
 
  fprintf(stdout, "\n</xml_snippet>\n") ;

 return(1) ; 
}

// prints the context at the beginning point of the textwindow extracted
int print_backward_context(int *backward_rank_context, 
			   int backward_rank_context_num,
			   int backward_position, 
			   char *backward_window, 
			   XCDE_Console_Type *XCDE_Console) 
{ unsigned
    char state=0 ;

  int res=0 ;

  register 
    int j=0 ;

  register
    int i=0 ;

  unsigned 
    char done=0 ;

  unsigned
    char found=0 ;

  int token_count=0 ; 

  int ncars=0 ; 

  int cw=0 ;

  int cwlen=0 ;

  char *string=NULL ; 

  int dimensionstring=100 ; 

  char *token=NULL ; 

  XCDE_IL_Couple_Type *IL=NULL ;

  int IL_num=0 ;

  int position=0 ;

  // checks if the tagstruct isn't empty
  if (backward_rank_context_num > 0) { // if enters this block the tagstruct isn't empty

    // gets the IL of the tag that is more grafted in the tagstruct 
    res=XCDE_Get_IL_ListPos(backward_rank_context[backward_rank_context_num-1],
			    &IL,
			    &IL_num,
			    XCDE_Console) ;
    
    if (res == -1) { 

      if (IL != NULL)
	
	free(IL) ;
      
      return(-1) ;
    }
    
    // finds the occurrence of the most grafted tag that contains the position in the variable backward_position 
    for (j=0 ; j<IL_num; j+=2) 
      
      if ((IL[j].XCDE_Position <= backward_position) && (IL[j+1].XCDE_Position >= backward_position))
	
	break ;
    
    // copie the starting position of the tag (that corresponds to the character '<')
    position=IL[j].XCDE_Position ; 

    done=0 ;
    
    ncars=0 ;
    
    // allocates the memory to contain the string 
    string=(char *) malloc(dimensionstring*sizeof(char)) ; 

    string[0]='\0' ; 

    while(!done) { // checks if the backward position is between the first '<' and the last '>' characters of 
                   // the tag opening 

      // in this case the position corresponds to the backward position. It's and exit condition 
      if (position==backward_position) { 

	// sets to on the found flag
	found=1 ;

	// deleted the most grafted element in the tagstruct because it has
	// alyed be printed (otherwise it should be printed two times)
	--backward_rank_context_num ; 
	
	break ;
      }

      // reads the previous codeword in the body file 
      res=XCDE_Get_Body_Forward_Codeword(position,
					 &cw,
					 &cwlen,
					 XCDE_Console) ;
      
      if (res == -1) { 

	if (IL != NULL)
	  
	  free(IL) ;
	
	return(-1) ;
      }

      // traslates the codeword to the token
      res=XCDE_Codeword2Token(cw,
			      cwlen,
			      &token, 
			      XCDE_Console) ;
      
      if (res == -1) { 

	if (IL != NULL)
	  
	  free(IL) ;
	
	if (token != NULL)
	  
	  free(token) ;
	
	return(-1) ;
      }
      
      // the array of character string is not great enough to contain 
      // the text read. Realloc it 
      if ((strlen(string)+strlen(token)) >= dimensionstring) { 

	string=realloc(string,
		       sizeof(char)*2*dimensionstring) ; 
	
	dimensionstring*=2 ;
      }

      // creates the string to be printed 
      strcat(string, 
	     token) ; 

      // analyzes the text extracted from the body file 
      for(i=0 ; (i<strlen(token) && (!done)); ++i) 
	
	// it's in an attribute definition 
	if ((token[i]=='\"') && (state==0))
	  
	  state=1 ;
      
	else
	  
	  // it's outside an attribute definition 
	  if ((token[i]=='\"') && (state==1))
	    
	    state=0 ;
      
	  else 
	    
	    // it's in an attribute definition 
	    if ((token[i]=='\'') && (state==0))
	      
	      state=2 ;
      
	    else
	      
	      // it's outside an attribute definition 
	      if ((token[i]=='\'') && (state==2))
		
		state=0 ;
      
	      else
		
		// it has reached a tag, so exit from the loop 
		if ((token[i]=='>') && (state==0))
		  
		  done=1 ;
      
      // increases the position in the body file 
      position+=cwlen ;

      // frees the memory used 
      if (token != NULL) { 

	free(token) ;
	
	token=NULL ;
      }
    }
    
    if (!found) { // the backward_position is not between the character '<' or '>' of the opening. Checks if it
                  // is between the same character of the closing position 

      done=0 ;
      
      ncars=0 ;

      string[0]='\0' ; 
      
      position=IL[j+1].XCDE_Position ;
      
      while(!done) { 

	// checks if the position has reached the backward_position 
	if (position == backward_position) { // in this case the position has reached the backward position
	                                     // it's the ending condition of the loop 

	  found=1 ;
	  
	  ncars=0 ; 
	}

	// reads the previous codeword in the body file 
	res=XCDE_Get_Body_Backward_Codeword(position,
					    &cw,
					    &cwlen,
					    XCDE_Console) ;
	
	
	if (res == -1) { 

	  if (IL != NULL)
	    
	    free(IL) ;
	  
	  return(-1) ;
	}  
	
	// traslates the codeword to the token 
	res=XCDE_Codeword2Token(cw,
				cwlen,
				&token, 
				XCDE_Console) ;
	
	if (res == -1) {
	  
	  if (IL != NULL)
	    
	    free(IL) ;
	  
	  if (token != NULL)
	    
	    free(token) ;
	  
	  return(-1) ;
	}

	// analyzes the token extracted from the body file 
	for(i=strlen(token)-1; ((i>=0) && (!done));--i, ++ncars) { 

	  if (token[i] == '<') // checks if it's the first character of the closing tag 
	    
	    done=1 ;
	}

	// frees the memory allocated to contain the tag 
	if (token != NULL) { 

	  free(token) ;
	  
	  token=NULL ;
	}
	
	// descreases the position 
	position-=cwlen ;
      }

      // goes forward to read the string between the backward position and the '>' character 
      while((ncars>0) && (position<=IL[j+1].XCDE_Position)) { 

	// reads the next codeword from the body file 
	res=XCDE_Get_Body_Forward_Codeword(position,
					   &cw,
					   &cwlen,
					   XCDE_Console) ;
	
	if (res == -1) {

	  if (IL != NULL)
	    
	    free(IL) ;
	  
	  return(-1) ;
	}
	
	// traslates the token to the codeword 
	res=XCDE_Codeword2Token(cw,
				cwlen,
				&token, 
				XCDE_Console) ;
	
	if (res == -1) { 

	  if (IL != NULL)
	    
	    free(IL) ;
	  
	  if (token != NULL)
	    
	    free(token) ;
	  
	  return(-1) ;
	}   
	
	// checks if the string length is enough to contain the token read
	if ((strlen(string)+strlen(token)) >= dimensionstring) { // if enter the string length isn't enough and 
	                                                         // it must realloc it 

	  string=realloc(string,
			 sizeof(char)*2*dimensionstring) ; 
	  
	  dimensionstring*=2 ;
	}
	
	// creates the string to be printed 
	strcat(string, 
	       token) ; 

	// increases the position
	position+=cwlen ; 
      }
    }
  }

  // loops over all the elements in the tagstruct
  for(j=0 ; j<backward_rank_context_num; ++j) { 

    // traslates the tag rank to its name 
    res=XCDE_Rank2Token(backward_rank_context[j],
			&token,
			XCDE_Console) ;
    
    if (res == -1)
      
      return(-1) ;
    
    token_count=0 ; 
  
    // delete the xml_ prefix from the logical recursive tag name 
    while ((strncmp(XCDE_RECURSIVE_TAG_PREFIX, token+token_count, strlen(XCDE_RECURSIVE_TAG_PREFIX) == 0) && (token_count < strlen(token))))
      
      token_count+=strlen(XCDE_RECURSIVE_TAG_PREFIX) ; 
    
    // prints the tagstruct
    for(i=0; i<j; ++i) 
      
      fprintf(stdout, " ") ; 
    
    if ((strcmp(token, "<![CDATA[") == 0) || (strcmp(token, "<!--") == 0))
      
      fprintf(stdout, "%s %s \n", token, XCDE_MORE_TEXT) ;
    
    else
      
      fprintf(stdout, "<%s> %s \n", token+token_count, XCDE_MORE_TEXT) ;

    fflush(stdout) ;

    // frees the memory allocated to contain the string 
    if (token != NULL) { 

      free(token) ;
      
      token=NULL ;
    }
  }
  
  // checks if the string has been created  
  if (found) 
    
    // the string has been created. Prints it 
    fprintf(stdout, "%s", string) ; 
  
  if (found)

    // the string has been created. Checks the number of character of the 
    // backward window that doesn't print
    fprintf(stdout, "%s", backward_window+ncars) ;
  
  else

    // the string hasn't been createt. Printed the backward_window as passed
    // to the function 
    fprintf(stdout, "%s", backward_window) ; 

  // frees the memory used to contain the IL 
  if (IL != NULL)
    
    free(IL) ;
  
  return(1) ;
}

// prints the context at the ending point of the textwindow extracted   
int print_forward_context(int *forward_rank_context,
			  int forward_rank_context_num,
			  int forward_position,
			  char *forward_window,
			  XCDE_Console_Type *XCDE_Console) 
{ unsigned
    char state=0 ;

  unsigned
    char found=0 ;

  unsigned

    char done=0 ;

  int res=0 ; 

  int ncars=0 ;

 register 
    int j=0 ;

  register
    int i=0 ;

  int cw=0 ;

  int cwlen=0 ;

  char *string=NULL ;

  int dimensionstring=100 ; 
  
  char *token=NULL ;

  XCDE_IL_Couple_Type *IL=NULL ;

  int IL_num=0 ;

  int position=0 ;

  int token_count=0 ; 
  
  // checks if the tagstruct at the ending point of the window isn't empty 
  if (forward_rank_context_num > 0) {
    
    // gets the IL of the more grafted tag in the tagstruct
    res=XCDE_Get_IL_ListPos(forward_rank_context[forward_rank_context_num-1], 
			    &IL, 
			    &IL_num, 
			    XCDE_Console) ;
      
    if (res == -1) { 

      if (IL != NULL) 
	
	free(IL) ; 
      
      return(-1) ; 
    } 
    
    // loops over the IL to find the tag occurrence that contain the forward position 
    for(j=0 ; j<IL_num; j+=2)
      
      if ((IL[j].XCDE_Position <= forward_position) && (IL[j+1].XCDE_Position >= forward_position))
	  
	break ;
    
    // copies the opening position of the tag that contains the forward position 
    position=IL[j].XCDE_Position ;
    
    done=0 ;

    // allocates the memory needed to contain the string extracted from the body 
    string=(char *) malloc(dimensionstring*sizeof(char)) ; 
    
    string[0]='\0' ; 
    
    // checks if it finds the position of the character of the initial tag '<'
    while(!done) { 
      
      if (position==forward_position) 
	
	found=1 ;
      
      // get the codeword from the file Body  
      res=XCDE_Get_Body_Forward_Codeword(position,
					 &cw,
					 &cwlen,
					 XCDE_Console) ;
      
      if (res == -1) { 

	if (IL != NULL)
	  
	  free(IL) ;
	
	return(-1) ;
      }
      
      // trasform the codeword read to the token 
      res=XCDE_Codeword2Token(cw,
			      cwlen,
			      &token, 
			      XCDE_Console) ;
      
      if (res == -1) { 
	
	if (IL != NULL)
	  
	  free(IL) ;
	
	if (token != NULL)
	  
	  free(token) ;
	
	return(-1) ;
      }
      
      // checks if has found the position of the initial '<' character
      if (found) { 

	if ((strlen(string)+strlen(token)) >= dimensionstring) { 
	  
	  string=realloc(string,
			 sizeof(char)*2*dimensionstring) ; 
	  
	  dimensionstring*=2 ;
	}
	
	strcat(string, token) ; 
      }
     
      // checks the state in which it's 
      for(i=0 ; (i<strlen(token) && (!done)); ++i) {
	
	if ((token[i]=='\"') && (state==0))
	  
	  state=1 ;
	
	else
	  
	  if ((token[i]=='\"') && (state==1))
	    
	    state=0 ;
	
	  else 
	    
	    if ((token[i]=='\'') && (state==0))
	      
	      state=2 ;
	
	    else
	      
	      if ((token[i]=='\'') && (state==2))
		
		state=0 ;
	
	      else
		
		if ((token[i]=='>') && (state==0))
		  
		  done=1 ;
      }
      
      // increases the position 
      position+=cwlen ;

      // frees the memory
      if (token != NULL) { 

	free(token) ;
	
	token=NULL ;
      }
    }

    // checks if has found the initial position   
    if (!found) { 

      done=0 ;

      ncars=0 ;
      
      position=IL[j+1].XCDE_Position ;
      
      while(!done) { 
	
	// checks if the position is equal to the initial one 
	if (position == forward_position) { // if enters this block the position is equal to the
	                                    // initial one, so it's found the position 

	  found=1 ;
	  
	  ncars=0 ;
	}
	
	// read the previous codeword from the body file 
	res=XCDE_Get_Body_Backward_Codeword(position,
					    &cw,
					    &cwlen,
					    XCDE_Console) ;
	
	if (res == -1) { 
	  
	  if (IL != NULL)
	    
	    free(IL) ;
	  
	  return(-1) ;
	}  
      
	// traslates the codeword to the token
	res=XCDE_Codeword2Token(cw,
				cwlen,
				&token, 
				XCDE_Console) ;
    
      if (res == -1) {

	if (IL != NULL)
	  
	  free(IL) ;
	
	if (token != NULL)
	  
	  free(token) ;
	
	return(-1) ;
      }
      
      // loops over all the characters of the token 
      for(i=0; (i<strlen(token) && (!done)); ++i, ++ncars) 
      
	if (token[i]=='<') 
	  
	  done=1 ;

      position-=cwlen ;

      if (token != NULL) { 

	free(token) ;
	
	token=NULL ;
      }
      }
    }
  }
  
  if (found)
    
    forward_window[strlen(forward_window)-ncars]='\0' ; 
  
  fprintf(stdout, "%s", forward_window) ;
  
  if (found)
    
    fprintf(stdout, "%s", string) ;
  
  fprintf(stdout, "\n") ;

  // decreases the number of element in the tagstruct
  // because the closing tag has alyet be printed 
  if ((found) && (ncars==0)) 

    --forward_rank_context_num ;  
  
  if (forward_rank_context_num > 0) 
    
    fprintf(stdout, "%s", XCDE_MORE_TEXT) ; 

  // prints the tagstruct 
  for(j=forward_rank_context_num-1; j>=0; --j) { 
    
    // traslates the rank to the token 
    res=XCDE_Rank2Token(forward_rank_context[j],
			&token,
			XCDE_Console) ;
    
    if (res == -1)
      
      return(-1) ;
    
    token_count=0 ; 
    
    // delete the xml_ prefix from the logical recursive tag name 
    while ((strncmp((XCDE_RECURSIVE_TAG_PREFIX), token+token_count, strlen(XCDE_RECURSIVE_TAG_PREFIX)) == 0) && (token_count < strlen(token)))
      
      token_count+=strlen((XCDE_RECURSIVE_TAG_PREFIX)) ; 
    
    for(i=0; i<j; ++i) 
      
      fprintf(stdout, " ") ;
    
    // eventually prints the CDATA section closing 
    if (strcmp(token, "<![CDATA[") == 0) 
      
      fprintf(stdout, "]]> %s \n", XCDE_MORE_TEXT) ; 
    
    else
      
      // eventually prints the comment opening
      if (strcmp(token, "<!--") == 0)
	
	fprintf(stdout, "--> %s \n", XCDE_MORE_TEXT) ; 
    
      else
      
	// prints the tagname without the characters that form the prefix
	// of the virtual tag
	fprintf(stdout, "</%s>", token+token_count) ;       
    
    fflush(stdout) ;    
    
    if (j>0) fprintf(stdout, "%s", XCDE_MORE_TEXT) ;
    
    fprintf(stdout, "\n") ;
    
    fflush(stdout) ;
    
    // frees the memory allocated 
    if (token != NULL) { 
      
      free(token) ;
      
      token=NULL ;
    }
  }
  
  // frees the memory allocated 
  if (IL != NULL)
    
    free(IL) ;
  
  return(1) ;
}

// function used by XCDE_Get_Body_Textwindow to count words 
int Counter_function(int cw,
		     int cwlen,
		     XCDE_Console_Type *XCDE_Console) 
{ char *token=NULL ; 
 
 int res=0 ;

 int rank=0 ; 
 
 unsigned 
   char dict ;
 
 int nelem=0 ;

 res=XCDE_Codeword2Rank(cw,
			cwlen,
			&rank, 
			XCDE_Console) ;

 if (res == -1)

   return(-1) ;

 res=XCDE_Codeword2Token(cw,
			 cwlen,
			 &token,
			 XCDE_Console) ;

 if (res == -1)

   return(-1) ;

 res=XCDE_Rank2Dict(rank,
		    &dict,
		    &nelem,
		    XCDE_Console) ;

 if (res == -1)

   return(-1) ;
 
 if ((dict == WORD) && (isalpha(token[0])))

   return(1) ;

 else 

   return(0) ;

}

