 /********************************************************************
			       xcde_libsearch.c
                 header file needed by the XCDE 2.0 API library
	         Xml Compressed Document Engine Query Language
	             Alessandro Perrucci - Rosanna Rossi
             Dipartimento di Informatica - Universita' di Pisa
	                   alexperrucci@libero.it
	                  rosannarossi79@libero.it
 ********************************************************************/

/********************************************************************
                            files header used
*********************************************************************/

#include "xcde_libsearch.h"
#include "xmlparse.h"


#define MACRO_FATAL_ERROR { \
	fflush(stderr) ; \
	exit(-1) ; \
}

/********************************************************************
                            constants used
 *********************************************************************/

#define _StartConfComment 35 // #
#define _StartConfEntDecl 38 // &
#define _EndConfEntDecl 59 // ;
#define _StartElement 60  // <
#define _SepTagAttr 38 // &
#define _SepValAttr 61 // =
#define _CloseTagElement 47 // '/'
#define _DoppiApici 34 //
#define _Apici 39 // '
#define _PuntoEVirgola 59 // ;

#define INIT_ENTITY_SEPARATORS_DIMENSION 20
#define XCDE_F_B_INITIAL_WINDOW_DIMENSION 1000
#define INIT_FORWARD_STRING_SIZE 100
#define XCDE_INITIAL_FROM_STRING_VECTOR_SIZE 3
#define XCDE_INITIAL_STACK_TAG_DIMENSION 10

/*******************************************************************
                   DECLARING GLOBAL VARIABLES
 *******************************************************************/

// variabile per la tabella hash delle variabili
HHash_table *global_hash_table = NULL;

// variabile usata per testare l'esito dei comandi
int res = 0 ;

query_tree_node * node_radice = NULL;

node_pila *pila_parsing = NULL;
unsigned int size_pila_parsing = 0;
unsigned int dimension_pila_parsing = INIT_SIZE_PILA_PARSING;

char *snippet_notag = NULL;
int dimension_snippet_notag = INIT_SIZE_SNIPPET_NOTAG;

char separators[XCDE_MAX_NUMBER_SEPARATORS]=" <>?!,.;:()[]{}\t\n\"" ;
char **entity_separators = NULL;
int num_entityseparators = 0;




/*******************************************************************
			       UTILITY functions
*******************************************************************/

int is_special_node(query_tree_node *node)
{

	if (strcmp(node->node_name, "xml_exact") == 0)
		return 1;

	if (strcmp(node->node_name, "xml_prefix") == 0)
		return 1;

	if (strcmp(node->node_name, "xml_suffix") == 0)
		return 1;

	if (strcmp(node->node_name, "xml_contained") == 0)
		return 1;

	if (strcmp(node->node_name, "xml_error") == 0)
		return 1;

	if (strcmp(node->node_name, "xml_regexp") == 0)
		return 1;

	return 0;

}



int is_special_attribute(char * attributo)
{

	if (strcmp(attributo, "xml_maxprox") == 0)
		return 1;

	if (strcmp(attributo, "xml_maxerr") == 0)
		return 1;

	if (strcmp(attributo, "xml_dist") == 0)
		return 1;

	if (strcmp(attributo, "xml_var") == 0)
		return 1;

	return 0;

}



int is_leaf(query_tree_node * node)
{

	return (node->num_childs == 0);

}


int is_entity(char *token) {

	int z;

	for (z=0; z<strlen(token); z++) {

		if (token[z] == '&') {

			return 1;
		}
	}

	return 0;
}



 /*******************************************************************
           functions that reads the configuration file contain
  *******************************************************************/

int read_configuration_file()
{

#define CLOSE_FILE   {  fclose(configFile) ;    \
						\
			configFile=NULL ;	\
						\
			free(configFilename) ;	\
						\
			configFilename=NULL ; }	\


	register int k=0;

	int number_element_separators_inserted=0 ;

	int dimension_entity_separators = INIT_ENTITY_SEPARATORS_DIMENSION;

	int ascii_code_separator=0 ;

	int number_line=0 ;

	char line[XCDE_MAX_CONF_FILE_LINE_LENGTH]="\0" ;

	char entity[XCDE_MAX_CONF_FILE_LINE_LENGTH]="\0" ;

	FILE *configFile;

	char *configFilename = NULL;



	configFilename=(char *) malloc(strlen(XCDE_CONFIGFILE_STANDARD_NAME)+1) ;

	if (configFilename == NULL) {

		fprintf(stderr,"\n\t<xcde:error> problems during memory allocation for configFilename</xcde:error>\n\n") ;

		CLOSE_FILE

		return(-1);
	}

	sprintf(configFilename, "%s", XCDE_CONFIGFILE_STANDARD_NAME) ;

	if ((configFile=fopen(configFilename, "r"))==NULL) {

		fprintf(stderr, "\n\t<xcde:error> error in open configuration file %s </xcde:error>\n\n", configFilename) ;

		free(configFilename) ;

		configFilename=NULL ;

		return(-1) ;
	}

	num_entityseparators = 0;

	entity_separators = (char **) malloc (dimension_entity_separators * sizeof(char *));

	if (entity_separators == NULL) {

		fprintf(stderr, "\n\t<xcde:error> problem during memory allocation for entity_separators </xcde:error>\n\n") ;

		CLOSE_FILE

		return(-1) ;

	}

	while(fgets(line, XCDE_MAX_CONF_FILE_LINE_LENGTH, configFile) != NULL) {

		++number_line ;

		// the line contains an external general entity or and internal general entity
		// declared as a separator
		if (((line[0] == ASCII_SPACE) || (line[0] == ASCII_E)) && (line[1] == _StartConfEntDecl)) {

			for(k=1; ( (k<strlen(line)) && (line[k] != _EndConfEntDecl)); ++k)
				entity[k-1]=line[k] ;

			if (k>strlen(line)) {

				fprintf(stderr, "\n\t<xcde:error> syntax error in configuration file al line %d : %s </xcde:error>\n\n", number_line, line) ;

				CLOSE_FILE

				return(-1) ;

			} else {

				entity[k-1]=line[k] ;

				entity[k]='\0' ;
			}

			// checks if the entity has been inserted into the separator table
			for (k=0; k<num_entityseparators; k++) {

				if (strcmp(entity_separators[k], entity) == 0) {

					fprintf(stderr, "\n\t<xcde:error> duplicated declaration in configuration file at line %d </xcde:error>\n\n", number_line) ;

					CLOSE_FILE

					return(-1) ;
				}
			}

			entity_separators[num_entityseparators] = (char *) malloc ((strlen(entity) + 1) * sizeof(char));

			if (entity_separators[num_entityseparators] == NULL) {

				fprintf(stderr, "\n\t<xcde:error> problem during memory allocation for entity_separators[num_entityseparators] </xcde:error>\n\n") ;

				CLOSE_FILE

				return(-1) ;

			}

			strcpy(entity_separators[num_entityseparators], entity);

			num_entityseparators++;

			// controllo sulla disponibilit di un elemento nella pila di parsing
			if (num_entityseparators >= dimension_entity_separators) {

				// aggiornamento della dimensione della pila di parsing
				dimension_entity_separators *= 2;

				entity_separators = (char **) realloc (entity_separators, dimension_entity_separators * sizeof(char *));

				if (entity_separators == NULL) {

					fprintf(stderr, "\n\t<xcde:error> problem during memory reallocation for entity_separators </xcde:error>\n\n") ;

					CLOSE_FILE

					return(-1) ;

				}
			}


		} else

			if ((line[0] != _StartConfComment) && (line[1] > 47) && (line[1] < 58)) {

				// checks if the separator has been inserted yet
				if ((line[0] == ASCII_SPACE) || (line[0] == ASCII_E)) {

					line[0]=' ' ;

					sscanf(line, "%d", &ascii_code_separator) ;

					for(k=0; k<number_element_separators_inserted; ++k)

						if (separators[k] == ascii_code_separator) {

							fprintf(stderr, "\n\t<xcde:error>  duplicated declaration in configuration file at line %d </xcde:error>\n\n", number_line) ;

							CLOSE_FILE

							return(-1) ;
						}

					separators[number_element_separators_inserted] = ascii_code_separator ;

					++number_element_separators_inserted ;
				}

				if ((line[0] != ASCII_SPACE) && (line[0] != ASCII_A) && (line[0] != ASCII_E)) {

					fprintf(stderr, "\n\t<xcde:error> syntax error in configuration file al line %d : %s </xcde:error>\n\n", number_line, line) ;

					CLOSE_FILE

					return(-1) ;
				}

			} else

				if ( (line[0] != _StartConfComment) && (line[0] != ASCII_A) ) {

					fprintf(stderr, "\n\t<xcde:error> syntax error in configuration file al line %d : %s </xcde:error>\n\n", number_line, line) ;

					CLOSE_FILE

					return(-1) ;
				}
	}

	separators[number_element_separators_inserted]='\0';

//	CLOSE_FILE

	return 1;
}



int is_separator(char *token) {

	unsigned char k;

	if (strlen(token) == 1) {

		// si controlla se  un carattere separatore
		for (k=0; k<strlen(separators); k++) {

			if (separators[k] == token[0]) {

				return(1) ;
			}
		}

	} else {

		if (strlen(token) == 2) {

			// si controlla se  un carattere separatore
			for (k=0; k<strlen(separators); k++) {

				if ( (separators[k] == token[0]) && (' ' == token[1]) ) {

					return(1) ;
				}
			}

		} else {

			// si controlla se  una entit separatrice
			for (k=0; k<num_entityseparators; k++) {

				if (strcmp(entity_separators[k], token) == 0) {

					return(1) ;
				}
			}
		}
	}

	return 0;
}




int reverse_stack(char **pila_tag, int size_pila_tag) {

	char *temp_string = NULL;

	int num_scambi = size_pila_tag / 2;
	
	int i;


	for(i=0; i < num_scambi; i++) {

		temp_string = pila_tag[i];

		pila_tag[i] = pila_tag[size_pila_tag - 1 - i];
		
		pila_tag[size_pila_tag - 1 - i] = temp_string;

	}
	
	return 1;


}





int free_node(query_tree_node * node)
{

	if ( (node->node_type == WORD) && (node->word_positions != NULL) && (strcmp(node->node_name, "xml_not") != 0) ) {

		free(node->word_positions);

	}

	if ( (node->node_type == TAG) && (node->tag_positions != NULL) && (strcmp(node->node_name, "xml_not") != 0) ) {

		free(node->tag_positions);

	}

	if (node->node_name)
		free(node->node_name);

	if (node->depth_vector)
		free(node->depth_vector);

	if (node->childs_vector)
		free(node->childs_vector);

	if (node->attributes_list)
		free(node->attributes_list);

	if (node)
		free(node);

	return 1;
}



char *get_attribute(query_tree_node *node, char *nome_attr)
{

	int i=0;

	char * valore_attributo = NULL;

	while (i < (node->num_attributes)) {

		if (strcmp(node->attributes_list[i], nome_attr) == 0) {

			// allocazione della memoria necessaria per la stringa che conterr il valore dell'attributo
			valore_attributo = (char *) malloc((strlen(node->attributes_list[i+1]) + 1) * sizeof(char));

			// controllo sulla corretta allocazione della memoria
			if (valore_attributo == NULL) {

				fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for valore_attributo </xcde:error>\n\n") ;

				fflush(stderr) ;

				return ((char * ) NULL);
			}

			strcpy(valore_attributo, node->attributes_list[i+1]);

			return(valore_attributo);
		}

		i += 2;
	}

	return ((char * ) NULL);

}



// questa funzione controlla che tutte le variabili inserite nella clausola return ci siano anche nella clausola select
//	ritorna 1 se le variabili inserite nella clausola return ci siano anche nella clausola select
//	ritorna 0 altrimenti

//	per ritornare 1 tutte le variabili inserite nella tabella hash devono avere associato un node diverso da NULL
int is_valid_return(HHash_table *ht)
{
	int i;
	Hash_node *p;

	for(i=0; i<ht->size;i++){
		for(p = ht->table[i]; p ; p = p->next){
			if (!p->node_name)
				return 0;
		}
	}

	return 1;
}


int remove_duplicated_spaces(char *string) {

	int i;
	int j = 1;

	if ( isspace(string[0]) )
		string[0] = ' ';

	for (i = 1; string[i] != '\0'; i++) {

		if ( !(isspace(string[i]) && (string[j-1] == ' ')) ) {

			if ( isspace(string[i]) && (string[j-1] != ' ') ){

				string[j] = ' ';

				j++;

			} else {

				string[j] = string[i];

				j++;
			}
		}
	}

	string[j] = '\0';



	return 1;
}



int replace_amp_occ(char **string) {

	int i;


	for (i=0; (*string)[i] != '\0'; i++) {

		if ((*string)[i] == '&') {

			(*string)[i] = XCDE_SPECIAL_ENTITY_CHARACTER;
		}
	}

	return 1;
}




int restore_amp_occ(char **string) {

	int i;


	for (i=0; (*string)[i] != '\0'; i++) {

		if ((*string)[i] == XCDE_SPECIAL_ENTITY_CHARACTER) {

			(*string)[i] = '&';
		}
	}

	return 1;
}




int read_int(int *destinazione, unsigned char *buffer) {

	int i;

	*destinazione = buffer[0];

	for (i=1; i < sizeof(int); i++) {

		*destinazione <<= 8;
		*destinazione += buffer[i];
	}


	return 1;

}




int fwrite_int (FILE *ptrFile, int number) {

	int i;
	char temp_number = 0;


	for (i = sizeof(int) - 1; i >= 0 ; i--) {

		temp_number = number >> (8 * i);

		// si scrive nel file il byte i-esimo dell'intero
		fwrite(&temp_number, sizeof(char), 1, ptrFile);

	}

	return 1;
}




/*******************************************************************
              funzione che visualizza il vettore risultato
*******************************************************************/

int visualizza_risultato(int numero_variabili, XCDE_IL_Couple_Type **risultato, int size_risultato) {

	int i,k;

	for (i = 0; i < size_risultato; i++) {

		for (k = 0; k < 2 * numero_variabili; k++) {

			if ((risultato[i][k]).XCDE_Position != 0) {

				printf("(%d, %d),", (risultato[i][k]).XCDE_Position, (risultato[i][k]).XCDE_Offset);

			} else {

				printf("(-, -),");
			}

		}

		printf("\n");
	}

}



/**********************************************************************************************
funzione che divide la stringa in input in due stringhe ottenute spezzando la stringa in input
	in corrispondenza della prima occorrenza del carattere separatore.
	la funzione ritorna:
		-1 in caso di errore;
		 0 nel caso in cui la seconda delle stringhe restituite  vuota
		 1 nel caso in vengono restituite due stringhe non vuote
**********************************************************************************************/

int my_split(const char * stringa_da_dividere, const char separatore, char **stringa_1, char **stringa_2)
{

	int j = 0;
	int i = 0;

	// allocazione della memoria necessaria per la stringa che conterr la prima parte della stringa in input
	*stringa_1 = (char *) malloc ((strlen(stringa_da_dividere) + 1) * sizeof(char));

	// controllo sulla corretta allocazione della memoria
	if (*stringa_1 == NULL) {

		fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for *stringa_1 in my_split() </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-1) ;
	}

	// si scorre la stringa in input fino a raggiungere il carattere speciale o la fine della stringa stessa
	while ( (stringa_da_dividere[i] != separatore) && (stringa_da_dividere[i] != '\0') ) {

		// viene copiato il carattere i-esimo nella stringa_1
		(*stringa_1)[i] = stringa_da_dividere[i];
		i++;
	}

	// si inserisce il carattere separatore nella prima stringa
	(*stringa_1)[i] = '\0';

	// allocazione della memoria necessaria per la stringa che conterr la seconda parte della stringa in input
	*stringa_2 = (char *) malloc ((strlen(stringa_da_dividere) + 1) * sizeof(char));

	// controllo sulla corretta allocazione della memoria
	if (*stringa_2 == NULL) {

		fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for *stringa_2 in my_split() </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-1) ;
	}

	// controllo sulla condizione di uscita del ciclo precedente
	if (stringa_da_dividere[i] == '\0') {

		// non si  incontrato il carattere speciale quindi la seconda delle due stringhe restituite deve essere vuota
		(*stringa_2)[j] = '\0';

		// si ritorna il valore 0 che indica che la seconda stringa  vuota
		return 0;
	}


	// se si  arrivati a questo punto l'i-esimo carattere della stringa in input contiene il
	//	carattere separatore quindi va saltato
	i++;

	// si continua a scorrere la stringa in input fino a raggiungerne la fine
	while (stringa_da_dividere[i] != '\0') {

		// si inserisce il carattere i-esimo della stringa in input nella seconda stringa
		(*stringa_2)[j] = stringa_da_dividere[i];
		i++;
		j++;
	}

	// si inserisce il carattere separatore nella seconda stringa
	(*stringa_2)[j] = '\0';

	// si controlla il contenuto della seconda stringa per decidere il valore da restituire
	if (strcmp(*stringa_2, "") == 0)
		return 0;
	else
		return 1;

}







int confronta_2_posizioni(XCDE_IL_Couple_Type posizione_1, XCDE_IL_Couple_Type posizione_2) {


	if (posizione_1.XCDE_Position < posizione_2.XCDE_Position) {

		return -1;

	} else {

		if (posizione_1.XCDE_Position == posizione_2.XCDE_Position) {

			if (posizione_1.XCDE_Offset < posizione_2.XCDE_Offset) {

				return -1;

			} else {

				if (posizione_1.XCDE_Offset == posizione_2.XCDE_Offset) {

					return 0;

				} else {

					return 1;

				}

			}

		} else {

			return 1;

		}

	}

	fprintf(stderr, "\n\t<xcde:error> inconsistent position </xcde:error>\n\n");

	fflush(stderr);

	return 2;

}




int confronta_posizioni(XCDE_IL_Couple_Type pos_apertura_padre, XCDE_IL_Couple_Type pos_chiusura_padre,
			XCDE_IL_Couple_Type *pos_apertura_figlio, XCDE_IL_Couple_Type *pos_chiusura_figlio,
			XCDE_IL_Triple_Type *pos_parola_figlio, unsigned int is_figlio_tag)
{

	XCDE_IL_Couple_Type temp_posizione_figlio;
	int res1 = 0;
	int res2 = 0;
	int res3 = 0;
	int res4 = 0;


	if (is_figlio_tag){

		if (
		(
		  ( (res1 = confronta_2_posizioni(*pos_apertura_figlio, pos_apertura_padre)) == 1) ||
		  ( (res2 = confronta_2_posizioni(*pos_apertura_figlio, pos_apertura_padre)) == 0)
		) && (
		  ( (res3 = confronta_2_posizioni(*pos_chiusura_figlio, pos_chiusura_padre)) == -1) ||
		  ( (res4 = confronta_2_posizioni(*pos_chiusura_figlio, pos_chiusura_padre)) == 0)
		)) {

			return 0;
		}
		
		if ( (res1 == 2) || (res2 == 2) || (res3 == 2) || (res4 == 2) )
			return 2;

		if ( (res1 = confronta_2_posizioni(*pos_chiusura_figlio, pos_apertura_padre)) == -1) {

			return -1;
		}

		if (res1 == 2)
			return 2;

		if ( (res1 = confronta_2_posizioni(*pos_apertura_figlio, pos_chiusura_padre)) == 1) {

			return 1;
		}

		if (res1 == 2)
			return 2;

	} else {

		temp_posizione_figlio.XCDE_Position = (*pos_parola_figlio).XCDE_Position;
		temp_posizione_figlio.XCDE_Offset = (*pos_parola_figlio).XCDE_Offset;


		if (
		(
		  ( (res1 = confronta_2_posizioni(temp_posizione_figlio, pos_apertura_padre)) == 1) ||
		  ( (res2 = confronta_2_posizioni(temp_posizione_figlio, pos_apertura_padre)) == 0)
		) && (
		  ( (res3 = confronta_2_posizioni(temp_posizione_figlio, pos_chiusura_padre)) == -1 ) ||
		  ( (res4 = confronta_2_posizioni(temp_posizione_figlio, pos_chiusura_padre)) == 0 )
		))  {

  			return 0;
		}

		if ( (res1 == 2) || (res2 == 2) || (res3 == 2) || (res4 == 2) )
			return 2;

		if ( (res1 = confronta_2_posizioni(temp_posizione_figlio, pos_apertura_padre)) == -1) {

			return -1;
		}

		if (res1 == 2)
			return 2;

		if ( (res1 = confronta_2_posizioni(temp_posizione_figlio, pos_chiusura_padre)) == 1) {

			return 1;
		}

		if (res1 == 2)
			return 2;

	}


	if (is_figlio_tag)

		fprintf(stderr, "\n\t<xcde:error> inconsistent position (%d, %d) (%d, %d) </xcde:error>\n\n", pos_apertura_padre.XCDE_Position, pos_chiusura_padre.XCDE_Position, (*pos_apertura_figlio).XCDE_Position, (*pos_chiusura_figlio).XCDE_Position);
	else
		fprintf(stderr, "\n\t<xcde:error> inconsistent position (%d, %d) (%d) </xcde:error>\n\n", pos_apertura_padre.XCDE_Position, pos_chiusura_padre.XCDE_Position, (*pos_parola_figlio).XCDE_Position) ;

	fflush(stderr);

	return 2;
}



int get_type_search_and_value(char * valore, unsigned char *tipo_ricerca, unsigned char *lettercase, int *num_err, char ** valore_da_ricercare)
{

	char * temp_seconda_parte = 0;

	char * temp_valore_da_ricercare = 0;

	char * temp_tipo_ricerca = 0;

	char * temp_numero_errori = 0;

	int i = 0;

	int j = 0;

	int res = 0;

	const char sep = ':';


	// la funzione my_split divide la stringa in input in due stringhe ottenute spezzando la stringa in input
	//	in corrispondenza della prima occorrenza del carattere separatore.
	//	la funzione ritorna:
	//		-1 in caso di errore;
	//		 0 nel caso in cui la seconda delle stringhe restituite  vuota
	//		 1 nel caso in vengono restituite due stringhe non vuote
	res = my_split(valore, sep, &temp_tipo_ricerca, valore_da_ricercare);

	// se c' stato un errore nella funzione my_split questo viene restituito al chiamante
	if (res == -1)
		return -1;

	// se la my_split ha restituito due stringhe non vuote bisogna analizzare la prima per sapere il tipo di ricerca da effettuare
	if (res == 1) {

		// il tipo di ricerca specificata  EXI (esatta e case insensitive)
		if ( strcmp(temp_tipo_ricerca, "EXI") == 0) {

			*tipo_ricerca = XCDE_EXACT;

			*lettercase = XCDE_NOCASE;

			*num_err = (int) NULL;

			return 1;
		}

		// il tipo di ricerca specificata  EXS (esatta e case sensitive)
		if ( strcmp(temp_tipo_ricerca, "EXS") == 0 ) {

			*tipo_ricerca = XCDE_EXACT;

			*lettercase = XCDE_CASE;

			*num_err = (int) NULL;

			return 1;
		}

		// il tipo di ricerca specificata  PREFI (per prefisso e case insensitive)
		if ( strcmp(temp_tipo_ricerca, "PREFI") == 0 ) {

			*tipo_ricerca = XCDE_PREFIX;

			*lettercase = XCDE_NOCASE;

			*num_err = (int) NULL;

			return 1;

		}

		// il tipo di ricerca specificata  PREFS (per prefisso e case sensitive)
		if ( strcmp(temp_tipo_ricerca, "PREFS") == 0 ) {

			*tipo_ricerca = XCDE_PREFIX;

			*lettercase = XCDE_CASE;

			*num_err = (int) NULL;

			return 1;

		}

		// il tipo di ricerca specificata  SUFI (per suffisso e case insensitive)
		if ( strcmp(temp_tipo_ricerca, "SUFI") == 0 ) {

			*tipo_ricerca = XCDE_SUFFIX;

			*lettercase = XCDE_NOCASE;

			*num_err = (int) NULL;

			return 1;

		}

		// il tipo di ricerca specificata  SUFS (per suffisso e case sensitive)
		if ( strcmp(temp_tipo_ricerca, "SUFS") == 0 ) {

			*tipo_ricerca = XCDE_SUFFIX;

			*lettercase = XCDE_CASE;

			*num_err = (int) NULL;

			return 1;

		}

		// il tipo di ricerca specificata  CONTI (per inclusione e case insensitive)
		if ( strcmp(temp_tipo_ricerca, "CONTI") == 0 ) {

			*tipo_ricerca = XCDE_CONTAINED;

			*lettercase = XCDE_NOCASE;

			*num_err = (int) NULL;

			return 1;

		}

		// il tipo di ricerca specificata  CONTS (per inclusione e case sensitive)
		if ( strcmp(temp_tipo_ricerca, "CONTS") == 0 ) {

			*tipo_ricerca = XCDE_CONTAINED;

			*lettercase = XCDE_CASE;

			*num_err = (int) NULL;

			return 1;

		}

		// il tipo di ricerca specificata  ERRI (con errori e case insensitive)
/*		if ( strcmp(temp_tipo_ricerca, "ERRI") == 0 ) {

			*tipo_ricerca = XCDE_ERROR;

			*lettercase = XCDE_NOCASE;

			// bisogna recuperare il numero di errori
			res = my_split(*valore_da_ricercare, sep, &temp_numero_errori, &temp_valore_da_ricercare);

			// se c' stato un errore nella funzione my_split questo viene restituito al chiamante
			if (res == -1)
				return -1;

			// se la seconda stringa restituita dalla my_split  vuota significa che non  stato specificato il valore da ricercare
			//	 e viene visualizzato un messaggio di errore e ritornato -1 al chiamante
			if (res == 0) {

				fprintf(stderr, "\n\t<xcde:error> missing value in '%s' </xcde:error>\n\n", valore) ;

				fflush(stderr) ;

				return -2;
			}

			// controllo che il numero di errori immesso sia valido.
			//	per essere valido l'importante  che cominci con una cifra
			//	in caso di numero della forma 2au il valore considerato  2
			if (!isdigit(temp_numero_errori[0])) {

				fprintf(stderr, "\n\t<xcde:error> invalid number of error (%s) in value '%s' </xcde:error>\n\n", temp_numero_errori, valore) ;

				fflush(stderr) ;

				return(-2) ;
			}

			// viene convertita la stringa contenente il numero di errori in un intero
			*num_err = atoi(temp_numero_errori);

			// si copia nella variabile *valore_da_ricercare la parte relativa alla stringa da ricercare
			//	che  la seconda restituita dalla my_split
			strcpy(*valore_da_ricercare, temp_valore_da_ricercare);

			return 1;

		}
*/
		// il tipo di ricerca specificata  ERR (con errori e case sensitive)
		if ( strcmp(temp_tipo_ricerca, "ERR") == 0 ) {

			*tipo_ricerca = XCDE_ERROR;

			*lettercase = XCDE_CASE;


			// bisogna recuperare il numero di errori
			res = my_split(*valore_da_ricercare, sep, &temp_numero_errori, &temp_valore_da_ricercare);

			// se c' stato un errore nella funzione my_split questo viene restituito al chiamante
			if (res == -1)
				return -1;

			// se la seconda stringa restituita dalla my_split  vuota significa che non  stato specificato il valore da ricercare
			//	 e viene visualizzato un messaggio di errore e ritornato -1 al chiamante
			if (res == 0) {

				fprintf(stderr, "\n\t<xcde:error> missing value in '%s' </xcde:error>\n\n", valore) ;

				fflush(stderr) ;

				return -2;
			}

			// controllo che il numero di errori immesso sia valido.
			//	per essere valido l'importante  che cominci con una cifra
			//	in caso di numero della forma 2au il valore considerato  2
			if (!isdigit(temp_numero_errori[0])) {

				fprintf(stderr, "\n\t<xcde:error> invalid number of error (%s) in value '%s' </xcde:error>\n\n", temp_numero_errori, valore) ;

				fflush(stderr) ;

				return(-2) ;
			}

			// viene convertita la stringa contenente il numero di errori in un intero
			*num_err = atoi(temp_numero_errori);

			// si copia nella variabile *valore_da_ricercare la parte relativa alla stringa da ricercare
			//	che  la seconda restituita dalla my_split
			strcpy(*valore_da_ricercare, temp_valore_da_ricercare);

			return 1;

		}

	}

	// se la seconda stringa restituita dalla funzione my_split  vuota significa che non  stato specificato
	//	nessun tipo di ricerca e quindi per default viene effettuata una ricerca di tipo EXI (esatta e case insensitive)
	if (res == 0) {

		*tipo_ricerca = XCDE_EXACT;

		*lettercase = XCDE_NOCASE;

		*num_err = (int) NULL;

		strcpy(*valore_da_ricercare, valore);

		return 1;
	}


	// se si  arrivati a questo punto nel valore dell'attributo  presente il carattere separatore
	//	ma la prima parte non corrisponde a nessun tipo di ricerca quindi si applica la ricerca
	//	di default (esatta e case insensitive) interpretando il carattere separatore come facente
	//	parte della stringa da ricercare.

	*tipo_ricerca = XCDE_EXACT;

	*lettercase = XCDE_NOCASE;

	*num_err = (int) NULL;

	strcpy(*valore_da_ricercare, valore);

	return 1;

}





/********************************************************************
			functions
*********************************************************************/


/* ---------------------------------------------------- */
/* -------------------  Hash Table    ----------------- */
/* ---------------------------------------------------- */



/** Initialize the hash table according to the number of estimated
 *   tokens; the load factor is set to 10 and the lists are managed
 *   via the MTF-rule.
 *
 * @param ht pointer to an (empty) hash table.
 * @param n estimated number of items to be inserted.
 */

int HHashtable_init(HHash_table **ht, int n)
{

	int i;

	*ht = (HHash_table *) malloc (sizeof(HHash_table));

	if ((*ht) == NULL) {

		fprintf(stderr,"\n\t<xcde:error> Hash table allocation </xcde:error>\n\n");

		fflush(stderr);

		return(-1);
	}

	(*ht)->size  = (int) 1.2 * n + 13;     // Load factor ~ 80%

	(*ht)->card = 0;

	(*ht)->table = (Hash_nodeptr_array) malloc((*ht)->size * sizeof(Hash_node *));

	if ((*ht)->table == NULL) {
	
		fprintf(stderr,"\n\t<xcde:error> Hash table allocation </xcde:error>\n\n");

		fflush(stderr);

		return(-1);
	}
	
	for (i=0; i < (*ht)->size; i++)
		(*ht)->table[i] = NULL;


	return 1;

}


/** Print the content of the hash table ht: token, length, count
 * occurrences, tagged cw and its length.
 *
 * @param ht pointer to the hash table to be printed out.
 */

void HHashtable_print(HHash_table *ht)
{
	int i;
	Hash_node *p;

	printf("\n\n================== Hash Table ====================\n");
	printf("Table size = %d, number of stored objects = %d\n\n",ht->size,ht->card);

	for(i=0; i<ht->size;i++){
		for(p = ht->table[i]; p ; p = p->next){
			printf("Nome variabile = '%s'\n", p->variable_name);
			printf("\tindex variabile = '%d'\n", p->index);
			printf("\tnome node variabile = '%s'\n", p->node_name);
		}
	}

	printf("\n==================================================\n\n");

}



/** Computes the hash value for the given string. The function was
    proposed by Ramakrishnais and Zobel in a paper appeared in:
    Int. Conf. on DB Systems for advanced applications, 1997.
 * @param s input string whose hash value has to be computed.
 * @param len length of s.
 * @param ht pointer to an hash table.
 */
int HHashtable_func(const char *s, int len, HHash_table *ht)
{
	register int hfn;
	int hfi;
	int table_size = ht->size;

	hfn = 11;

	for (hfi=0; hfi<len ; hfi++)
		hfn = hfn ^ ((hfn<<5) + (hfn>>2) + (unsigned char) *s++);

	hfn = abs(hfn % table_size);

	return(hfn);
}


/** Searches for the given string into the passed hash table (NULL if not).
 * @param s string to be searched.
 * @param slen length of s (to manage also the NULL char).
 * @param ht pointer to the hash table to be searched.
 */
Hash_node *HHashtable_search(const char *s, int slen, HHash_table *ht)
{
	Hash_node *hsp;

	for (hsp = ht->table[HHashtable_func(s,slen,ht)]; hsp; hsp = hsp->next)
		if (strcmp(s,hsp->variable_name) == 0)
			return(hsp);

	return((Hash_node *) NULL);
}



/** Inserts the token in the hash table and returns 1 if new, 0
 *  otherwise; it also updates the counter of occurrences for that
 *  token.
 *
 * @param s string to be inserted.
 * @param slen length of s (to manage also the NULL char).
 * @param ht pointer to an hash table, it will be updated.
 */
int HHashtable_insert(const char *s, int slen, int index_var, char *node_var, HHash_table *ht)
{
	int hiv;
	Hash_node *hip;
	Hash_node *table_head;
	Hash_node *hip_pred;
	char *token;

	hiv = HHashtable_func(s,slen,ht);  // compute the hash value
	hip = HHashtable_search(s,slen,ht);   // check string occurrence

	if (hip) {

		return(0);

	} else {    //------ The token is new -------

		hip = (Hash_node *) malloc(sizeof(Hash_node));

		if (hip == NULL){
			
			fprintf(stderr,"\n\t<xcde:error> Insert hash table </xcde:error>\n\n");

			fflush(stderr);

			return(-1);
		}

		hip->variable_name = (char *) malloc ((slen + 1) * sizeof(char));

		if (hip->variable_name == NULL){

			fprintf(stderr,"\n\t<xcde:error> Insert hash table </xcde:error>\n\n");

			fflush(stderr);

			return(-1);

		}

		strcpy(hip->variable_name,s);

		hip->index = index_var;

		if (node_var) {

			hip->node_name = (char *) malloc ((strlen(node_var) + 1) * sizeof(char));

			if (hip->node_name == NULL){

				fprintf(stderr,"\n\t<xcde:error> Insert hash table </xcde:error>\n\n");

				fflush(stderr);

				return(-1);
			}

			strcpy(hip->node_name, node_var);

		} else {

			hip->node_name = NULL;

		}


  		hip->next = ht->table[hiv];
		ht->table[hiv] = hip;
		ht->card += 1;
		return(1);
	}

	return 1;
}


/** Frees all elements of a hashtable. After this call, ht is an empty,
 * uninitialized HHash_table.
 *
 * @param ht pointer to a HHash_table
 */
int HHashtable_clear(HHash_table *ht)
{
	int i;

	for ( i = 0; i < ht->size; i++ ) {

		Hash_node *hn = ht->table[i];

		while ( hn ) {
			Hash_node *toFree = hn;
			hn = hn->next;
			free(toFree->variable_name);
			free(toFree);
		}
	}

	ht->card = 0;
	ht->size = 0;
	free(ht->table);
	free(ht);

	return 1;
}







/*******************************************************************
              HANDLERS FOR PARSING QUERY WITH EXPAT
 *******************************************************************/


/*******************************************************************
   		         tag start handlers
 *******************************************************************/

void
startTag(void *data,
      const char *el,
      const char **attr)
{

  	int i;
	int size_vett_attr = 0;

	Hash_node *node_hash;


	// puntatore al node che si sta costruendo per il Tag corrente
	query_tree_node *node;


	// allocazione della memoria necessaria per il node relativo al Tag corrente
	node = (query_tree_node *) malloc(sizeof(query_tree_node));

	if (node == NULL) {

		fprintf(stderr, "\n\tFatal Error: problems during memory allocation for query_tree_node in xcde_search.c - startTag()\n\n\n") ;

		MACRO_FATAL_ERROR
	}

	// riempimento di alcuni campi che possono essere gi riempiti, gli altri vengono messi a NULL
	node->node_name = (char *) malloc((strlen(el) + 1) * sizeof(char));

	if (node->node_name == NULL) {

		fprintf(stderr, "\n\tFatal Error: problems during memory allocation for node->node_name in xcde_search.c - startTag()\n\n\n") ;

		MACRO_FATAL_ERROR
	}

	strcpy(node->node_name, el);
	node->node_type = TAG;
	node->num_childs = 0;
	node->childs_vector = NULL;
	node->num_positions = 0;
	node->tag_positions = NULL;
	node->depth_vector = NULL;
	node->num_depth = 0;
	node->range_results = NULL;
	node->is_setted_number_result = 0;
	node->is_potable = 0;


	// conto il numero di elementi da allocare nel vettore degli attributi
	for (i=0; attr[i]; i++)
		size_vett_attr++;

	node->num_attributes = size_vett_attr;

	// allocazione della memoria necessaria per il vettore degli attributi
	node->attributes_list = (char **) malloc(size_vett_attr * sizeof(char *));

	if (node->attributes_list == NULL) {

		fprintf(stderr, "\n\tFatal Error: problems during memory allocation for node->attributes_list in xcde_search.c - startTag()\n\n\n") ;

		MACRO_FATAL_ERROR
	}


	// copia della lista degli attributi
	for (i=0; i < node->num_attributes; i++) {

		// se il nome dell'attributo  'xml_var' e il node non  speciale (se  speciale si inserisce nel node figlio) si inserisce la variabile relativa ed il puntatore al node nella tabella hash
		if( ((i % 2) == 0) && (strcmp(attr[i], "xml_var") == 0) && (!is_special_node(node)) ) {


			node_hash = HHashtable_search(attr[i+1], strlen(attr[i+1]), global_hash_table);

			if (node_hash) {

				if (node_hash->node_name != NULL) {

					fprintf(stderr, "\n\tFatal Error: multiple association of the same variable (%s = '%s') - in xcde_search.c - startTag()\n\n\n", attr[i], attr[i + 1]) ;

					MACRO_FATAL_ERROR

				} else {

					node_hash->node_name = (char *) malloc ((strlen(node->node_name) + 1) * sizeof(char));

					if (node_hash->node_name == NULL){
						fprintf(stderr,"Fatal Error: Insert hash table\n\n\n");
						exit(-1);
					}

					strcpy(node_hash->node_name, node->node_name);
				}
			}
		}

		node->attributes_list[i] = (char *) malloc((strlen(attr[i]) + 1) * sizeof(char));

		if (node->attributes_list[i] == NULL) {

			fprintf(stderr, "\n\tFatal Error: problems during memory allocation for attribute in xcde_search.c - startTag()\n\n\n") ;

			MACRO_FATAL_ERROR
		}

		strcpy(node->attributes_list[i],attr[i]);

	}


	// riempimento dell'elemento della pila di parsing relativo al tag corrente
	pila_parsing[size_pila_parsing].node = node;
	pila_parsing[size_pila_parsing].childs_vector_dimension = 0;

	// se non  il primo Tag nella pila devo aggiornare le informazioni riguardanti la lista dei
	// figli del Tag sottostante nella pila
	if (size_pila_parsing != 0) {

		// controllo sulla disponibilit di un elemento nel vettore figli del Tag sottostante nella pila
		if ( (pila_parsing[size_pila_parsing - 1].childs_vector_dimension) <= (pila_parsing[size_pila_parsing - 1].node)->num_childs ) {

			pila_parsing[size_pila_parsing - 1].childs_vector_dimension += INIT_SIZE_VETTORE_FIGLI;

			// non ho pi posto nel vettore dei figli del Tag sottostante nella pila e devo
			// riallocare uno spazio maggiore (dimensione vecchia + INIT_SIZE_VETTORE_FIGLI)
			(pila_parsing[size_pila_parsing - 1].node)->childs_vector = (query_tree_node **) realloc( (pila_parsing[size_pila_parsing - 1].node)->childs_vector, pila_parsing[size_pila_parsing - 1].childs_vector_dimension * sizeof(query_tree_node *));

			if ( (pila_parsing[size_pila_parsing - 1].node)->childs_vector == NULL ) {

				fprintf(stderr, "\n\tFatal Error: problems during memory allocation for childs_vector in xcde_search.c - startTag()\n\n\n") ;

				MACRO_FATAL_ERROR
			}
		}

		// inserisco il puntatore del node relativo al tag corrente nella lista dei figli del tag sottostante nella pila di parsing
		(pila_parsing[size_pila_parsing - 1].node)->childs_vector[(pila_parsing[size_pila_parsing - 1].node)->num_childs] = node;

		// aumento il numero di figli associati al tag sottostante nella pila di parsing
		(pila_parsing[size_pila_parsing - 1].node)->num_childs++;

	} else {

		// se si sta costruendo il primo node, questo si fa puntare anche da node_radice
		node_radice = pila_parsing[0].node;
	}

	// incremento delle dimensioni della pila
	size_pila_parsing++;

	// controllo sulla disponibilit di un elemento nella pila di parsing
	if (size_pila_parsing >= dimension_pila_parsing) {


		// aggiornamento della dimensione della pila di parsing
		dimension_pila_parsing += INIT_SIZE_PILA_PARSING;

		// non ho pi posto nella pila necessaria durante la fase di costruzione dell'albero della
		// query e devo riallocare uno spazio maggiore (dimensione vecchia + INIT_SIZE_PILA_PARSING)
		pila_parsing = (node_pila *) realloc(pila_parsing, dimension_pila_parsing * sizeof(node_pila));

		if (pila_parsing == NULL) {

			fprintf(stderr, "\n\tFatal Error: problems during memory allocation for pila_parsing in xcde_search.c - startTag()\n\n\n") ;

			MACRO_FATAL_ERROR
		}


	}


	return ;
}




/*******************************************************************
		          tag end handlers
*******************************************************************/

void endTag(void *data,
	    const char *el)
{

	// se si sta chiudendo un tag speciale si controlla che abbia esattamente un figlio
	if ( 	(strcmp(el, "xml_exact") == 0) ||
		(strcmp(el, "xml_prefix") == 0) ||
		(strcmp(el, "xml_suffix") == 0) ||
		(strcmp(el, "xml_contained") == 0) ||
		(strcmp(el, "xml_error") == 0) ||
		(strcmp(el, "xml_regexp") == 0) ||
		(strcmp(el, "xml_not") == 0) ) {

		if ( ((pila_parsing[size_pila_parsing - 1].node)->num_childs) != 1) {

			fprintf(stderr, "\nFatal Error: special tag '%s' must have one child\n\n\n", el) ;

			MACRO_FATAL_ERROR

		}
	}

	// se si sta chiudendo il tag speciale 'xml_proximity' si controlla che abbia almeno due figli
	if (strcmp(el, "xml_proximity") == 0) {

		if ( ((pila_parsing[size_pila_parsing - 1].node)->num_childs) < 2) {

			fprintf(stderr, "\nFatal Error: special tag '%s' must have at least two childs\n\n\n", el) ;

			MACRO_FATAL_ERROR

		}
	}


  	// eliminazione del puntatore al node del tag corrente
	pila_parsing[size_pila_parsing - 1].node = NULL;
	pila_parsing[size_pila_parsing - 1].childs_vector_dimension = 0;

	// decremento delle dimensioni della pila
	size_pila_parsing--;

	return;
}

/********************************************************************
		text handlers for select_string
*********************************************************************/

void
textElementSelect(void *data,
     const char *t,
     int len)
{
	int i;

	int numero_parola = 1;

	unsigned char is_padre_proximity = 0;

	char seps[] = " ";

	char *token;

	char *text, *temp_text;

	// puntatore al node che si sta costruendo per la Parola corrente
	query_tree_node *node;

	// puntatore al node speciale che bisogna inserire per default se la parola corrente non  racchiusa in un tag speciale
	query_tree_node *node_speciale;

	char *variable_name;

	Hash_node *node_hash;




	// allocazione memoria per la parola eventualmente ripulita dagli spazi
	text = (char *) malloc((len + 1) * sizeof(char));

	// allocazione memoria per la parola eventualmente ripulita dagli spazi
	temp_text = (char *) malloc((len + 1) * sizeof(char));

	// inserisco nella variabile text il testo
	for (i=0; i < len; i++) {
		
		if (t[i] == XCDE_SPECIAL_ENTITY_CHARACTER) {

			text[i] = '&';

		} else {

			text[i] = t[i];
		}
	}

	text[i] = '\0';

	strcpy(temp_text, text);

	// dal testo prendo la prima parola
	token = strtok( text, seps );

	while (token != NULL) {

/*		if ( (numero_parola != 1) && (strcmp((pila_parsing[size_pila_parsing - 1].node)->node_name, "xml_proximity") != 0) ) {

			fprintf(stderr, "\n\tFatal Error: presenza di piu' parole (%s) all'interno di uno stesso tag diverso da 'xml_proximity', in xcde_search.c - textElement()\n\n\n", temp_text) ;

			MACRO_FATAL_ERROR
		}
*/
		// allocazione della memoria necessaria per il node relativo alla Parola corrente
		node = (query_tree_node *) malloc(sizeof(query_tree_node));

		if (node == NULL) {

			fprintf(stderr, "\n\tFatal Error: problems during memory allocation for query_tree_node in xcde_search.c - textElement()\n\n\n") ;

			MACRO_FATAL_ERROR
		}


		// riempimento di alcuni campi che possono essere gi riempiti, gli altri vengono messi a NULL

		// allocazione della stringa che conterr la Parola
		node->node_name = (char *) malloc((strlen(token) + 1) * sizeof(char));

		// controllo sulla corretta allocazione della memoria
		if (node->node_name == NULL) {

			fprintf(stderr, "\n\tFatal Error: problems during memory allocation for query_tree_node in xcde_search.c - textElement()\n\n\n") ;

			MACRO_FATAL_ERROR
		}

		strcpy(node->node_name, token);
		node->node_type = WORD;
		node->num_childs = 0;
		node->childs_vector = NULL;
		node->num_positions = 0;
		node->word_positions = NULL;
		node->num_attributes = 0;
		node->attributes_list = NULL;
		node->range_results = NULL;
		node->depth_vector = NULL;
		node->num_depth = 0;
		node->is_setted_number_result = 0;
		node->is_potable = 0;


		// se ci sono elementi nella pila devo aggiornare le informazioni riguardanti la lista dei
		// figli dell'ultimo tag
		if (size_pila_parsing != 0) {


			// controllo sulla disponibilit di un elemento nel vettore figli del Tag sottostante nella pila
			if ( (pila_parsing[size_pila_parsing - 1].childs_vector_dimension) <= (pila_parsing[size_pila_parsing - 1].node)->num_childs ) {

				pila_parsing[size_pila_parsing - 1].childs_vector_dimension += INIT_SIZE_VETTORE_FIGLI;

				// non ho pi posto nel vettore dei figli del Tag sottostante nella pila e devo
				// riallocare uno spazio maggiore (dimensione vecchia + INIT_SIZE_VETTORE_FIGLI)
				(pila_parsing[size_pila_parsing - 1].node)->childs_vector = (query_tree_node **) realloc( (pila_parsing[size_pila_parsing - 1].node)->childs_vector, pila_parsing[size_pila_parsing - 1].childs_vector_dimension * sizeof(query_tree_node *));

				if ( (pila_parsing[size_pila_parsing - 1].node)->childs_vector == NULL ) {

					fprintf(stderr, "\n\tFatal Error: problems during memory allocation for childs_vector in xcde_search.c - startTag()\n\n\n") ;

					MACRO_FATAL_ERROR
				}
			}

			// se il padre del node di testo corrente non  un tag speciale bisogna inserire un tag speciale di tipo xml_exact
			//	come padre del node di testo corrente.
			if (!is_special_node(pila_parsing[size_pila_parsing - 1].node)) {

				// allocazione della memoria necessaria per il node speciale
				node_speciale = (query_tree_node *) malloc(sizeof(query_tree_node));

				if (node_speciale == NULL) {

					fprintf(stderr, "\n\tFatal Error: problems during memory allocation for node_speciale in xcde_search.c - textElement()\n\n\n") ;

					MACRO_FATAL_ERROR
				}

				// riempimento di alcuni campi che possono essere gi riempiti, gli altri vengono messi a NULL
				node_speciale->node_name = "xml_exact";
				node_speciale->node_type = TAG;
				node_speciale->num_childs = 1;

				node_speciale->childs_vector = (query_tree_node **) malloc(sizeof(query_tree_node *));

				if (node_speciale->childs_vector == NULL) {

					fprintf(stderr, "\n\tFatal Error: problems during memory allocation for childs_vector in xcde_search.c - textElement()\n\n\n") ;

					MACRO_FATAL_ERROR
				}



				node_speciale->childs_vector[0] = node;
				node_speciale->num_positions = 0;
				node_speciale->tag_positions = NULL;
				node_speciale->num_attributes = 0;
				node_speciale->depth_vector = NULL;
				node_speciale->num_depth = 0;
				node_speciale->attributes_list = NULL;
				node_speciale->range_results = NULL;
				node_speciale->is_setted_number_result = 0;
				node_speciale->is_potable = 0;


				// inserisco il puntatore del node relativo al tag speciale nella lista dei figli del tag sottostante nella pila di parsing
				(pila_parsing[size_pila_parsing - 1].node)->childs_vector[(pila_parsing[size_pila_parsing - 1].node)->num_childs] = node_speciale;

				// aumento il numero di figli associati al tag sottostante nella pila di parsing
				(pila_parsing[size_pila_parsing - 1].node)->num_childs++;

			} else {

				// il tag padre della parola corrente  un tag speciale e si controlla se a tale node
				//  associata una variabile questa va associata al node figlio
				variable_name = get_attribute(pila_parsing[size_pila_parsing - 1].node,"xml_var");

				if (variable_name != NULL) {

					node_hash = HHashtable_search(variable_name, strlen(variable_name), global_hash_table);

					// se la variabile associata al node speciale padre  presente nella tabella hash vuol dire che viene usata nella clausola
					//	RETURN e quindi alla variabile va associato il node corrente
					if (node_hash) {

						if (node_hash->node_name != NULL) {

							fprintf(stderr, "\n\tFatal Error: multiple association of the same variable (xml_var = '%s') - in xcde_search.c - startTag()\n\n\n", variable_name) ;

							MACRO_FATAL_ERROR

						} else {

							node_hash->node_name = (char *) malloc ((strlen(node->node_name) + 1) * sizeof(char));

							if (node_hash->node_name == NULL){
								fprintf(stderr,"Fatal Error: Insert hash table\n");
								exit(-1);
							}

							strcpy(node_hash->node_name, node->node_name);
						}
					}

					// allocazione della memoria necessaria per il vettore degli attributi
					node->attributes_list = (char **) malloc(2 * sizeof(char *));

					if (node->attributes_list == NULL) {

						fprintf(stderr, "\n\tFatal Error: problems during memory allocation for node->attributes_list \n in xcde_search.c - textElement()\n\n\n") ;

						MACRO_FATAL_ERROR
					}


					node->attributes_list[0] = "xml_var";
					node->attributes_list[1] = (char *) malloc((strlen(variable_name) + 1) * sizeof(char));

					if (node->attributes_list[1] == NULL) {

						fprintf(stderr, "\n\t problems during memory allocation for attribute xml_var \n in xcde_search.c - textElement()\n\n\n") ;

						MACRO_FATAL_ERROR
					}

					strcpy(node->attributes_list[1],variable_name);

					node->num_attributes = 2;

				}

				// inserisco il puntatore del node relativo alla parola corrente nella lista dei figli del tag sottostante nella pila di parsing
				(pila_parsing[size_pila_parsing - 1].node)->childs_vector[(pila_parsing[size_pila_parsing - 1].node)->num_childs] = node;

				// aumento il numero di figli associati al tag sottostante nella pila di parsing
				(pila_parsing[size_pila_parsing - 1].node)->num_childs++;

			}
		}

		numero_parola++;
		token = strtok( NULL, seps );

	}

	free(text);
	free(temp_text);
	return;

}



/********************************************************************
		text handlers for return_string
*********************************************************************/

void
textElementReturn(void *data,
     const char *t,
     int len)
{
	int i;

	char seps[] = " ";

	char *token;

	char *text;

	// puntatore al node che si sta costruendo per la Parola corrente
	query_tree_node *node;

	// puntatore al node speciale che bisogna inserire per default se la parola corrente non  racchiusa in un tag speciale
	query_tree_node *node_speciale;

	Hash_node *node_hash;


	// allocazione memoria per la parola eventualmente ripulita dagli spazi
	text = (char *) malloc((len + 1) * sizeof(char));

	// inserisco nella variabile text il testo
	for (i=0; i < len; i++) {
		text[i] = t[i];
	}

	text[i] = '\0';

	// dal testo prendo la prima parola
	token = strtok( text, seps );

	while (token != NULL) {

		// allocazione della memoria necessaria per il node relativo alla Parola corrente
		node = (query_tree_node *) malloc(sizeof(query_tree_node));

		if (node == NULL) {

			fprintf(stderr, "\n\tFatal Error: problems during memory allocation for query_tree_node in xcde_search.c - textElement()\n\n\n") ;

			MACRO_FATAL_ERROR
		}


		// riempimento di alcuni campi che possono essere gi riempiti, gli altri vengono messi a NULL

		// allocazione della stringa che conterr la Parola
		node->node_name = (char *) malloc((strlen(token) + 1) * sizeof(char));

		// controllo sulla corretta allocazione della memoria
		if (node->node_name == NULL) {

			fprintf(stderr, "\n\tFatal Error: problems during memory allocation for query_tree_node in xcde_search.c - textElement()\n\n\n") ;

			MACRO_FATAL_ERROR
		}

		strcpy(node->node_name, token);
		node->node_type = WORD;
		node->num_childs = 0;
		node->childs_vector = NULL;
		node->num_positions = 0;
		node->word_positions = NULL;
		node->depth_vector = NULL;
		node->num_depth = 0;
		node->num_attributes = 0;
		node->attributes_list = NULL;
		node->range_results = NULL;
		node->is_setted_number_result = 0;
		node->is_potable = 0;


		// se ci sono elementi nella pila devo aggiornare le informazioni riguardanti la lista dei
		// 	figli dell'ultimo tag
		if (size_pila_parsing != 0) {


			// controllo sulla disponibilit di un elemento nel vettore figli del Tag sottostante nella pila
			if ( (pila_parsing[size_pila_parsing - 1].childs_vector_dimension) <= (pila_parsing[size_pila_parsing - 1].node)->num_childs ) {

				pila_parsing[size_pila_parsing - 1].childs_vector_dimension += INIT_SIZE_VETTORE_FIGLI;

				// non ho pi posto nel vettore dei figli del Tag sottostante nella pila e devo
				// riallocare uno spazio maggiore (dimensione vecchia + INIT_SIZE_VETTORE_FIGLI)
				(pila_parsing[size_pila_parsing - 1].node)->childs_vector = (query_tree_node **) realloc( (pila_parsing[size_pila_parsing - 1].node)->childs_vector, pila_parsing[size_pila_parsing - 1].childs_vector_dimension * sizeof(query_tree_node *));


				if ( (pila_parsing[size_pila_parsing - 1].node)->childs_vector == NULL ) {

					fprintf(stderr, "\n\tFatal Error: problems during memory allocation for childs_vector in xcde_search.c - startTag()\n\n\n") ;

					MACRO_FATAL_ERROR
				}
			}

		}

		// inserisco il puntatore del node relativo alla parola corrente nella lista dei figli del tag sottostante nella pila di parsing
		(pila_parsing[size_pila_parsing - 1].node)->childs_vector[(pila_parsing[size_pila_parsing - 1].node)->num_childs] = node;

		// aumento il numero di figli associati al tag sottostante nella pila di parsing
		(pila_parsing[size_pila_parsing - 1].node)->num_childs++;

		// se l'elemento di testo analizzato  una variabile si aggiunge alla tabella hash. Se la variabile  gia presente
		// 	la HHashtable_insert non fa nulla.
		if (token[0] == '$') {

			res = HHashtable_insert(token, strlen(token), global_hash_table->card, (char *) NULL, global_hash_table);

/*			if (res == -1)
				return -1;

*/
		}

		token = strtok( NULL, seps );

	}

	free(text);
	return;

}



/*******************************************************************
			       default handler
*******************************************************************/

void defaultHandler(void *data,
	  const XML_Char *s,
	  int len)
{

	return ;

}


/*******************************************************************
     funzione che dato un node elimina tutti i nodi speciali
*******************************************************************/

int elimina_figli_speciali(query_tree_node * node)
{

	int i;
	query_tree_node * node_speciale;

	// se non  un node foglia si scorrono i figli per vedere quali possono essere potati
	if (!is_leaf(node)) {

		for (i=0; i<node->num_childs; i++) {

			if (is_special_node(node->childs_vector[i])) {

				// riferimento che permetter di liberare la memoria relativa a tale node
				node_speciale = node->childs_vector[i];

				// si assegna il nipote direttamente al posto del figlio speciale
				node->childs_vector[i] = node_speciale->childs_vector[0];

				// viene liberata la memoria relativa al node speciale eliminato
				res = free_node(node_speciale);

				if (res == -1)
					return (-1);

			}
		}
	}

	return (1);
}


/*******************************************************************
	funzione che crea la lista_pos per il tag xml_proximity
	         utilizzando l'algoritmo 'plane sweep'
*******************************************************************/

// function used to compute the proximity windows with the plane
// sweep algorithm
int calcola_proximity_con_plane_sweep(int number_proximal_lists,
			  XCDE_IL_Triple_Type **Proximal_lists,
			  int *Proximal_Lists_Length,
			  XCDE_IL_Couple_Type **windows_result,
			  int *number_windows_result,
			  int proximity)
{

 register
    int k=0 ;

 unsigned
   char done=0 ;

 int *position=NULL ;

 int leftmost_element=0 ;

 int rightmost_element=0 ;

 int leftmost_list_position=0 ;

 int rightmost_list_position=0 ;

 int result_list_dimension=0 ;

 // allocates the memory needed for the list pointers
 position=(int *) calloc(number_proximal_lists,
			 sizeof(int)) ;

 result_list_dimension=20 ;

 // allocates the memory for result windows
 *windows_result=(XCDE_IL_Couple_Type *) malloc(result_list_dimension*sizeof(XCDE_IL_Couple_Type)) ;

 *number_windows_result=0 ;

 while(!done)

   { leftmost_element=INT_MAX ;

     rightmost_element=-1 ;

     // find leftmost and rightmost keyword, and their positions in the lists
     for (k=0 ; k<number_proximal_lists; ++k)

       { if (Proximal_lists[k][position[k]].XCDE_Proximity < leftmost_element)

	 { leftmost_element=Proximal_lists[k][position[k]].XCDE_Proximity ;

	 leftmost_list_position=k ;
	 }

       if (Proximal_lists[k][position[k]].XCDE_Proximity > rightmost_element)

	 { rightmost_element= Proximal_lists[k][position[k]].XCDE_Proximity ;

	 rightmost_list_position=k ;
	 }
       }

     if ((position[leftmost_list_position]+1) >= Proximal_Lists_Length[leftmost_list_position])

       { if ((proximity == 0) || ((proximity > 0) && ((rightmost_element-leftmost_element) <= proximity)))

	   { (*windows_result)[(*number_windows_result)].XCDE_Position = Proximal_lists[leftmost_list_position][position[leftmost_list_position]].XCDE_Position ;

	   (*windows_result)[(*number_windows_result)].XCDE_Offset = Proximal_lists[leftmost_list_position][position[leftmost_list_position]].XCDE_Offset ;

	   (*windows_result)[(*number_windows_result+1)].XCDE_Position = Proximal_lists[rightmost_list_position][position[rightmost_list_position]].XCDE_Position ;

	   (*windows_result)[(*number_windows_result+1)].XCDE_Offset = Proximal_lists[rightmost_list_position][position[rightmost_list_position]].XCDE_Offset ;

	   (*number_windows_result)+=2 ;
	   }

	 if ((*number_windows_result) >= result_list_dimension)

	   { *windows_result=(XCDE_IL_Couple_Type *) realloc(*windows_result,
							     result_list_dimension*2*sizeof(XCDE_IL_Couple_Type)) ;

	   result_list_dimension*=2 ;
	   }
	 done=1 ;
       }

     else

       if (Proximal_lists[leftmost_list_position][position[leftmost_list_position]+1].XCDE_Proximity > rightmost_element)

	 // interval [l, r] is minimal, and we insert it in the result ;
	 { if ((proximity == 0) || ((proximity > 0) && ((rightmost_element-leftmost_element) <= proximity)))

	   { (*windows_result)[(*number_windows_result)].XCDE_Position = Proximal_lists[leftmost_list_position][position[leftmost_list_position]].XCDE_Position ;

	   (*windows_result)[(*number_windows_result)].XCDE_Offset = Proximal_lists[leftmost_list_position][position[leftmost_list_position]].XCDE_Offset ;

	   (*windows_result)[(*number_windows_result+1)].XCDE_Position = Proximal_lists[rightmost_list_position][position[rightmost_list_position]].XCDE_Position ;

	   (*windows_result)[(*number_windows_result+1)].XCDE_Offset = Proximal_lists[rightmost_list_position][position[rightmost_list_position]].XCDE_Offset ;

	   (*number_windows_result)+=2 ;
	   }

	 if ((*number_windows_result) >= result_list_dimension)

	   { *windows_result=(XCDE_IL_Couple_Type *) realloc(*windows_result,
							     result_list_dimension*2*sizeof(XCDE_IL_Couple_Type)) ;

	   result_list_dimension*=2 ;
	   }

	 // when there're exactly 2 lists there's a special case, increases also the rightmost position
/*	eliminiamo per una prova la FURBATA!!!

	 if (number_proximal_lists == 2)

	   { ++position[rightmost_list_position] ;

	   if (position[rightmost_list_position] >= Proximal_Lists_Length[rightmost_list_position])

	     done=1 ;
	   }
*/

	 }

     // if [l, r] is minimal, let r=p, where p is the next element in the leftmost element list
     ++position[leftmost_list_position] ;

     if (position[leftmost_list_position] >= Proximal_Lists_Length[leftmost_list_position])
       done=1 ;
   }

 if (position != NULL)

   free(position) ;

 return(1) ;
}




int aggiungi_a_lista_tag(XCDE_IL_Couple_Type ** vettore_da_aggiungere, int Num_El_da_aggiungere, XCDE_IL_Couple_Type **vettore, int *Num_El)
{

	int i = 0;
	int j = 0;
	int k = 0;

	// vettore temporaneo utilizzato per aggiungere le posizioni in maniera ordinata
	XCDE_IL_Couple_Type *temp_vettore = NULL;




	// se non si ha nessun elemento nel vettore
	if (*Num_El == 0) {

		*vettore = *vettore_da_aggiungere;
		*Num_El = Num_El_da_aggiungere;
		*vettore_da_aggiungere = NULL;


	} else { // se il vettore non  vuoto

		// allocazione del vettore temporaneo
		temp_vettore = (XCDE_IL_Couple_Type *) malloc ((Num_El_da_aggiungere + *Num_El) * sizeof(XCDE_IL_Couple_Type));

		if (temp_vettore == NULL) {

			fprintf(stderr,"\n\t<xcde:error> problems during memory reallocation for temp_vettore </xcde:error>\n\n");

			fflush(stderr);

			return(-1);

		}

		// si scorrono i due vettori e si inseriscono le posizioni in maniera ordinata nel vettore temporaneo
		while ( (i < *Num_El) && (j < Num_El_da_aggiungere) ) {

			if ((*vettore)[i].XCDE_Position < (*vettore_da_aggiungere)[j].XCDE_Position) {

				temp_vettore[k] = (*vettore)[i];
				k++;
				i++;

			} else {

				if ((*vettore)[i].XCDE_Position > (*vettore_da_aggiungere)[j].XCDE_Position) {

					temp_vettore[k] = (*vettore_da_aggiungere)[j];
					k++;
					j++;

				} else {

					if ( ((*vettore)[i].XCDE_Position == (*vettore_da_aggiungere)[j].XCDE_Position) && ((*vettore)[i].XCDE_Offset < (*vettore_da_aggiungere)[j].XCDE_Offset) ) {

						temp_vettore[k] = (*vettore)[i];
						k++;
						i++;

					} else {

						if ( ((*vettore)[i].XCDE_Position == (*vettore_da_aggiungere)[j].XCDE_Position) && ((*vettore)[i].XCDE_Offset > (*vettore_da_aggiungere)[j].XCDE_Offset) ) {

							temp_vettore[k] = (*vettore_da_aggiungere)[j];
							k++;
							j++;

						} else {

							temp_vettore[k] = (*vettore)[i];
							k++;
							i++;
							j++;

						}

					}
				}
			}
		}

		// se ci sono ancora elementi nel vettore si finiscono di inserire nel vettore temporaneo
		while (i < *Num_El) {

			temp_vettore[k] = (*vettore)[i];
			k++;
			i++;

		}

		// se ci sono ancora elementi nel vettore_da_aggiungere si finiscono di inserire nel vettore temporaneo
		while (j < Num_El_da_aggiungere) {

			temp_vettore[k] = (*vettore_da_aggiungere)[j];
			k++;
			j++;

		}

		free(*vettore);
		*vettore = temp_vettore;
		*Num_El = k;
		temp_vettore = NULL;

	}

	return (1);
}





//int fondi_liste_or_tag(XCDE_IL_Couple_Type * vettore_da_aggiungere, int Num_El_da_aggiungere, XCDE_IL_Couple_Type **vettore, int *Num_El)
int fondi_liste_or_tag(query_tree_node *node_da_fondere, query_tree_node *node_or)
{
	int i = 0;
	int j = 0;
	int k = 0;

	// vettore temporaneo utilizzato per aggiungere le posizioni in maniera ordinata
	XCDE_IL_Couple_Type *temp_vettore = NULL;

	// vettore temporaneo utilizzato per aggiungere le profondit in maniera ordinata
	int *temp_depth = NULL;




	// se non si ha nessun elemento nel vettore
	if (node_or->num_positions == 0) {

		// allocazione del vettore delle posizioni
		node_or->tag_positions = (XCDE_IL_Couple_Type *) malloc (node_da_fondere->num_positions * sizeof(XCDE_IL_Couple_Type));

		if (node_or->tag_positions == NULL) {

			fprintf(stderr,"\n\t<xcde:error> problems during memory allocation for node_or->tag_positions </xcde:error>\n\n");

			fflush(stderr);

			return(-1);

		}

		// allocazione del vettore delle posizioni
		node_or->depth_vector = (int *) malloc (node_da_fondere->num_positions / 2 * sizeof(int));

		if (node_or->depth_vector == NULL) {

			fprintf(stderr,"\n\t<xcde:error> problems during memory allocation for node_or->depth_vector </xcde:error>\n\n");

			fflush(stderr);

			return(-1);

		}

		// si scorrono i due vettori e si inseriscono le posizioni in maniera ordinata nel vettore temporaneo
		for ( j = 0; j < node_da_fondere->num_positions; j += 2, k++ ) {

			(node_or->tag_positions)[j].XCDE_Position = (node_da_fondere->tag_positions)[j].XCDE_Position;
			(node_or->tag_positions)[j].XCDE_Offset = (node_da_fondere->tag_positions)[j].XCDE_Offset;
			(node_or->tag_positions)[j+1].XCDE_Position = (node_da_fondere->tag_positions)[j+1].XCDE_Position;
			(node_or->tag_positions)[j+1].XCDE_Offset = (node_da_fondere->tag_positions)[j+1].XCDE_Offset;

			(node_or->depth_vector)[k] = (node_da_fondere->depth_vector)[k];

		}

		node_or->num_positions = j;
		node_or->num_depth = k;

	} else { // se il vettore non  vuoto

		// allocazione del vettore temporaneo per le posizioni
		temp_vettore = (XCDE_IL_Couple_Type *) malloc ((node_or->num_positions + node_da_fondere->num_positions) * sizeof(XCDE_IL_Couple_Type));

		if (temp_vettore == NULL) {

			fprintf(stderr,"\n\t<xcde:error> problems during memory allocation for node_or->depth_vector </xcde:error>\n\n");

			fflush(stderr);

			return(-1);

		}

		// allocazione del vettore temporaneo per le depth
		temp_depth = (int *) malloc ((node_or->num_positions + node_da_fondere->num_positions) / 2 * sizeof(int));

		if (temp_depth == NULL) {

			fprintf(stderr,"\n\t<xcde:error> problems during memory allocation for temp_depth </xcde:error>\n\n");

			fflush(stderr);

			return(-1);


		}

		// si scorrono i due vettori e si inseriscono le posizioni  e le depth in maniera ordinata nei vettore temporanei
		while ( (i < node_or->num_positions) && (j < node_da_fondere->num_positions) ) {

			if ((node_or->tag_positions)[i].XCDE_Position < (node_da_fondere->tag_positions)[j].XCDE_Position) {

				temp_vettore[k].XCDE_Position = (node_or->tag_positions)[i].XCDE_Position;
				temp_vettore[k].XCDE_Offset = (node_or->tag_positions)[i].XCDE_Offset;
				temp_vettore[k+1].XCDE_Position = (node_or->tag_positions)[i+1].XCDE_Position;
				temp_vettore[k+1].XCDE_Offset = (node_or->tag_positions)[i+1].XCDE_Offset;

				temp_depth[k/2] = node_or->depth_vector[i/2];

				k += 2;
				i += 2;

			} else {

				if ((node_or->tag_positions)[i].XCDE_Position > (node_da_fondere->tag_positions)[j].XCDE_Position) {

					temp_vettore[k].XCDE_Position = (node_da_fondere->tag_positions)[j].XCDE_Position;
					temp_vettore[k].XCDE_Offset = (node_da_fondere->tag_positions)[j].XCDE_Offset;
					temp_vettore[k+1].XCDE_Position = (node_da_fondere->tag_positions)[j+1].XCDE_Position;
					temp_vettore[k+1].XCDE_Offset = (node_da_fondere->tag_positions)[j+1].XCDE_Offset;

					temp_depth[k/2] = node_da_fondere->depth_vector[j/2];

					k += 2;
					j += 2;

				} else { // sicuramente (*vettore)[i].XCDE_Position == vettore_da_aggiungere[j].XCDE_Position

					if ( (node_or->tag_positions)[i].XCDE_Offset < (node_da_fondere->tag_positions)[j].XCDE_Offset ) {

						temp_vettore[k].XCDE_Position = (node_or->tag_positions)[i].XCDE_Position;
						temp_vettore[k].XCDE_Offset = (node_or->tag_positions)[i].XCDE_Offset;
						temp_vettore[k+1].XCDE_Position = (node_or->tag_positions)[i+1].XCDE_Position;
						temp_vettore[k+1].XCDE_Offset = (node_or->tag_positions)[i+1].XCDE_Offset;

						temp_depth[k/2] = node_or->depth_vector[i/2];

						k += 2;
						i += 2;

					} else {

						temp_vettore[k].XCDE_Position = (node_da_fondere->tag_positions)[j].XCDE_Position;
						temp_vettore[k].XCDE_Offset = (node_da_fondere->tag_positions)[j].XCDE_Offset;
						temp_vettore[k+1].XCDE_Position = (node_da_fondere->tag_positions)[j+1].XCDE_Position;
						temp_vettore[k+1].XCDE_Offset = (node_da_fondere->tag_positions)[j+1].XCDE_Offset;

						temp_depth[k/2] = node_da_fondere->depth_vector[j/2];

						k += 2;
						j += 2;
						i += 2;
					}
				}
			}
		}

		// se ci sono ancora elementi nel vettore si finiscono di inserire nel vettore temporaneo
		while (i < node_or->num_positions) {

			temp_vettore[k].XCDE_Position = (node_or->tag_positions)[i].XCDE_Position;
			temp_vettore[k].XCDE_Offset = (node_or->tag_positions)[i].XCDE_Offset;
			temp_vettore[k+1].XCDE_Position = (node_or->tag_positions)[i+1].XCDE_Position;
			temp_vettore[k+1].XCDE_Offset = (node_or->tag_positions)[i+1].XCDE_Offset;

			temp_depth[k/2] = node_or->depth_vector[i/2];

			k += 2;
			i += 2;

		}

		// se ci sono ancora elementi nel vettore_da_aggiungere si finiscono di inserire nel vettore temporaneo
		while (j < node_da_fondere->num_positions) {

			temp_vettore[k].XCDE_Position = (node_da_fondere->tag_positions)[j].XCDE_Position;
			temp_vettore[k].XCDE_Offset = (node_da_fondere->tag_positions)[j].XCDE_Offset;
			temp_vettore[k+1].XCDE_Position = (node_da_fondere->tag_positions)[j+1].XCDE_Position;
			temp_vettore[k+1].XCDE_Offset = (node_da_fondere->tag_positions)[j+1].XCDE_Offset;

			temp_depth[k/2] = node_da_fondere->depth_vector[j/2];

			k += 2;
			j += 2;

		}

		free(node_or->tag_positions);
		(node_or->tag_positions) = temp_vettore;
		(node_or->num_positions) = k;
		temp_vettore = NULL;

		free(node_or->depth_vector);
		(node_or->depth_vector) = temp_depth;
		(node_or->num_depth) = k/2;
		temp_depth = NULL;

	}

	return 1;

}



int fondi_liste_or_parole(query_tree_node *node_da_fondere, query_tree_node *node_or)
{
	int i = 0;
	int j = 0;
	int k = 0;

	// vettore temporaneo utilizzato per aggiungere le posizioni in maniera ordinata
	XCDE_IL_Couple_Type *temp_vettore = NULL;

	// vettore temporaneo utilizzato per aggiungere le profondit in maniera ordinata
	int *temp_depth = NULL;




	// se non si ha nessun elemento nel vettore
	if (node_or->num_positions == 0) {

		// allocazione del vettore delle posizioni
		node_or->tag_positions = (XCDE_IL_Couple_Type *) malloc (node_da_fondere->num_positions  * 2 * sizeof(XCDE_IL_Couple_Type));

		if (node_or->tag_positions == NULL) {

			fprintf(stderr,"\n\t<xcde:error> problems during memory allocation for node_or->tag_positions </xcde:error>\n\n");

			fflush(stderr);

			return(-1);

		}

		// allocazione del vettore delle posizioni
		node_or->depth_vector = (int *) malloc (node_da_fondere->num_positions * sizeof(int));

		if (node_or->depth_vector == NULL) {

			fprintf(stderr,"\n\t<xcde:error> problems during memory allocation for node_or->depth_vector </xcde:error>\n\n");

			fflush(stderr);

			return(-1);

		}

		// si scorrono i due vettori e si inseriscono le posizioni in maniera ordinata nel vettore temporaneo
		for ( j = 0; j < node_da_fondere->num_positions; j++, k += 2 ) {

			(node_or->tag_positions)[k].XCDE_Position = (node_da_fondere->word_positions)[j].XCDE_Position;
			(node_or->tag_positions)[k].XCDE_Offset = (node_da_fondere->word_positions)[j].XCDE_Offset;
			(node_or->tag_positions)[k+1].XCDE_Position = (node_da_fondere->word_positions)[j].XCDE_Position;
			(node_or->tag_positions)[k+1].XCDE_Offset = (node_da_fondere->word_positions)[j].XCDE_Offset;

			(node_or->depth_vector)[j] = -1;

		}

		node_or->num_positions = k;
		node_or->num_depth = j;

	} else { // se il vettore non  vuoto

		// allocazione del vettore temporaneo per le posizioni
		temp_vettore = (XCDE_IL_Couple_Type *) malloc ((node_or->num_positions + node_da_fondere->num_positions * 2) * sizeof(XCDE_IL_Couple_Type));

		if (temp_vettore == NULL) {

			fprintf(stderr,"\n\t<xcde:error> problems during memory allocation for temp_vettore </xcde:error>\n\n");

			fflush(stderr);

			return(-1);

		}

		// allocazione del vettore temporaneo per le depth
		temp_depth = (int *) malloc ((node_or->num_positions / 2 + node_da_fondere->num_positions) * sizeof(int));

		if (temp_depth == NULL) {

			fprintf(stderr,"\n\t<xcde:error> problems during memory allocation for temp_depth </xcde:error>\n\n");

			fflush(stderr) ;

			return(-1) ;
		}

		// si scorrono i due vettori e si inseriscono le posizioni  e le depth in maniera ordinata nei vettore temporanei
		while ( (i < node_or->num_positions) && (j < node_da_fondere->num_positions) ) {

			if ((node_or->tag_positions)[i].XCDE_Position < (node_da_fondere->word_positions)[j].XCDE_Position) {

				temp_vettore[k].XCDE_Position = (node_or->tag_positions)[i].XCDE_Position;
				temp_vettore[k].XCDE_Offset = (node_or->tag_positions)[i].XCDE_Offset;
				temp_vettore[k+1].XCDE_Position = (node_or->tag_positions)[i+1].XCDE_Position;
				temp_vettore[k+1].XCDE_Offset = (node_or->tag_positions)[i+1].XCDE_Offset;

				temp_depth[k/2] = node_or->depth_vector[i/2];

				k += 2;
				i += 2;

			} else {

				if ((node_or->tag_positions)[i].XCDE_Position > (node_da_fondere->word_positions)[j].XCDE_Position) {

					temp_vettore[k].XCDE_Position = (node_da_fondere->word_positions)[j].XCDE_Position;
					temp_vettore[k].XCDE_Offset = (node_da_fondere->word_positions)[j].XCDE_Offset;
					temp_vettore[k+1].XCDE_Position = (node_da_fondere->word_positions)[j].XCDE_Position;
					temp_vettore[k+1].XCDE_Offset = (node_da_fondere->word_positions)[j].XCDE_Offset;

					temp_depth[k/2] = -1;

					k += 2;
					j++;

				} else { // sicuramente (*vettore)[i].XCDE_Position == vettore_da_aggiungere[j].XCDE_Position

					if ( (node_or->tag_positions)[i].XCDE_Offset < (node_da_fondere->word_positions)[j].XCDE_Offset ) {

						temp_vettore[k].XCDE_Position = (node_or->tag_positions)[i].XCDE_Position;
						temp_vettore[k].XCDE_Offset = (node_or->tag_positions)[i].XCDE_Offset;
						temp_vettore[k+1].XCDE_Position = (node_or->tag_positions)[i+1].XCDE_Position;
						temp_vettore[k+1].XCDE_Offset = (node_or->tag_positions)[i+1].XCDE_Offset;

						temp_depth[k/2] = node_or->depth_vector[i/2];

						k += 2;
						i += 2;

					} else {

						temp_vettore[k].XCDE_Position = (node_da_fondere->word_positions)[j].XCDE_Position;
						temp_vettore[k].XCDE_Offset = (node_da_fondere->word_positions)[j].XCDE_Offset;
						temp_vettore[k+1].XCDE_Position = (node_da_fondere->word_positions)[j].XCDE_Position;
						temp_vettore[k+1].XCDE_Offset = (node_da_fondere->word_positions)[j].XCDE_Offset;

						temp_depth[k/2] = -1;

						k += 2;
						j++;
						i += 2;
					}
				}
			}
		}

		// se ci sono ancora elementi nel vettore si finiscono di inserire nel vettore temporaneo
		while (i < node_or->num_positions) {

			temp_vettore[k].XCDE_Position = (node_or->tag_positions)[i].XCDE_Position;
			temp_vettore[k].XCDE_Offset = (node_or->tag_positions)[i].XCDE_Offset;
			temp_vettore[k+1].XCDE_Position = (node_or->tag_positions)[i+1].XCDE_Position;
			temp_vettore[k+1].XCDE_Offset = (node_or->tag_positions)[i+1].XCDE_Offset;

			temp_depth[k/2] = node_or->depth_vector[i/2];

			k += 2;
			i += 2;

		}

		// se ci sono ancora elementi nel vettore_da_aggiungere si finiscono di inserire nel vettore temporaneo
		while (j < node_da_fondere->num_positions) {

			temp_vettore[k].XCDE_Position = (node_da_fondere->word_positions)[j].XCDE_Position;
			temp_vettore[k].XCDE_Offset = (node_da_fondere->word_positions)[j].XCDE_Offset;
			temp_vettore[k+1].XCDE_Position = (node_da_fondere->word_positions)[j].XCDE_Position;
			temp_vettore[k+1].XCDE_Offset = (node_da_fondere->word_positions)[j].XCDE_Offset;

			temp_depth[k/2] = -1;

			k += 2;
			j++;

		}

		free(node_or->tag_positions);
		(node_or->tag_positions) = temp_vettore;
		(node_or->num_positions) = k;
		temp_vettore = NULL;

		free(node_or->depth_vector);
		(node_or->depth_vector) = temp_depth;
		(node_or->num_depth) = k/2;
		temp_depth = NULL;

	}

	return 1;

}





int fondi_liste_tag(XCDE_IL_Couple_Type ** vettore_da_aggiungere, int Num_El_da_aggiungere, XCDE_IL_Couple_Type **vettore, int *Num_El)
{
	int i = 0;
	int j = 0;
	int k = 0;

	// vettore temporaneo utilizzato per aggiungere le posizioni in maniera ordinata
	XCDE_IL_Couple_Type *temp_vettore = NULL;




	// se non si ha nessun elemento nel vettore
	if (*Num_El == 0) {

		*vettore = *vettore_da_aggiungere;
		*Num_El = Num_El_da_aggiungere;
		*vettore_da_aggiungere = NULL;


	} else { // se il vettore non  vuoto

		// allocazione del vettore temporaneo
		temp_vettore = (XCDE_IL_Couple_Type *) malloc (*Num_El * sizeof(XCDE_IL_Couple_Type));

		if (temp_vettore == NULL) {

			fprintf(stderr,"\n\t<xcde:error> problems during memory allocation for temp_vettore </xcde:error>\n\n");

			fflush(stderr) ;

			return(-1) ;
		}

		// si scorrono i due vettori e si inseriscono le posizioni in maniera ordinata nel vettore temporaneo
		while ( (i < *Num_El) && (j < Num_El_da_aggiungere) ) {

			if ((*vettore)[i].XCDE_Position < (*vettore_da_aggiungere)[j].XCDE_Position) {

				i++;

			} else {

				if ((*vettore)[i].XCDE_Position > (*vettore_da_aggiungere)[j].XCDE_Position) {

					j++;

				} else {

					if ( ((*vettore)[i].XCDE_Position == (*vettore_da_aggiungere)[j].XCDE_Position) && ((*vettore)[i].XCDE_Offset < (*vettore_da_aggiungere)[j].XCDE_Offset) ) {

						i++;

					} else {

						if ( ((*vettore)[i].XCDE_Position == (*vettore_da_aggiungere)[j].XCDE_Position) && ((*vettore)[i].XCDE_Offset > (*vettore_da_aggiungere)[j].XCDE_Offset) ) {

							j++;

						} else {

							temp_vettore[k] = (*vettore)[i];
							k++;
							i++;
							j++;

						}
					}
				}
			}
		}

		free(*vettore);
		*vettore = temp_vettore;
		*Num_El = k;
		temp_vettore = NULL;

	}

	return 1;

}



int aggiungi_triple_lista_parole(XCDE_IL_Triple_Type ** vettore_da_aggiungere, int Num_El_da_aggiungere, XCDE_IL_Triple_Type **vettore, int *Num_El, char *string_to_search, unsigned char lettercase, XCDE_Console_Type *XCDE_Console)
{

	int i = 0;
	int j = 0;
	int k = 0;
	int z = 0;

	// vettore temporaneo utilizzato per aggiungere le posizioni in maniera ordinata
	XCDE_IL_Triple_Type *temp_vettore = NULL;

	int codeword=0 ;

	int codeword_len=0 ;

	char *token=0 ;
	

	// se non c' nessun elemento nel vettore
	if (*Num_El == 0) {

		// allocazione del vettore
		*vettore = (XCDE_IL_Triple_Type *) malloc (Num_El_da_aggiungere * sizeof(XCDE_IL_Triple_Type));

		if (*vettore == NULL) {

			fprintf(stderr,"\n\t<xcde:error> problems during memory allocation for *vettore </xcde:error>\n\n");

			fflush(stderr) ;

			return(-1) ;
		}


		// si scorre il vettore e si inseriscono le posizioni nel vettore temporaneo settando la proximity a -1
		while ( j < Num_El_da_aggiungere ) {

			// se la ricerca  di tipo case sensitive vanno decompresse le singole parole per controllarle
			if ( (!is_entity(string_to_search)) && (lettercase == XCDE_CASE) ) {

				// si recupera la parola alla posizione data e la si confronta con string_to_search
				res = XCDE_Get_Body_Forward_Codeword((*vettore_da_aggiungere)[j].XCDE_Position, &codeword, &codeword_len, XCDE_Console) ;

				if (res == -1) {

					return(-1) ;

				}

				res = XCDE_Codeword2Token(codeword, codeword_len, &token, XCDE_Console) ;

				if (res == -1) {

					return(-1) ;
				}

				if ( (is_entity(token)) || (strcmp(token, string_to_search) == 0) ) {

					(*vettore)[k] = (*vettore_da_aggiungere)[j];
					k++;

				}

			} else {

				(*vettore)[k] = (*vettore_da_aggiungere)[j];
				k++;

			}

			j++;

		}

		*Num_El = k;

	} else { // se il vettore non  vuoto

		// allocazione del vettore temporaneo
		temp_vettore = (XCDE_IL_Triple_Type *) malloc ((Num_El_da_aggiungere + *Num_El) * sizeof(XCDE_IL_Triple_Type));

		if (temp_vettore == NULL) {

			fprintf(stderr,"\n\t<xcde:error> problems during memory reallocation for temp_vettore </xcde:error>\n\n");

			fflush(stderr) ;

			return(-1) ;
		}

		// si scorrono i due vettori e si inseriscono le posizioni in maniera ordinata nel vettore temporaneo
		while ( (i < *Num_El) && (j < Num_El_da_aggiungere) ) {

			if ((*vettore)[i].XCDE_Position < (*vettore_da_aggiungere)[j].XCDE_Position) {

				temp_vettore[k] = (*vettore)[i];
				k++;
				i++;

			} else {

				if ((*vettore)[i].XCDE_Position > (*vettore_da_aggiungere)[j].XCDE_Position) {

					// se la ricerca  di tipo case sensitive vanno decompresse le singole parole per controllarle
					if ( (!is_entity(string_to_search)) && (lettercase == XCDE_CASE) ) {

						// si recupera la parola alla posizione data e la si confronta con string_to_search
						res = XCDE_Get_Body_Forward_Codeword((*vettore_da_aggiungere)[j].XCDE_Position, &codeword, &codeword_len, XCDE_Console) ;

						if (res == -1) {

							return(-1) ;
						}

						res = XCDE_Codeword2Token(codeword, codeword_len, &token, XCDE_Console) ;

						if (res == -1) {

							return(-1) ;
						}

						if ( (is_entity(token)) || (strcmp(token, string_to_search) == 0) ) {

							(*vettore)[k] = (*vettore_da_aggiungere)[j];
							k++;

						}

					} else {

						(*vettore)[k] = (*vettore_da_aggiungere)[j];
						k++;

					}

					j++;

				} else {

					if ( ((*vettore)[i].XCDE_Position == (*vettore_da_aggiungere)[j].XCDE_Position) && ((*vettore)[i].XCDE_Offset < (*vettore_da_aggiungere)[j].XCDE_Offset) ) {

						temp_vettore[k] = (*vettore)[i];
						k++;
						i++;

					} else {

						if ( ((*vettore)[i].XCDE_Position == (*vettore_da_aggiungere)[j].XCDE_Position) && ((*vettore)[i].XCDE_Offset > (*vettore_da_aggiungere)[j].XCDE_Offset) ) {

							// se la ricerca  di tipo case sensitive vanno decompresse le singole parole per controllarle
							if ( (!is_entity(string_to_search)) && (lettercase == XCDE_CASE) ) {

								// si recupera la parola alla posizione data e la si confronta con string_to_search
								res = XCDE_Get_Body_Forward_Codeword((*vettore_da_aggiungere)[j].XCDE_Position, &codeword, &codeword_len, XCDE_Console) ;

								if (res == -1) {

									return(-1) ;
								}

								res = XCDE_Codeword2Token(codeword, codeword_len, &token, XCDE_Console) ;

								if (res == -1) {

									return(-1) ;
								}

								if ( (is_entity(token)) || (strcmp(token, string_to_search) == 0) ) {

									(*vettore)[k] = (*vettore_da_aggiungere)[j];
									k++;

								}

							} else {

								(*vettore)[k] = (*vettore_da_aggiungere)[j];
								k++;

							}
							j++;

						} else {

							temp_vettore[k] = (*vettore)[i];
							k++;
							i++;
							j++;

						}
					}
				}
			}
		}

		// se ci sono ancora elementi nel vettore si finiscono di inserire nel vettore temporaneo
		while (i < *Num_El) {

			temp_vettore[k] = (*vettore)[i];
			k++;
			i++;

		}

		// se ci sono ancora elementi nel vettore_da_aggiungere si finiscono di inserire nel vettore temporaneo
		while (j < Num_El_da_aggiungere) {

			// se la ricerca  di tipo case sensitive vanno decompresse le singole parole per controllarle
			if ( (!is_entity(string_to_search)) && (lettercase == XCDE_CASE) ) {

				// si recupera la parola alla posizione data e la si confronta con string_to_search
				res = XCDE_Get_Body_Forward_Codeword((*vettore_da_aggiungere)[j].XCDE_Position, &codeword, &codeword_len, XCDE_Console) ;

				if (res == -1) {

					return(-1) ;
				}

				res = XCDE_Codeword2Token(codeword, codeword_len, &token, XCDE_Console) ;

				if (res == -1) {

					return(-1) ;
				}

				if ( (is_entity(token)) || (strcmp(token, string_to_search) == 0) ) {

					(*vettore)[k] = (*vettore_da_aggiungere)[j];
					k++;

				}

			} else {

				(*vettore)[k] = (*vettore_da_aggiungere)[j];
				k++;

			}
			j++;

		}

		free(*vettore);
		*vettore = temp_vettore;
		*Num_El = k;
		temp_vettore = NULL;

	}

	return 1;
}




int aggiungi_couple_lista_parole(XCDE_IL_Couple_Type *vettore_da_aggiungere, int Num_El_da_aggiungere, XCDE_IL_Triple_Type **vettore, int *Num_El, char *string_to_search, unsigned char lettercase, XCDE_Console_Type *XCDE_Console)
{

	int i = 0;
	int j = 0;
	int k = 0;
	int z = 0;

	// vettore temporaneo utilizzato per aggiungere le posizioni in maniera ordinata
	XCDE_IL_Triple_Type *temp_vettore = NULL;

	int codeword=0 ;

	int codeword_len=0 ;

	char *token=0 ;

	
	// se non c' nessun elemento nel vettore
	if (*Num_El == 0) {

		// allocazione del vettore
		*vettore = (XCDE_IL_Triple_Type *) malloc (Num_El_da_aggiungere * sizeof(XCDE_IL_Triple_Type));

		if (*vettore == NULL) {

			fprintf(stderr,"\n\t<xcde:error> problems during memory reallocation for *vettore </xcde:error>\n\n");

			fflush(stderr) ;

			return(-1) ;
		}


		// si scorre il vettore e si inseriscono le posizioni nel vettore temporaneo settando la proximity a -1
		while ( j < Num_El_da_aggiungere ) {

			// se la ricerca  di tipo case sensitive vanno decompresse le singole parole per controllarle
			if ( (!is_entity(string_to_search)) && (lettercase == XCDE_CASE) ) {

				// si recupera la parola alla posizione data e la si confronta con string_to_search
				res = XCDE_Get_Body_Forward_Codeword(vettore_da_aggiungere[j].XCDE_Position, &codeword, &codeword_len, XCDE_Console) ;

				if (res == -1) {

					return(-1) ;
				}

				res = XCDE_Codeword2Token(codeword, codeword_len, &token, XCDE_Console) ;

				if (res == -1) {

					return(-1) ;
				}

				if ( (is_entity(token)) || (strcmp(token, string_to_search) == 0) ) {

					(*vettore)[k].XCDE_Position = vettore_da_aggiungere[j].XCDE_Position;
					(*vettore)[k].XCDE_Offset = vettore_da_aggiungere[j].XCDE_Offset;
					(*vettore)[k].XCDE_Proximity = -1;
					k++;
				}

			} else {

				(*vettore)[k].XCDE_Position = vettore_da_aggiungere[j].XCDE_Position;
				(*vettore)[k].XCDE_Offset = vettore_da_aggiungere[j].XCDE_Offset;
				(*vettore)[k].XCDE_Proximity = -1;
				k++;

			}

			j++;

		}

		*Num_El = k;

	} else { // se il vettore non  vuoto

		// allocazione del vettore temporaneo
		temp_vettore = (XCDE_IL_Triple_Type *) malloc ((Num_El_da_aggiungere + *Num_El) * sizeof(XCDE_IL_Triple_Type));

		if (temp_vettore == NULL) {

			fprintf(stderr,"\n\t<xcde:error> problems during memory allocation for temp_vettore </xcde:error>\n\n");

			fflush(stderr) ;

			return(-1) ;
		}

		// si scorrono i due vettori e si inseriscono le posizioni in maniera ordinata nel vettore temporaneo
		while ( (i < *Num_El) && (j < Num_El_da_aggiungere) ) {

			if ((*vettore)[i].XCDE_Position < vettore_da_aggiungere[j].XCDE_Position) {

				temp_vettore[k] = (*vettore)[i];
				k++;
				i++;

			} else {

				if ((*vettore)[i].XCDE_Position > vettore_da_aggiungere[j].XCDE_Position) {

					// se la ricerca  di tipo case sensitive vanno decompresse le singole parole per controllarle
					if ( (!is_entity(string_to_search)) && (lettercase == XCDE_CASE) ) {

						// si recupera la parola alla posizione data e la si confronta con string_to_search
						res = XCDE_Get_Body_Forward_Codeword(vettore_da_aggiungere[j].XCDE_Position, &codeword, &codeword_len, XCDE_Console) ;

						if (res == -1) {

							return(-1) ;
						}

						res = XCDE_Codeword2Token(codeword, codeword_len, &token, XCDE_Console) ;

						if (res == -1) {

							return(-1) ;
						}

						if ( (is_entity(token)) || (strcmp(token, string_to_search) == 0) ) {

							temp_vettore[k].XCDE_Position = vettore_da_aggiungere[j].XCDE_Position;
							temp_vettore[k].XCDE_Offset = vettore_da_aggiungere[j].XCDE_Offset;
							temp_vettore[k].XCDE_Proximity = -1;
							k++;
						}

					} else {

						temp_vettore[k].XCDE_Position = vettore_da_aggiungere[j].XCDE_Position;
						temp_vettore[k].XCDE_Offset = vettore_da_aggiungere[j].XCDE_Offset;
						temp_vettore[k].XCDE_Proximity = -1;
						k++;

					}
					j++;

				} else {

					if ( ((*vettore)[i].XCDE_Position == vettore_da_aggiungere[j].XCDE_Position) && ((*vettore)[i].XCDE_Offset < vettore_da_aggiungere[j].XCDE_Offset) ) {

						temp_vettore[k] = (*vettore)[i];
						k++;
						i++;

					} else {

						if ( ((*vettore)[i].XCDE_Position == vettore_da_aggiungere[j].XCDE_Position) && ((*vettore)[i].XCDE_Offset > vettore_da_aggiungere[j].XCDE_Offset) ) {

							// se la ricerca  di tipo case sensitive vanno decompresse le singole parole per controllarle
							if ( (!is_entity(string_to_search)) && (lettercase == XCDE_CASE) ) {

								// si recupera la parola alla posizione data e la si confronta con string_to_search
								res = XCDE_Get_Body_Forward_Codeword(vettore_da_aggiungere[j].XCDE_Position, &codeword, &codeword_len, XCDE_Console) ;

								if (res == -1) {

									return(-1) ;
								}

								res = XCDE_Codeword2Token(codeword, codeword_len, &token, XCDE_Console) ;

								if (res == -1) {

									return(-1) ;
								}

								if ( (is_entity(token)) || (strcmp(token, string_to_search) == 0) ) {

									temp_vettore[k].XCDE_Position = vettore_da_aggiungere[j].XCDE_Position;
									temp_vettore[k].XCDE_Offset = vettore_da_aggiungere[j].XCDE_Offset;
									temp_vettore[k].XCDE_Proximity = -1;
									k++;
								}

							} else {

								temp_vettore[k].XCDE_Position = vettore_da_aggiungere[j].XCDE_Position;
								temp_vettore[k].XCDE_Offset = vettore_da_aggiungere[j].XCDE_Offset;
								temp_vettore[k].XCDE_Proximity = -1;
								k++;

							}

							j++;

						} else {

							temp_vettore[k] = (*vettore)[i];
							k++;
							i++;
							j++;

						}
					}
				}
			}
		}

		// se ci sono ancora elementi nel vettore si finiscono di inserire nel vettore temporaneo
		while (i < *Num_El) {

			temp_vettore[k] = (*vettore)[i];
			k++;
			i++;

		}

		// se ci sono ancora elementi nel vettore_da_aggiungere si finiscono di inserire nel vettore temporaneo
		while (j < Num_El_da_aggiungere) {

			// se la ricerca  di tipo case sensitive vanno decompresse le singole parole per controllarle
			if ( (!is_entity(string_to_search)) && (lettercase == XCDE_CASE) ) {

				// si recupera la parola alla posizione data e la si confronta con string_to_search
				res = XCDE_Get_Body_Forward_Codeword(vettore_da_aggiungere[j].XCDE_Position, &codeword, &codeword_len, XCDE_Console) ;

				if (res == -1) {

					return(-1) ;
				}

				res = XCDE_Codeword2Token(codeword, codeword_len, &token, XCDE_Console) ;

				if (res == -1) {

					return(-1) ;
				}

				if ( (is_entity(token)) || (strcmp(token, string_to_search) == 0) ) {

					temp_vettore[k].XCDE_Position = vettore_da_aggiungere[j].XCDE_Position;
					temp_vettore[k].XCDE_Offset = vettore_da_aggiungere[j].XCDE_Offset;
					temp_vettore[k].XCDE_Proximity = -1;
					k++;
				}

			} else {

				temp_vettore[k].XCDE_Position = vettore_da_aggiungere[j].XCDE_Position;
				temp_vettore[k].XCDE_Offset = vettore_da_aggiungere[j].XCDE_Offset;
				temp_vettore[k].XCDE_Proximity = -1;
				k++;

			}

			j++;

		}

		free(*vettore);
		*vettore = temp_vettore;
		*Num_El = k;
		temp_vettore = NULL;

	}

	return 1;
}





/**********************************************************************************************
		funzione che data la posizione di apertura di un tag,
		restituisce la relativa posizione di chiusura
**********************************************************************************************/

int get_pos_chiusura_e_depth(XCDE_IL_Couple_Type posizione_apertura, XCDE_IL_Couple_Type *posizione_chiusura, int *depth, XCDE_Console_Type *XCDE_Console)
{

	// vettore di coppie di tipo XCDE_IL_Couple_Type che conterr la lista
	// invertita delle posizioni;
	XCDE_IL_Couple_Type *IL_Couple;

	// variabile che conterr il numero di elementi della lista invertita
	int IL_Num_El;

	int *temp_depth;
	int temp_num_depth;

	int *vettore_ranghi;
	int num_elementi;
	unsigned char is_setted_tagstruct;
	int res;
	unsigned char dict;
	int rango = 0;
	int j;


	res = XCDE_SettedTagstruct(&is_setted_tagstruct, XCDE_Console);

	// controllo sull'esito della funzione XCDE_SettedTagstruct
	if (res == -1) {

		fprintf(stderr,"\n\t<xcde:error> problems with the XCDE_SettedTagstruct function </xcde:error>\n\n");

		fflush(stderr) ;

		return(-1) ;
	}

	// controllo sulla presenza della Tagstruct
	if (!is_setted_tagstruct) {

		fprintf(stderr, "\n\t<xcde:error> Tagstruct must be setted if there is special tag xml_anyvalue in query </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-1) ;
	}

	// recupero dei ranghi globali dei tag che includono la posizione_apertura
	res = XCDE_Get_Tagstruct_Tags_Enclosing(&posizione_apertura,
				      &vettore_ranghi,
				      &num_elementi,
				      XCDE_Console) ;

	// controllo sull'esito della funzione XCDE_Get_Tagstruct_Tags_Enclosing
	if (res == -1) {

		fprintf(stderr, "\n\t<xcde:error> problems with the XCDE_Get_Tagstruct_Tags_Enclosing function </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-1) ;
	}

	res = XCDE_Get_IL_ListPos(vettore_ranghi[num_elementi - 1], &IL_Couple, &IL_Num_El, XCDE_Console);

	// controllo sull'esito della funzione XCDE_Get_IL_ListPos
	if (res == -1) {

		fprintf(stderr, "\n\t<xcde:error> problems with the XCDE_Get_IL_ListPos function </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-1) ;
	}

	// recupero le profondit del tag trovato

	// recupero del rango nel dizionario globale a partire da quello nel dizionario dei TAG
	res = XCDE_Rank2Dict(vettore_ranghi[num_elementi - 1], &dict, &rango, XCDE_Console);

	// controllo sull'esito della funzione XCDE_Dict2Rank
	if (res == -1) {

		fprintf(stderr, "\n\t<xcde:error> problems with the XCDE_Rank2Dict function </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-1) ;
	}

	res = XCDE_Get_Tag_Depth(rango, &temp_depth, &temp_num_depth, XCDE_Console);

	// controllo sull'esito della funzione XCDE_Get_Tag_Depth
	if (res == -1) {

		fprintf(stderr, "\n\t<xcde:error> problems with the XCDE_Get_Tag_Depth function </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-1) ;
	}

	// si scorre il vettore delle posizioni risultato e si riempie la lista delle posizioni del node corrente
	for (j=0; j<IL_Num_El; j+=2) {

		if (posizione_apertura.XCDE_Position == IL_Couple[j].XCDE_Position) {

			posizione_chiusura->XCDE_Position = IL_Couple[j + 1].XCDE_Position;
			posizione_chiusura->XCDE_Offset = IL_Couple[j + 1].XCDE_Offset;

			*depth = temp_depth[j/2];

			break;

		}
	}

	free(IL_Couple);
	IL_Couple = NULL;
	free(temp_depth);
	temp_depth = NULL;

	return (1);
}




/*******************************************************************
               funzione che crea la lista dei TAG
*******************************************************************/

int crea_lista_tag(query_tree_node * node, XCDE_Console_Type *XCDE_Console)
{

	// variabili utilizzate come indici di scorrimento
	unsigned int i,k;
	unsigned int j = 0;

	// rank of the tag in the tags dictionary. The APIs that search in
	// different dictionaries (TAG, ATTR, WORD) returns the position (or
	// positions, because for attributes and words dictionaries its possibile
	// to return more values) of the element in the specific dictionary. This value
	// (or these values) is called the "rank in the dictionary" and can be a number
	// greater than 0 and less than the total number of elements in the dictionary.
	// To obtain the IL of the element the rank in the dictionary must be traslated
	// to the general rank, that is the rank in the global dictionary, using a specific
	// API.
	int tagdictrank=0 ;

	// lista temporanea delle posizioni relative ad un attributo
	XCDE_IL_Couple_Type *temp_aperture_tag = NULL;

	// numero di elementi della lista temporanea
	int temp_elementi = 0;

	// vettore contenente i ranghi nel dizionario delle occorrenze trovate
	int * token_dict = NULL;

	// variabile contenente il numero di occorrenze trovate
	int num_token = 0;

	// variabile contenente il rango nel dizionario globale
	int rango_globale = 0;

	// vettore di coppie di tipo XCDE_IL_Couple_Type che conterr la lista
	// invertita delle posizioni;
	XCDE_IL_Couple_Type *IL_Couple = NULL;

	// variabile che conterr il numero di elementi della lista invertita
	int IL_Num_El = 0;

	int offset_chiusura, posizione_chiusura;

	// variabili utilizzate per l'algoritmo di plane sweep
	XCDE_IL_Triple_Type **Proximal_lists = NULL;
	int *Proximal_Lists_Length = NULL;
	char *max_prox_stringa;
	int proximity;

	// Variabili che conterranno il tipo di ricerca da effettuare sul nome e sul valore dell'attributo in caso di tag xml_anyvalue
	unsigned char tipo_ricerca_nome = XCDE_EXACT;
	unsigned char lettercase_nome = XCDE_CASE;
	unsigned int num_error_nome = 0;
	unsigned char tipo_ricerca_valore = XCDE_EXACT;
	unsigned char lettercase_valore = XCDE_CASE;
	unsigned int num_error_valore = 0;
    	char * valore_da_ricercare;

	// variabile utilizzata durante l'inserimento nella lista_aperture_tag
	unsigned char inserito = 0;



	/*******************************************************************
		       Trattamento dei tag di tipo xml_not
	*******************************************************************/

	// nel caso di TAG speciale xml_not la lista del node punta a quella del node figlio
	// e se il figlio  di tipo WORD il tipo di xml_not diventa WORD
	if (strcmp(node->node_name, "xml_not") == 0) {

		if (node->childs_vector[0]->node_type == TAG) {
			node->tag_positions = node->childs_vector[0]->tag_positions;
			node->num_positions = node->childs_vector[0]->num_positions;
			node->attributes_list = node->childs_vector[0]->attributes_list;
			node->num_attributes = node->childs_vector[0]->num_attributes;
			node->depth_vector =  node->childs_vector[0]->depth_vector;
			node->num_depth =  node->childs_vector[0]->num_depth;

		} else {
			node->word_positions = node->childs_vector[0]->word_positions;
			node->num_positions = node->childs_vector[0]->num_positions;
			node->attributes_list = node->childs_vector[0]->attributes_list;
			node->num_attributes = node->childs_vector[0]->num_attributes;
			node->node_type = WORD;
		}

		if (node->childs_vector[0]->is_potable) {

			node->childs_vector[0]->tag_positions = NULL;
			node->childs_vector[0]->attributes_list = NULL;
			node->childs_vector[0]->depth_vector = NULL;
			node->childs_vector[0]->word_positions = NULL;

		}



	} else {


		/*******************************************************************
			Trattamento dei tag di tipo xml_or
		*******************************************************************/

		// nel caso di TAG speciale xml_or la lista del node  di tipo "tag_positions" ed 
		//	ottenuta fondendo le liste dei figli. Nel caso di figli di tipo WORD si
		// 	inserisce la posizione della parola sia come apertura che chiusura
		if (strcmp(node->node_name, "xml_or") == 0) {

			for (i=0; i<node->num_childs; i++) {

				if (node->childs_vector[i]->node_type == TAG) {

					res = fondi_liste_or_tag(node->childs_vector[i], node);

				} else {

					res = fondi_liste_or_parole(node->childs_vector[i], node);

				}

				// controllo sull'esito della funzione fondi_liste_or_tag() o fondi_liste_or_parole()
				if (res == -1) {

					return(-1) ;
				}
			}

		} else {

			/*******************************************************************
				Trattamento dei tag di tipo xml_proximity
			*******************************************************************/

			if (strcmp(node->node_name, "xml_proximity") == 0) {

				// si recupera il valore della proximity del node xml_proximity
				max_prox_stringa = get_attribute(node, "xml_maxprox");

				// si controlla la presenza dell'attributo 'xml_maxprox'
				if (max_prox_stringa == NULL) {

					fprintf(stderr, "\n\t<xcde:error> missing xml_maxprox attribute in proximity tag </xcde:error>\n\n") ;

					fflush(stderr) ;

					return(-2) ;
				}

				// si controlla che la proximity immessa sia valida.
				//	per essere valida l'importante  che cominci con una cifra
				//	in caso di numero della forma 2au il valore considerato  2
				if (!isdigit(max_prox_stringa[0])) {

					fprintf(stderr, "\n\t<xcde:error> invalid number of proximity (%s) in xml_maxprox attribute </xcde:error>\n\n", max_prox_stringa) ;

					fflush(stderr) ;

					return(-2) ;
				}

				// viene convertita la stringa contenente la proximity in un intero
				proximity = atoi(max_prox_stringa);

				// si controlla che la proximity immessia sia maggiore di 0
				if (proximity <= 0) {

					fprintf(stderr, "\n\t<xcde:error> invalid number of proximity (%d) in xml_maxprox attribute </xcde:error>\n\n", proximity) ;

					fflush(stderr) ;

					return(-2) ;
				}

				// allocazione della memoria necessaria per il vettore che conterr i puntatori
				//	alle liste di posizioni dei figli
				Proximal_lists = (XCDE_IL_Triple_Type **) malloc(node->num_childs * sizeof(XCDE_IL_Triple_Type *));

				if (Proximal_lists == NULL) {

					fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for Proximal_lists </xcde:error>\n\n") ;

					fflush(stderr) ;

					return(-1) ;
				}

				// allocazione della memoria necessaria per il vettore che conterr i puntatori
				//	alle liste di posizioni dei figli
				Proximal_Lists_Length = (int *) malloc(node->num_childs * sizeof(int));

				if (Proximal_Lists_Length == NULL) {

					fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for Proximal_Lists_Length </xcde:error>\n\n") ;

					fflush(stderr) ;

					return(-1) ;
				}

				for (i=0; i < node->num_childs; i++) {

					Proximal_lists[i] = node->childs_vector[i]->word_positions;
					Proximal_Lists_Length[i] = node->childs_vector[i]->num_positions;
				}

				res = calcola_proximity_con_plane_sweep(node->num_childs, Proximal_lists, Proximal_Lists_Length, &(node->tag_positions), &(node->num_positions), proximity);

				if (res == -1)
					return (-1);

				if (node->num_positions == 0) {

					return(0) ;

				}

				node->depth_vector =  (int *) malloc (node->num_positions / 2 * sizeof(int));
				
				if (node->depth_vector == NULL) {

					fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for node->depth_vector </xcde:error>\n\n") ;

					fflush(stderr) ;

					return(-1) ;
				}

				node->num_depth =  node->num_positions / 2;

				for (i=0; i < node->num_depth; i++) {

					node->depth_vector[i] = -1;

				}



			} else {


				/*******************************************************************
					Trattamento dei tag di tipo xml_anyvalue
				*******************************************************************/


				// nel caso in cui si tratta di un tag di tipo xml_anyvalue si richiama la funzione XCDE_Search_Attr_Dictionary
				if (strcmp(node->node_name, "xml_anyvalue") == 0) {

					// si recuperano i ranghi degli attributi del tag xml_anyvalue
					for (i=0; i < node->num_attributes; i+=2) {

						// se l'attributo non  speciale si esegue la ricerca sul valore dell'attributo
						if (!is_special_attribute(node->attributes_list[i])) {

							// si recuperano il tipo di ricerca da effettuare sul valore, se effettuare la ricerca case sensitive o insensitive ed in caso di ricerca con errori il numero di errori
							res = get_type_search_and_value(node->attributes_list[i+1], &tipo_ricerca_valore, &lettercase_valore, &num_error_valore, &valore_da_ricercare);

							if (res == -1) {

								return(-1) ;
							}

							if (res == -2) {

								return(-2) ;
							}

							// ricerca i ranghi nel dizionario degli Attributi degli attributi specificati
							res = XCDE_Search_Attr_Dictionary(node->attributes_list[i], valore_da_ricercare, &token_dict, &num_token, tipo_ricerca_nome, tipo_ricerca_valore,
											lettercase_nome, lettercase_valore, num_error_nome, num_error_valore, XCDE_Console);

							// controllo sull'esito della funzione XCDE_Search_Attr_Dictionary
							if (res == -1) {

								fprintf(stderr, "\n<xcde:error> problems with the XCDE_Search_Attr_Dictionary function </xcde:error>\n\n") ;

								fflush(stderr) ;

								return(-1) ;
							}

							// per ogni elemento del vettore contenente i ranghi vengono recuperate le posizioni ed aggiunte alla lista
							for (j=0; j<num_token; j++) {

								// recupero del rango nel dizionario globale a partire da quello nel dizionario dei TAG
								res = XCDE_Dict2Rank(ATTR, token_dict[j], &rango_globale, XCDE_Console);

								// controllo sull'esito della funzione XCDE_Dict2Rank
								if (res == -1) {

									fprintf(stderr, "\n\t<xcde:error> problems with the XCDE_Dict2Rank function </xcde:error>\n\n") ;

									fflush(stderr) ;

									return(-1) ;
								}

								res = XCDE_Get_IL_ListPos(rango_globale, &IL_Couple, &IL_Num_El, XCDE_Console);

								// controllo sull'esito della funzione XCDE_Get_IL_ListPos
								if (res == -1) {

									fprintf(stderr, "\n\t<xcde:error> problems with the XCDE_Get_IL_ListPos function </xcde:error>\n\n") ;

									fflush(stderr) ;

									return(-1) ;
								}

								res = aggiungi_a_lista_tag(&IL_Couple, IL_Num_El, &temp_aperture_tag, &temp_elementi);

								// controllo sull'esito della funzione aggiungi_a_lista_tag()
								if (res == -1) {

									return(-1) ;
								}

								IL_Couple = NULL;
								IL_Num_El = 0;

							}




							// a questo punto per l'attributo i-esimo in temp_aperture_tag ho la lista delle aperture dei tag
							// 	che soddisfano le condizioni richieste sull'attributo

							if (i == 0) {

								node->tag_positions = temp_aperture_tag;
								node->num_positions = temp_elementi;
								temp_aperture_tag = NULL;
								temp_elementi = 0;

							} else {

								res = fondi_liste_tag(&temp_aperture_tag, temp_elementi, &(node->tag_positions), &(node->num_positions));

								// controllo sull'esito della funzione fondi_liste_tag()
								if (res == -1) {

									return(-1) ;
								}

								free(temp_aperture_tag);
								temp_aperture_tag = NULL;
								temp_elementi = 0;
							}
						}
					}


					/*******************************************************************
						bisogna trasformare la lista di sole aperture in
						lista di aperture e chiusure
					*******************************************************************/

					// allocazione della memoria necessaria per il vettore che conterr le posizioni
					//	di apertura e chiusura dei tag che soddisfano la ricerca
					temp_aperture_tag = (XCDE_IL_Couple_Type *) malloc(2 * node->num_positions * sizeof(XCDE_IL_Couple_Type));

					if (temp_aperture_tag == NULL) {

						fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for temp_aperture_tag </xcde:error>\n\n") ;

						fflush(stderr) ;

						return(-1) ;
					}

					node->depth_vector = (int *) malloc(node->num_positions * sizeof(int));

					if (node->depth_vector == NULL) {

						fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for node->depth_vector </xcde:error>\n\n") ;

						fflush(stderr) ;

						return(-1) ;
					}

					node->num_depth = node->num_positions;

					for (i=0; i < node->num_positions; i++) {

						// si ricopia la posizione di apertura nella relativa posizione del vettore temporaneo
						temp_aperture_tag[2 * i].XCDE_Position = node->tag_positions[i].XCDE_Position;
						temp_aperture_tag[2 * i].XCDE_Offset = node->tag_positions[i].XCDE_Offset;

						// si inserisce la posizione di chiusura relativa alla posizione di apertura corrente nella
						//	posizione successiva a quella di apertura nel vettore temporaneo
						res = get_pos_chiusura_e_depth(node->tag_positions[i], &(temp_aperture_tag[(2 * i) + 1]), &(node->depth_vector[i]), XCDE_Console);

						if (res == -1)
							return (-1);
					}

					free(node->tag_positions);
					node->tag_positions = temp_aperture_tag;
					node->num_positions *= 2;
					temp_aperture_tag = NULL;

					if (node->num_positions == 0) {

						return(0) ;

					}

				} else {

					/*******************************************************************
						Trattamento dei tag non speciali
					*******************************************************************/

					// ricerca il rango nel dizionario dei Tag del tag specificato
					res = XCDE_Search_Tag_Dictionary(node->node_name, &tagdictrank, XCDE_Console);

					// controllo sull'esito della funzione XCDE_Search_Tag_Dictionary
					if (res == -1) {

						fprintf(stderr, "\n\t<xcde:error> problems with the XCDE_Search_Tag_Dictionary function </xcde:error>\n\n") ;

						fflush(stderr) ;

						return(-1) ;
					}

					if ( tagdictrank != 0 ) {

						// recupero del rango nel dizionario globale a partire da quello nel dizionario dei TAG
						res = XCDE_Dict2Rank(TAG, tagdictrank, &rango_globale, XCDE_Console);

						// controllo sull'esito della funzione XCDE_Dict2Rank
						if (res == -1) {

							fprintf(stderr, "\n\t<xcde:error> problems with the XCDE_Dict2Rank function </xcde:error>\n\n") ;

							fflush(stderr) ;

							return(-1) ;
						}

						res = XCDE_Get_IL_ListPos(rango_globale, &(node->tag_positions), &(node->num_positions), XCDE_Console);

						// controllo sull'esito della funzione XCDE_Get_IL_ListPos
						if (res == -1) {

							fprintf(stderr, "\n\t<xcde:error> problems with the XCDE_Get_IL_ListPos function </xcde:error>\n\n") ;

							fflush(stderr) ;

							return(-1) ;
						}

						// recupero le profondit del tag trovato

						res = XCDE_Get_Tag_Depth(tagdictrank, &(node->depth_vector), &(node->num_depth), XCDE_Console);

						// controllo sull'esito della funzione XCDE_Get_Tag_Depth
						if (res == -1) {

							fprintf(stderr, "\n\t<xcde:error> problems with the XCDE_Get_Tag_Depth function </xcde:error>\n\n") ;

							fflush(stderr) ;

							return(-1) ;
						}


					}

					if (node->num_positions == 0) {

						return(0) ;

					}
				}
			}
		}
	}
	
	return 1;
}





/*******************************************************************
            funzione che crea la lista delle PAROLE
*******************************************************************/

int crea_lista_parole(query_tree_node * node, unsigned char tipo_ricerca, int num_error, unsigned char lettercase, unsigned char proximity, XCDE_Console_Type *XCDE_Console)
{


	char *lower_search;

	// variabili utilizzate come indici di scorrimento
	int i, j;

	// pointer to the array that contains the rank of the first word occurrences in the words dictionary. It's
	// an array because depending on the type of search it may contain more than one value.
	int * word_dict = NULL;

	// length of the word_dict array
	int num_word;

	// variabile contenente il rango nel dizionario globale
	int rango_globale;

	// variabile contenente il flag sulla proximity
	unsigned char proximity_settata;

	// vettore di triple di tipo XCDE_IL_Triple_Type che conterr la lista
	// invertita delle posizioni e della proximity;
	XCDE_IL_Triple_Type *IL_Triple;

	// vettore di coppie di tipo XCDE_IL_Couple_Type che conterr la lista
	// invertita delle posizioni;
	XCDE_IL_Couple_Type *IL_Couple;

	// variabile che conterr il numero di elementi della lista invertita
	int IL_Num_El;


	// controllo in caso di proximity settato a XCDE_PROXIMITY sulla presenza della lista delle proximity
	if ( proximity == XCDE_PROXIMITY ) {

		res = XCDE_SettedProximity(&proximity_settata, XCDE_Console);

		// controllo sull'esito della funzione XCDE_Setted_Proximity

		if (res == -1) {

			fprintf(stderr, "\n\t<xcde:error> problems with the XCDE_Setted_Proximity function </xcde:error>\n\n") ;

			fflush(stderr) ;

			return(-1) ;
	 	}

		if ( !proximity_settata ) {

			fprintf(stderr, "\n\t<xcde:error> for proximity search is required proximity list. Rebuild the source file with -t option </xcde:error>\n\n") ;

			fflush(stderr) ;

			return(-1) ;
	 	}
	}

	// la stringa temp_search conterr la versione lower-case della stringa da ricercare
	lower_search = (char *) malloc ((strlen(node->node_name) + 1) * sizeof(char));

	if (lower_search == NULL) {

		fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for lower_search string </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-1) ;
	}

	for (i=0; i<strlen(node->node_name); i++) {

		lower_search[i] = tolower((node->node_name)[i]);
	}

	lower_search[i] = '\0';

	// ricerca sul dizionario delle parole dei ranghi delle occorrenze della parola
	res = XCDE_Search_Word_Dictionary(lower_search, &word_dict, &num_word, tipo_ricerca, XCDE_NOCASE, num_error, XCDE_Console);

	// controllo sull'esito della funzione XCDE_Search_Word_Dictionary
	if (res == -1) {

		fprintf(stderr, "\n\t<xcde:error> problems with the XCDE_Search_Word_Dictionary function </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-1) ;
	}

	// per ogni elemento del vettore contenente i ranghi vengono recuperate le posizioni ed aggiunte alla lista
	for (i=0; i<num_word; i++) {

		// recupero del rango nel dizionario globale a partire da quello nel dizionario delle parole
		res = XCDE_Dict2Rank(WORD, word_dict[i], &rango_globale, XCDE_Console);

		// controllo sull'esito della funzione XCDE_Dict2Rank
		if (res == -1) {

			fprintf(stderr, "\n\t<xcde:error> problems with the XCDE_Dict2Rank function </xcde:error>\n\n") ;

			fflush(stderr) ;

			return(-1) ;
		}


		// in caso di ricerca con proximity si richiama la funzione XCDE_Get_IL_ListPosProx,
		// altrimenti si richiama la funzione XCDE_Get_IL_ListPos
		if ( proximity == XCDE_PROXIMITY ) {

			res = XCDE_Get_IL_ListPosProx(rango_globale, &IL_Triple, &IL_Num_El, XCDE_Console);

			// controllo sull'esito della funzione XCDE_Get_IL_ListPosProx
			if (res == -1) {

				fprintf(stderr, "\n\t<xcde:error> problems with the XCDE_Get_IL_ListPosProx function </xcde:error>\n\n") ;

				fflush(stderr) ;

				return(-1) ;
			}

			// se il tipo di ricerca  con errori si chiama la aggiungi_triple_lista_parole con lettercase = XCDE_NOCASE
			//	altrimenti le parole con errori verrebbero scartate
			if (tipo_ricerca == XCDE_ERROR) {

				res = aggiungi_triple_lista_parole(&IL_Triple, IL_Num_El, &node->word_positions, &node->num_positions, node->node_name, XCDE_NOCASE, XCDE_Console);

				if (res == -1)
					return (-1);

			} else {

				res = aggiungi_triple_lista_parole(&IL_Triple, IL_Num_El, &node->word_positions, &node->num_positions, node->node_name, lettercase, XCDE_Console);

				if (res == -1)
					return (-1);

			}

			if (IL_Triple)
				free(IL_Triple);

		} else { // no proximity

			res = XCDE_Get_IL_ListPos(rango_globale, &IL_Couple, &IL_Num_El, XCDE_Console);

			// controllo sull'esito della funzione XCDE_Get_IL_ListPos
			if (res == -1) {

				fprintf(stderr, "\n\t<xcde:error> problems with the XCDE_Get_IL_ListPos function </xcde:error>\n\n") ;

				fflush(stderr) ;

				return(-1) ;
			}

			// se il tipo di ricerca  con errori si chiama la aggiungi_couple_lista_parole con lettercase = XCDE_NOCASE
			//	altrimenti le parole con errori verrebbero scartate
			if ( (tipo_ricerca == XCDE_ERROR) || (tipo_ricerca == XCDE_REGEXP) ) {

				res = aggiungi_couple_lista_parole(IL_Couple, IL_Num_El, &node->word_positions, &node->num_positions, node->node_name, XCDE_NOCASE, XCDE_Console);

				if (res == -1)
					return (-1);

			} else {

				res = aggiungi_couple_lista_parole(IL_Couple, IL_Num_El, &node->word_positions, &node->num_positions, node->node_name, lettercase, XCDE_Console);

				if (res == -1)
					return (-1);

			}

			if (IL_Couple)
				free(IL_Couple);
		}
	}

	if (word_dict)
		free(word_dict);


	// si controlla che ci sia almeno una occorrenza della parola cercata.
	//	nel caso in cui non ci sia nessuna occorrenza della parola la query non produce nessun risultato
	if (node->num_positions == 0) {

		return(0) ;

	}

	return 1;

}




/*******************************************************************
       funzione che filtra la lista delle posizioni di un node
       in base agli attributi associati a quel node
*******************************************************************/

int filtra_attributes_list(query_tree_node *node, XCDE_IL_Couple_Type *aperture_tag, int elementi_aperture_tag){

	XCDE_IL_Couple_Type *vettore_posizioni_filtrate = NULL;
	int num_positions_filtrate = 0;

	int *vettore_depth_filtrate = NULL;
	int numero_depth_filtrate = 0;

	int i, j;


	vettore_posizioni_filtrate = (XCDE_IL_Couple_Type *) malloc(2 * node->num_positions * sizeof(XCDE_IL_Couple_Type));

	if (vettore_posizioni_filtrate == NULL) {

		fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for vettore_posizioni_filtrate </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-1) ;

	}

	vettore_depth_filtrate = (int *) malloc(node->num_positions * sizeof(int));

	if (vettore_depth_filtrate == NULL) {

		fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for vettore_depth_filtrate </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-1) ;

	}


	for (i = 0, j = 0; (i < node->num_positions) && (j < elementi_aperture_tag); ) {

		if (node->tag_positions[i].XCDE_Position == aperture_tag[j].XCDE_Position)  {

			if (node->tag_positions[i].XCDE_Offset == aperture_tag[j].XCDE_Offset)  {

				// viene aggiunta la posizione di apertura alla nuova lista delle posizioni
				vettore_posizioni_filtrate[num_positions_filtrate].XCDE_Position = node->tag_positions[i].XCDE_Position;
				vettore_posizioni_filtrate[num_positions_filtrate].XCDE_Offset = node->tag_positions[i].XCDE_Offset;

				num_positions_filtrate++;

				// viene aggiunta la posizione di chiusura alla nuova lista delle posizioni
				vettore_posizioni_filtrate[num_positions_filtrate].XCDE_Position = node->tag_positions[i+1].XCDE_Position;
				vettore_posizioni_filtrate[num_positions_filtrate].XCDE_Offset = node->tag_positions[i+1].XCDE_Offset;

				num_positions_filtrate++;

				// viene aggiunta la posizione alla nuova lista delle posizioni del figlio
				vettore_depth_filtrate[numero_depth_filtrate] = node->depth_vector[i/2];
				numero_depth_filtrate++;

				i += 2;
				j++;

			} else {

				if (node->tag_positions[i].XCDE_Offset > aperture_tag[j].XCDE_Offset)  {

					j++;

				} else {

					i += 2;

				}
			}
		} else  {

			if (node->tag_positions[i].XCDE_Position > aperture_tag[j].XCDE_Position)  {

				j++;

			} else {

				i += 2;

			}
		}
	}

	free(node->tag_positions);
	free(node->depth_vector);

	node->tag_positions = vettore_posizioni_filtrate;
	node->num_positions = num_positions_filtrate;
	vettore_posizioni_filtrate = NULL;

	node->depth_vector = vettore_depth_filtrate;
	node->num_depth = numero_depth_filtrate;
	vettore_depth_filtrate = NULL;

	return (1);
}





/*******************************************************************
       funzione che ...........
*******************************************************************/

int filtra_con_attributi(query_tree_node * node, XCDE_Console_Type *XCDE_Console)
{

	// vettore contenente i ranghi nel dizionario delle occorrenze trovate
	int * token_dict = NULL;

	// variabile contenente il numero di occorrenze trovate
	int num_token = 0;

	// variabile contenente il rango nel dizionario globale
	int rango_globale = 0;

	// vettore di coppie di tipo XCDE_IL_Couple_Type che conterr la lista
	// invertita delle posizioni;
	XCDE_IL_Couple_Type *IL_Couple = NULL;

	// variabile che conterr il numero di elementi della lista invertita
	int IL_Num_El = 0;

	// lista temporanea delle posizioni relative ad un attributo
	XCDE_IL_Couple_Type *temp_aperture_tag = NULL;

	// numero di elementi della lista temporanea
	int temp_elementi = 0;

	int i,j;

	unsigned char tipo_ricerca_valore = XCDE_EXACT;
	unsigned char lettercase_valore = XCDE_CASE;
	unsigned int num_error_valore = 0;
    	char * valore_da_ricercare;

	
	if (node->node_type == TAG) {

		for (i=0; i < node->num_attributes; i+=2) {

			if (!is_special_attribute(node->attributes_list[i])){

				// si recuperano il tipo di ricerca da effettuare sul valore, se effettuare la ricerca case sensitive o insensitive ed in caso di ricerca con errori il numero di errori
				res = get_type_search_and_value(node->attributes_list[i+1], &tipo_ricerca_valore, &lettercase_valore, &num_error_valore, &valore_da_ricercare);

				if (res == -1) {

					return(-1) ;
				}

				if (res == -2) {

					return(-2) ;
				}

				// ricerca i ranghi nel dizionario degli Attributi degli attributi specificati
				res = XCDE_Search_Attr_Dictionary(node->attributes_list[i], valore_da_ricercare, &token_dict, &num_token, XCDE_EXACT, tipo_ricerca_valore,
								XCDE_CASE, lettercase_valore, 0, num_error_valore, XCDE_Console);

				// controllo sull'esito della funzione XCDE_Search_Attr_Dictionary
				if (res == -1) {

					fprintf(stderr, "\n\t<xcde:error> problems with the XCDE_Search_Attr_Dictionary function </xcde:error>\n\n") ;

					fflush(stderr) ;

					return(-1) ;

				}

				// per ogni elemento del vettore contenente i ranghi vengono recuperate le posizioni ed aggiunte alla lista
				for (j=0; j<num_token; j++) {

					// recupero del rango nel dizionario globale a partire da quello nel dizionario dei TAG
					res = XCDE_Dict2Rank(ATTR, token_dict[j], &rango_globale, XCDE_Console);

					// controllo sull'esito della funzione XCDE_Dict2Rank
					if (res == -1) {

						fprintf(stderr, "\n\t<xcde:error> problems with the XCDE_Dict2Rank function </xcde:error>\n\n") ;

						fflush(stderr) ;

						return(-1) ;

					}

					res = XCDE_Get_IL_ListPos(rango_globale, &IL_Couple, &IL_Num_El, XCDE_Console);

					// controllo sull'esito della funzione XCDE_Get_IL_ListPos
					if (res == -1) {

						fprintf(stderr, "\n\t<xcde:error> problems with the XCDE_Get_IL_ListPos function </xcde:error>\n\n") ;

						fflush(stderr) ;

						return(-1) ;
					}

					res = aggiungi_a_lista_tag(&IL_Couple, IL_Num_El, &temp_aperture_tag, &temp_elementi);

					// controllo sull'esito della funzione aggiungi_a_lista_tag()
					if (res == -1) {

						return(-1) ;
					}

					IL_Couple = NULL;
					IL_Num_El = 0;
				}

				// a questo punto per l'attributo i-esimo in temp_aperture_tag ho la lista delle aperture dei tag
				// 	che soddisfano le condizioni richieste sull'attributo

				res = filtra_attributes_list(node, temp_aperture_tag, temp_elementi);

				if (res == -1)
					return (-1);

			}
		}
	}
	
	return (1);
}








/*******************************************************************
       funzione che filtra la lista delle posizioni di un node
       in base alle posizioni dei nodi figli
*******************************************************************/

int filtra_lista_node(query_tree_node * node)
{
	// il node oggetto di questa procedura  sicuramente un TAG
	// tutti i TAG speciali sono stati eliminati tranne quelli xml_proximity e xml_not


	// indici temporanei per lo scorrimento delle liste delle posizioni
	int i, j, k;

	XCDE_IL_Couple_Type *vettore_posizioni_filtrate = NULL;
	XCDE_IL_Couple_Type *temp_posizioni = NULL;

	int num_positions_filtrate = 0;

	int *vettore_depth_filtrate = NULL;
	int *temp_depth = NULL;

	int numero_depth_filtrate = 0;

	char * distanza_string;
	int distanza_int;

	char * proximity_string;
	int proximity_int;

	int res;


	// in questa fase i node speciali semplici non vengono considerati poich questa funzione non viene richiamata su tali nodi
	//	e sui node speciali di tipo xml_or

	// nel caso di TAG speciale xml_proximity vanno gestiti gli SNIPPET
	if (strcmp(node->node_name, "xml_proximity") == 0) {

		// la lista associata al tag xml_proximity  gi stata riempita correttamente da crea_lista_tag

	} else {

		// nel caso di TAG speciale xml_not la lista del node va fatta puntare a quella del node figlio
		if (strcmp(node->node_name, "xml_not") == 0) {

			// la lista del node figlio  gia stata associata nella crea_lista_tag

		} else {

			// caso di TAG non speciali.

			vettore_posizioni_filtrate = (XCDE_IL_Couple_Type *) malloc(node->num_positions * sizeof(XCDE_IL_Couple_Type));

			if (vettore_posizioni_filtrate == NULL) {

				fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for vettore_posizioni_filtrate </xcde:error>\n\n") ;

				fflush(stderr) ;

				return(-1) ;
			}

			vettore_depth_filtrate = (int *) malloc(node->num_positions / 2 * sizeof(int));

			if (vettore_depth_filtrate == NULL) {

				fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for vettore_depth_filtrate </xcde:error>\n\n") ;

				fflush(stderr) ;

				return(-1) ;
			}

			for (i = 0; i < node->num_childs; i++) {

				// si estrae il valore dell'attributo xml_dist del node figlio che indica a quale distanza deve
				// trovarsi il figlio dal padre. Se l'attributo non  settato si assume una distanza qualsiasi
				distanza_string = get_attribute(node->childs_vector[i],"xml_dist");

				if (distanza_string != NULL) {

					distanza_int = atoi(distanza_string);
				} else {

					// il valore -1 indica un valore qualsiasi
					distanza_int = -1;

				}

				// inizializzazione dell'index per lo scorrimento della lista di posizioni del figlio
				k = 0;

				/*******************************************************************
						caso di figlio di tipo xml_not
				*******************************************************************/

				if (strcmp(node->childs_vector[i]->node_name, "xml_not") == 0) {

					if (node->childs_vector[i]->node_type == WORD) {

						for (j = 0; (j < node->num_positions) && (k < node->childs_vector[i]->num_positions); ) {

							res = confronta_posizioni(node->tag_positions[j], node->tag_positions[j+1], NULL, NULL, &(node->childs_vector[i]->word_positions[k]), 0);

							// se la funzione confronta_posizioni ritorna 2 significa che c' stato un problema
							if (res == 2)
								return -1;


							if (res < 0) { // la parola del figlio  prima del tag padre

								// bisogna avanzare nella lista del figlio
								k++;

							} else {

								if (res > 0) { // la parola del figlio  dopo del tag padre quindi soddisfa la clausola xml_not

									// viene aggiunta la posizione alla nuova lista delle posizioni
									vettore_posizioni_filtrate[num_positions_filtrate] = node->tag_positions[j];
									num_positions_filtrate++;

									vettore_posizioni_filtrate[num_positions_filtrate] = node->tag_positions[j+1];
									num_positions_filtrate++;

									// si passa alla coppia di posizioni successive del padre
									j += 2;

								} else { // la posizione del figlio  contenuta nel tag padre

									j += 2;
									k += 1;
								}
							}
						}

						while (j < node->num_positions) {

							// viene aggiunta la posizione alla nuova lista delle posizioni
							vettore_posizioni_filtrate[num_positions_filtrate] = node->tag_positions[j];
							num_positions_filtrate++;

							vettore_posizioni_filtrate[num_positions_filtrate] = node->tag_positions[j+1];
							num_positions_filtrate++;

							// si passa alla coppia di posizioni successive del padre
							j += 2;

						}

						// viene fatto uno scambio di puntatori tra vettore_posizioni_filtrate e node->tag_positions
						temp_posizioni = node->tag_positions;
						node->tag_positions = vettore_posizioni_filtrate;
						vettore_posizioni_filtrate = temp_posizioni;

						node->num_positions = num_positions_filtrate;
						num_positions_filtrate = 0;

					} else { // il figlio  di tipo TAG

						for (j = 0; (j < node->num_positions) && (k < node->childs_vector[i]->num_positions); ) {

							res = confronta_posizioni(node->tag_positions[j], node->tag_positions[j+1], &(node->childs_vector[i]->tag_positions[k]), &(node->childs_vector[i]->tag_positions[k+1]), NULL, 1);

							// se la funzione confronta_posizioni ritorna 2 significa che c' stato un problema
							if (res == 2)
								return -1;

								
							if (res < 0) { // il TAG figlio  prima del tag padre

								// bisogna avanzare nella lista del figlio
								k += 2;

							} else {

								if (res > 0) { // il TAG figlio  dopo del tag padre quindi soddisfa la clausola xml_not

									// viene aggiunta la posizione alla nuova lista delle posizioni
									vettore_posizioni_filtrate[num_positions_filtrate] = node->tag_positions[j];
									num_positions_filtrate++;

									vettore_posizioni_filtrate[num_positions_filtrate] = node->tag_positions[j+1];
									num_positions_filtrate++;

									// viene aggiunta la posizione alla nuova lista delle posizioni del figlio
									vettore_depth_filtrate[numero_depth_filtrate] = node->childs_vector[i]->depth_vector[j/2];
									numero_depth_filtrate++;

									// si passa alla coppia di posizioni successive del padre
									j += 2;

								} else { // la posizione del figlio  contenuta nel tag padre

									if ( (distanza_int != -1) && ((node->childs_vector[i]->depth_vector[k/2] - node->depth_vector[j/2]) > distanza_int) ) {

										// viene aggiunta la posizione alla nuova lista delle posizioni
										vettore_posizioni_filtrate[num_positions_filtrate] = node->tag_positions[j];
										num_positions_filtrate++;

										vettore_posizioni_filtrate[num_positions_filtrate] = node->tag_positions[j+1];
										num_positions_filtrate++;

										// viene aggiunta la posizione alla nuova lista delle posizioni del figlio
										vettore_depth_filtrate[numero_depth_filtrate] = node->childs_vector[i]->depth_vector[j/2];
										numero_depth_filtrate++;
									}

									// si passa alla coppia di posizioni successive del padre
									j += 2;

									k += 2;
								}
							}
						}

						while (j < node->num_positions) {

							// viene aggiunta la posizione alla nuova lista delle posizioni
							vettore_posizioni_filtrate[num_positions_filtrate] = node->tag_positions[j];
							num_positions_filtrate++;

							vettore_posizioni_filtrate[num_positions_filtrate] = node->tag_positions[j+1];
							num_positions_filtrate++;

							// viene aggiunta la posizione alla nuova lista delle posizioni del figlio
							vettore_depth_filtrate[numero_depth_filtrate] = node->childs_vector[i]->depth_vector[j/2];
							numero_depth_filtrate++;

							// si passa alla coppia di posizioni successive del padre
							j += 2;

						}

						// viene fatto uno scambio di puntatori tra vettore_posizioni_filtrate e node->tag_positions
						temp_posizioni = node->tag_positions;
						node->tag_positions = vettore_posizioni_filtrate;
						vettore_posizioni_filtrate = temp_posizioni;

						node->num_positions = num_positions_filtrate;
						num_positions_filtrate = 0;


						// viene fatto uno scambio di puntatori tra vettore_depth_filtrate e node->depth_vector
						temp_depth = node->depth_vector;
						node->depth_vector = vettore_depth_filtrate;
						vettore_depth_filtrate = temp_depth;

						node->num_depth = numero_depth_filtrate;
						numero_depth_filtrate = 0;
					}

				} else {

					/*******************************************************************
							caso di figlio generico di tipo TAG
					*******************************************************************/

					if (node->childs_vector[i]->node_type == TAG) {

						if ( (strcmp(node->childs_vector[i]->node_name, "xml_proximity") == 0) && (distanza_int != -1) ) {

							fprintf(stderr, "\n\t<xcde_warning> the xml_dist attribute is not required in xml_proximity tag. xml_dist ignored. </xml_warning>\n\n") ;

						}

						for (j = 0; (j < node->num_positions) && (k < node->childs_vector[i]->num_positions); ) {

							res = confronta_posizioni(node->tag_positions[j], node->tag_positions[j+1], &(node->childs_vector[i]->tag_positions[k]), &(node->childs_vector[i]->tag_positions[k+1]), NULL, 1);

							// se la funzione confronta_posizioni ritorna 2 significa che c' stato un problema
							if (res == 2)
								return -1;


							if (res < 0) { // il TAG figlio  prima del tag padre

								// bisogna avanzare nella lista del figlio
								k += 2;

							} else {

								if (res > 0) { // il TAG figlio  dopo del tag padre quindi bisogna scorrere nella lista del padre

									j += 2;

								} else { // la posizione del figlio  contenuta nel tag padre

									if ( (node->childs_vector[i]->depth_vector[k/2] == -1) || (distanza_int == -1) || ((node->childs_vector[i]->depth_vector[k/2] - node->depth_vector[j/2]) <= distanza_int) ) {

										// viene aggiunta la posizione alla nuova lista delle posizioni
										vettore_posizioni_filtrate[num_positions_filtrate] = node->tag_positions[j];
										num_positions_filtrate++;

										vettore_posizioni_filtrate[num_positions_filtrate] = node->tag_positions[j+1];
										num_positions_filtrate++;

										// viene aggiunta la posizione alla nuova lista delle posizioni del figlio
										vettore_depth_filtrate[numero_depth_filtrate] = node->depth_vector[j/2];
										numero_depth_filtrate++;

										j += 2;
									}

									// si passa alla coppia di posizioni successive del figlio
									k += 2;
								}
							}
						}

						// viene fatto uno scambio di puntatori tra vettore_posizioni_filtrate e node->tag_positions
						temp_posizioni = node->tag_positions;
						node->tag_positions = vettore_posizioni_filtrate;
						vettore_posizioni_filtrate = temp_posizioni;

						node->num_positions = num_positions_filtrate;
						num_positions_filtrate = 0;


						// viene fatto uno scambio di puntatori tra vettore_depth_filtrate e node->depth_vector
						temp_depth = node->depth_vector;
						node->depth_vector = vettore_depth_filtrate;
						vettore_depth_filtrate = temp_depth;

						node->num_depth = numero_depth_filtrate;
						numero_depth_filtrate = 0;

					} else {

						/*******************************************************************
								caso di figlio di tipo WORD
						*******************************************************************/

						for (j = 0; (j < node->num_positions) && (k < node->childs_vector[i]->num_positions); ) {

							res = confronta_posizioni(node->tag_positions[j], node->tag_positions[j+1], NULL, NULL, &(node->childs_vector[i]->word_positions[k]), 0);

							// se la funzione confronta_posizioni ritorna 2 significa che c' stato un problema
							if (res == 2)
								return -1;


							if (res < 0) { // la parola del figlio  prima del tag padre

								// bisogna avanzare nella lista del figlio
								k++;

							} else {

								if (res > 0) { // la parola del figlio  dopo del tag padre

									// si passa alla coppia di posizioni successive del padre
									j += 2;

								} else { // la posizione del figlio  contenuta nel tag padre

									// viene aggiunta la posizione alla nuova lista delle posizioni
									vettore_posizioni_filtrate[num_positions_filtrate] = node->tag_positions[j];
									num_positions_filtrate++;

									vettore_posizioni_filtrate[num_positions_filtrate] = node->tag_positions[j+1];
									num_positions_filtrate++;

									// viene aggiunta la posizione alla nuova lista delle posizioni del figlio
									vettore_depth_filtrate[numero_depth_filtrate] = node->depth_vector[j/2];
									numero_depth_filtrate++;

									j += 2;
									k++;
								}
							}
						}

						// viene fatto uno scambio di puntatori tra vettore_posizioni_filtrate e node->tag_positions
						temp_posizioni = node->tag_positions;
						node->tag_positions = vettore_posizioni_filtrate;
						vettore_posizioni_filtrate = temp_posizioni;

						node->num_positions = num_positions_filtrate;
						num_positions_filtrate = 0;

						// viene fatto uno scambio di puntatori tra vettore_depth_filtrate e node->depth_vector
						temp_depth = node->depth_vector;
						node->depth_vector = vettore_depth_filtrate;
						vettore_depth_filtrate = temp_depth;

						node->num_depth = numero_depth_filtrate;
						numero_depth_filtrate = 0;
					}

				}

				if (node->num_positions == 0) {

					return(0) ;

				}
			}
		}
	}
	
	return (1);
}




/*******************************************************************
  funzione che corregge le liste delle posizioni dei figli del node
*******************************************************************/

int correggi_lista(query_tree_node * node)
{

	XCDE_IL_Couple_Type *vettore_posizioni_filtrate = NULL;
	XCDE_IL_Couple_Type *temp_posizioni = NULL;

	XCDE_IL_Triple_Type *vettore_pos_parole_filtrate = NULL;
	XCDE_IL_Triple_Type *temp_pos_parole = NULL;

	int num_positions_filtrate = 0;

	int *vettore_depth_filtrate = NULL;
	int *temp_depth = NULL;

	char * distanza_string;
	int distanza_int;

	int numero_depth_filtrate = 0;

	int i, j, k;

	int res;


	// se non  un node foglia bisogna correggere la lista di tutti i figli
	if (!is_leaf(node)) {

		// si corregge la lista di tutti i figli
		for (i=0; i<node->num_childs; i++) {

			// si estrae il valore dell'attributo xml_dist del node figlio che indica a quale distanza deve
			// trovarsi il figlio dal padre. Se l'attributo non  settato si assume una distanza qualsiasi
			distanza_string = get_attribute(node->childs_vector[i],"xml_dist");

			if (distanza_string != NULL) {

				distanza_int = atoi(distanza_string);
			} else {

				// il valore -1 indica un valore qualsiasi
				distanza_int = -1;

			}

			// inizializzazione dell'index per lo scorrimento della lista di posizioni del figlio
			k = 0;

			/*******************************************************************
					caso di figlio di tipo xml_not
			*******************************************************************/

			if (strcmp(node->childs_vector[i]->node_name, "xml_not") == 0) {

				// questo tipo di tag va ignorato perch nel not non va eliminato niente

			} else {

				/*******************************************************************
						caso di figlio di tipo TAG
				*******************************************************************/

				if (node->childs_vector[i]->node_type == TAG) {

					vettore_posizioni_filtrate = (XCDE_IL_Couple_Type *) realloc(vettore_posizioni_filtrate, node->childs_vector[i]->num_positions * sizeof(XCDE_IL_Couple_Type));

					if (vettore_posizioni_filtrate == NULL) {

						fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for vettore_posizioni_filtrate </xcde:error>\n\n") ;

						fflush(stderr) ;

						return(-1) ;
					}

					num_positions_filtrate = 0;

					vettore_depth_filtrate = (int *) realloc(vettore_depth_filtrate, node->childs_vector[i]->num_depth * sizeof(int));

					if (vettore_depth_filtrate == NULL) {

						fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for vettore_depth_filtrate </xcde:error>\n\n") ;

						fflush(stderr) ;

						return(-1) ;
					}

					numero_depth_filtrate = 0;

					// si scorrono le posizioni per correggere la lista di posizioni del figlio
					for (j = 0; (j < node->num_positions) && (k < node->childs_vector[i]->num_positions); ) {

						res = confronta_posizioni(node->tag_positions[j], node->tag_positions[j+1], &(node->childs_vector[i]->tag_positions[k]), &(node->childs_vector[i]->tag_positions[k+1]), NULL, 1);

						// se la funzione confronta_posizioni ritorna 2 significa che c' stato un problema
						if (res == 2)
							return -1;


						if (res < 0) { // il TAG figlio  prima del tag padre

							// bisogna avanzare nella lista del figlio
							k += 2;

						} else {

							if (res > 0) { // il TAG figlio  dopo del tag padre quindi bisogna scorrere nella lista del padre

								j += 2;

							} else { // la posizione del figlio  contenuta nel tag padre

								if ( (node->childs_vector[i]->depth_vector[k/2] == -1) || (distanza_int == -1) || ((node->childs_vector[i]->depth_vector[k/2] - node->depth_vector[j/2]) <= distanza_int) ) {

									// viene aggiunta la posizione alla nuova lista delle posizioni del figlio
									vettore_posizioni_filtrate[num_positions_filtrate] = node->childs_vector[i]->tag_positions[k];
									num_positions_filtrate++;

									vettore_posizioni_filtrate[num_positions_filtrate] = node->childs_vector[i]->tag_positions[k+1];
									num_positions_filtrate++;

									// viene aggiunta la posizione alla nuova lista delle posizioni del figlio
									vettore_depth_filtrate[numero_depth_filtrate] = node->childs_vector[i]->depth_vector[k/2];
									numero_depth_filtrate++;

								}

								// si passa alla coppia di posizioni successive del figlio
								k += 2;

							}
						}
					}

					// viene fatto uno scambio di puntatori tra vettore_posizioni_filtrate e node->tag_positions
					temp_posizioni = node->childs_vector[i]->tag_positions;
					node->childs_vector[i]->tag_positions = vettore_posizioni_filtrate;
					vettore_posizioni_filtrate = temp_posizioni;

					node->childs_vector[i]->num_positions = num_positions_filtrate;

					// viene fatto uno scambio di puntatori tra vettore_depth_filtrate e node->depth_vector
					temp_depth = node->childs_vector[i]->depth_vector;
					node->childs_vector[i]->depth_vector = vettore_depth_filtrate;
					vettore_depth_filtrate = temp_depth;

					node->childs_vector[i]->num_depth = numero_depth_filtrate;

				} else {


					/*******************************************************************
							caso di figlio di tipo WORD
					*******************************************************************/

					vettore_pos_parole_filtrate = (XCDE_IL_Triple_Type *) realloc(vettore_pos_parole_filtrate, node->childs_vector[i]->num_positions * sizeof(XCDE_IL_Triple_Type));

					if (vettore_pos_parole_filtrate == NULL) {

						fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for vettore_pos_parole_filtrate </xcde:error>\n\n") ;

						fflush(stderr) ;

						return(-1) ;
					}

					num_positions_filtrate = 0;


					for (j = 0; (j < node->num_positions) && (k < node->childs_vector[i]->num_positions); ) {

						res = confronta_posizioni(node->tag_positions[j], node->tag_positions[j+1], NULL, NULL, &(node->childs_vector[i]->word_positions[k]), 0);

						// se la funzione confronta_posizioni ritorna 2 significa che c' stato un problema
						if (res == 2)
							return -1;


						if (res < 0) { // la parola del figlio  prima del tag padre

							// si avanza nella lista del figlio
							k++;

						} else {

							if (res > 0) { // la parola del figlio  dopo del tag padre

								// si passa alla coppia di posizioni successive del padre
								j += 2;

							} else { // la posizione del figlio  contenuta nel tag padre

								// viene aggiunta la posizione alla nuova lista delle posizioni
								vettore_pos_parole_filtrate[num_positions_filtrate] = node->childs_vector[i]->word_positions[k];
								num_positions_filtrate++;

								// si avanza nella lista del figlio
								k++;
							}
						}
					}

					// viene fatto uno scambio di puntatori tra vettore_posizioni_filtrate e node->childs_vector[i]->word_positions
					temp_pos_parole = node->childs_vector[i]->word_positions;
					node->childs_vector[i]->word_positions = vettore_pos_parole_filtrate;
					vettore_pos_parole_filtrate = temp_pos_parole;

					node->childs_vector[i]->num_positions = num_positions_filtrate;

				}

			}

			// si richiama ricorsivamente la correggi_lista sul figlio

			// nel caso in cui il figlio  un tag xml_not non viene richiamata ricorsivamente la correggi_lista
			//	perch il sottoalbero con radice xml_not non deve essere visualizzato nel risultato
			//	in quanto le posizioni associate ai nodi del sottoalbero sono le posizioni di elementi
			//	che non devono essere presenti nel risultato
			if (strcmp(node->childs_vector[i]->node_name, "xml_not") != 0) {

				res = correggi_lista(node->childs_vector[i]);
				
				if (res == -1)
					return (-1);

			}

		}

	}

	return 1;
}



/*******************************************************************
       funzione che crea la lista delle posizioni di un node
*******************************************************************/

int crea_lista_node(query_tree_node * node, XCDE_Console_Type *XCDE_Console)
{

	if (node->node_type == TAG) {

		res = crea_lista_tag(node, XCDE_Console);
		
		if (res == -1)
			return (-1);
			
		if (res == -2)
			return -2;

		if (res == 0)
			return (0);

		if (strcmp(node->node_name, "xml_or") != 0) {

			res = filtra_lista_node(node);

			if (res == -1)
				return (-1);

			if (res == 0)
				return (0);

			if (strcmp(node->node_name, "xml_anyvalue") != 0) {

				res = filtra_con_attributi(node, XCDE_Console);

				if (res == -1)
					return (-1);
			}
		}

	} else {

		fprintf(stderr, "\n\t<xcde:error> an internal node must be a TAG </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-2) ;
	}
	
	return (1);
}




/*******************************************************************
      funzione che libera la memoria del sottoalbero radice
*******************************************************************/
int elimina_sottoalbero(query_tree_node * node) {


	int i;

	if (!is_leaf(node)) {

		for (i=0; i<node->num_childs; i++) {
			
			res = elimina_sottoalbero(node->childs_vector[i]);
			
			if (res == -1)
				return (-1);

		}
	}
	
	free_node(node);

	return 1;
}





/*******************************************************************
      funzione che esegue la potatura dell'albero passato in input
      eliminando tutti i nodi speciali ed i sottoalberi che non
      hanno nessun node interno al quale  associata una variabile
*******************************************************************/
int potatura_albero(query_tree_node * radice)
{

	unsigned int num_childs_dopo_potatura = 0;

	query_tree_node ** vettore_figli_dopo_potatura = NULL;

	int i;

	query_tree_node *node_speciale;


	vettore_figli_dopo_potatura = (query_tree_node **) malloc(radice->num_childs * sizeof(query_tree_node *));

	if (vettore_figli_dopo_potatura == NULL) {

		fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for vettore_figli_dopo_potatura </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-1) ;
	}

	for (i=0; i<radice->num_childs; i++) {

		if (radice->childs_vector[i]->is_potable) {

			// viene liberata la memoria relativa al sottoalbero che si sta potando
			res = elimina_sottoalbero(radice->childs_vector[i]);
			
			if (res == -1)
				return (-1);

		} else {

			if (is_special_node(radice->childs_vector[i])) {

				// riferimento che permetter di liberare la memoria relativa a tale node
				node_speciale = radice->childs_vector[i];

				// si assegna il nipote direttamente al posto del figlio speciale
				vettore_figli_dopo_potatura[num_childs_dopo_potatura] = node_speciale->childs_vector[0];
				num_childs_dopo_potatura++;

				// viene liberata la memoria relativa al node speciale eliminato
				free_node(node_speciale);

			} else {

				vettore_figli_dopo_potatura[num_childs_dopo_potatura] = radice->childs_vector[i];
				num_childs_dopo_potatura++;

			}
		}

		radice->childs_vector[i] = NULL;

	}

	if (radice->childs_vector)
		free(radice->childs_vector);

	radice->childs_vector = vettore_figli_dopo_potatura;
	radice->num_childs = num_childs_dopo_potatura;

	return (1);

}




/*******************************************************************
     funzione che riempie le liste delle posizioni associate a node
*******************************************************************/

int riempi_lista(query_tree_node * node, unsigned char tipo_ricerca, int num_errori, unsigned char lettercase, unsigned char proximity, HHash_table *input_hash_table, XCDE_Console_Type *XCDE_Console)
{

	int i, errore;

	char * string_err;

	char * string_case;

	char * variabile_string;

	char * dist_string;

	unsigned char flag_case, temp_is_potable;

	Hash_node * node_hash;



	// controllo se nella tabella hash  presente la variabile, in tal caso index_variabile prende il valore dell'index
	//	associato alla variabile, altrimenti prende valore -1
	variabile_string = get_attribute(node,"xml_var");

	if (variabile_string != NULL) {

		node_hash = HHashtable_search(variabile_string, strlen(variabile_string), input_hash_table);

		if (node_hash != NULL)

			node->is_potable = 0;
		else
			node->is_potable = 1;
	} else {

		node->is_potable = 1;
	}

	free(variabile_string);


	// se  un node foglia viene creata la lista opportuna
	if (is_leaf(node)) {


		if (node->node_type == TAG) {

			res = crea_lista_tag(node, XCDE_Console);

			if (res == -1)
				return (-1);

			if (res == -2)
				return -2;

			if (res == 0)
				return (0);

			if (strcmp(node->node_name, "xml_anyvalue") != 0) {

				res = filtra_con_attributi(node, XCDE_Console);

				if (res == -1)
					return (-1);
			}

		} else {

			res = crea_lista_parole(node, tipo_ricerca, num_errori, lettercase, proximity, XCDE_Console);

			if (res == -1)
				return (-1);

			if (res == 0)
				return (0);


		}

	} else {
	// se non  un node foglia viene chiamata ricorsivamente la funzione su ciascuno dei figli

		for (i=0; i<node->num_childs; i++) {

			string_case = get_attribute(node,"xml_case");

			if ( (string_case == NULL) || (strcmp(string_case, "0") == 0) || (strcmp(string_case, "insensitive") == 0) )
				flag_case = XCDE_NOCASE;
			
			else
				if ( (strcmp(string_case, "1") == 0) || (strcmp(string_case, "sensitive") == 0) )
					flag_case = XCDE_CASE;

				else {

					fprintf(stderr, "\n\t<xcde:error> invalid value of 'xml_case' attribute (%s). Value must be { 1 | sensitive | 0 | insensitive } </xcde:error>\n\n", string_case) ;

					fflush(stderr) ;

					return(-2) ;

				}

			if (string_case)
				free(string_case);

			if (strcmp(node->node_name, "xml_exact") == 0) {

				res = riempi_lista(node->childs_vector[i], XCDE_EXACT, 0, flag_case, proximity, input_hash_table, XCDE_Console);

				if (res == -1)
					return (-1);

				if (res == -2)
					return -2;

				if (res == 0)
					return (0);

				if (node->childs_vector[i]->is_potable == 0) {
					node->is_potable = 0;
				}

			} else {
				if (strcmp(node->node_name, "xml_prefix") == 0) {

					res = riempi_lista(node->childs_vector[i], XCDE_PREFIX, num_errori, flag_case, proximity, input_hash_table, XCDE_Console);

					if (res == -1)
						return (-1);

					if (res == -2)
						return -2;

					if (res == 0)
						return (0);

					if (node->childs_vector[i]->is_potable == 0) {
						node->is_potable = 0;
					}

				} else {

					if (strcmp(node->node_name, "xml_suffix") == 0) {

						res = riempi_lista(node->childs_vector[i], XCDE_SUFFIX, num_errori, flag_case, proximity, input_hash_table, XCDE_Console);

						if (res == -1)
							return (-1);

						if (res == -2)
							return -2;

						if (res == 0)
							return (0);

						if (node->childs_vector[i]->is_potable == 0) {
							node->is_potable = 0;
						}

					} else {

						if (strcmp(node->node_name, "xml_contained") == 0) {

							res = riempi_lista(node->childs_vector[i], XCDE_CONTAINED, num_errori, flag_case, proximity, input_hash_table, XCDE_Console);

							if (res == -1)
								return (-1);

							if (res == -2)
								return -2;

							if (res == 0)
								return (0);

							if (node->childs_vector[i]->is_potable == 0) {
								node->is_potable = 0;
							}

						} else {

							if (strcmp(node->node_name, "xml_regexp") == 0) {

								res = riempi_lista(node->childs_vector[i], XCDE_REGEXP, num_errori, flag_case, proximity, input_hash_table, XCDE_Console);

								if (res == -1)
									return (-1);

								if (res == -2)
									return -2;

								if (res == 0)
									return (0);

								if (node->childs_vector[i]->is_potable == 0) {
									node->is_potable = 0;
								}

							} else {

								if (strcmp(node->node_name, "xml_error") == 0) {

									string_err = get_attribute(node,"xml_maxerr");

									if ( string_err == NULL)
										errore = 0;
									else
										errore = atoi(string_err);

									free(string_err);

									res = riempi_lista(node->childs_vector[i], XCDE_ERROR, errore, flag_case, proximity, input_hash_table, XCDE_Console);

									if (res == -1)
										return (-1);

									if (res == -2)
										return -2;

									if (res == 0)
										return (0);

									if (node->childs_vector[i]->is_potable == 0) {
										node->is_potable = 0;
									}

								} else {

									if (strcmp(node->node_name, "xml_proximity") == 0) {

										res = riempi_lista(node->childs_vector[i], XCDE_PROXIMITY, num_errori, flag_case, XCDE_PROXIMITY, input_hash_table, XCDE_Console);

										if (res == -1)
											return (-1);

										if (res == -2)
											return -2;

										if (res == 0)
											return (0);

										if (node->childs_vector[i]->is_potable == 0) {
											node->is_potable = 0;
										}

									} else {

										if (strcmp(node->node_name, "xml_anyvalue") == 0) {

											res = riempi_lista(node->childs_vector[i], XCDE_ANYVALUE, num_errori, flag_case, proximity, input_hash_table, XCDE_Console);

											if (res == -1)
												return (-1);

											if (res == -2)
												return -2;

											if (res == 0)
												return (0);

											if (node->childs_vector[i]->is_potable == 0) {
												node->is_potable = 0;
											}

										} else {

											if (strcmp(node->node_name, "xml_not") == 0) {

												res = riempi_lista(node->childs_vector[i], XCDE_NOT, num_errori, flag_case, proximity, input_hash_table, XCDE_Console);

												if (res == -1)
													return (-1);

												if (res == -2)
													return -2;

												if (node->childs_vector[i]->is_potable == 0) {
													node->is_potable = 0;
												}

											} else {

												if (strcmp(node->node_name, "xml_or") == 0) {

													res = riempi_lista(node->childs_vector[i], XCDE_OR, num_errori, flag_case, proximity, input_hash_table, XCDE_Console);

													if (res == -1)
														return (-1);

													if (res == -2)
														return -2;

													if (node->childs_vector[i]->is_potable == 0) {
														node->is_potable = 0;
													}

												} else {

													res = riempi_lista(node->childs_vector[i], XCDE_EXACT, 0, flag_case, proximity, input_hash_table, XCDE_Console);

													if (res == -1)
														return (-1);

													if (res == -2)
														return -2;

													if (res == 0)
														return (0);

													if (node->childs_vector[i]->is_potable == 0) {
														node->is_potable = 0;
													}
												}
											}
										}
									}
								}
							}
						}
					}
				}
			}
		}

		// si eliminano eventuali figli speciali del node corrente
		res = elimina_figli_speciali(node);

		if (res == -1)
			return (-1);

		// se il node non  speciale bisogna creare la lista del node
		if ( !is_special_node(node) ) {

			res = crea_lista_node(node, XCDE_Console);

			if (res == -1)
				return (-1);

			if (res == -2)
				return -2;

			if (res == 0)
				return (0);

		}

		// dopo aver creato la lista del node si pu effettuare la potatura del sottoalbero radicato nel node corrente
		if (!is_special_node(node)) {
		
			res = potatura_albero(node);
		
			if (res == -1)
				return (-1);

		}


	}


	return 1;



}




/* ---------------------------------------------------- */
/* -----------  function for build results  ----------- */
/* ---------------------------------------------------- */



/*******************************************************************
	funzione che inserisce nei nodi le informazioni relative
	al numero di risultati del proprio sottoalbero.
*******************************************************************/

int fill_number_results_from_position(query_tree_node *node, XCDE_IL_Couple_Type posizione_apertura, XCDE_IL_Couple_Type posizione_chiusura, int *number_result) {

	int temp_res = 0;
	int i, j, res;

	if (is_leaf(node)) {

		for (i=0; i < node->num_positions; i++) {

			res = confronta_posizioni(posizione_apertura, posizione_chiusura, NULL, NULL, &(node->word_positions[i]), 0);

			// se la funzione confronta_posizioni ritorna 2 significa che c' stato un problema
			if (res == 2)
				return -1;


			if (res == 0) { // la parola  contenuta nelle posizioni in input

				temp_res++;

			}
		}

	} else {

		for (i=0; i < node->num_positions; i+=2) {

			res = confronta_posizioni(posizione_apertura, posizione_chiusura, &(node->tag_positions[i]), &(node->tag_positions[i+1]), NULL, 1);

			// se la funzione confronta_posizioni ritorna 2 significa che c' stato un problema
			if (res == 2)
				return -1;


			if (res == 0) { // il tag  contenuto nelle posizioni in input

				temp_res += node->range_results[i/2];

			}
		}
	}

	*number_result = temp_res;

	return 1;

}





/*******************************************************************
	funzione che inserisce nei nodi le informazioni relative
	al numero di risultati del proprio sottoalbero.
*******************************************************************/

int fill_number_results(query_tree_node *node) {


	int temp_res = 1;
	int num_res;
	int i, j;

	if (is_leaf(node)) {

		if (node->node_type == WORD) {

			// allocates the memory needed by the field range_results of node
			node->range_results = (int *) malloc(node->num_positions * sizeof(int)) ;

			if (node->range_results == NULL) {

				fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for node->range_results </xcde:error>\n\n") ;

				fflush(stderr) ;

				return(-1) ;
			}

			for (i=0; i < node->num_positions; i++) {

				node->range_results[i] = 1;
			}

			node->is_setted_number_result = 1;

		} else {

			// allocates the memory needed by the field range_results of node
			node->range_results = (int *) malloc(node->num_positions / 2 * sizeof(int)) ;

			if (node->range_results == NULL) {

				fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for node->range_results </xcde:error>\n\n") ;

				fflush(stderr) ;

				return(-1) ;
			}

			for (i=0; i < (node->num_positions / 2); i++) {

				node->range_results[i] = 1;
			}

			node->is_setted_number_result = 1;

		}

	} else {


		// allocates the memory needed by the field range_results of node
		node->range_results = (int *) malloc(node->num_positions / 2 * sizeof(int)) ;

		if (node->range_results == NULL) {

			fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for node->range_results </xcde:error>\n\n") ;

			fflush(stderr) ;

			return(-1) ;
		}

		for (i=0; i < node->num_childs; i++) {

			res = fill_number_results(node->childs_vector[i]);

			if (res == -1)
				return (-1);

		}

		for (i=0; i < node->num_positions; i += 2) {

			for (j=0; j < node->num_childs; j++) {

				res = fill_number_results_from_position(node->childs_vector[j], node->tag_positions[i], node->tag_positions[i+1], &num_res);

				if (res == -1)
					return -1;

				temp_res *= num_res;

			}

			node->range_results[i/2] = temp_res;
		}

		node->is_setted_number_result = 1;
	}

	return 1;

}



/*******************************************************************
	funzione che restituisce il numero di risultati
*******************************************************************/

int get_number_results_from_position(query_tree_node *node, XCDE_IL_Couple_Type posizione_apertura, XCDE_IL_Couple_Type posizione_chiusura, int *numero_risultati, HHash_table *input_hash_table) {

	char * variabile;
	int numero_variabili;
	int index_variabile;
	Hash_node * node_hash;

	int temp_res = 0;
	int temp_res_figli = 1;
	int num_res;
	int i, j;



	// controllo se nella tabella hash  presente la variabile, in tal caso index_variabile prende il valore dell'index
	//	associato alla variabile, altrimenti prende valore -1
	variabile = get_attribute(node,"xml_var");

	if (variabile != NULL) {

		node_hash = HHashtable_search(variabile, strlen(variabile), input_hash_table);

		if (node_hash != NULL)

			index_variabile = node_hash->index;
		else
			index_variabile = -1;
	} else {

		index_variabile = -1;
	}

	if (is_leaf(node)) {

		// si devono inserire le posizioni nel risultato solo se al node  associata una variabile che  presente anche nella calusola RETURN
		if (index_variabile != -1) {

			if (node->node_type == WORD) {

				for (i=0; i < node->num_positions; i++) {

					res = confronta_posizioni(posizione_apertura, posizione_chiusura, NULL, NULL, &(node->word_positions[i]), 0);

					// se la funzione confronta_posizioni ritorna 2 significa che c' stato un problema
					if (res == 2)
						return -1;


					if (res == 0) { // la parola  contenuta nelle posizioni in input

						temp_res++;

					}
				}

			} else {

				for (i=0; i < node->num_positions; i += 2) {

					res = confronta_posizioni(posizione_apertura, posizione_chiusura, &(node->tag_positions[i]), &(node->tag_positions[i+1]), NULL, 1);

					// se la funzione confronta_posizioni ritorna 2 significa che c' stato un problema
					if (res == 2)
						return -1;


					if (res == 0) { // la parola  contenuta nelle posizioni in input

						temp_res++;

					}
				}

			}



		} else {

			temp_res = 0;
		}

	} else { // il node non  una foglia

		if ( strcmp(node->node_name, "xml_or") == 0) {

			for (j=0; j < node->num_childs; j++) {

				res = get_number_results_from_position(node->childs_vector[j], posizione_apertura, posizione_chiusura, &num_res, input_hash_table);

				if (res == -1)
					return -1;

				if (num_res != 0)
					temp_res_figli *= num_res;

			}

			temp_res += temp_res_figli;

		} else {

			for (i=0; i < node->num_positions; i+=2) {

				temp_res_figli = 1;

				res = confronta_posizioni(posizione_apertura, posizione_chiusura, &(node->tag_positions[i]), &(node->tag_positions[i+1]), NULL, 1);

				// se la funzione confronta_posizioni ritorna 2 significa che c' stato un problema
				if (res == 2)
					return -1;


				if (res == 0) { // il tag  contenuto nelle posizioni in input

					for (j=0; j < node->num_childs; j++) {

						res = get_number_results_from_position(node->childs_vector[j], node->tag_positions[i], node->tag_positions[i+1], &num_res, input_hash_table);

						if (res == -1)
							return -1;

						temp_res_figli *= num_res;

					}

					temp_res += temp_res_figli;

				}
			}
		}

	}

	*numero_risultati = temp_res;

	return 1;
}



/*******************************************************************
funzione che aggiunge un vettore di risultati ad un altro vettore risultati
*******************************************************************/

int aggiungi_a_risultato(XCDE_IL_Couple_Type ***ris_da_agg, int *size_ris_da_agg,
			 XCDE_IL_Couple_Type ***risultato, int *size_risultato ) {



	int i;
	int j;
	int k;



	if (*size_risultato == 0)
		*risultato = NULL;

	(*risultato) = (XCDE_IL_Couple_Type **) realloc((*risultato), ((*size_risultato) + (*size_ris_da_agg)) * sizeof(XCDE_IL_Couple_Type *));

	if ((*risultato) == NULL) {

		fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for (*risultato) </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-1) ;
	}


	for (i = 0; i < *size_ris_da_agg; i++) {

		(*risultato)[(*size_risultato)] = (*ris_da_agg)[i];

		(*size_risultato)++;

		(*ris_da_agg)[i] = NULL;

	}

	free(*ris_da_agg);
	*ris_da_agg = NULL;
	*size_ris_da_agg = 0;
	
	return 1;
}



/*******************************************************************
funzione che aggiunge un vettore di risultati ad un altro vettore risultati
*******************************************************************/

int fondi_risultati(XCDE_IL_Couple_Type ***primo_risultato, int *size_primo_ris,
		    XCDE_IL_Couple_Type ***secondo_risultato, int *size_secondo_ris,
		    XCDE_IL_Couple_Type ***risultato_fuso, int *size_risultato_fuso,
		    int numero_variabili ) {


	int i;
	int j;
	int k;


	if (*size_risultato_fuso == 0)
		*risultato_fuso = NULL;
	else {

		for (i = 0; i < *size_risultato_fuso; i++)
			free((*risultato_fuso)[i]);

		*size_risultato_fuso = 0;
		*risultato_fuso = NULL;

	}

	(*risultato_fuso) = (XCDE_IL_Couple_Type **) realloc((*risultato_fuso), ((*size_primo_ris) * (*size_secondo_ris)) * sizeof(XCDE_IL_Couple_Type *));

	if ((*risultato_fuso) == NULL) {

		fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for (**risultato_fuso) </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-1) ;
	}

	for (i = 0; i < *size_primo_ris; i++) {

		for (j = 0; j < *size_secondo_ris; j++) {

			(*risultato_fuso)[*size_risultato_fuso] = (XCDE_IL_Couple_Type *) calloc(2 * numero_variabili, sizeof(XCDE_IL_Couple_Type));

			if ((*risultato_fuso)[*size_risultato_fuso] == NULL) {

				fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for (**risultato_fuso)[size_risultato_fuso] </xcde:error>\n\n") ;

				fflush(stderr) ;

				return(-1) ;
			}

			for (k = 0; k < 2 * numero_variabili; k+=2) {

				if ((*primo_risultato)[i][k].XCDE_Position != 0) {

					(*risultato_fuso)[*size_risultato_fuso][k].XCDE_Position = (*primo_risultato)[i][k].XCDE_Position;
					(*risultato_fuso)[*size_risultato_fuso][k].XCDE_Offset = (*primo_risultato)[i][k].XCDE_Offset;

					(*risultato_fuso)[*size_risultato_fuso][k+1].XCDE_Position = (*primo_risultato)[i][k+1].XCDE_Position;
					(*risultato_fuso)[*size_risultato_fuso][k+1].XCDE_Offset = (*primo_risultato)[i][k+1].XCDE_Offset;

				}

				if ((*secondo_risultato)[j][k].XCDE_Position != 0) {

					(*risultato_fuso)[*size_risultato_fuso][k].XCDE_Position = (*secondo_risultato)[j][k].XCDE_Position;
					(*risultato_fuso)[*size_risultato_fuso][k].XCDE_Offset = (*secondo_risultato)[j][k].XCDE_Offset;

					(*risultato_fuso)[*size_risultato_fuso][k+1].XCDE_Position = (*secondo_risultato)[j][k+1].XCDE_Position;
					(*risultato_fuso)[*size_risultato_fuso][k+1].XCDE_Offset = (*secondo_risultato)[j][k+1].XCDE_Offset;

				}
			}


			(*size_risultato_fuso)++;

		}

		free((*primo_risultato)[i]);
		//(*primo_risultato)[i] = NULL;

	}

	free(*primo_risultato);
	*primo_risultato = NULL;
	*size_primo_ris = 0;

	for (j = 0; j < *size_secondo_ris; j++) {

		free((*secondo_risultato)[j]);
		(*secondo_risultato)[j] = NULL;
	}

	free(*secondo_risultato);
	*secondo_risultato = NULL;
	*size_secondo_ris = 0;

	return (1);
}





/*******************************************************************
funzione che aggiunge un vettore di risultati ad un altro vettore risultati
*******************************************************************/

int fondi_risultati_or (XCDE_IL_Couple_Type ***primo_risultato, int *size_primo_ris,
			XCDE_IL_Couple_Type ***secondo_risultato, int *size_secondo_ris,
			XCDE_IL_Couple_Type ***risultato_fuso, int *size_risultato_fuso,
			int numero_variabili ) {

	int i;
	int j;
	int k;



	if ((*size_primo_ris != 0) && (*size_secondo_ris != 0) ) {

		res = fondi_risultati(primo_risultato, size_primo_ris, secondo_risultato, size_secondo_ris, risultato_fuso, size_risultato_fuso, numero_variabili);
		
		if (res == -1)
			return (-1);

	} else {

		if (*size_risultato_fuso == 0) {

			*risultato_fuso = NULL;
		} else {

			for (i = 0; i < *size_risultato_fuso; i++)
				free((*risultato_fuso)[i]);

			*size_risultato_fuso = 0;
			*risultato_fuso = NULL;

		}

		if (*size_primo_ris != 0)  {

			(*risultato_fuso) = (*primo_risultato);
			*size_risultato_fuso = *size_primo_ris;

			*primo_risultato = NULL;
			*size_primo_ris = 0;

		} else {

			if (*size_secondo_ris != 0)  {

				(*risultato_fuso) = (*secondo_risultato);
				*size_risultato_fuso = *size_secondo_ris;

				*secondo_risultato = NULL;
				*size_secondo_ris = 0;
			}
		}
	}
	
	return (1);
}






int aggiungi_riga_a_risultato(XCDE_IL_Couple_Type *riga_da_inserire, XCDE_IL_Couple_Type ***risultato, int *size_vettore_risultato, int numero_variabili) {

	int k,i;


	if (*size_vettore_risultato == 0) {

		*risultato = NULL;

	}

	// controllo sulla disponibilit di un elemento nel vettore (*risultato)
	if ( (*size_vettore_risultato % INIT_SIZE_VETTORE_RISULTATO) == 0) {

		// non ho pi posto nel vettore (*risultato) e si deve
		// riallocare uno spazio maggiore (dimensione vecchia + INIT_SIZE_VETTORE_RISULTATO)
		(*risultato) = (XCDE_IL_Couple_Type **) realloc(*risultato, (*size_vettore_risultato + INIT_SIZE_VETTORE_RISULTATO) * sizeof(XCDE_IL_Couple_Type *));

		if ((*risultato) == NULL) {

			fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for (*risultato) </xcde:error>\n\n") ;

			fflush(stderr) ;

			return(-1) ;
		}
	}

	(*risultato)[*size_vettore_risultato] = riga_da_inserire;
	(*size_vettore_risultato)++;

	riga_da_inserire = NULL;

	return (1);
}





/*******************************************************************
funzione che costruisce il vettore risultato da una posizione
*******************************************************************/

int get_result_from_position(query_tree_node *node, XCDE_IL_Couple_Type posizione_apertura_padre, XCDE_IL_Couple_Type posizione_chiusura_padre,
			     int index_padre, XCDE_IL_Couple_Type ***risultato, int *size_vettore_risultato, HHash_table *input_hash_table) {


	char * variabile;
	int numero_variabili;
	int index_variabile;
	Hash_node * node_hash;

	XCDE_IL_Couple_Type **risultato_figlio;
	int size_risultato_figlio = 0;

	XCDE_IL_Couple_Type ***risultato_parziale;
	int *size_risultato_parziale;

	XCDE_IL_Couple_Type *riga_da_inserire;

	int i, k;

	int j = 0;



	// controllo se nella tabella hash  presente la variabile, in tal caso index_variabile prende il valore dell'index
	//	associato alla variabile, altrimenti prende valore -1
	variabile = get_attribute(node,"xml_var");

	if (variabile != NULL) {

		node_hash = HHashtable_search(variabile, strlen(variabile), input_hash_table);

		if (node_hash != NULL)

			index_variabile = node_hash->index;
		else
			index_variabile = -1;
	} else {

		index_variabile = -1;
	}

	// si recupara dalla tabella hash il numero di variabili
	numero_variabili = input_hash_table->card;

	if (is_leaf(node)) {

		// si devono inserire le posizioni nel risultato solo se al node  associata una variabile che  presente anche nella calusola RETURN
		if (index_variabile != -1) {

			if (node->node_type == WORD) {

				for (i = 0; i < node->num_positions; i++) {

					res = confronta_posizioni(posizione_apertura_padre, posizione_chiusura_padre, NULL, NULL, &(node->word_positions[i]), 0);

					// se la funzione confronta_posizioni ritorna 2 significa che c' stato un problema
					if (res == 2)
						return -1;


					if (res == 0) {

						// creo la riga della matrice relativa alla posizione corrente
						// devo allocare il vettore che contiene le posizioni del (*risultato) corrente (la dimensione di numero_variabili  moltiplicata
						//	per 2 perch vanno considerate le posizioni di apertura e chiusura
						riga_da_inserire = (XCDE_IL_Couple_Type *) calloc(2 * numero_variabili, sizeof(XCDE_IL_Couple_Type));

						if (riga_da_inserire == NULL) {

							fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for riga_da_inserire </xcde:error>\n\n") ;

							fflush(stderr) ;

							return(-1) ;
						}

						// inserisco la posizione della parola sia come valore di apertura che come valore di chiusura nella riga_da_inserire
						//	nella posizione corrispondente alla variabile associata al node
						riga_da_inserire[2 * index_variabile].XCDE_Position = node->word_positions[i].XCDE_Position;
						riga_da_inserire[2 * index_variabile].XCDE_Offset = node->word_positions[i].XCDE_Offset;

						riga_da_inserire[2 * index_variabile + 1].XCDE_Position = node->word_positions[i].XCDE_Position;
						riga_da_inserire[2 * index_variabile + 1].XCDE_Offset = node->word_positions[i].XCDE_Offset;

						if (index_padre != -1 ) {

							// inserisco la posizione del padre nel vettore risultato
							//	nella posizione corrispondente alla variabile associata al node padre
							riga_da_inserire[2 * index_padre].XCDE_Position = posizione_apertura_padre.XCDE_Position;
							riga_da_inserire[2 * index_padre].XCDE_Offset = posizione_apertura_padre.XCDE_Offset;

							riga_da_inserire[2 * index_padre + 1].XCDE_Position = posizione_chiusura_padre.XCDE_Position;
							riga_da_inserire[2 * index_padre + 1].XCDE_Offset = posizione_chiusura_padre.XCDE_Offset;

						}

						res = aggiungi_riga_a_risultato(riga_da_inserire, risultato, size_vettore_risultato, numero_variabili);
						
						if (res == -1)
							return (-1);


					} else {

						if (res > 0) { // la posizione del figlio  dopo quella del padre

							// essendo le posizioni del figlio in ordine, se la posizione corrente  dopo quella del padre
							//	si pu uscire dal for perch anche tutte quelle successive non saranno buone
							i = node->num_positions;
						}

					}
				}

			} else { // la foglia  di tipo TAG

				for (i = 0; i < node->num_positions; i += 2) {

					res = confronta_posizioni(posizione_apertura_padre, posizione_chiusura_padre, &(node->tag_positions[i]), &(node->tag_positions[i+1]), NULL, 1);

					// se la funzione confronta_posizioni ritorna 2 significa che c' stato un problema
					if (res == 2)
						return -1;


					if (res == 0) {

						// creo la riga della matrice relativa alla posizione corrente
						// devo allocare il vettore che contiene le posizioni del (*risultato) corrente (la dimensione di numero_variabili  moltiplicata
						//	per 2 perch vanno considerate le posizioni di apertura e chiusura
						riga_da_inserire = (XCDE_IL_Couple_Type *) calloc(2 * numero_variabili, sizeof(XCDE_IL_Couple_Type));

						if (riga_da_inserire == NULL) {

							fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for riga_da_inserire </xcde:error>\n\n") ;

							fflush(stderr) ;

							return(-1) ;
						}

						// inserisco la posizione della parola sia come valore di apertura che come valore di chiusura nella riga_da_inserire
						//	nella posizione corrispondente alla variabile associata al node
						riga_da_inserire[2 * index_variabile].XCDE_Position = node->tag_positions[i].XCDE_Position;
						riga_da_inserire[2 * index_variabile].XCDE_Offset = node->tag_positions[i].XCDE_Offset;

						riga_da_inserire[2 * index_variabile + 1].XCDE_Position = node->tag_positions[i+1].XCDE_Position;
						riga_da_inserire[2 * index_variabile + 1].XCDE_Offset = node->tag_positions[i+1].XCDE_Offset;

						if (index_padre != -1 ) {

							// inserisco la posizione del padre nel vettore risultato
							//	nella posizione corrispondente alla variabile associata al node padre
							riga_da_inserire[2 * index_padre].XCDE_Position = posizione_apertura_padre.XCDE_Position;
							riga_da_inserire[2 * index_padre].XCDE_Offset = posizione_apertura_padre.XCDE_Offset;

							riga_da_inserire[2 * index_padre + 1].XCDE_Position = posizione_chiusura_padre.XCDE_Position;
							riga_da_inserire[2 * index_padre + 1].XCDE_Offset = posizione_chiusura_padre.XCDE_Offset;

						}

						res = aggiungi_riga_a_risultato(riga_da_inserire, risultato, size_vettore_risultato, numero_variabili);

						if (res == -1)
							return (-1);

					} else {

						if (res > 0) { // la posizione del figlio  dopo quella del padre

							// essendo le posizioni del figlio in ordine, se la posizione corrente  dopo quella del padre
							//	si pu uscire dal for perch anche tutte quelle successive non saranno buone
							i = node->num_positions;
						}

					}
				}
			}

		} else {

			if (index_padre != -1 ) {

				// creo la riga della matrice relativa alla posizione corrente
				// devo allocare il vettore che contiene le posizioni del (*risultato) corrente (la dimensione di numero_variabili  moltiplicata
				//	per 2 perch vanno considerate le posizioni di apertura e chiusura
				riga_da_inserire = (XCDE_IL_Couple_Type *) calloc(2 * numero_variabili, sizeof(XCDE_IL_Couple_Type));

				if (riga_da_inserire == NULL) {

					fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for riga_da_inserire </xcde:error>\n\n") ;

					fflush(stderr) ;

					return(-1) ;
				}

				// inserisco la posizione del padre nel vettore risultato
				//	nella posizione corrispondente alla variabile associata al node padre
				riga_da_inserire[2 * index_padre].XCDE_Position = posizione_apertura_padre.XCDE_Position;
				riga_da_inserire[2 * index_padre].XCDE_Offset = posizione_apertura_padre.XCDE_Offset;

				riga_da_inserire[2 * index_padre + 1].XCDE_Position = posizione_chiusura_padre.XCDE_Position;
				riga_da_inserire[2 * index_padre + 1].XCDE_Offset = posizione_chiusura_padre.XCDE_Offset;

				res = aggiungi_riga_a_risultato(riga_da_inserire, risultato, size_vettore_risultato, numero_variabili);

				if (res == -1)
					return (-1);

			} else {

				// le posizioni non devono essere inserite nel risultato
				*risultato = NULL;
				size_vettore_risultato = 0;
			}
		}

	} else { // il node non  una foglia


		if ( strcmp(node->node_name, "xml_or") == 0) {

			if (node->num_childs == 1) {

				res = get_result_from_position(node->childs_vector[0], posizione_apertura_padre, posizione_chiusura_padre, -1, &risultato_figlio, &size_risultato_figlio, input_hash_table);

				if (res == -1)
					return (-1);

				res = aggiungi_a_risultato(&risultato_figlio, &size_risultato_figlio, risultato, size_vettore_risultato );

				if (res == -1)
					return (-1);

			} else { // numero figli  maggiore di zero altrimenti sarebbe stata una foglia

				// devo allocare il vettore che contiene le posizioni del risultato corrente (la dimensione di numero_variabili  moltiplicata per 2 perch vanno considerate le posizioni di apertura e chiusura
				risultato_parziale = (XCDE_IL_Couple_Type ***) malloc(3 * sizeof(XCDE_IL_Couple_Type**));

				if (risultato_parziale == NULL) {

					fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for risultato_parziale </xcde:error>\n\n") ;

					fflush(stderr) ;

					return(-1) ;
				}

				// devo allocare il vettore che contiene le posizioni del risultato corrente (la dimensione di numero_variabili  moltiplicata per 2 perch vanno considerate le posizioni di apertura e chiusura
				size_risultato_parziale = (int *) calloc(3, sizeof(int));

				if (size_risultato_parziale == NULL) {

					fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for size_risultato_parziale </xcde:error>\n\n") ;

					fflush(stderr) ;

					return(-1) ;
				}

				res = get_result_from_position(node->childs_vector[0], posizione_apertura_padre, posizione_chiusura_padre, -1, &(risultato_parziale[j]), &(size_risultato_parziale[j]), input_hash_table);

				if (res == -1)
					return (-1);

				res = get_result_from_position(node->childs_vector[1], posizione_apertura_padre, posizione_chiusura_padre, -1, &(risultato_parziale[((j + 1) % 3)]), &(size_risultato_parziale[((j + 1) % 3)]), input_hash_table);

				if (res == -1)
					return (-1);

				if (strcmp(node->node_name, "xml_or") == 0) {

					res = fondi_risultati_or(&(risultato_parziale[j]), &(size_risultato_parziale[j]), &(risultato_parziale[((j + 1) % 3)]), &(size_risultato_parziale[((j + 1) % 3)]), &(risultato_parziale[((j + 2) % 3)]), &(size_risultato_parziale[((j + 2) % 3)]), numero_variabili);

					if (res == -1)
						return (-1);

				} else {

					res = fondi_risultati(&(risultato_parziale[j]), &(size_risultato_parziale[j]), &(risultato_parziale[((j + 1) % 3)]), &(size_risultato_parziale[((j + 1) % 3)]), &(risultato_parziale[((j + 2) % 3)]), &(size_risultato_parziale[((j + 2) % 3)]), numero_variabili);

					if (res == -1)
						return (-1);

				}

				j = ((j + 2) % 3);


				for (k=2; k<node->num_childs; k++) {

					res = get_result_from_position(node->childs_vector[k], posizione_apertura_padre, posizione_chiusura_padre, -1, &(risultato_parziale[((j + 1) % 3)]), &(size_risultato_parziale[((j + 1) % 3)]), input_hash_table);

					if (res == -1)
						return (-1);

					if (strcmp(node->node_name, "xml_or") == 0) {

						res = fondi_risultati_or(&(risultato_parziale[j]), &(size_risultato_parziale[j]), &(risultato_parziale[((j + 1) % 3)]), &(size_risultato_parziale[((j + 1) % 3)]), &(risultato_parziale[((j + 2) % 3)]), &(size_risultato_parziale[((j + 2) % 3)]), numero_variabili);

						if (res == -1)
							return (-1);

					} else {

						res = fondi_risultati(&(risultato_parziale[j]), &(size_risultato_parziale[j]), &(risultato_parziale[((j + 1) % 3)]), &(size_risultato_parziale[((j + 1) % 3)]), &(risultato_parziale[((j + 2) % 3)]), &(size_risultato_parziale[((j + 2) % 3)]), numero_variabili);

						if (res == -1)
							return (-1);

					}

					j = ((j + 2) % 3);

				}

				res = aggiungi_a_risultato(&(risultato_parziale[j]), &(size_risultato_parziale[j]), risultato, size_vettore_risultato );

				if (res == -1)
					return (-1);

				if (risultato_parziale)
					free(risultato_parziale);

				if (size_risultato_parziale)
					free(size_risultato_parziale);

			}


		} else {

			for (i = 0; i < node->num_positions; i += 2) {

				res = confronta_posizioni(posizione_apertura_padre, posizione_chiusura_padre, &(node->tag_positions[i]), &(node->tag_positions[i+1]), NULL, 1);

				// se la funzione confronta_posizioni ritorna 2 significa che c' stato un problema
				if (res == 2)
					return -1;


				if (res == 0) {

					if (node->num_childs == 1) {

						res = get_result_from_position(node->childs_vector[0], node->tag_positions[i], node->tag_positions[i+1], index_variabile, &risultato_figlio, &size_risultato_figlio, input_hash_table);

						if (res == -1)
							return (-1);

						res = aggiungi_a_risultato(&risultato_figlio, &size_risultato_figlio, risultato, size_vettore_risultato );

						if (res == -1)
							return (-1);

					} else { // numero figli  maggiore di zero altrimenti sarebbe stata una foglia

						// devo allocare il vettore che contiene le posizioni del risultato corrente (la dimensione di numero_variabili  moltiplicata per 2 perch vanno considerate le posizioni di apertura e chiusura
						risultato_parziale = (XCDE_IL_Couple_Type ***) malloc(3 * sizeof(XCDE_IL_Couple_Type**));

						if (risultato_parziale == NULL) {

							fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for risultato_parziale </xcde:error>\n\n") ;

							fflush(stderr) ;

							return(-1) ;
						}

						// devo allocare il vettore che contiene le posizioni del risultato corrente (la dimensione di numero_variabili  moltiplicata per 2 perch vanno considerate le posizioni di apertura e chiusura
						size_risultato_parziale = (int *) calloc(3, sizeof(int));

						if (size_risultato_parziale == NULL) {

							fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for size_risultato_parziale </xcde:error>\n\n") ;

							fflush(stderr) ;

							return(-1) ;
						}

						res = get_result_from_position(node->childs_vector[0], node->tag_positions[i], node->tag_positions[i+1], index_variabile, &(risultato_parziale[j]), &(size_risultato_parziale[j]), input_hash_table);

						if (res == -1)
							return (-1);

						res = get_result_from_position(node->childs_vector[1], node->tag_positions[i], node->tag_positions[i+1], index_variabile, &(risultato_parziale[((j + 1) % 3)]), &(size_risultato_parziale[((j + 1) % 3)]), input_hash_table);

						if (res == -1)
							return (-1);

						if (strcmp(node->node_name, "xml_or") == 0) {

							res = fondi_risultati_or(&(risultato_parziale[j]), &(size_risultato_parziale[j]), &(risultato_parziale[((j + 1) % 3)]), &(size_risultato_parziale[((j + 1) % 3)]), &(risultato_parziale[((j + 2) % 3)]), &(size_risultato_parziale[((j + 2) % 3)]), numero_variabili);

							if (res == -1)
								return (-1);

						} else {

							res = fondi_risultati(&(risultato_parziale[j]), &(size_risultato_parziale[j]), &(risultato_parziale[((j + 1) % 3)]), &(size_risultato_parziale[((j + 1) % 3)]), &(risultato_parziale[((j + 2) % 3)]), &(size_risultato_parziale[((j + 2) % 3)]), numero_variabili);

							if (res == -1)
								return (-1);

						}

						j = ((j + 2) % 3);


						for (k=2; k<node->num_childs; k++) {

							res = get_result_from_position(node->childs_vector[k], node->tag_positions[i], node->tag_positions[i+1], index_variabile, &(risultato_parziale[((j + 1) % 3)]), &(size_risultato_parziale[((j + 1) % 3)]), input_hash_table);

							if (res == -1)
								return (-1);

							if (strcmp(node->node_name, "xml_or") == 0) {

								res = fondi_risultati_or(&(risultato_parziale[j]), &(size_risultato_parziale[j]), &(risultato_parziale[((j + 1) % 3)]), &(size_risultato_parziale[((j + 1) % 3)]), &(risultato_parziale[((j + 2) % 3)]), &(size_risultato_parziale[((j + 2) % 3)]), numero_variabili);

								if (res == -1)
									return (-1);

							} else {

								res = fondi_risultati(&(risultato_parziale[j]), &(size_risultato_parziale[j]), &(risultato_parziale[((j + 1) % 3)]), &(size_risultato_parziale[((j + 1) % 3)]), &(risultato_parziale[((j + 2) % 3)]), &(size_risultato_parziale[((j + 2) % 3)]), numero_variabili);

								if (res == -1)
									return (-1);

							}

							j = ((j + 2) % 3);

						}

						res = aggiungi_a_risultato(&(risultato_parziale[j]), &(size_risultato_parziale[j]), risultato, size_vettore_risultato );

						if (res == -1)
							return (-1);

						if (risultato_parziale)
							free(risultato_parziale);

						if (size_risultato_parziale)
							free(size_risultato_parziale);

					}

				} else {

					if (res > 0) { // la posizione del figlio  dopo quella del padre

						// essendo le posizioni del figlio in ordine, se la posizione corrente  dopo quella del padre
						//	si pu uscire dal for perch anche tutte quelle successive non saranno buone
						i = node->num_positions;
					}
				}
			}
		}


		if (index_padre != -1 ) {

			for (i = 0; i < *size_vettore_risultato; i++) {

				// inserisco la posizione del padre nel vettore risultato
				//	nella posizione corrispondente alla variabile associata al node padre
				(*risultato)[i][2 * index_padre].XCDE_Position = posizione_apertura_padre.XCDE_Position;
				(*risultato)[i][2 * index_padre].XCDE_Offset = posizione_apertura_padre.XCDE_Offset;

				(*risultato)[i][2 * index_padre + 1].XCDE_Position = posizione_chiusura_padre.XCDE_Position;
				(*risultato)[i][2 * index_padre + 1].XCDE_Offset = posizione_chiusura_padre.XCDE_Offset;
			}
		}

	}


	return 1;
}



// prints the header of results
int print_header_return(int number)
{
	
	fprintf(stdout, "<");
	fprintf(stdout, XCDE_RETURN_RESULT_ELEMENT);
	fprintf(stdout, " result_number = '%d'>\n", number + 1);

	fflush (stdout);

	return(1) ;
}

// prints the foot of results
int print_footer_return(void)
{

	fprintf(stdout, "</");
	fprintf(stdout, XCDE_RETURN_RESULT_ELEMENT);
	fprintf(stdout, ">\n\n");

	fflush(stdout);

	return(1) ;
}



// prints the header of results
int print_header(const char *filename, int number)
{
	fprintf(stdout, "<");
	fprintf(stdout, XCDE_FROM_RESULT_ELEMENT);
	fprintf(stdout, ">\n");

	fprintf(stdout, "<xcde:filename>%s</xcde:filename>\n\n", filename) ;

	fflush (stdout);

	return(1) ;
}

// prints the foot of results
int print_footer(void)
{

	fprintf(stdout, "</");
	fprintf(stdout, XCDE_FROM_RESULT_ELEMENT);
	fprintf(stdout, ">\n\n");

	fflush(stdout);

	return(1) ;
}



int estrai_text_element(char *string) {

	int i;
	int j = -1;
	int string_len;

	unsigned char is_printable = 1;
	unsigned char discover_phase = 1;

	
	
	string_len = strlen(string);

	for (i = 0; i < string_len; i++) {

		if (string[i] == '>') {

			j = i+1;
			is_printable = 1;
			i = string_len;

		} else {

			if (string[i] == '<') {

				j = 0;
				i = string_len;
			}
		}

	}

	if (j == -1) {

		is_printable = 1;
		j = 0;
	}

	for (; j < string_len; j++) {

		if (string[j] == '>') {

			is_printable = 1;

		} else {

			if (string[j] == '<') {

				is_printable = 0;

			} else {

				if (is_printable) {

					fprintf(stdout, "%c", string[j]);

				}
			}
		}
	}

	return(1);

}



int estrai_contesto_backward_stop_parent(int posizione_iniziale, int ampiezza, unsigned char flag_tag, char **backward_string, char ***pila_tag, int *size_pila_tag, int *dimension_pila_tag, XCDE_Console_Type *XCDE_Console) {

	char **backward_window=NULL ;

	int backward_window_dimension=0 ;

	int backward_window_pointer=0 ;

	int backward_body_position=0 ;

	int window_counter = 0;

	int codeword=0 ;

	int codeword_len=0 ;

	char *token=0 ;

	int function_result=0 ;

	int backward_len = 0;

	int  k, i, indice;

	unsigned char is_tag_complete = 1;



	char tmp_string[256] = "/";


	/* ------------------------ */
	/* initializes data structs */
	/* ------------------------ */

	backward_window_dimension = XCDE_F_B_INITIAL_WINDOW_DIMENSION ;

	backward_window = (char **) malloc(sizeof(char *) * backward_window_dimension) ;

	if (backward_window == NULL){

		fprintf(stderr, "\n\t<xcde:error> problems with memory allocation for backward_window </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-1) ;
	}

	backward_window_pointer = 0 ;

	window_counter=0 ;

	backward_body_position = posizione_iniziale;

	/* ------------------------------ */
	/* reads window backward codewords */
	/* ------------------------------ */

	while ( ((window_counter <= ampiezza) || (!is_tag_complete)) && (backward_body_position > 0)){

		function_result = XCDE_Get_Body_Backward_Codeword(backward_body_position, &codeword, &codeword_len, XCDE_Console);

		if (function_result == -1) {

			for(k=0 ; k < backward_window_pointer; ++k)

				free(backward_window[k]) ;

			free(backward_window) ;

			return(-1) ;
		}

		function_result = XCDE_Codeword2Token(codeword, codeword_len, &token, XCDE_Console) ;

		if (function_result == -1) {

			for(k=0 ; k < backward_window_pointer; ++k)

				free(backward_window[k]) ;

			free(backward_window) ;

			return(-1) ;
		}

		backward_body_position -= codeword_len ;

		backward_len += strlen(token) + 1 ;

		fflush(stdout) ;

		backward_window[backward_window_pointer] = token ;

		++backward_window_pointer ;

		if (backward_window_pointer > backward_window_dimension) {

			backward_window_dimension *= 2 ;

			backward_window = (char **) realloc(backward_window, backward_window_dimension) ;
		}

		++window_counter ;

		if (strcmp(token, ">") == 0)
			is_tag_complete = 0;

		if (strcmp(token, "<") == 0) {

			is_tag_complete = 1;

			if ( strcmp(backward_window[backward_window_pointer-2], "/") == 0 ) {

				indice = backward_window_pointer - 3;

				while( is_separator(backward_window[indice]) ) {

					indice--;

				}

				(*pila_tag)[*size_pila_tag] = (char *) malloc ( (strlen(backward_window[indice]) + 2) * sizeof(char) );

				if ((*pila_tag)[*size_pila_tag] == NULL){

					fprintf(stderr, "\n\t<xcde:error> problems with memory allocation for (*pila_tag)[*size_pila_tag] </xcde:error>\n\n") ;

					fflush(stderr) ;

					return(-1) ;
				}

				strcpy( (*pila_tag)[*size_pila_tag], "/");

				strcat((*pila_tag)[*size_pila_tag],backward_window[indice]);

				(*size_pila_tag)++;

				if (*size_pila_tag >= *dimension_pila_tag) {

					(*dimension_pila_tag) *= 2;

					*pila_tag = (char **) realloc (*pila_tag, sizeof(char *) * (*dimension_pila_tag)) ;

					if (*pila_tag == NULL){

						fprintf(stderr, "\n\t<xcde:error> problems with memory reallocation for *pila_tag </xcde:error>\n\n") ;

						fflush(stderr) ;

						return(-1) ;
					}
				}

			}  else { // si sta leggendo un tag di apertura

				indice = backward_window_pointer - 2;

				while( is_separator(backward_window[indice]) ) {

					indice--;

				}

				if (*size_pila_tag == 0) {

					(*pila_tag)[*size_pila_tag] = (char *) malloc ( (strlen(backward_window[indice]) + 2) * sizeof(char) );

					if ((*pila_tag)[*size_pila_tag] == NULL){

						fprintf(stderr, "\n\t<xcde:error> problems with memory allocation for (*pila_tag)[*size_pila_tag] </xcde:error>\n\n") ;

						fflush(stderr) ;

						return(-1) ;
					}

					strcpy((*pila_tag)[*size_pila_tag],backward_window[indice]);

					(*size_pila_tag)++;

					if (*size_pila_tag >= *dimension_pila_tag) {

						(*dimension_pila_tag) *= 2;

						*pila_tag = (char **) realloc (*pila_tag, sizeof(char *) * (*dimension_pila_tag)) ;

						if (*pila_tag == NULL){

							fprintf(stderr, "\n\t<xcde:error> problems with memory reallocation for *pila_tag </xcde:error>\n\n") ;

							fflush(stderr) ;

							return(-1) ;
						}
					}


					// si  arrivati a leggere il tag del padre
					break;

				} else {

					strcat(tmp_string, backward_window[indice]);

					if ( strcmp((*pila_tag)[*size_pila_tag - 1], tmp_string) == 0 ) {

						free((*pila_tag)[*size_pila_tag - 1]);

						(*size_pila_tag)--;

					}

					strcpy(tmp_string, "/");
				}
			}
		}
	}

	*backward_string = (char *) malloc ( (backward_len + 1) * sizeof(char) );

	if (*backward_string == NULL){

		return(-1) ;
	}

	strcpy(*backward_string, "");

	for (i = (backward_window_pointer - 1); i >= 0 ; i--) {

		strcat(*backward_string, backward_window[i]);
	}


	return 1;



}




int estrai_contesto_backward(int posizione_iniziale, int ampiezza, unsigned char flag_tag, char **backward_string, char ***pila_tag, int *size_pila_tag, int *dimension_pila_tag, XCDE_Console_Type *XCDE_Console) {

	char **backward_window=NULL ;

	int backward_window_dimension=0 ;

	int backward_window_pointer=0 ;

	int backward_body_position=0 ;

	int window_counter = 0;

	int codeword=0 ;

	int codeword_len=0 ;

	char *token=0 ;

	int function_result=0 ;

	int backward_len = 0;

	int  k, i, indice;

	unsigned char is_tag_complete = 1;



	char tmp_string[256] = "/";


	/* ------------------------ */
	/* initializes data structs */
	/* ------------------------ */

	backward_window_dimension = XCDE_F_B_INITIAL_WINDOW_DIMENSION ;

	backward_window = (char **) malloc(sizeof(char *) * backward_window_dimension) ;

	if (backward_window == NULL){

		fprintf(stderr, "\n\t<xcde:error> problems with memory allocation for backward_window </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-1) ;
	}

	backward_window_pointer = 0 ;

	window_counter=0 ;

	backward_body_position = posizione_iniziale;

	/* ------------------------------ */
	/* reads window backward codewords */
	/* ------------------------------ */

	while ( ((window_counter <= ampiezza) || (!is_tag_complete)) && (backward_body_position > 0)){

		function_result = XCDE_Get_Body_Backward_Codeword(backward_body_position, &codeword, &codeword_len, XCDE_Console);

		if (function_result == -1) {

			for(k=0 ; k < backward_window_pointer; ++k)

				free(backward_window[k]) ;

			free(backward_window) ;

			return(-1) ;
		}

		function_result = XCDE_Codeword2Token(codeword, codeword_len, &token, XCDE_Console) ;

		if (function_result == -1) {

			for(k=0 ; k < backward_window_pointer; ++k)

				free(backward_window[k]) ;

			free(backward_window) ;

			return(-1) ;
		}

		backward_body_position -= codeword_len ;

		backward_len += strlen(token) + 1 ;

		fflush(stdout) ;

		backward_window[backward_window_pointer] = token ;

		++backward_window_pointer ;

		if (backward_window_pointer > backward_window_dimension) {

			backward_window_dimension *= 2 ;

			backward_window = (char **) realloc(backward_window, backward_window_dimension) ;
		}

		++window_counter ;

		if (strcmp(token, ">") == 0)
			is_tag_complete = 0;

		if (strcmp(token, "<") == 0) {

			is_tag_complete = 1;

			if ( strcmp(backward_window[backward_window_pointer-2], "/") == 0 ) {

				indice = backward_window_pointer - 3;

				while( is_separator(backward_window[indice]) ) {

					indice--;

				}

				(*pila_tag)[*size_pila_tag] = (char *) malloc ( (strlen(backward_window[indice]) + 2) * sizeof(char) );

				if ((*pila_tag)[*size_pila_tag] == NULL){

					fprintf(stderr, "\n\t<xcde:error> problems with memory allocation for (*pila_tag)[*size_pila_tag] </xcde:error>\n\n") ;

					fflush(stderr) ;

					return(-1) ;
				}

				strcpy( (*pila_tag)[*size_pila_tag], "/");

				strcat((*pila_tag)[*size_pila_tag],backward_window[indice]);

				(*size_pila_tag)++;

				if (*size_pila_tag >= *dimension_pila_tag) {

					(*dimension_pila_tag) *= 2;

					*pila_tag = (char **) realloc (*pila_tag, sizeof(char *) * (*dimension_pila_tag)) ;

					if (*pila_tag == NULL){

						fprintf(stderr, "\n\t<xcde:error> problems with memory reallocation for *pila_tag </xcde:error>\n\n") ;

						fflush(stderr) ;

						return(-1) ;
					}
				}

			}  else {

				indice = backward_window_pointer - 2;

				while( is_separator(backward_window[indice]) ) {

					indice--;

				}

				if (*size_pila_tag == 0) {

					(*pila_tag)[*size_pila_tag] = (char *) malloc ( (strlen(backward_window[indice]) + 2) * sizeof(char) );

					if ((*pila_tag)[*size_pila_tag] == NULL){

						fprintf(stderr, "\n\t<xcde:error> problems with memory allocation for (*pila_tag)[*size_pila_tag] </xcde:error>\n\n") ;

						fflush(stderr) ;

						return(-1) ;
					}

					strcpy((*pila_tag)[*size_pila_tag],backward_window[indice]);

					(*size_pila_tag)++;

					if (*size_pila_tag >= *dimension_pila_tag) {

						(*dimension_pila_tag) *= 2;

						*pila_tag = (char **) realloc (*pila_tag, sizeof(char *) * (*dimension_pila_tag)) ;

						if (*pila_tag == NULL){

							fprintf(stderr, "\n\t<xcde:error> problems with memory reallocation for *pila_tag </xcde:error>\n\n") ;

							fflush(stderr) ;

							return(-1) ;
						}
					}

				} else {

					strcat(tmp_string, backward_window[indice]);

					if ( strcmp((*pila_tag)[*size_pila_tag - 1], tmp_string) == 0 ) {

						free((*pila_tag)[*size_pila_tag - 1]);

						(*size_pila_tag)--;

					} else {

						(*pila_tag)[*size_pila_tag] = (char *) malloc ( (strlen(backward_window[indice]) + 2) * sizeof(char) );

						if ((*pila_tag)[*size_pila_tag] == NULL){

							fprintf(stderr, "\n\t<xcde:error> problems with memory allocation for (*pila_tag)[*size_pila_tag] </xcde:error>\n\n") ;

							fflush(stderr) ;

							return(-1) ;
						}

						strcpy((*pila_tag)[*size_pila_tag],backward_window[indice]);

						(*size_pila_tag)++;

						if (*size_pila_tag >= *dimension_pila_tag) {

							(*dimension_pila_tag) *= 2;

							*pila_tag = (char **) realloc (*pila_tag, sizeof(char *) * (*dimension_pila_tag)) ;

							if (*pila_tag == NULL){

								fprintf(stderr, "\n\t<xcde:error> problems with memory reallocation for *pila_tag </xcde:error>\n\n") ;

								fflush(stderr) ;

								return(-1) ;
							}
						}
					}

					strcpy(tmp_string, "/");
				}
			}
		}
	}

	*backward_string = (char *) malloc ( (backward_len + 1) * sizeof(char) );

	if (*backward_string == NULL){

		return(-1) ;
	}

	strcpy(*backward_string, "");

	for (i = (backward_window_pointer - 1); i >= 0 ; i--) {

		strcat(*backward_string, backward_window[i]);
	}


	return 1;



}




int estrai_contesto_forward(int posizione_iniziale, int ampiezza, char **forward_string, char ***pila_tag, int *size_pila_tag, int *dimension_pila_tag, XCDE_Console_Type *XCDE_Console) {

	int forward_window_pointer=0 ;

	int forward_body_position ;

	int window_counter = 0;

	int codeword=0 ;

	int codeword_len=0 ;

	char *token=0 ;

	int function_result=0 ;

	int dimension_forward_string = INIT_FORWARD_STRING_SIZE;

	unsigned char is_tag_complete = 1;

	unsigned char res = 0;
	
	int i;

	

	// la variabile 'stato' pu assumere i seguenti valori:
	//
	//	0: non siamo in un tag;
	//	1: se l'ultimo token letto  '<'
	//	2: se gli ultimi token letti sono '</'
	//	3: se il token dopo il '<' non  '/'
	//	4: si deve leggere il nome del tag
	//
	unsigned char stato = 0;

	unsigned char is_open_tag = 0;

	unsigned char is_close_tag = 0;


	/* ------------------------ */
	/* initializes data structs */
	/* ------------------------ */




	reverse_stack(*pila_tag, *size_pila_tag);





	forward_body_position = posizione_iniziale;

	forward_window_pointer = 0 ;

	window_counter = 0 ;

	/* ------------------------------ */
	/* reads window forkward codewords */
	/* ------------------------------ */

	*forward_string = (char *) malloc ( dimension_forward_string * sizeof(char) );

	if (*forward_string == NULL){

		fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for *forward_string </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-1) ;
	}

	strcpy(*forward_string, "");

	while ( ((window_counter <= ampiezza) || (!is_tag_complete)) && (forward_body_position < ((XCDE_Console)->XCDE_Body_Size_Without_Zlib))) {

		function_result = XCDE_Get_Body_Forward_Codeword(forward_body_position, &codeword, &codeword_len, XCDE_Console) ;

		if (function_result == -1) {

			fprintf(stderr, "\n\t<xcde:error> problems with the XCDE_Get_Body_Forward_Codeword function </xcde:error>\n\n") ;

			fflush(stderr) ;

			return(-1) ;
		}

		function_result = XCDE_Codeword2Token(codeword, codeword_len, &token, XCDE_Console) ;

		if (function_result == -1) {

			fprintf(stderr, "\n\t<xcde:error> problems with the XCDE_Codeword2Token function </xcde:error>\n\n") ;

			fflush(stderr) ;

			return(-1) ;
		}

		forward_body_position += codeword_len ;

		if ( (strlen(*forward_string) + (strlen(token) + 1) > dimension_forward_string ) ) {

			dimension_forward_string *= 2;

			*forward_string = (char *) realloc (*forward_string, dimension_forward_string * sizeof(char) );

			if (forward_string == NULL){

				fprintf(stderr, "\n\n\t<xcde:error> problems during memory allocation for forward_string </xcde:error>\n\n") ;

				fflush(stderr) ;

				return(-1) ;
			}
		}

		strcat(*forward_string, token);

		++window_counter;

		if ( (stato == 0) && (strcmp(token, "<") == 0) ) {

			is_tag_complete = 0;
			stato = 1;
			continue;

		}

		if ( (stato == 1) && (strcmp(token, "/") == 0) ) {

			stato = 2;
			continue;
		}

		if ( (stato == 1) && (strcmp(token, "/") != 0) && (is_separator(token)) ) {

			stato = 3;
			continue;
		}

		// se si  nello stato 3 mi aspetto o caratteri separatori o un tag di apertura
		if ( (stato == 3) || ((stato == 1) && (strcmp(token, "/") != 0) && (!(res = is_separator(token)))) )  {

			if (!res) {

				(*pila_tag)[*size_pila_tag] = token;

				token = NULL;

				(*size_pila_tag)++;

				if (*size_pila_tag >= *dimension_pila_tag) {

					(*dimension_pila_tag) *= 2;

					*pila_tag = (char **) realloc (*pila_tag, sizeof(char *) * (*dimension_pila_tag)) ;

					if (*pila_tag == NULL){

						fprintf(stderr, "\n\t<xcde:error> problems with memory reallocation for *pila_tag </xcde:error>\n\n") ;

						fflush(stderr) ;

						return(-1) ;
					}
				}

				stato = 4;
				continue;

			}
		}


		// se si  nello stato 2 mi aspetto o caratteri separatori o un tag di chiusura
		if (stato == 2) {

			if (!is_separator(token)) {

				if (*size_pila_tag == 0) {

					(*pila_tag)[*size_pila_tag] = (char *) malloc ( (strlen(token) + 2) * sizeof(char) );

					if ((*pila_tag)[*size_pila_tag] == NULL){

						fprintf(stderr, "\n\t<xcde:error> problems with memory allocation for (*pila_tag)[*size_pila_tag] </xcde:error>\n\n") ;

						fflush(stderr) ;

						return(-1) ;
					}

					strcpy((*pila_tag)[*size_pila_tag],"/");

					strcat((*pila_tag)[*size_pila_tag],token);

					(*size_pila_tag)++;

					if (*size_pila_tag >= *dimension_pila_tag) {

						(*dimension_pila_tag) *= 2;

						*pila_tag = (char **) realloc (*pila_tag, sizeof(char *) * (*dimension_pila_tag)) ;

						if (*pila_tag == NULL){

							fprintf(stderr, "\n\t<xcde:error> problems with memory reallocation for *pila_tag </xcde:error>\n\n") ;

							fflush(stderr) ;

							return(-1) ;
						}
					}

				} else {

					if ( strcmp((*pila_tag)[*size_pila_tag - 1], token) == 0 ) {

						free((*pila_tag)[*size_pila_tag - 1]);

						(*size_pila_tag)--;

					} else {

						(*pila_tag)[*size_pila_tag] = (char *) malloc ( (strlen(token) + 2) * sizeof(char) );

						if ((*pila_tag)[*size_pila_tag] == NULL){

							fprintf(stderr, "\n\t<xcde:error> problems with memory allocation for (*pila_tag)[*size_pila_tag] </xcde:error>\n\n") ;

							fflush(stderr) ;

							return(-1) ;
						}

						strcpy((*pila_tag)[*size_pila_tag],"/");

						strcat((*pila_tag)[*size_pila_tag],token);

						(*size_pila_tag)++;

						if (*size_pila_tag >= *dimension_pila_tag) {

							(*dimension_pila_tag) *= 2;

							*pila_tag = (char **) realloc (*pila_tag, sizeof(char *) * (*dimension_pila_tag)) ;

							if (*pila_tag == NULL){

								fprintf(stderr, "\n\t<xcde:error> problems with memory reallocation for *pila_tag </xcde:error>\n\n") ;

								fflush(stderr) ;

								return(-1) ;
							}
						}

					}
				}

				stato = 4;
				continue;
			}

		}

		if ( (stato == 4) && (strcmp(token, ">") == 0) ) {
			is_tag_complete = 1;
			stato = 0;
			continue;
		}
	}
}



int estrai_contesto_forward_stop_parent(int posizione_iniziale, int ampiezza, char **forward_string, char ***pila_tag, int *size_pila_tag, int *dimension_pila_tag, XCDE_Console_Type *XCDE_Console) {

	int forward_window_pointer=0 ;

	int forward_body_position ;

	int window_counter = 0;

	int codeword=0 ;

	int codeword_len=0 ;

	char *token=0 ;

	int function_result=0 ;

	int dimension_forward_string = INIT_FORWARD_STRING_SIZE;

	unsigned char is_tag_complete = 1;

	unsigned char res = 0;

	int i;

	char **temp_pila_tag = NULL;



	// la variabile 'stato' pu assumere i seguenti valori:
	//
	//	0: non siamo in un tag;
	//	1: se l'ultimo token letto  '<'
	//	2: se gli ultimi token letti sono '</'
	//	3: se il token dopo il '<' non  '/'
	//	4: si deve leggere il nome del tag
	//
	unsigned char stato = 0;

	unsigned char is_open_tag = 0;

	unsigned char is_close_tag = 0;


	/* ------------------------ */
	/* initializes data structs */
	/* ------------------------ */



	if ( *size_pila_tag != 0 )
		reverse_stack(*pila_tag, *size_pila_tag);


	if (*size_pila_tag == 1) {

		temp_pila_tag = (*pila_tag);
		(*pila_tag) = (*pila_tag)++;
		(*size_pila_tag)--;
		(*dimension_pila_tag)--;
	}



	forward_body_position = posizione_iniziale;

	forward_window_pointer = 0 ;

	window_counter = 0 ;

	/* ------------------------------ */
	/* reads window forkward codewords */
	/* ------------------------------ */

	*forward_string = (char *) malloc ( dimension_forward_string * sizeof(char) );

	if (*forward_string == NULL){

		fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for *forward_string </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-1) ;
	}

	strcpy(*forward_string, "");

	while ( ((window_counter <= ampiezza) || (!is_tag_complete)) && (forward_body_position < ((XCDE_Console)->XCDE_Body_Size_Without_Zlib))) {

		function_result = XCDE_Get_Body_Forward_Codeword(forward_body_position, &codeword, &codeword_len, XCDE_Console) ;

		if (function_result == -1) {

			fprintf(stderr, "\n\t<xcde:error> problems with the XCDE_Get_Body_Forward_Codeword function </xcde:error>\n\n") ;

			fflush(stderr) ;

			return(-1) ;
		}

		function_result = XCDE_Codeword2Token(codeword, codeword_len, &token, XCDE_Console) ;

		if (function_result == -1) {

			fprintf(stderr, "\n\t<xcde:error> problems with the XCDE_Codeword2Token function </xcde:error>\n\n") ;

			fflush(stderr) ;

			return(-1) ;
		}

		forward_body_position += codeword_len ;

		if ( (strlen(*forward_string) + (strlen(token) + 1) > dimension_forward_string ) ) {

			dimension_forward_string *= 2;

			*forward_string = (char *) realloc (*forward_string, dimension_forward_string * sizeof(char) );

			if (forward_string == NULL){

				fprintf(stderr, "\n\n\t<xcde:error> problems during memory allocation for forward_string </xcde:error>\n\n") ;

				fflush(stderr) ;

				return(-1) ;
			}
		}

		strcat(*forward_string, token);

		++window_counter;

		if ( (stato == 0) && (strcmp(token, "<") == 0) ) {

			is_tag_complete = 0;
			stato = 1;
			continue;

		}

		if ( (stato == 1) && (strcmp(token, "/") == 0) ) {

			stato = 2;
			continue;
		}

		if ( (stato == 1) && (strcmp(token, "/") != 0) && (is_separator(token)) ) {

			stato = 3;
			continue;
		}

		// se si  nello stato 3 mi aspetto o caratteri separatori o un tag di apertura
		if ( (stato == 3) || ((stato == 1) && (strcmp(token, "/") != 0) && (!(res = is_separator(token)))) )  {

			if (!res) {

				(*pila_tag)[*size_pila_tag] = token;

				token = NULL;

				(*size_pila_tag)++;

				if (*size_pila_tag >= *dimension_pila_tag) {

					(*dimension_pila_tag) *= 2;

					*pila_tag = (char **) realloc (*pila_tag, sizeof(char *) * (*dimension_pila_tag)) ;

					if (*pila_tag == NULL){

						fprintf(stderr, "\n\t<xcde:error> problems with memory reallocation for *pila_tag </xcde:error>\n\n") ;

						fflush(stderr) ;

						return(-1) ;
					}
				}

				stato = 4;
				continue;

			}
		}


		// se si  nello stato 2 mi aspetto o caratteri separatori o un tag di chiusura
		if (stato == 2) {

			if (!is_separator(token)) {

				if (*size_pila_tag == 0) {

					(*pila_tag)[*size_pila_tag] = (char *) malloc ( (strlen(token) + 2) * sizeof(char) );

					if ((*pila_tag)[*size_pila_tag] == NULL){

						fprintf(stderr, "\n\t<xcde:error> problems with memory allocation for (*pila_tag)[*size_pila_tag] </xcde:error>\n\n") ;

						fflush(stderr) ;

						return(-1) ;
					}

					strcpy((*pila_tag)[*size_pila_tag],"/");

					strcat((*pila_tag)[*size_pila_tag],token);

					(*size_pila_tag)++;

					if (*size_pila_tag >= *dimension_pila_tag) {

						(*dimension_pila_tag) *= 2;

						*pila_tag = (char **) realloc (*pila_tag, sizeof(char *) * (*dimension_pila_tag)) ;

						if (*pila_tag == NULL){

							fprintf(stderr, "\n\t<xcde:error> problems with memory reallocation for *pila_tag </xcde:error>\n\n") ;

							fflush(stderr) ;

							return(-1) ;
						}
					}

					// si  arrivati a leggere la chiusura del tag padre
					window_counter = ampiezza + 1;

				} else {

					if ( strcmp((*pila_tag)[*size_pila_tag - 1], token) == 0 ) {

						free((*pila_tag)[*size_pila_tag - 1]);

						(*size_pila_tag)--;

					} else {

						(*pila_tag)[*size_pila_tag] = (char *) malloc ( (strlen(token) + 2) * sizeof(char) );

						if ((*pila_tag)[*size_pila_tag] == NULL){

							fprintf(stderr, "\n\t<xcde:error> problems with memory allocation for (*pila_tag)[*size_pila_tag] </xcde:error>\n\n") ;

							fflush(stderr) ;

							return(-1) ;
						}

						strcpy((*pila_tag)[*size_pila_tag],"/");

						strcat((*pila_tag)[*size_pila_tag],token);

						(*size_pila_tag)++;

						if (*size_pila_tag >= *dimension_pila_tag) {

							(*dimension_pila_tag) *= 2;

							*pila_tag = (char **) realloc (*pila_tag, sizeof(char *) * (*dimension_pila_tag)) ;

							if (*pila_tag == NULL){

								fprintf(stderr, "\n\t<xcde:error> problems with memory reallocation for *pila_tag </xcde:error>\n\n") ;

								fflush(stderr) ;

								return(-1) ;
							}
						}

						// si  arrivati a leggere la chiusura del tag padre
						window_counter = ampiezza + 1;

					}

				}

				stato = 4;
				continue;
			}

		}

		if ( (stato == 4) && (strcmp(token, ">") == 0) ) {
			is_tag_complete = 1;
			stato = 0;
			continue;
		}
	}

	if ( (*size_pila_tag == 0) && (temp_pila_tag != NULL) ){

		(*pila_tag) = temp_pila_tag;
		(*size_pila_tag)++;
		temp_pila_tag = NULL;
		(*dimension_pila_tag)++;

		return 1;

	}

	if ( (*size_pila_tag == 1) && (temp_pila_tag != NULL) ) {

	 	if ( strcmp(temp_pila_tag[0], ((*pila_tag)[0]) + 1) == 0 ) {

			free(temp_pila_tag[0]);
			free((*pila_tag)[0]);
			temp_pila_tag[0] = NULL;
			(*pila_tag)[0] = NULL;
			(*pila_tag) = temp_pila_tag;
			(*size_pila_tag) = 0;
			temp_pila_tag = NULL;
			(*dimension_pila_tag)++;

		} else {

			(*pila_tag) = temp_pila_tag;
			(*size_pila_tag)++;
			temp_pila_tag = NULL;
			(*dimension_pila_tag)++;

		}

		return 1;

	}

	if ( (*size_pila_tag > 1) && (temp_pila_tag != NULL) ) {

		(*pila_tag) = temp_pila_tag;
		(*size_pila_tag)++;
		temp_pila_tag = NULL;
		(*dimension_pila_tag)++;

		return 1;
	}
	
	return 1;


}



int estrai_contesto(query_tree_node *node, char *nomeFile, int ampiezza, unsigned char flag_tag, unsigned char direction, XCDE_IL_Couple_Type **risultato, int size_vettore_risultato, HHash_table *input_hash_table, int riga_da_visualizzare, XCDE_Console_Type *XCDE_Console) {

	char *sys_comand = NULL;

	char *snippet_estratto = NULL;

	int forward_body_position=0 ;

	char **pila_tag = NULL;

	int size_pila_tag = 0;

	int dimension_pila_tag = 0;

	int codeword=0 ;

	int codeword_len=0 ;

	char *token=0 ;

	int function_result=0 ;

	int posizione_iniziale;

	int posizione_finale;

	char * variabile = NULL;
	int index_variabile;
	Hash_node * node_hash;

	int i, k;

	char *snippet_risultato = NULL;

	char *forward_string = NULL;
	char *backward_string = NULL;



	variabile = node->node_name;

	if (variabile != NULL) {

		node_hash = HHashtable_search(variabile, strlen(variabile), input_hash_table);

		if (node_hash != NULL)

			index_variabile = node_hash->index;
		else
			index_variabile = -1;

	} else {

		index_variabile = -1;
	}

	if (index_variabile != -1) {

		posizione_iniziale = risultato[riga_da_visualizzare][2 * index_variabile].XCDE_Position;

		posizione_finale = risultato[riga_da_visualizzare][2 * index_variabile + 1].XCDE_Position;


		// Si controlla se alla variabile sono associate delle posizioni (in caso di query in OR alcune variabili
		//	potrebbero non avere alcune posizioni
		if ( !((posizione_iniziale == 0) && (risultato[riga_da_visualizzare][2 * index_variabile].XCDE_Offset == 0)) ) {


			if (flag_tag == 2) {

				//richiamo la view

				sys_comand = (char *) malloc ((strlen(nomeFile) + strlen("xcde_view ") + 5 + 30) * sizeof(char));

				if (sys_comand == NULL) {

					fprintf(stderr, "\n\t<xcde:error> problems with memory allocation for sys_comand </xcde:error>\n\n") ;

					fflush(stderr) ;

					return(-1) ;
				}

				// build comand that will passed to the system function
				sprintf(sys_comand, "xcde_view %s %d %d %d", nomeFile, risultato[riga_da_visualizzare][2 * index_variabile].XCDE_Position, risultato[riga_da_visualizzare][2 * index_variabile + 1].XCDE_Position, ampiezza) ;

				system(sys_comand);

				// releases the memory for the system comand
				if (sys_comand != NULL) {

					free(sys_comand) ;

					sys_comand=NULL ;
				}

			} else {


				// allocazione della memoria necessaria per la pila che servir per
				//	estrarre lo snippet ben formato.
				size_pila_tag = 0;

				dimension_pila_tag = XCDE_INITIAL_STACK_TAG_DIMENSION ;

				pila_tag = (char **) malloc(sizeof(char *) * dimension_pila_tag) ;

				if (pila_tag == NULL){

					fprintf(stderr, "\n\t<xcde:error> problems with memory allocation for pila_tag </xcde:error>\n\n") ;

					fflush(stderr) ;

					return(-1) ;
				}

				// se settato l'attributo xml_view a 'backward' o 'bidir' va estratto il backward context.
				if ( ((direction == 0) || (direction == 1)) && (ampiezza != 0) ) {

					if (flag_tag == 3) {

						res = estrai_contesto_backward_stop_parent(posizione_iniziale, ampiezza, flag_tag, &backward_string, &pila_tag, &size_pila_tag, &dimension_pila_tag, XCDE_Console);

					} else {

						res = estrai_contesto_backward(posizione_iniziale, ampiezza, flag_tag, &backward_string, &pila_tag, &size_pila_tag, &dimension_pila_tag, XCDE_Console);

					}

					if (res == -1)
						return (-1);

				}


				// si estrae lo snippet associato alla variabile

				// se la variabile da visualizzare non  un TAG si effettua
				if ( (risultato[riga_da_visualizzare][2 * index_variabile].XCDE_Position == risultato[riga_da_visualizzare][2 * index_variabile + 1].XCDE_Position)
				&& (risultato[riga_da_visualizzare][2 * index_variabile].XCDE_Offset == risultato[riga_da_visualizzare][2 * index_variabile + 1].XCDE_Offset) ) {

					res = extract_word(&snippet_estratto, &forward_body_position, risultato[riga_da_visualizzare][2 * index_variabile].XCDE_Position, XCDE_Console);

					if (res == -1)
						return(-1) ;

				} else {

					res = extract_snippet(&snippet_estratto, risultato[riga_da_visualizzare][2 * index_variabile], risultato[riga_da_visualizzare][2 * index_variabile + 1], XCDE_Console);

					if (res == -1)
						return(-1) ;

					// la codeword relativa alla posizione di chiusura  gi stata visualizzata dalla extract_snippet, quindi si legge ma non si visualizza
					function_result = XCDE_Get_Body_Forward_Codeword(posizione_finale, &codeword, &codeword_len, XCDE_Console) ;

					if (function_result == -1) {

						fprintf(stderr, "\n\t<xcde:error> problems with the XCDE_Get_Body_Forward_Codeword function </xcde:error>\n\n") ;

						fflush(stderr) ;

						return(-1) ;
					}

					forward_body_position = posizione_finale + codeword_len ;

				}


				if ( ((direction == 1) || (direction == 2)) && (ampiezza != 0) ) {

					if (flag_tag == 3) {

						res = estrai_contesto_forward_stop_parent(forward_body_position, ampiezza, &forward_string, &pila_tag, &size_pila_tag, &dimension_pila_tag, XCDE_Console);

					} else {

						res = estrai_contesto_forward(forward_body_position, ampiezza, &forward_string, &pila_tag, &size_pila_tag, &dimension_pila_tag, XCDE_Console);

					}

					if (res == -1)
						return (-1);

				}


				// Visualizzazione dello snippet ben formato

				if (flag_tag) {

					for (i=size_pila_tag-1; (i>=0) && ((pila_tag[i][0] != '/')); i--);

					// si legge la pila e si bilanciano i tag chiusi ma non aperti
					for ( k = i; k>=0; k-- ) {

						fprintf (stdout, "<%s>\n", (pila_tag[k])+1);
						fprintf (stdout, "%s\n", XCDE_MORE_TEXT);
					}


					if ((direction == 1) || (direction == 0) ) {

						fprintf (stdout, "%s", backward_string);

					}

					fprintf (stdout, "%s", snippet_estratto);


					if ((direction == 1) || (direction == 2) ) {

						fprintf (stdout, "%s", forward_string);

					}

					// si legge la pila e si bilanciano i tag chiusi ma non aperti
					for (k=size_pila_tag-1; k>i; k--) {

						fprintf (stdout, "\n%s", XCDE_MORE_TEXT);
						fprintf (stdout, "\n</%s>", pila_tag[k]);
					}

				} else {

					res = remove_duplicated_spaces(backward_string);

					if (res == -1)
						return (-1);

					res = estrai_text_element(backward_string);

					if (res == -1)
						return (-1);

					fprintf(stdout, " ");




					res = remove_duplicated_spaces(snippet_estratto);

					if (res == -1)
						return (-1);

					res = estrai_text_element(snippet_estratto);

					if (res == -1)
						return (-1);




					res = remove_duplicated_spaces(forward_string);

					if (res == -1)
						return (-1);

					res = estrai_text_element(forward_string);

					if (res == -1)
						return (-1);

				}

				fprintf(stdout, "\n");

				// free della memoria necessaria per la pila che servir per
				//	estrarre lo snippet ben formato.
				for (i=0; i<size_pila_tag; i++) {

					free(pila_tag[i]);
					pila_tag[i] = NULL;
				}

				free(pila_tag);

			}
		}

	} else {

		fprintf(stderr, "\n\t<xcde:error> wrong use of xml_context TAG. All childs must be WORDS </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-2) ;

	}

	return 1;
}

int view_snippet_result(query_tree_node * node, char *nomeFile, XCDE_IL_Couple_Type **risultato, int size_vettore_risultato, HHash_table *input_hash_table, int riga_da_visualizzare, XCDE_Console_Type *XCDE_Console) {

	char * variabile;
	int index_variabile;
	Hash_node * node_hash;
	char *snippet_estratto = NULL;
	int i;
	char *ampiezza_string;
	char *flag_tag_string;
	char *direction_string;
	int ampiezza;
	unsigned char flag_tag;
	unsigned char direction;
	int final_position = 0;

	if (node->node_type == TAG) {

		if (strcmp(node->node_name, "xml_context") == 0) {

			ampiezza_string = get_attribute(node,"xml_size");

			if (ampiezza_string != NULL) {

				if ( !isdigit(ampiezza_string[0]) ) {

					fprintf(stderr, "\n\n\t<xcde:error> xml_size value must be integer (xml_size = '%s') </xcde:error>\n\n", ampiezza_string) ;

					fflush(stderr) ;

					return(-2) ;


				} else {

					ampiezza = atoi(ampiezza_string);
				}

			} else {

				ampiezza = 0;
			}

			flag_tag_string = get_attribute(node,"xml_format");

			if (flag_tag_string != NULL) {

				if ( (strcmp(flag_tag_string, "notag") == 0) || (strcmp(flag_tag_string, "1") == 0) ) {

					flag_tag = 0;

				} else {

					if ( (strcmp(flag_tag_string, "tag") == 0) || (strcmp(flag_tag_string, "0") == 0) ) {

						flag_tag = 1;

					} else {

						if ( (strcmp(flag_tag_string, "full") == 0) || (strcmp(flag_tag_string, "2") == 0) ) {

							flag_tag = 2;

						} else {

							if ( (strcmp(flag_tag_string, "stop_parent") == 0) || (strcmp(flag_tag_string, "3") == 0) ) {

								flag_tag = 3;

							} else {

								fprintf(stderr, "\n\n\t<xcde:error> xml_format value must be { 0 | tag | 1 | notag | 2 | full | 3 | stop_parent }. (xml_format = '%s'). </xcde:error>\n\n", flag_tag_string) ;

								fflush(stderr) ;

								return(-2) ;

							}
						}

					}
				}

			} else {

				flag_tag = 1;

			}

			direction_string = get_attribute(node,"xml_view");

			if (direction_string != NULL) {

				if (strcmp(direction_string, "backward") == 0) {

					direction = 0;

				} else {

					if (strcmp(direction_string, "bidir") == 0) {

						direction = 1;

					} else {

						if (strcmp(direction_string, "forward") == 0) {

							direction = 2;

						} else {

							fprintf(stderr, "\n\n\t<xcde:error> xml_view value must be { backward | forward | bidir}. (xml_view = '%s'). </xcde:error>\n\n", direction_string) ;

							fflush(stderr) ;

							return(-2) ;

						}
					}
				}

			} else {

				direction = 1;

			}

			if ( (ampiezza != 0) || (flag_tag == 0) ) {

				for (i=0; i < node->num_childs; i++) {

					res = estrai_contesto(node->childs_vector[i], nomeFile, ampiezza, flag_tag, direction, risultato, size_vettore_risultato, input_hash_table, riga_da_visualizzare, XCDE_Console);

					if (res == -1)
						return (-1);

					if (res == -2)
						return (-2);

				}

			} else {

				for (i=0; i < node->num_childs; i++) {

					res = view_snippet_result(node->childs_vector[i], nomeFile, risultato, size_vettore_risultato, input_hash_table, riga_da_visualizzare, XCDE_Console);

					if (res == -1)
						return (-1);

					if (res == -2)
						return (-2);

				}
			}

		} else {

			if (strcmp(node->node_name, XCDE_RETURN_RESULT_ELEMENT) != 0) {

				fprintf(stdout, "<%s", node->node_name);

				for (i=0; i < node->num_attributes; i+=2) {

					fprintf(stdout, " %s = '%s'", node->attributes_list[i],node->attributes_list[i+1]);
				}

				fprintf(stdout, ">\n");

				// se non  un node foglia viene richiamata la funzione su ciascuno dei figli
				if (!is_leaf(node)) {

					for (i=0; i < node->num_childs; i++) {

						res = view_snippet_result(node->childs_vector[i], nomeFile, risultato, size_vettore_risultato, input_hash_table, riga_da_visualizzare, XCDE_Console);

						if (res == -1)
							return (-1);

						if (res == -2)
							return (-2);

					}

				}

				fprintf(stdout, "</%s>\n", node->node_name);

			} else {

				// se non  un node foglia viene richiamata la funzione su ciascuno dei figli
				if (!is_leaf(node)) {

					for (i=0; i < node->num_childs; i++) {

						res = view_snippet_result(node->childs_vector[i], nomeFile, risultato, size_vettore_risultato, input_hash_table, riga_da_visualizzare, XCDE_Console);

						if (res == -1)
							return (-1);

						if (res == -2)
							return (-2);

					}

				}

			}
		}

	} else { // caso di una foglia

		variabile = node->node_name;

		if (variabile != NULL) {

			node_hash = HHashtable_search(variabile, strlen(variabile), input_hash_table);

			if (node_hash != NULL)

				index_variabile = node_hash->index;
			else
				index_variabile = -1;
		} else {

			index_variabile = -1;
		}

		if (index_variabile != -1) {

			// se la variabile ha un valore da stampare (potrebbe non averlo in caso di query in OR
			if ( ( (risultato[riga_da_visualizzare][2 * index_variabile].XCDE_Position != 0)
			       ||(risultato[riga_da_visualizzare][2 * index_variabile].XCDE_Offset != 0) ) ) {

				// se la variabile da visualizzare non  un TAG si effettua
				if ( (risultato[riga_da_visualizzare][2 * index_variabile].XCDE_Position == risultato[riga_da_visualizzare][2 * index_variabile + 1].XCDE_Position)
				&& (risultato[riga_da_visualizzare][2 * index_variabile].XCDE_Offset == risultato[riga_da_visualizzare][2 * index_variabile + 1].XCDE_Offset) ) {

					res = extract_word(&snippet_estratto, &final_position, risultato[riga_da_visualizzare][2 * index_variabile].XCDE_Position, XCDE_Console);

					if (res == -1)
						return(-1) ;

					fprintf (stdout, "%s\n", snippet_estratto);

				} else {

					res = extract_snippet(&snippet_estratto, risultato[riga_da_visualizzare][2 * index_variabile], risultato[riga_da_visualizzare][2 * index_variabile + 1], XCDE_Console);

					if (res == -1)
						return(-1) ;

					fprintf (stdout, "%s\n", snippet_estratto);
				}
			}

		} else {

			fprintf(stdout, "%s\n", node->node_name);
		}
	}

	return 1;
}



int extract_word(char **word, int *final_position, int position, XCDE_Console_Type *XCDE_Console) {

	int forward_window_pointer=0 ;

	int codeword=0 ;

	int codeword_len=0 ;

	char *token=0 ;

	int function_result=0 ;

	int dimension_forward_string = INIT_FORWARD_STRING_SIZE;

	unsigned char done = 0;

	/* ------------------------ */
	/* initializes data structs */
	/* ------------------------ */

	*final_position = position ;

	/* ------------------------------ */
	/* reads window forkward codewords */
	/* ------------------------------ */

	*word = (char *) malloc ( dimension_forward_string * sizeof(char) );

	if (*word == NULL){

		fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for *word </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-1) ;
	}

	strcpy(*word, "");

	while ((!done) && (*final_position < ((XCDE_Console)->XCDE_Body_Size_Without_Zlib))) {

		function_result = XCDE_Get_Body_Forward_Codeword(*final_position, &codeword, &codeword_len, XCDE_Console) ;

		if (function_result == -1) {

			fprintf(stderr, "\n\t<xcde:error> problems with the XCDE_Get_Body_Forward_Codeword function </xcde:error>\n\n") ;

			fflush(stderr) ;

			return(-1) ;
		}

		function_result = XCDE_Codeword2Token(codeword, codeword_len, &token, XCDE_Console) ;

		if (function_result == -1) {

			fprintf(stderr, "\n\t<xcde:error> problems with the XCDE_Codeword2Token function </xcde:error>\n\n") ;

			fflush(stderr) ;

			return(-1) ;
		}

		if ( is_separator(token) ) {

			done = 1;

		} else {

			*final_position += codeword_len ;

			if ( (strlen(*word) + (strlen(token) + 1) > dimension_forward_string ) ) {

				dimension_forward_string *= 2;

				*word = (char *) realloc (*word, dimension_forward_string * sizeof(char) );

				if (*word == NULL){

					fprintf(stderr, "\n\n\t<xcde:error> problems during memory allocation for *word </xcde:error>\n\n") ;

					fflush(stderr) ;

					return(-1) ;
				}
			}

			strcat(*word, token);
		}
	}
	
	return 1;

}




int extract_snippet(char **snippet, XCDE_IL_Couple_Type posizione_iniziale, XCDE_IL_Couple_Type posizione_finale, XCDE_Console_Type *XCDE_Console) {

	int dimension_forward_string = INIT_FORWARD_STRING_SIZE;

	int forward_window_pointer=0 ;

	int forward_body_position=0 ;

	int codeword=0 ;

	int codeword_len=0 ;

	char *token=0 ;

	int function_result=0 ;



	/* ------------------------ */
	/* initializes data structs */
	/* ------------------------ */

	forward_window_pointer = 0 ;

	forward_body_position = posizione_iniziale.XCDE_Position ;


	/* ------------------------------- */
	/* reads window forkward codewords */
	/* ------------------------------- */

	if (*snippet)
		free(*snippet);

	*snippet = (char *) malloc ( dimension_forward_string * sizeof(char) );

	if (*snippet == NULL){

		fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for *snippet </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-1) ;
	}

	strcpy(*snippet, "");

	while ((forward_body_position <= posizione_finale.XCDE_Position) && (forward_body_position < ((XCDE_Console)->XCDE_Body_Size_Without_Zlib))) {

		function_result = XCDE_Get_Body_Forward_Codeword(forward_body_position, &codeword, &codeword_len, XCDE_Console) ;

		if (function_result == -1) {

			fprintf(stderr, "\n\t<xcde:error> problems with the XCDE_Get_Body_Forward_Codeword function </xcde:error>\n\n") ;

			fflush(stderr) ;

			return(-1) ;
		}

		function_result = XCDE_Codeword2Token(codeword, codeword_len, &token, XCDE_Console) ;

		if (function_result == -1) {

			fprintf(stderr, "\n\t<xcde:error> problems with the XCDE_Codeword2Token function </xcde:error>\n\n") ;

			fflush(stderr) ;

			return(-1) ;
		}

		forward_body_position += codeword_len ;

		if ( (strlen(*snippet) + (strlen(token) + 1) > dimension_forward_string ) ) {

			dimension_forward_string *= 2;

			*snippet = (char *) realloc (*snippet, dimension_forward_string * sizeof(char) );

			if (*snippet == NULL){

				fprintf(stderr, "\n\n\t<xcde:error> problems during memory allocation for *snippet </xcde:error>\n\n") ;

				fflush(stderr) ;

				return(-1) ;
			}
		}

		strcat(*snippet, token);

	}



	return(1) ;
}



int view_position_result(query_tree_node * node, XCDE_IL_Couple_Type **risultato, int size_vettore_risultato, HHash_table *input_hash_table, int riga_da_visualizzare) {

	char * variabile;
	int index_variabile;
	Hash_node * node_hash;
	char * ampiezza_finestra;
	int i;


	if (node->node_type == TAG) {

		if (strcmp(node->node_name, "xml_context") == 0) {

			fprintf(stderr, "<xcde:error> TAG xml_context can't occurs with flag -p </xcde:error>\n\n") ;

			fflush(stderr) ;

			return (-2);

		}

		if (strcmp(node->node_name, XCDE_RETURN_RESULT_ELEMENT) != 0) {

			fprintf(stdout, "<%s", node->node_name);

			for (i=0; i < node->num_attributes; i+=2) {

				fprintf(stdout, " %s = '%s'", node->attributes_list[i],node->attributes_list[i+1]);
			}

			fprintf(stdout, ">\n");

			// se non  un node foglia viene richiamata la funzione su ciascuno dei figli
			if (!is_leaf(node)) {

				for (i=0; i < node->num_childs; i++) {

					res = view_position_result(node->childs_vector[i], risultato, size_vettore_risultato, input_hash_table, riga_da_visualizzare);

					if (res == -1)
						return (-1);

					if (res == -2)
						return (-2);

				}

			}

			fprintf(stdout, "</%s>\n", node->node_name);

		} else {

			// se non  un node foglia viene richiamata la funzione su ciascuno dei figli
			if (!is_leaf(node)) {

				for (i=0; i < node->num_childs; i++) {

					res = view_position_result(node->childs_vector[i], risultato, size_vettore_risultato, input_hash_table, riga_da_visualizzare);

					if (res == -1)
						return (-1);

					if (res == -2)
						return (-2);

				}
			}
		}

	} else { // caso di una parola

		variabile = node->node_name;

		if (variabile != NULL) {

			node_hash = HHashtable_search(variabile, strlen(variabile), input_hash_table);

			if (node_hash != NULL)

				index_variabile = node_hash->index;
			else
				index_variabile = -1;
		} else {

			index_variabile = -1;
		}

		if (index_variabile != -1) {

			// se la variabile ha un valore da stampare (potrebbe non averlo in caso di query in OR
			if ( ( (risultato[riga_da_visualizzare][2 * index_variabile].XCDE_Position != 0)
			       ||(risultato[riga_da_visualizzare][2 * index_variabile].XCDE_Offset != 0) ) ) {

				// se la variabile da visualizzare non  un TAG si effettua
				if ( (risultato[riga_da_visualizzare][2 * index_variabile].XCDE_Position == risultato[riga_da_visualizzare][2 * index_variabile + 1].XCDE_Position)
				&& (risultato[riga_da_visualizzare][2 * index_variabile].XCDE_Offset == risultato[riga_da_visualizzare][2 * index_variabile + 1].XCDE_Offset) ) {

					fprintf(stdout, "\t[(%d, %d)]\n", risultato[riga_da_visualizzare][2 * index_variabile].XCDE_Position, risultato[riga_da_visualizzare][2 * index_variabile].XCDE_Offset);

				} else {

					fprintf(stdout, "\t[(%d, %d), (%d, %d)]\n", risultato[riga_da_visualizzare][2 * index_variabile].XCDE_Position, risultato[riga_da_visualizzare][2 * index_variabile].XCDE_Offset, risultato[riga_da_visualizzare][2 * index_variabile + 1].XCDE_Position, risultato[riga_da_visualizzare][2 * index_variabile + 1].XCDE_Offset);

				}
			}

		} else {

			fprintf(stdout, "\t%s\n", node->node_name);
		}
	}
	
	return (1);
}





int split_query_string(char *query_string, char **select_string, char ***from_file_vector, int *number_of_files, char **return_string, int *first_range, int *last_range)
{

	char seps[]   = " ";
	int i;
	char *token = NULL;
	
	char *main_return_tag;

	// dimensione del vettore contenente i nomei dei file nei quali effettuare la ricerca
	int from_vetctor_dimension = XCDE_INITIAL_FROM_STRING_VECTOR_SIZE;

	//inizializzazione del contatore del numero di file
	*number_of_files = 0;


	if ( (query_string == NULL) || (strcmp(query_string,"") == 0) ) {

		fprintf(stderr, "\n\t<xcde:error> missing query_string </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-2);
	}

	// allocazione della memoria necessaria per le stringhe che conterrano le varie parti della query_string
	*select_string = (char *) malloc((strlen(query_string) + 1) * sizeof(char));
	*return_string = (char *) malloc((strlen(query_string) + 1) * sizeof(char));
	main_return_tag = (char *) malloc( (strlen(XCDE_RETURN_RESULT_ELEMENT) + 3) * sizeof(char));

	// allocazione del vettore contenente i file in cui effettuare la ricerca
	*from_file_vector = (char **) malloc(from_vetctor_dimension * sizeof(char *));

	
	strcpy(*select_string, "");

	strcpy(*return_string, "<");
	strcat(*return_string, XCDE_RETURN_RESULT_ELEMENT);
	strcat(*return_string, ">");

	strcpy(main_return_tag, "<");
	strcat(main_return_tag, XCDE_RETURN_RESULT_ELEMENT);
	strcat(main_return_tag, ">");

	token = strtok( query_string, seps );

	if ( strcmp(token, "SELECT") == 0 ) {

		token = strtok( NULL, seps );

		while( (token != NULL) && (strcmp (token, "FROM") != 0) ) {

			strcat( *select_string, token);

			token = strtok( NULL, seps );

			if ((token) && (strcmp(token, "FROM") != 0) )
				strcat( *select_string, " ");

		}

		if (token == NULL) {

			fprintf(stderr, "\n\t<xcde:error> missing FROM clause in query_string  </xcde:error>\n\n") ;

			fflush(stderr) ;

			return(-2);
		}

		if (strcmp(*select_string,"") == 0) {

			fprintf(stderr, "\n\t<xcde:error> SELECT clause is empty  </xcde:error>\n\n") ;

			fflush(stderr) ;

			return(-2);
		}

		token = strtok( NULL, seps );

		while( (token != NULL) && (strcmp(token, "RETURN") != 0) ) {

			if (*number_of_files >= from_vetctor_dimension) {

				from_vetctor_dimension *= 2;

				// allocazione del vettore contenente i file in cui effettuare la ricerca
				*from_file_vector = (char **) realloc(*from_file_vector, from_vetctor_dimension * sizeof(char *));

				if (*from_file_vector == NULL) {

					fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for *from_file_vector </xcde:error>\n\n") ;

					fflush(stderr) ;

					return(-1) ;
				}

			}

			(*from_file_vector)[*number_of_files] = (char *) malloc((strlen(token) + 1) * sizeof(char));

			if ((*from_file_vector)[*number_of_files] == NULL) {

				fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for (*from_file_vector)[%d] </xcde:error>\n\n", *number_of_files) ;

				fflush(stderr) ;

				return(-1) ;
			}

			strcpy((*from_file_vector)[*number_of_files], token);

			(*number_of_files)++;

			token = strtok( NULL, seps );
/*
			if ((token) && (strcmp(token, "RETURN") != 0) )
				strcat( *from_file_vector, " ");
*/
		}

		if (token == NULL) {

			fprintf(stderr, "\n\t<xcde:error> missing RETURN clause  </xcde:error>\n\n") ;

			fflush(stderr) ;

			return(-2);
		}

		if (*number_of_files == 0) {

			fprintf(stderr, "\n\t<xcde:error> FROM clause is empty  </xcde:error>\n\n") ;

			fflush(stderr) ;

			return(-2);
		}

		token = strtok( NULL, seps );

		while( (token != NULL)  && (strcmp(token, "RANGE") != 0) ) {

			strcat( *return_string, token);

			token = strtok( NULL, seps );

			if (token  && (strcmp(token, "RANGE") != 0) )
				strcat( *return_string, " ");
		}

		if (strcmp(*return_string, main_return_tag) == 0) {

			fprintf(stderr, "\n\t<xcde:error> RETURN clause is empty </xcde:error>\n\n") ;

			fflush(stderr) ;

			return(-2);
		}

		if (token && (strcmp(token, "RANGE") == 0) ) {

			token = strtok( NULL, seps );

			if (token != NULL) {

				if (isdigit(token[0]) || (token[0] == '-') ) {

					*first_range = atoi(token);

				} else {

					fprintf(stderr, "\n\t<xcde:error> wrong RANGE value. RANGE usage: 'RANGE number number' </xcde:error>\n\n") ;

					fflush(stderr) ;

					return(-2);

				}

				token = strtok( NULL, seps );

				if (token != NULL) {

					if (isdigit(token[0]) || (token[0] == '-') ) {

						*last_range = atoi(token);

					} else {

						fprintf(stderr, "\n\t<xcde:error> wrong RANGE value. RANGE usage: 'RANGE number number' </xcde:error>\n\n") ;

						fflush(stderr) ;

						return(-2);


					}

				} else {

					fprintf(stderr, "\n\t<xcde:error> missing second number of RANGE clause. RANGE usage: 'RANGE number number' </xcde:error>\n\n") ;

					fflush(stderr) ;

					return(-2);

				}

			} else {

				fprintf(stderr, "\n\t<xcde:error> missing first number of RANGE clause. RANGE usage: 'RANGE number number' </xcde:error>\n\n") ;

				fflush(stderr) ;

				return(-2);

			}

			// controllo sulla consistenza dei valori dell'intervallo inseriti
			if (*first_range > *last_range) {

				fprintf(stderr, "\n\t<xcde:error> first number of RANGE clause must be greater than second one </xcde:error>\n\n") ;

				fflush(stderr) ;

				return(-2);

			}

			if ( (*first_range == 0) || (*last_range == 0) ) {

				fprintf(stderr, "\n\t<xcde:error> numbers in RANGE clause must be different from 0 </xcde:error>\n\n") ;

				fflush(stderr) ;

				return(-2);

			}

			if ( (*first_range < 0) || (*last_range < 0) ) {

				fprintf(stderr, "\n\t<xcde:error> numbers in RANGE clause must be greater than 0 </xcde:error>\n\n") ;

				fflush(stderr) ;

				return(-2);

			}
		}


	} else {

		fprintf(stderr, "\n\t<xcde:error> missing SELECT clause in query_string </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-2);


	}

	strcat(*return_string, "</");
	strcat(*return_string, XCDE_RETURN_RESULT_ELEMENT);
	strcat(*return_string, ">");

	replace_amp_occ(select_string);

	return 1;
}



// Funzione che dato un albero ne restituisce una copia che ha la stessa struttura
int copia_albero(query_tree_node *albero_da_copiare, query_tree_node **albero_copiato) {

	int i;


	// allocazione della memoria necessaria per il node dell'albero copiato
	*albero_copiato = (query_tree_node *) malloc(sizeof(query_tree_node));

	if (*albero_copiato == NULL) {

		fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for *albero_copiato </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-1);

	}

	// allocazione della memoria necessaria per il nome del node dell'albero copiato
	(*albero_copiato)->node_name = (char *) malloc((strlen(albero_da_copiare->node_name) + 1) * sizeof(char));

	if ((*albero_copiato)->node_name  == NULL) {

		fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for (*albero_copiato)->node_name </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-1);

	}

	strcpy((*albero_copiato)->node_name, albero_da_copiare->node_name);
	(*albero_copiato)->node_type = albero_da_copiare->node_type;
	(*albero_copiato)->num_childs = albero_da_copiare->num_childs;
	(*albero_copiato)->num_positions = 0;
	(*albero_copiato)->num_depth = 0;
	(*albero_copiato)->depth_vector = NULL;

	if (albero_da_copiare->node_type == TAG)
		(*albero_copiato)->tag_positions = NULL;
	else
		(*albero_copiato)->word_positions = NULL;

	(*albero_copiato)->range_results = NULL;
	(*albero_copiato)->is_setted_number_result = 0;
	(*albero_copiato)->is_potable = 0;
	(*albero_copiato)->num_attributes = albero_da_copiare->num_attributes;

	// allocazione della memoria necessaria per il vettore degli attributi
	(*albero_copiato)->attributes_list = (char **) malloc(albero_da_copiare->num_attributes * sizeof(char *));

	if ((*albero_copiato)->attributes_list == NULL) {

		fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for (*albero_copiato)->attributes_list </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-1);

	}


	// copia della lista degli attributi
	for (i=0; i < albero_da_copiare->num_attributes; i++) {

		(*albero_copiato)->attributes_list[i] = (char *) malloc((strlen(albero_da_copiare->attributes_list[i]) + 1) * sizeof(char));

		if ((*albero_copiato)->attributes_list[i] == NULL) {

			fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for (*albero_copiato)->attributes_list[%d] </xcde:error>\n\n", i) ;

			fflush(stderr) ;

			return(-1);

		}

		strcpy((*albero_copiato)->attributes_list[i],albero_da_copiare->attributes_list[i]);

	}

	(*albero_copiato)->childs_vector = (query_tree_node **) malloc (albero_da_copiare->num_childs * sizeof(query_tree_node *));

	if ((*albero_copiato)->childs_vector == NULL) {

		fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for (*albero_copiato)->childs_vector </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-1);

	}

	for (i=0; i < albero_da_copiare->num_childs; i++) {

		res = copia_albero(albero_da_copiare->childs_vector[i], &((*albero_copiato)->childs_vector[i]));

		// controllo sull'esito della funzione copia_albero
		if (res == -1) {

			return(-1);

		}
	}

	return 1;
}






int riempi_hash(char *vettore_hash, int numero_variabili, XCDE_Console_Query_Type *Console_Query) {

	int i;
	int pointer = 0;

	int lungh_node_name, lungh_nome, index;
	char *variable_name;


	if (Console_Query->hash_table) {

		res = HHashtable_clear(Console_Query->hash_table);
		
		if (res == -1)
			return -1;

	}


	/*******************************************************************
	                    initialize Hash Table
	 *******************************************************************/

	// inizializzazione della global_hash_table con dimensione 2
	//	nella init la dimensione totale  calcolata come size  = (int) 1.2 * n + 13; (con n = 2) (Load factor ~ 80%)
	res = HHashtable_init(&(Console_Query->hash_table), 2);

	if (res == -1)
		return -1;

	// vengono recuperate le informazioni relative alla tabella hash
	for(i=0; i<numero_variabili;i++){

		// si recupera dal file mappato la lunghezza del nome della variabile
		read_int(&lungh_nome, vettore_hash + pointer);

		pointer += sizeof(int);

		// allocazione della memoria necessaria per il nome del file in cui scrivere i risultati
		variable_name = (char *) malloc( lungh_nome * sizeof(char));

		// controllo sulla corretta allocazione della memoria
		if (variable_name == NULL) {

			fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for variable_name </xcde:error>\n\n") ;

			fflush(stderr) ;

			return(-1);
		}

		strcpy(variable_name, "");

		strcpy(variable_name, vettore_hash + pointer);

		pointer += lungh_nome;

		// si recupera dal file mappato l'index associalt alla variabile
		read_int(&index, vettore_hash + pointer);

		pointer += sizeof(int);

		// si recupera dal file mappato la lunghezza del nome della variabile
		read_int(&lungh_node_name, vettore_hash + pointer);

		pointer += sizeof(int);

		res = HHashtable_insert(variable_name, lungh_node_name, index, vettore_hash + pointer, Console_Query->hash_table);

		if (res == -1)
			return -1;

		pointer += lungh_node_name;

	}

	return (1);
}



/* -------------------------------------------------------- */
/* ---------  		API FUNCTIONS		 ---------- */
/* -------------------------------------------------------- */


// Funzione che data la query_string ne fa il parsing restituendo la il puntatore all'albero della SELECT,
//	il puntatore all'albero della RESULT, un vettore contenente i nomi dei files sui quali effettuare la ricerca,
//	un intero che indica il numero di files e gli estremi dell'intervallo della range se presente,
//	(la coppia 0, 0 altrimenti), e la tabella hash associata alla query.
int XCDE_Parse_Query_String(char *query_string, XCDE_Console_Query_Type *Console_Query) {

	int i;

	// data structure used by Expat
	XML_Parser parser ;


	if (query_string == NULL)
		return -1;

	if(Console_Query == NULL) {

		Console_Query = (XCDE_Console_Query_Type *) malloc (sizeof(XCDE_Console_Query_Type));
	}

	// Inizializzazione della console
	Console_Query->select_string = NULL;
	Console_Query->file_vector = NULL;
	Console_Query->number_of_files = 0;
	Console_Query->return_string = NULL;
	Console_Query->first_range = 0;
	Console_Query->last_range = 0;

	Console_Query->select_tree = NULL;
	Console_Query->trees_vector = NULL;
	Console_Query->number_of_trees = 0;
	Console_Query->return_tree = NULL;
	Console_Query->hash_table = NULL;
	
	Console_Query->XCDE_query_parsed = 0;
	Console_Query->XCDE_query_executed = 0;



	res = split_query_string(query_string, &(Console_Query->select_string), &(Console_Query->file_vector), &(Console_Query->number_of_files), &(Console_Query->return_string), &(Console_Query->first_range), &(Console_Query->last_range));

	if (res == -1) {

		return(-1);

	}

	if (res == -2)
		return (-2);


	/*******************************************************************
	                    initialize Hash Table
	 *******************************************************************/

	// inizializzazione della global_hash_table con dimensione 2
	//	nella init la dimensione totale  calcolata come size  = (int) 1.2 * n + 13; (con n = 2) (Load factor ~ 80%)
	res = HHashtable_init(&global_hash_table, 2);

	if (res == -1)
		return -1;

	/*******************************************************************
			configures EXPAT system for parsing RETURN
	 *******************************************************************/

 	// allocates the memory needed by the Pila
	pila_parsing = (node_pila *) malloc(dimension_pila_parsing * sizeof(node_pila)) ;

	if (pila_parsing == NULL) {

		fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for pila_parsing </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-1);

	}

	// create the EXPAT parser
	parser = XML_ParserCreate(NULL) ;

	if (!parser) {

		fprintf(stderr, "\n\t<xcde:error> problems with EXPAT Parser create </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-1);

	}

	// handlers for the tag start and end
	XML_SetElementHandler(parser, startTag, endTag);

	// handler for the text
	XML_SetCharacterDataHandler(parser, textElementReturn);

	// handler for the default elements
//	XML_SetDefaultHandler(parser, defaultHandler);

	// This enables parsing of parameter entities, including the external parameter entity that is the external DTD subset, according to code
	XML_SetParamEntityParsing(parser, XML_PARAM_ENTITY_PARSING_NEVER);

	/*******************************************************************
                          parsing della return_string e
                       costruzione dell'albero della return
	 *******************************************************************/

	if (!XML_Parse(parser, Console_Query->return_string, strlen(Console_Query->return_string), 1)) {

		fprintf(stderr, "\n\t<xcde:error> problems in parsing raw %d %s of RETURN clause </xcde:error>\n\n",

			  XML_GetCurrentLineNumber(parser),

			  XML_ErrorString(XML_GetErrorCode(parser))) ;

		fflush(stderr) ;

		return(-2);

	}


	// deallocazione della pila e reinizializzazione degli indici
	free(pila_parsing);
	size_pila_parsing = 0;
	dimension_pila_parsing = INIT_SIZE_PILA_PARSING;


	Console_Query->return_tree = node_radice;

	node_radice = (query_tree_node *) NULL;

	// deallocazione del parser
	XML_ParserFree(parser) ;


	/*******************************************************************
	                    configures EXPAT system for select
	 *******************************************************************/

	// allocates the memory needed by the Pila
	pila_parsing = (node_pila *) malloc(dimension_pila_parsing * sizeof(node_pila)) ;

	if (pila_parsing == NULL) {

		fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for pila_parsing </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-1);

	}

	// create the EXPAT parser
	parser = XML_ParserCreate(NULL) ;

	if (!parser) {

		fprintf(stderr, "\n\t<xcde:error> problems with EXPAT Parser create </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-1);

	}

	// handlers for the tag start and end
	XML_SetElementHandler(parser, startTag, endTag);

	// handler for the text
	XML_SetCharacterDataHandler(parser, textElementSelect);

	// handler for the default elements
//	XML_SetDefaultHandler(parser, defaultHandler);

	// This enables parsing of parameter entities, including the external parameter entity that is the external DTD subset, according to code
	XML_SetParamEntityParsing(parser, XML_PARAM_ENTITY_PARSING_NEVER);


	/*******************************************************************
                          parsing della select_string e
                       costruzione dell'albero della select
	 *******************************************************************/

	if (!XML_Parse(parser, Console_Query->select_string, strlen(Console_Query->select_string), 1)) {

		fprintf(stderr, "\n\t<xcde:error> problems in parsing raw %d %s of SELECT clause  </xcde:error>\n\n",

			  XML_GetCurrentLineNumber(parser),

			  XML_ErrorString(XML_GetErrorCode(parser))) ;

		fflush(stderr) ;

		return(-2);

	}


	// deallocazione della pila e reinizializzazione degli indici
	free(pila_parsing);
	size_pila_parsing = 0;
	dimension_pila_parsing = INIT_SIZE_PILA_PARSING;


	Console_Query->select_tree = node_radice;

	node_radice = (query_tree_node *) NULL;

	// deallocazione del parser
	XML_ParserFree(parser) ;


	if ((global_hash_table)->card == 0) {

		fprintf(stderr, "\n\t<xcde:error> there aren't variable in query </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-2);

	}


	if ( !is_valid_return(global_hash_table) ) {

		fprintf(stderr, "\n\t<xcde:error> RETURN clause contain an unknow variable. </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-2);

	}

	Console_Query->hash_table = global_hash_table;
	Console_Query->number_of_trees = Console_Query->number_of_files;
	global_hash_table = NULL;

	Console_Query->XCDE_query_parsed = 1;

	return 1;
}



// Funzione che dato un albero restituisce il numero di risultati associati senza costruire effettivamente il vettore risultato
int XCDE_Get_Number_Results(query_tree_node *node, int *numero_risultati, XCDE_Console_Query_Type *Console_Query) {

	int temp_res = 0;
	int temp_res_figli = 1;
	int num_res;
	int i, j;


	if (node == NULL)
		return -1;



	if (is_leaf(node)) {

		if (node->node_type == WORD) {

			temp_res = node->num_positions;

		} else {

			temp_res = (node->num_positions) / 2;

		}

	} else {

		for (i=0; i < node->num_positions; i += 2) {

			temp_res_figli = 1;

			for (j=0; j < node->num_childs; j++) {

				res = get_number_results_from_position(node->childs_vector[j], node->tag_positions[i], node->tag_positions[i+1], &num_res, Console_Query->hash_table);

				if (res == -1)
					return -1;

				temp_res_figli *= num_res;

			}

			temp_res += temp_res_figli;
		}
	}

	*numero_risultati = temp_res;

	return 1;
}



// Funzione che data la QUERY restituisce l'albero della SELECT dopo la ricerca
//	restituisce un albero per ogni file

int XCDE_Execute_Query(XCDE_Console_Query_Type *Console_Query) {

	int i, k, j;
	XCDE_Console_Type XCDE_Console;
	unsigned char is_setted_depthstruct;
	

	if (Console_Query == NULL)
		return -1;


	if (Console_Query->XCDE_query_parsed != 1)
		return -1;


	Console_Query->trees_vector = (query_tree_node **) malloc ((Console_Query->number_of_files) * sizeof(query_tree_node *));

	if (Console_Query->trees_vector == NULL) {

		fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for Console_Query->trees_vector </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-1);

	}


	/*******************************************************************
			Costruzione dei risultati
	*******************************************************************/


	// per ogni file indicato nella clausola FROM (quindi presenti nel vettore vettore_files)
	//	viene fatta una copia della struttura dell'albero della select e costruito
	for (i=0; i < (Console_Query->number_of_files); i++) {


		res = copia_albero(Console_Query->select_tree, &((Console_Query->trees_vector)[i]));

		if (res == -1) {

			continue;

		}

		/*******************************************************************
				Inizializzazione della console
		*******************************************************************/

		// opening console
		res=XCDE_Init_Console((Console_Query->file_vector)[i], &XCDE_Console) ;

		if (res == -1) {

			continue;

		}

		res = XCDE_SettedDepthstruct(&is_setted_depthstruct, &XCDE_Console) ;

		if (res == -1) {

			fprintf(stderr, "\n\t<xcde_error> problem with XCDE_SettedDepthstruct funtion </xcde_error>\n\n") ;

			fflush(stderr) ;

			continue;

		}

		if (is_setted_depthstruct != 1) {

			continue;

		}




		/*******************************************************************
			Operazioni che richiedono l'utilizzo della console
		*******************************************************************/

		// visita bottom-up per riempire le liste associate ai nodi di cui alla fine solo
		// quella associata al node radice  affidabile
		res = riempi_lista((Console_Query->trees_vector)[i], XCDE_EXACT, 0, XCDE_CASE, XCDE_NOPROXIMITY, Console_Query->hash_table, &XCDE_Console);

		if (res == -1) {

			XCDE_Finalize_Console(&XCDE_Console) ;

			continue;

		}

		if (res == -2) {

			XCDE_Finalize_Console(&XCDE_Console) ;

			return -2;

		}


		if (is_special_node((Console_Query->trees_vector)[i])) {

			// allocazione del vettore delle posizioni
			((Console_Query->trees_vector)[i])->tag_positions = (XCDE_IL_Couple_Type *) malloc (2 * ((Console_Query->trees_vector)[i])->childs_vector[0]->num_positions * sizeof(XCDE_IL_Couple_Type));

			if (((Console_Query->trees_vector)[i])->tag_positions == NULL) {

				fprintf(stderr,"\n\t<xcde:error> problems during memory allocation for (Console_Query->trees_vector)[%d])->tag_positions </xcde:error>\n\n", i);

				fflush(stderr);

				continue;

			}

			for(j=0, k=0; j<((Console_Query->trees_vector)[i])->childs_vector[0]->num_positions; j++, k += 2) {

				((Console_Query->trees_vector)[i])->tag_positions[k].XCDE_Position = ((Console_Query->trees_vector)[i])->childs_vector[0]->word_positions[j].XCDE_Position;
				((Console_Query->trees_vector)[i])->tag_positions[k].XCDE_Offset = ((Console_Query->trees_vector)[i])->childs_vector[0]->word_positions[j].XCDE_Offset;

				((Console_Query->trees_vector)[i])->tag_positions[k+1].XCDE_Position = ((Console_Query->trees_vector)[i])->childs_vector[0]->word_positions[j].XCDE_Position;
				((Console_Query->trees_vector)[i])->tag_positions[k+1].XCDE_Offset = ((Console_Query->trees_vector)[i])->childs_vector[0]->word_positions[j].XCDE_Offset;

			}

			((Console_Query->trees_vector)[i])->num_positions = 2 * ((Console_Query->trees_vector)[i])->childs_vector[0]->num_positions;

/*			// riferimento che permetter di liberare la memoria relativa a tale node
			node_speciale = (Console_Query->trees_vector)[i];

			// si assegna il nipote direttamente al posto del figlio speciale
			(Console_Query->trees_vector)[i] = node_speciale->childs_vector[0];

			// viene liberata la memoria relativa al node speciale eliminato
			res = free_node(node_speciale);

			if (res == -1)
				return (-1);
*/
		}


		if (res != 0) {

			// visita top-down per filtrare dalle liste dei nodi le posizioni che non fanno parte del risultato
			res = correggi_lista((Console_Query->trees_vector)[i]);

			if (res == -1) {

				XCDE_Finalize_Console(&XCDE_Console) ;

				continue;

			}

		}

		/*******************************************************************
				Finalizzazione della console
		*******************************************************************/

		res = XCDE_Finalize_Console(&XCDE_Console) ;

		if (res == -1) {

			fprintf(stderr, "\n\t<xcde:error> problems with the XCDE_Finalize_Console function </xcde:error>\n\n") ;

			fflush(stderr) ;

			continue ;
		}
	}


	Console_Query->XCDE_query_executed = 1;

	return 1;

}



// Funzione che dato l'albero della select dopo l'esecuzione della query, la tabella hash e l'albero della return
//	restituisce il vettore risultato

int XCDE_Compute_TableResult(int index_albero, XCDE_IL_Couple_Type ***risultato, int *size_vettore_risultato, XCDE_Console_Query_Type *Console_Query) {


	char * variabile;
	int numero_variabili;
	int index_variabile;
	Hash_node * node_hash;
	query_tree_node *radice_select;

	XCDE_IL_Couple_Type **risultato_figlio = NULL;
	int size_risultato_figlio = 0;

	XCDE_IL_Couple_Type ***risultato_parziale;
	int *size_risultato_parziale;

	int i, k;

	int j = 0;



	if (Console_Query == NULL)
		return -1;


	if (Console_Query->XCDE_query_executed != 1)
		return -1;


	if ( (index_albero < 0) || (index_albero >= Console_Query->number_of_trees) )
		return -1;


	radice_select = Console_Query->trees_vector[index_albero];

	// se la tabella dei risultati non  vuota viene svuotata
	if (*size_vettore_risultato != 0) {

		for (i = 0; i < *size_vettore_risultato; i++)
			if ((*risultato)[i])
				free((*risultato)[i]);

		*size_vettore_risultato = 0;
		if ((*risultato))
			*risultato = NULL;

	}


	// controllo se nella tabella hash  presente la variabile, in tal caso index_variabile prende il valore dell'index
	//	associato alla variabile, altrimenti prende valore -1
	variabile = get_attribute(radice_select,"xml_var");

	if (variabile != NULL) {

		node_hash = HHashtable_search(variabile, strlen(variabile), Console_Query->hash_table);

		if (node_hash != NULL)

			index_variabile = node_hash->index;
		else
			index_variabile = -1;
	} else {

		index_variabile = -1;
	}

	// si recupara dalla tabella hash il numero di variabili
	numero_variabili = Console_Query->hash_table->card;


	// se il numero i figli  uguale a zero bisogna inserire nel risultato le posizioni del radice_select se la variabile associata
	//	 presente nella tabella hash
	if (radice_select->num_childs == 0) {

		// si devono inserire le posizioni nel risultato solo se al radice_select  associata una variabile che  presente anche nella calusola RETURN
		if (index_variabile != -1) {

			if (radice_select->node_type == TAG) { // la foglia  di tipo TAG

				(*risultato) = (XCDE_IL_Couple_Type **) malloc(radice_select->num_positions * sizeof(XCDE_IL_Couple_Type *));

				if ((*risultato) == NULL) {

					fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for (*risultato) </xcde:error>\n\n") ;

					fflush(stderr) ;

					return(-1);

				}


				*size_vettore_risultato = 0;

				for (i = 0; i < radice_select->num_positions; i += 2) {

					// creo la riga della matrice relativa alla posizione corrente
					// devo allocare il vettore che contiene le posizioni del (*risultato) corrente (la dimensione di numero_variabili  moltiplicata
					//	per 2 perch vanno considerate le posizioni di apertura e chiusura
					(*risultato)[*size_vettore_risultato] = (XCDE_IL_Couple_Type *) calloc(2 * numero_variabili, sizeof(XCDE_IL_Couple_Type));

					if ((*risultato)[*size_vettore_risultato] == NULL) {

						fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for (*risultato)[*size_vettore_risultato] </xcde:error>\n\n") ;

						fflush(stderr) ;

						return(-1);

					}

					// inserisco la posizione della parola sia come valore di apertura che come valore di chiusura nella riga_da_inserire
					//	nella posizione corrispondente alla variabile associata al radice_select
					(*risultato)[*size_vettore_risultato][2 * index_variabile].XCDE_Position = radice_select->tag_positions[i].XCDE_Position;
					(*risultato)[*size_vettore_risultato][2 * index_variabile].XCDE_Offset = radice_select->tag_positions[i].XCDE_Offset;

					(*risultato)[*size_vettore_risultato][2 * index_variabile + 1].XCDE_Position = radice_select->tag_positions[i+1].XCDE_Position;
					(*risultato)[*size_vettore_risultato][2 * index_variabile + 1].XCDE_Offset = radice_select->tag_positions[i+1].XCDE_Offset;

					(*size_vettore_risultato)++;
				}

			} else {

				fprintf(stderr, "\n\t<xcde:error> root node of radice_select must be a TAG </xcde:error>\n\n") ;

				fflush(stderr) ;

				return(-2);


			}

		}

	} else {


		if (radice_select->node_type == TAG) {

			for (i=0; i < radice_select->num_positions; i += 2) {

				if (radice_select->num_childs == 1) {

					res = get_result_from_position(radice_select->childs_vector[0], radice_select->tag_positions[i], radice_select->tag_positions[i+1], index_variabile, &risultato_figlio, &size_risultato_figlio, Console_Query->hash_table);

					if (res == -1)
						return -1;

					res = aggiungi_a_risultato(&risultato_figlio, &size_risultato_figlio, risultato, size_vettore_risultato );

					if (res == -1)
						return -1;

				} else { // numero figli  maggiore di zero altrimenti sarebbe stata una foglia

					// devo allocare il vettore che contiene le posizioni del risultato corrente (la dimensione di numero_variabili  moltiplicata per 2 perch vanno considerate le posizioni di apertura e chiusura
					risultato_parziale = (XCDE_IL_Couple_Type ***) malloc(3 * sizeof(XCDE_IL_Couple_Type**));

					if (risultato_parziale == NULL) {

						fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for risultato_parziale </xcde:error>\n\n") ;

						fflush(stderr) ;

						return(-1);

					}

					// devo allocare il vettore che contiene le posizioni del risultato corrente (la dimensione di numero_variabili  moltiplicata per 2 perch vanno considerate le posizioni di apertura e chiusura
					size_risultato_parziale = (int *) calloc(3, sizeof(int));

					if (size_risultato_parziale == NULL) {

						fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for size_risultato_parziale </xcde:error>\n\n") ;

						fflush(stderr) ;

						return(-1);

					}

					res = get_result_from_position(radice_select->childs_vector[0], radice_select->tag_positions[i], radice_select->tag_positions[i+1], index_variabile, &(risultato_parziale[j]), &(size_risultato_parziale[j]), Console_Query->hash_table);

					if (res == -1)
						return -1;

					res = get_result_from_position(radice_select->childs_vector[1], radice_select->tag_positions[i], radice_select->tag_positions[i+1], index_variabile, &(risultato_parziale[((j + 1) % 3)]), &(size_risultato_parziale[((j + 1) % 3)]), Console_Query->hash_table);

					if (res == -1)
						return -1;

					if (strcmp(radice_select->node_name, "xml_or") == 0) {

						res = fondi_risultati_or(&(risultato_parziale[j]), &(size_risultato_parziale[j]), &(risultato_parziale[((j + 1) % 3)]), &(size_risultato_parziale[((j + 1) % 3)]), &(risultato_parziale[((j + 2) % 3)]), &(size_risultato_parziale[((j + 2) % 3)]), numero_variabili);

						if (res == -1)
							return -1;

					} else {

						res = fondi_risultati(&(risultato_parziale[j]), &(size_risultato_parziale[j]), &(risultato_parziale[((j + 1) % 3)]), &(size_risultato_parziale[((j + 1) % 3)]), &(risultato_parziale[((j + 2) % 3)]), &(size_risultato_parziale[((j + 2) % 3)]), numero_variabili);

						if (res == -1)
							return -1;

					}

					j = ((j + 2) % 3);


					for (k=2; k<radice_select->num_childs; k++) {

						res = get_result_from_position(radice_select->childs_vector[k], radice_select->tag_positions[i], radice_select->tag_positions[i+1], index_variabile, &(risultato_parziale[((j + 1) % 3)]), &(size_risultato_parziale[((j + 1) % 3)]), Console_Query->hash_table);

						if (res == -1)
							return -1;

						if (strcmp(radice_select->node_name, "xml_or") == 0) {

							res = fondi_risultati_or(&(risultato_parziale[j]), &(size_risultato_parziale[j]), &(risultato_parziale[((j + 1) % 3)]), &(size_risultato_parziale[((j + 1) % 3)]), &(risultato_parziale[((j + 2) % 3)]), &(size_risultato_parziale[((j + 2) % 3)]), numero_variabili);

							if (res == -1)
								return -1;

						} else {

							res = fondi_risultati(&(risultato_parziale[j]), &(size_risultato_parziale[j]), &(risultato_parziale[((j + 1) % 3)]), &(size_risultato_parziale[((j + 1) % 3)]), &(risultato_parziale[((j + 2) % 3)]), &(size_risultato_parziale[((j + 2) % 3)]), numero_variabili);

							if (res == -1)
								return -1;

						}

						j = ((j + 2) % 3);

					}

					res = aggiungi_a_risultato(&(risultato_parziale[j]), &(size_risultato_parziale[j]), risultato, size_vettore_risultato );

					if (res == -1)
						return -1;

					if (risultato_parziale)
						free(risultato_parziale);

					if (size_risultato_parziale)
						free(size_risultato_parziale);

				}

			}

		} else {

			fprintf(stderr, "\n\t<xcde:error> XCDE_Compute_TableResult function can be call only if node is TAG </xcde:error>\n\n") ;

			fflush(stderr) ;

			return(-1);


		}
	}

	return (1);
}



int XCDE_View_Result(int index_file, XCDE_IL_Couple_Type **risultato, int size_vettore_risultato, unsigned char flag_position, XCDE_Console_Query_Type *Console_Query) {

	int i;
	XCDE_Console_Type XCDE_Console;
	unsigned char is_setted_depthstruct;


	if (Console_Query == NULL)
		return -1;


	if (Console_Query->XCDE_query_parsed != 1)
		return -1;


	if ( (index_file < 0) || (index_file >= Console_Query->number_of_files) )
		return -1;


	// controllo sulla consistenza dei valori dell'intervallo inseriti
	if (Console_Query->first_range > Console_Query->last_range) {

		fprintf(stderr, "\n\t<xcde:error> first number of range must be greater than second one - XCDE_View_Result() </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-2);

	}

	if ( ((Console_Query->first_range == 0) && (Console_Query->last_range != 0)) || ((Console_Query->first_range != 0) && (Console_Query->last_range == 0)) ) {

		fprintf(stderr, "\n\t<xcde:error> gli estremi dell'intervallo inseriti devono essere entrambi 0 (se si vogliono visualizzare tutti i risultati) o entrambi diversi da zero. - XCDE_View_Result() </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-2);

	}

	if ( (Console_Query->first_range == 0) && (Console_Query->last_range == 0) && (size_vettore_risultato > 0) ) {

		// bisogna visualizzare tutti i risultati
		Console_Query->first_range = 1;
		Console_Query->last_range = size_vettore_risultato;

	}

	if ( (Console_Query->first_range < 0) || (Console_Query->last_range < 0) ) {

		fprintf(stderr, "\n\t<xcde:error> numbers must be greater than 0 - XCDE_View_Result() </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-2);

	}

	if ( (size_vettore_risultato != 0) && (risultato == NULL) ) {

		return(-1);

	}


	/*******************************************************************
			Inizializzazione della console
	*******************************************************************/

	// opening console
	res=XCDE_Init_Console((Console_Query->file_vector[index_file]), &XCDE_Console) ;

	if (res == -1) {

		fprintf(stderr, "\n\t<xcde:error> problems with the XCDE_Init_Console function - XCDE_View_Result() </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-1);

	}

	res = XCDE_SettedDepthstruct(&is_setted_depthstruct, &XCDE_Console) ;

	if (res == -1) {

		fprintf(stderr, "\n\t<xcde_error> problem XCDE_SettedDepthstruct funtion </xcde_error>\n\n") ;

		fflush(stderr) ;

		XCDE_Finalize_Console(&XCDE_Console) ;

		return -1;

	}

	if (is_setted_depthstruct != 1) {

		fprintf(stderr, "\n\t<xcde_error> Depthstruct must be setted. Rebuild the file with -d option. </xcde_error>\n\n") ;

		fflush(stderr) ;

		XCDE_Finalize_Console(&XCDE_Console) ;

		return 2;

	}

	if ( size_vettore_risultato == 0 ) {

		XCDE_Finalize_Console(&XCDE_Console) ;

		return(0);

	}

	if ( Console_Query->first_range > size_vettore_risultato ) {

		fprintf(stderr, "\n\t<xml_warning> no result in current range. Results are only %d </xml_warning>\n\n", size_vettore_risultato) ;

		fflush(stderr) ;

		XCDE_Finalize_Console(&XCDE_Console) ;

		return(1);

	}

	if (flag_position == 0) {

		// vengono lette dal file di configurazione le informazioni relative ai separatori
		res = read_configuration_file();

		if (res == -1) {

			XCDE_Finalize_Console(&XCDE_Console) ;

			return (-1);

		}

		for (i = (Console_Query->first_range - 1); ( (i < Console_Query->last_range) && (i < size_vettore_risultato) ); i++) {

			print_header_return(i);

			res = view_snippet_result(Console_Query->return_tree, Console_Query->file_vector[index_file], risultato, size_vettore_risultato, Console_Query->hash_table, i, &XCDE_Console);

			print_footer_return();

			if (res == -1) {

				XCDE_Finalize_Console(&XCDE_Console) ;

				return (-1);

			}

			if (res == -2) {

				XCDE_Finalize_Console(&XCDE_Console) ;

				return (-2);

			}

		}

	} else {

		for (i = (Console_Query->first_range - 1); ( (i < Console_Query->last_range) && (i < size_vettore_risultato) ); i++) {

			print_header_return(i);

			res = view_position_result(Console_Query->return_tree, risultato, size_vettore_risultato, Console_Query->hash_table, i);

			print_footer_return();

			if (res == -1) {

				XCDE_Finalize_Console(&XCDE_Console) ;

				return (-1);

			}

			if (res == -2) {

				XCDE_Finalize_Console(&XCDE_Console) ;

				return (-2);

			}

		}

	}


	/*******************************************************************
			Finalizzazione della console
	*******************************************************************/

	res=XCDE_Finalize_Console(&XCDE_Console) ;

	if (res == -1) {

		fprintf(stderr, "\n\t<xcde:error> problems with the XCDE_Finalize_Console function </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-1);
	}

	return (1);
}



// Funzione che scarica il contenuto del vettore result e della tabella hash in un file
//
// Write in a file ".result" the result vector and hash table with an HEADER
//
// Header contains:
// 	1. number of result in risultato
// 	2. number of stored objects in hash table (number of variables)
//	3. pointer vector (which contain (number of result)/SKIP_RESULT integer)
//	4. pointer to hash table byte
//	5. pointer to name of file byte
//
// Body contains:
//
// 	for each result in risultato:
//		1. length in byte of comressed result
//		2. for each stored variable
//			1. open position (compressed)
//			2. open offset (compressed)
//			3. close position (compressed)
//			4. close offset (compressed)
//
// 	for each object stored in hash table:
// 		1. variable name length
//		2. variable name
// 		3. variable index (compressed)
// 		4. name of associated node length
// 		5. name of associated node

int XCDE_Write_ResultFile(XCDE_IL_Couple_Type **risultato, int size_vettore_risultato, char *filename, XCDE_Console_Query_Type *Console_Query) {

	int i,k, l;
	int j = 0;
	FILE *ptrFile;

	char * nome_file;
	Hash_node *p;

	int *pointer_vector = NULL;
	int pointer_dimension = 0;

	unsigned long pointer = 0ul ;
	unsigned long counter_byte = 0ul ;
	unsigned long len_continuation_bit = 0ul ;

	char *buffer_compr_result = NULL;
	int lungh_node_name_ass, lungh_nome;



	if (Console_Query == NULL)
		return -1;


	if (Console_Query->XCDE_query_parsed != 1)
		return -1;


	if ( filename == NULL )
		return -1;


	if ( (size_vettore_risultato != 0) && (risultato == NULL) ) {

		return(-1);

	}


	// allocazione della memoria necessaria per il nome del file in cui scrivere i risultati
	nome_file = (char *) malloc( (strlen(filename) + strlen(XCDE_RESULT_SUFFIX) + 1) * sizeof(char));

	// controllo sulla corretta allocazione della memoria
	if (nome_file == NULL) {

		fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for nome_file in xcde_search.c - XCDE_Write_ResultFile() </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-1);
	}

	strcpy(nome_file, filename);

	strcat(nome_file, XCDE_RESULT_SUFFIX);

	if ((ptrFile = fopen(nome_file, "wb")) == NULL) {

		fprintf(stderr, "\n\t<xcde:error> file %s could not be opened </xcde:error>\n\n", nome_file) ;

		fflush(stderr) ;

		return(-1);
	}


	pointer_dimension = ((size_vettore_risultato - 1) / SKIP_RESULT) + 1;

	// allocazione della memoria necessaria per il vettore dei puntatori per lo SKIP_RESULT
	//	siapprossima per eccesso
	pointer_vector = (int *) malloc( pointer_dimension * sizeof(int) );

	// controllo sulla corretta allocazione della memoria
	if (pointer_vector == NULL) {

		fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for pointer_vector in xcde_search.c - XCDE_Write_ResultFile() </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-1);
	}

	// si scrive nel file la dimensione del vettore risultato
	fwrite_int(ptrFile, size_vettore_risultato);

	// si scrive nel file il numero di variabili presenti nel risultato (2 * numero di variabili  il numero di elementi per ogni risultato)
	fwrite_int(ptrFile, Console_Query->hash_table->card);

	// si riserva nel file lo spazio necessario a contenere il vettore dei puntatori ai bytes campioni (uno ogni SKIP_RESULT) pi il puntatore al byte della tabella hash
	fwrite(pointer_vector, sizeof(int), pointer_dimension + 1, ptrFile);

	// l'Header ha una lunghezza fissa di (4 + 4 + (4 * size_vettore_risultato / SKIP_RESULT) + 4) byte
	counter_byte = (sizeof(int) * (3 + pointer_dimension));

	// allocazione della memoria necessaria per il buffer in cui verranno scritti gli interi compressi
	buffer_compr_result = (char *) malloc((sizeof(int) + 1) + 4 * Console_Query->hash_table->card);

	// controllo sulla corretta allocazione della memoria
	if (buffer_compr_result == NULL) {

		fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for buffer_compr_result in xcde_search.c - XCDE_Write_ResultFile() </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-1);
	}

	pointer = 0ul;

	/*******************************************************************
				loops over all results
	*******************************************************************/

	// per ogni risultato si scrivono nel file risultato le informazioni necessarie
	for (i = 0; i < size_vettore_risultato; i++) {

		if ( (i % (SKIP_RESULT)) == 0 ) {

			pointer_vector[j] = counter_byte;

			j++;

		}

		for (k = 0; k < (2 * Console_Query->hash_table->card); k++) {

			continuation_bit_c((unsigned long) risultato[i][k].XCDE_Position + 1,
					(char *) buffer_compr_result + pointer,
					(unsigned int) 0,
					(unsigned long *) &len_continuation_bit) ;

			pointer += (len_continuation_bit/8) ;

			continuation_bit_c((unsigned long) risultato[i][k].XCDE_Offset + 1,
					(char *) buffer_compr_result + pointer,
					(unsigned int) 0,
					(unsigned long *) &len_continuation_bit) ;

			pointer += (len_continuation_bit/8) ;

		}

		counter_byte += pointer + sizeof(int);

		// si scrive nel file la lunghezza in byte che serve per saltare al prossimo risultato
		fwrite_int(ptrFile, pointer);

		// si scrive nel file il buffer_compr_result
		fwrite(buffer_compr_result, sizeof(char), pointer, ptrFile);

		pointer = 0ul;

	}

	// vengono scritte le informazioni della tabella hash
	for(i=0; i<Console_Query->hash_table->size;i++){

		for(p = Console_Query->hash_table->table[i]; p ; p = p->next){

			lungh_nome = strlen(p->variable_name) + 1;

			// si scrive nel file la lunghezza del nome della variabile
			fwrite_int(ptrFile, lungh_nome);

			// si scrive nel file il nome della variabile
			fwrite(p->variable_name, sizeof(char), lungh_nome, ptrFile);

			// si scrive nel file l'index corrispondente alla variabile
			fwrite_int(ptrFile, p->index);

			lungh_node_name_ass = strlen(p->node_name) + 1;

			// si scrive nel file la lunghezza del nome della variabile
			fwrite_int(ptrFile, lungh_node_name_ass);

			// si scrive nel file il nome della variabile
			fwrite(p->node_name, sizeof(char), lungh_node_name_ass, ptrFile);
		}
	}

	// si riposiziona il puntatore per poter riscrivere il pointer_vector con le informazioni aggiornate
	fseek(ptrFile, 2 * sizeof(int), SEEK_SET);

	for (i=0; i < pointer_dimension; i++) {

		// si riscrive il pointer_vector con le informazioni aggiornate
		fwrite_int(ptrFile, pointer_vector[i]);

	}

	// si riscrive il puntatore al byte dove comincia la tabella hash
	fwrite_int(ptrFile, counter_byte);

	fclose(ptrFile);

	return(1);

}



int XCDE_Read_Range_ResultFile(char *filename, int first_range, int last_range, XCDE_IL_Couple_Type ***risultato, int *size_vettore_risultato, XCDE_Console_Query_Type *Console_Query) {


	int i,k;
	int j = 0;
	FILE *ptrFile;

	char * nome_file;
	Hash_node *p;

	int pointer = 0 ;
	int temp_pointer = 0;
	int pointer_dimension = 0;
	unsigned long len_continuation_bit = 0ul ;

	int byte_tabella_hash = 0 ;

	int index_risultato_marcato = 0;
	int numero_variabili = 0;

	char *result_mapped_pointer = NULL;
	unsigned long size_file_result = 0;


	
	
	if (Console_Query == NULL)
		return -1;


	if ( filename == NULL )
		return -1;


	// controllo sulla consistenza dei valori dell'intervallo inseriti
	if (first_range > last_range) {

		fprintf(stderr, "\n\t<xcde:error> first number of range must be greater than second one - XCDE_View_Result() </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-2);

	}

	if ( (first_range < 0) || (last_range < 0) ) {

		fprintf(stderr, "\n\t<xcde:error> range must be greater than 0 - XCDE_View_Result() </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-2);

	}


	if ( (first_range == 0) && (last_range == 0) ) {

		*size_vettore_risultato = 0;

		*risultato = NULL;

		return(-1);

	}


	// allocazione della memoria necessaria per il nome del file in cui scrivere i risultati
	nome_file = (char *) malloc( (strlen(filename) + strlen(XCDE_RESULT_SUFFIX) + 1) * sizeof(char));

	// controllo sulla corretta allocazione della memoria
	if (nome_file == NULL) {

		fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for nome_file in xcde_search.c - XCDE_Read_Range_ResultFile() </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-1);
	}

	strcpy(nome_file, filename);

	strcat(nome_file, XCDE_RESULT_SUFFIX);

	if ((ptrFile = fopen(nome_file, "rb")) == NULL) {

		fprintf(stderr, "\n\t<xcde:error> file %s could not be opened - XCDE_Read_Range_ResultFile </xcde:error>\n\n", nome_file) ;

		fflush(stderr) ;

		return(-1);
	}


	if (fseek(ptrFile, 0L, SEEK_END)) {

		fprintf(stderr, "<xcde:error> Seek error on result file - XCDE_Read_Range_ResultFile() </xcde:error>\n\n");

		fflush(stderr) ;

		return(-1);

	}

	size_file_result = ftell(ptrFile);

	if (fseek(ptrFile, 0L, SEEK_SET)) {

		fprintf(stderr, "<xcde:error> Seek error on result file - XCDE_Read_Range_ResultFile() </xcde:error>\n\n");

		fflush(stderr) ;

		return(-1);
	}

	/* -------------------------------- */
	/*   result file mapping operation  */
	/* -------------------------------- */

	result_mapped_pointer = (char *) mmap(NULL, size_file_result, PROT_READ, MAP_SHARED, fileno(ptrFile), 0) ;

	if (result_mapped_pointer == MAP_FAILED) {

		fprintf(stderr, "\n\t<xcde:error> problems in map operation on file %s - XCDE_Read_Range_ResultFile() </xcde:error>\n\n", nome_file) ;

		fflush(stderr) ;

		return(-1);
	}

	fclose (ptrFile);

	// si recupera dal file mappato il numero di risultati
	read_int(size_vettore_risultato, result_mapped_pointer + pointer);

	if ( first_range > *size_vettore_risultato ) {

		fprintf(stderr, "\n\t<xcde:error> no result in current range. Results are only %d </xcde:error>\n\n", *size_vettore_risultato) ;

		fflush(stderr) ;

		return(1);

	}

	pointer += sizeof(int);

	// si recupera dal file mappato il numero di variabili presenti nel risultato (2 * numero di variabili  il numero di elementi per ogni risultato)
	read_int(&(numero_variabili), result_mapped_pointer + pointer);

	pointer += sizeof(int);


	if (*size_vettore_risultato == 0) {

		fprintf(stderr, "\n\t<xcde:error> number of results in file is zero - XCDE_Read_Range_ResultFile() </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-1);
	}

	// se last_range va oltre il numero di risultati si limita al numero di risultati
	if (last_range > *size_vettore_risultato) {

		last_range = *size_vettore_risultato;
	}

	// controllo sulla consistenza dei valori dell'intervallo inseriti
	if (first_range > last_range) {

		fprintf(stderr, "\n\t<xcde:error> first number must be greater than second one - XCDE_Read_Range_ResultFile() </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-2);
	}

	if ( (first_range == 0) || (last_range == 0) ) {

		fprintf(stderr, "\n\t<xcde:error> numbers must be different from 0 - XCDE_Read_Range_ResultFile() </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-2);
	}

	if ( (first_range < 0) || (last_range < 0) ) {

		fprintf(stderr, "\n\t<xcde:error> numbers must be greater than 0 - XCDE_Read_Range_ResultFile() </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-2);
	}



	pointer_dimension = ((*size_vettore_risultato - 1) / SKIP_RESULT) + 1;

	byte_tabella_hash = (sizeof(int) * (2 + pointer_dimension));

	index_risultato_marcato = ((first_range - 1) / SKIP_RESULT);

	// si recupera dal file mappato il byte del primo risultato del blocco contenente il primo risultato
	//	da ricercare
	read_int(&(pointer), result_mapped_pointer + (sizeof(int) * (2 + index_risultato_marcato)));


	for (i=0; i<((first_range - 1) % SKIP_RESULT); i++) {

		// si recupera dal file mappato il numero di byte da saltare per raggiungere il successivo risultato
		read_int(&(temp_pointer), result_mapped_pointer + pointer);

		pointer += temp_pointer + sizeof(int);

	}


	*size_vettore_risultato = last_range -  first_range + 1;

	(*risultato) = (XCDE_IL_Couple_Type **) malloc((*size_vettore_risultato) * sizeof(XCDE_IL_Couple_Type *));

	if ((*risultato) == NULL) {

		fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for (*risultato) - XCDE_Read_Range_ResultFile() </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-1);
	}


	/*******************************************************************
				loops over all results
	*******************************************************************/

	// per ogni risultato si leggono dal file risultato le informazioni necessarie
	for (i = 0; i < *size_vettore_risultato; i++) {

		// si salta l'informazione relativa al byte del prossimo risultato
		pointer += sizeof(int);

		(*risultato)[i] = (XCDE_IL_Couple_Type *) malloc((2 * numero_variabili) * sizeof(XCDE_IL_Couple_Type));

		if ((*risultato)[i] == NULL) {

			fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for (*risultato)[i] - XCDE_Read_Range_ResultFile() </xcde:error>\n\n") ;

			fflush(stderr) ;

			return(-1);
		}

		for (k = 0; k < (2 * numero_variabili); k++) {

			continuation_bit_d((unsigned long *) &((*risultato)[i][k].XCDE_Position),
					(char *) result_mapped_pointer + pointer,
					(unsigned int) 0,
					(unsigned long *) &len_continuation_bit) ;

			// viene decrementato il valore di Position poich nella compressione si era aggiunto 1 per
			//	evitare la codifica dello 0
			((*risultato)[i][k].XCDE_Position)--;

			pointer += (len_continuation_bit/8) ;


			continuation_bit_d((unsigned long *) &((*risultato)[i][k].XCDE_Offset),
					(char *) result_mapped_pointer + pointer,
					(unsigned int) 0,
					(unsigned long *) &len_continuation_bit) ;

			// viene decrementato il valore di Position poich nella compressione si era aggiunto 1 per
			//	evitare la codifica dello 0
			((*risultato)[i][k].XCDE_Offset)--;

			pointer += (len_continuation_bit/8) ;


		}
	}


	// si recupera dal file mappato il byte dove inizia la tabella hash
	read_int(&pointer, result_mapped_pointer + byte_tabella_hash);

	res = riempi_hash(result_mapped_pointer + pointer, numero_variabili, Console_Query);

	if (res == -1)
		return (-1);

	return 1;


}


int XCDE_Read_ResultFile(char *filename, XCDE_IL_Couple_Type ***risultato, int *size_vettore_risultato, XCDE_Console_Query_Type *Console_Query) {


	FILE *ptrFile;
	char * nome_file;



	if (filename == NULL)
		return -1;

	if (Console_Query == NULL)
		return -1;


	// allocazione della memoria necessaria per il nome del file in cui scrivere i risultati
	nome_file = (char *) malloc( (strlen(filename) + strlen(XCDE_RESULT_SUFFIX) + 1) * sizeof(char));

	// controllo sulla corretta allocazione della memoria
	if (nome_file == NULL) {

		fprintf(stderr, "\n\t<xcde:error> problems during memory allocation for nome_file in xcde_search.c - XCDE_Write_ResultFile() </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-1);
	}

	strcpy(nome_file, filename);

	strcat(nome_file, XCDE_RESULT_SUFFIX);

	if ((ptrFile = fopen(nome_file, "rb")) == NULL) {

		fprintf(stderr, "\n\t<xcde:error> file %s could not be opened </xcde:error>\n\n", nome_file) ;

		fflush(stderr) ;

		return(-1);
	}

	fread (size_vettore_risultato, sizeof(int), 1, ptrFile);

	if (*size_vettore_risultato == 0) {

		fprintf(stderr, "\n\t<xcde:error> number of results in file is zero - XCDE_Read_Range_ResultFile() </xcde:error>\n\n") ;

		fflush(stderr) ;

		return(-1);
	}

	res = XCDE_Read_Range_ResultFile(filename, 1, *size_vettore_risultato, risultato, size_vettore_risultato, Console_Query);

	if (res == -1)
		return (-1);

	if (res == -2)
		return (-2);

	return 1;

}




int XCDE_Search(char *query_string, unsigned char flag_position) {


	int i;

	int query_len = 0;

	FILE *filePtr;

	// variabile console_query
	XCDE_Console_Query_Type Console_Query;

	XCDE_IL_Couple_Type **risultato = NULL;
	int size_vettore_risultato = 0;

	char *Header_Filename = NULL;



	if (query_string == NULL) {

		return -1;

	}

	// vengono sostituiti eventuali caratteri '\n' nella query con ' '
	query_len = strlen(query_string);

	for (i = 0; i < query_len; i++) {

		if ( (query_string[i] == '\n') || (query_string[i] == '\t') )
			query_string[i] = ' ';
	}


	res = XCDE_Parse_Query_String(query_string, &Console_Query);

	if (res == -1)
		return (-1);

	if (res == -2)
		return (-2);

	res = XCDE_Execute_Query(&Console_Query);

	if (res == -1)
		return (-1);

	if (res == -2)
		return (-2);

	fprintf(stdout, "<");
	fprintf(stdout, XCDE_MAIN_RESULT_ELEMENT);
	fprintf(stdout, ">\n\n");

	for (i=0; i<Console_Query.number_of_trees; i++) {

		print_header((Console_Query.file_vector)[i], i+1);

		Header_Filename = malloc(strlen((Console_Query.file_vector)[i]) + 1 + strlen(XCDE_HEADER_SUFFIX)) ;

		if (Header_Filename == NULL) {

			fprintf(stderr, "\n\t<xcde_error> error in memory allocation for Header_Filename </xcde_error>\n\n") ;

			fflush(stderr) ;

			continue;
		}

		strcpy(Header_Filename, (Console_Query.file_vector)[i]) ;

		strcat(Header_Filename, XCDE_HEADER_SUFFIX) ;

		if ((filePtr = fopen(Header_Filename, "r")) != NULL) {

			fclose(filePtr);

			res = XCDE_Compute_TableResult(i, &risultato, &size_vettore_risultato, &Console_Query);

			if (res == -1) {

				print_footer();

				continue;

			} else {

				res = XCDE_View_Result(i, risultato, size_vettore_risultato, flag_position, &Console_Query);

				if (res == -1) {

					print_footer();

					continue;

				}

				if (res == -2) {

					print_footer();

					continue;

				}

				if (res == 0) {

					fprintf(stderr, "\n\t<xcde_warning> no result found for this file </xcde_warning>\n\n") ;

					fflush(stderr) ;

				}

			}
		} else {

			fprintf(stderr, "\n\t<xcde_error> file not found. </xcde_error>\n\n") ;

			fflush(stderr) ;

		}

		print_footer();

	}

	fprintf(stdout, "</");
	fprintf(stdout, XCDE_MAIN_RESULT_ELEMENT);
	fprintf(stdout, ">\n\n");

	return (1);

}

